@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22035 $
+ * $Revision: 24114 $
  * $Author: rjongbloed $
- * $Date: 2009-02-11 05:20:48 +0000 (Wed, 11 Feb 2009) $
+ * $Date: 2010-03-11 20:29:31 -0600 (Thu, 11 Mar 2010) $
  */
 
 #ifndef OPAL_CODEC_OPALPLUGIN_H
@@ -124,11 +124,11 @@ enum PluginCodec_Flags {
   PluginCodec_MediaTypeFax           = 0x0003,
 
   PluginCodec_InputTypeMask          = 0x0010,
-  PluginCodec_InputTypeRaw           = 0x0000,
+  PluginCodec_InputTypeRaw           = 0x0000, // Note video is always RTP
   PluginCodec_InputTypeRTP           = 0x0010,
 
   PluginCodec_OutputTypeMask         = 0x0020,
-  PluginCodec_OutputTypeRaw          = 0x0000,
+  PluginCodec_OutputTypeRaw          = 0x0000, // Note video is always RTP
   PluginCodec_OutputTypeRTP          = 0x0020,
 
   PluginCodec_RTPTypeMask            = 0x0040,
@@ -169,9 +169,11 @@ enum PluginCodec_CoderFlags {
 };
 
 enum PluginCodec_ReturnCoderFlags {
-  PluginCodec_ReturnCoderLastFrame     = 1,    // indicates when video codec returns last data for frame
-  PluginCodec_ReturnCoderIFrame        = 2,    // indicates when video returns I frame
-  PluginCodec_ReturnCoderRequestIFrame = 4     // indicates when video decoder request I frame for resync
+  PluginCodec_ReturnCoderLastFrame      = 1,    // indicates when video codec returns last data for frame
+  PluginCodec_ReturnCoderIFrame         = 2,    // indicates when video returns I frame
+  PluginCodec_ReturnCoderRequestIFrame  = 4,    // indicates when video decoder request I frame for resync
+  PluginCodec_ReturnCoderBufferTooSmall = 8     // indicates when output buffer is not large enough to receive
+                                                // the data, another call to get_output_data_size is required
 };
 
 struct PluginCodec_Definition;
@@ -187,6 +189,8 @@ struct PluginCodec_Definition;
 #define PLUGINCODEC_CONTROL_TO_CUSTOMISED_OPTIONS "to_customised_options"
 #define PLUGINCODEC_CONTROL_SET_INSTANCE_ID       "set_instance_id"
 #define PLUGINCODEC_CONTROL_SET_LOG_FUNCTION      "set_log_function"
+#define PLUGINCODEC_CONTROL_GET_STATISTICS        "get_statistics"
+#define PLUGINCODEC_CONTROL_TERMINATE_CODEC       "terminate_codec"
 
 
 /* Log function, plug in gets a pointer to this function which allows
@@ -278,8 +282,6 @@ struct PluginCodec_Option {
 #define PLUGINCODEC_OPTION_MAX_RX_FRAME_HEIGHT        "Max Rx Frame Height"
 #define PLUGINCODEC_OPTION_TEMPORAL_SPATIAL_TRADE_OFF "Temporal Spatial Trade Off"
 #define PLUGINCODEC_OPTION_TX_KEY_FRAME_PERIOD        "Tx Key Frame Period"
-#define PLUGINCODEC_OPTION_TIAS                       "Bandwidth-TIAS"
-#define PLUGINCODEC_OPTION_MAXPRATE                   "MaxPRate"
 
 
 
@@ -541,6 +543,7 @@ enum {
 #define PLUGINCODEC_CIF_MPI       "CIF MPI"
 #define PLUGINCODEC_CIF4_MPI     "CIF4 MPI"
 #define PLUGINCODEC_CIF16_MPI   "CIF16 MPI"
+#define PLUGINCODEC_CUSTOM_MPI "Custom MPI"
 
 #define PLUGINCODEC_MPI_DISABLED 33
 
@@ -548,9 +551,13 @@ enum {
 #define PLUGINCODEC_MEDIA_PACKETIZATIONS "Media Packetizations"
 
 #define H261_ANNEX_D "Annex D - Still Image Transmit"
+#define H263_ANNEX_D "Annex D - Unrestricted Motion Vector"
 #define H263_ANNEX_F "Annex F - Advanced Prediction"
 #define H263_ANNEX_I "Annex I - Advanced INTRA Coding"
 #define H263_ANNEX_J "Annex J - Deblocking Filter"
+#define H263_ANNEX_K "Annex K - Slice Structure"
+#define H263_ANNEX_N "Annex N - Reference Picture Selection"
+#define H263_ANNEX_S "Annex S - Alternative INTER VLC"
 #define H263_ANNEX_T "Annex T - Modified Quantization"
 
 #ifndef STRINGIZE
