@@ -26,13 +26,13 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21004 $
+ * $Revision: 23115 $
  * $Author: rjongbloed $
- * $Date: 2008-09-16 07:08:56 +0000 (Tue, 16 Sep 2008) $
+ * $Date: 2009-07-17 08:13:11 -0500 (Fri, 17 Jul 2009) $
  */
 
-#ifndef __OPAL_H323CAPS_H
-#define __OPAL_H323CAPS_H
+#ifndef OPAL_H323_H323CAPS_H
+#define OPAL_H323_H323CAPS_H
 
 #ifdef P_USE_PRAGMA
 #pragma interface
@@ -40,6 +40,8 @@
 
 #include <opal/buildopts.h>
 
+#if OPAL_H323
+
 #include <opal/mediafmt.h>
 #include <h323/channels.h>
 
@@ -119,6 +121,8 @@ class H323Capability : public PObject
       e_Data,
       /// User Input capability
       e_UserInput,
+      /// Generic Control
+      e_GenericControl,
       /// Count of main types
       e_NumMainTypes
     };
@@ -149,10 +153,17 @@ class H323Capability : public PObject
       const PString & name     ///<  Name of capability
     );
 
+    enum {
+      DefaultAudioSessionID = 1,
+      DefaultVideoSessionID = 2,
+      DefaultDataSessionID  = 3,
+      DefaultH224SessionID  = 4
+    };
+
     /**Get the default RTP session.
        This function gets the default RTP session ID for the capability
        type. For example audio capabilities return the value
-       RTP_Session::DefaultAudioSessionID etc.
+       H323Capability::DefaultAudioSessionID etc.
 
        The default behaviour returns zero, indicating it is not an RTP
        based capability.
@@ -317,7 +328,7 @@ class H323Capability : public PObject
 
     /** Set media format option for the media data this class represents.
       */
-    bool SetMediaFormatOptions(
+    bool UpdateMediaFormat(
       const OpalMediaFormat & format
     );
 
@@ -337,14 +348,14 @@ class H323Capability : public PObject
 #endif
 
   protected:
-    OpalMediaFormat & GetWritableMediaFormat();
+    OpalMediaFormat & GetWritableMediaFormat() const;
 
     unsigned            assignedCapabilityNumber;  /// Unique ID assigned to capability
     CapabilityDirection capabilityDirection;
     RTP_DataFrame::PayloadTypes rtpPayloadType;
 
   private:
-    mutable OpalMediaFormat mediaFormat;
+    mutable OpalMediaFormat m_mediaFormat;
 
   friend class H323Capabilities;
 };
@@ -470,10 +481,10 @@ class H323GenericCapabilityInfo
   public:
     H323GenericCapabilityInfo(
       const PString & id,     ///< generic codec identifier
-      PINDEX maxBitRate = 0   ///< maxBitRate parameter for the GenericCapability
+      unsigned maxBitRate = 0   ///< maxBitRate parameter for the GenericCapability
     );
-    H323GenericCapabilityInfo(const H323GenericCapabilityInfo & obj);
-    virtual ~H323GenericCapabilityInfo();
+
+    virtual ~H323GenericCapabilityInfo() { }
 
   protected:
     virtual PBoolean OnSendingGenericPDU(
@@ -495,8 +506,8 @@ class H323GenericCapabilityInfo
     ) const;
 
 
-    H245_CapabilityIdentifier * identifier;
-    unsigned                    maxBitRate;
+    PString  m_identifier;
+    unsigned maxBitRate;
 };
 
 /**This class describes the interface to a codec that has channels based on
@@ -853,7 +864,7 @@ class H323GenericAudioCapability : public H323AudioCapability,
       */
     H323GenericAudioCapability(
       const PString & capabilityId,    ///< generic codec identifier
-      PINDEX maxBitRate = 0	           ///< maxBitRate parameter for the GenericCapability
+      PINDEX maxBitRate = 0               ///< maxBitRate parameter for the GenericCapability
     );
   //@}
 
@@ -1204,7 +1215,7 @@ class H323GenericVideoCapability : public H323VideoCapability,
       */
     H323GenericVideoCapability(
       const PString & capabilityId,    ///< generic codec identifier (OID)
-      PINDEX maxBitRate = 0	       ///< maxBitRate parameter for the GenericCapability
+      PINDEX maxBitRate = 0           ///< maxBitRate parameter for the GenericCapability
     );
   //@}
 
@@ -1265,17 +1276,294 @@ class H323GenericVideoCapability : public H323VideoCapability,
       const H245_VideoCapability & pdu,  ///< PDU to get information from
       CommandType type                   ///<  Type of PDU to send in
     );
-  //@}
 
     /**Compare the generic part of the capability, if applicable.
      */
     virtual PBoolean IsMatch(
       const PASN_Choice & subTypePDU  ///<  sub-type PDU of H323Capability
     ) const;
+  //@}
 };
 
+
+#if OPAL_H239
+
+/**This class descibes an extended ideo capability, as used in H.239.
+  */
+class H323ExtendedVideoCapability : public H323GenericVideoCapability
+{
+  PCLASSINFO(H323ExtendedVideoCapability, H323GenericVideoCapability);
+
+  public:
+  /**@name Construction */
+  //@{
+    /**Create a new Extended Video capability.
+      */
+    H323ExtendedVideoCapability(
+      const PString & identifier   ///< generic codec identifier (OID)
+    );
+  //@}
+
+  /**@name Identification functions */
+  //@{
+    /**Get the sub-type of the capability. This is a code dependent on the
+       main type of the capability.
+
+       This returns H245_VideoCapability::e_extendedVideoCapability.
+     */
+    virtual unsigned GetSubType() const;
+  //@}
+    
+  /**@name Protocol manipulation */
+  //@{
+    /**This function is called whenever and outgoing TerminalCapabilitySet
+       or OpenLogicalChannel PDU is being constructed for the control channel.
+       It allows the capability to set the PDU fields from information in
+       members specific to the class.
+
+       The default behaviour calls H323GenericCapabilityinfo::OnSendingPDU()
+       to handle the PDU.
+     */
+    virtual PBoolean OnSendingPDU(
+      H245_VideoCapability & pdu,  ///<  PDU to set information on
+      CommandType type             ///<  Type of PDU to send in
+    ) const;
+
+    /**This function is called whenever and outgoing RequestMode
+       PDU is being constructed for the control channel. It allows the
+       capability to set the PDU fields from information in members specific
+       to the class.
+
+       The default behaviour sets the PDUs tag according to the GetSubType()
+       function (translated to different enum).
+     */
+    virtual PBoolean OnSendingPDU(
+      H245_VideoMode & pdu  ///<  PDU to set information on
+    ) const;
+
+    /**This function is called whenever and incoming TerminalCapabilitySet
+       or OpenLogicalChannel PDU has been used to construct the control
+       channel. It allows the capability to set from the PDU fields,
+       information in members specific to the class.
+
+       The default behaviour calls H323GenericCapabilityinfo::OnReceivedPDU()
+       to handle the provided PDU.
+     */
+    virtual PBoolean OnReceivedPDU(
+      const H245_VideoCapability & pdu,  ///< PDU to get information from
+      CommandType type                   ///<  Type of PDU to send in
+    );
+
+    /**Compare the generic part of the capability, if applicable.
+     */
+    virtual PBoolean IsMatch(
+      const PASN_Choice & subTypePDU  ///<  sub-type PDU of H323Capability
+    ) const;
+  //@}
+
+  protected:
+    OpalMediaFormatList m_videoFormats;
+};
+
+
+///////////////////////////////////////////////////////////////////////////////
+
+/**This class describes a generic control capability, as used in H.239
+ */
+class H323GenericControlCapability : public H323Capability,
+                                     public H323GenericCapabilityInfo
+{
+  PCLASSINFO(H323GenericControlCapability, H323Capability);
+
+  public:
+  /**@name Construction */
+  //@{
+    /**Create a new data capability.
+      */
+    H323GenericControlCapability(
+      const PString & identifier  ///< Indentifer (OID) for generic control
+    );
+  //@}
+
+  /**@name Identification functions */
+  //@{
+    /**Get the main type of the capability.
+       Always returns e_Data.
+     */
+    virtual MainTypes GetMainType() const;
+
+    /**Get the sub-type of the capability. This is a code dependent on the
+       main type of the capability.
+
+       This returns H245_VideoCapability::e_extendedVideoCapability.
+     */
+    virtual unsigned GetSubType() const;
+  //@}
+
+  /**@name Protocol manipulation */
+  //@{
+    /**This function is called whenever and outgoing TerminalCapabilitySet
+       PDU is being constructed for the control channel. It allows the
+       capability to set the PDU fields from information in members specific
+       to the class.
+
+       The default behaviour calls the OnSendingPDU() function with a more
+       specific PDU type.
+     */
+    virtual PBoolean OnSendingPDU(
+      H245_Capability & pdu  ///<  PDU to set information on
+    ) const;
+
+    /**This function is called whenever and outgoing RequestMode
+       PDU is being constructed for the control channel. It allows the
+       capability to set the PDU fields from information in members specific
+       to the class.
+
+       The default behaviour is pure.
+     */
+    virtual PBoolean OnSendingPDU(
+      H245_ModeElement & pdu  ///<  PDU to set information on
+    ) const;
+
+    /**This function is called whenever and incoming TerminalCapabilitySet
+       PDU is received on the control channel, and a new H323Capability
+       descendent was created. This completes reading fields from the PDU
+       into the classes members.
+
+       If the function returns PFalse then the received PDU codec description
+       is not supported, so will be ignored. The default behaviour simply
+       returns PTrue.
+     */
+    virtual PBoolean OnReceivedPDU(
+      const H245_Capability & pdu  ///<  PDU to get information from
+    );
+
+    /**Compare the generic part of the capability, if applicable.
+     */
+    virtual PBoolean IsMatch(
+      const PASN_Choice & subTypePDU  ///<  sub-type PDU of H323Capability
+    ) const;
+  //@}
+
+  /**@name Operations */
+  //@{
+    /**Create the channel instance, allocating resources as required.
+       This creates a logical channel object appropriate for the parameters
+       provided. Not if param is NULL, sessionID must be provided, otherwise
+       this is taken from the fields in param.
+     */
+    virtual H323Channel * CreateChannel(
+      H323Connection & connection,    ///<  Owner connection for channel
+      H323Channel::Directions dir,    ///<  Direction of channel
+      unsigned sessionID,             ///<  Session ID for RTP channel
+      const H245_H2250LogicalChannelParameters * param
+                                      ///<  Parameters for channel
+    ) const;
+  //@}
+};
+
+
+///////////////////////////////////////////////////////////////////////////////
+
+class H323H239VideoCapability : public H323ExtendedVideoCapability
+{
+  PCLASSINFO(H323H239VideoCapability, H323ExtendedVideoCapability);
+
+  public:
+  /**@name Construction */
+  //@{
+    /**Create a new Extended Video capability.
+      */
+    H323H239VideoCapability(
+      const OpalMediaFormat & videoFormat
+    );
+  //@}
+
+  /**@name Overrides from class PObject */
+  //@{
+    /**Compare two capability instances. This compares the main and sub-types
+       of the capability.
+     */
+    Comparison Compare(const PObject & obj) const;
+
+    /**Create a copy of the object.
+      */
+    virtual PObject * Clone() const;
+  //@}
+
+  /**@name Identification functions */
+  //@{
+    /**Get the name of the media data format this class represents.
+     */
+    virtual PString GetFormatName() const;
+  //@}
+
+  /**@name Identification functions */
+  //@{
+    /**This function is called whenever and outgoing TerminalCapabilitySet
+       or OpenLogicalChannel PDU is being constructed for the control channel.
+       It allows the capability to set the PDU fields from information in
+       members specific to the class.
+
+       The default behaviour calls H323GenericCapabilityinfo::OnSendingPDU()
+       to handle the PDU.
+     */
+    virtual PBoolean OnSendingPDU(
+      H245_VideoCapability & pdu,  ///<  PDU to set information on
+      CommandType type             ///<  Type of PDU to send in
+    ) const;
+
+    /**This function is called whenever and incoming TerminalCapabilitySet
+       or OpenLogicalChannel PDU has been used to construct the control
+       channel. It allows the capability to set from the PDU fields,
+       information in members specific to the class.
+
+       The default behaviour calls H323GenericCapabilityinfo::OnReceivedPDU()
+       to handle the provided PDU.
+     */
+    virtual PBoolean OnReceivedPDU(
+      const H245_VideoCapability & pdu,  ///< PDU to get information from
+      CommandType type                   ///<  Type of PDU to send in
+    );
+  //@}
+};
+
+
+///////////////////////////////////////////////////////////////////////////////
+
+class H323H239ControlCapability : public H323GenericControlCapability
+{
+  PCLASSINFO(H323H239ControlCapability, H323GenericControlCapability);
+
+  public:
+  /**@name Construction */
+  //@{
+    /**Create a new Extended Video capability.
+      */
+    H323H239ControlCapability();
+  //@}
+
+  /**@name Overrides from class PObject */
+  //@{
+    /**Create a copy of the object.
+      */
+    virtual PObject * Clone() const;
+  //@}
+
+  /**@name Identification functions */
+  //@{
+    /**Get the name of the media data format this class represents.
+     */
+    virtual PString GetFormatName() const;
+  //@}
+};
+
+
+#endif  // OPAL_H239
+
 #endif  // OPAL_VIDEO
 
+
 /**This class describes the interface to a data channel used to transfer data
    via the logical channels opened and managed by the H323 control channel.
 
@@ -1605,237 +1893,6 @@ class H323_G711Capability : public H323AudioCapability
     Speed    speed;
 };
 
-#if 0
-
-/**This class describes the G.728 codec capability.
- */
-class H323_G728Capability : public H323AudioCapability
-{
-  PCLASSINFO(H323_G728Capability, H323AudioCapability)
-
-  public:
-  /**@name Construction */
-  //@{
-    /**Create a new G.728 capability.
-     */
-    H323_G728Capability();
-  //@}
-
-  /**@name Overrides from class PObject */
-  //@{
-    /**Create a copy of the object.
-      */
-    virtual PObject * Clone() const;
-  //@}
-
-  /**@name Identification functions */
-  //@{
-    /**Get the sub-type of the capability. This is a code dependent on the
-       main type of the capability.
-     */
-    virtual unsigned GetSubType() const;
-
-    /**Get the name of the media data format this class represents.
-     */
-    virtual PString GetFormatName() const;
-  //@}
-};
-
-
-/**This class describes the G.729 codec capability.
- */
-class H323_G729Capability : public H323AudioCapability
-{
-  PCLASSINFO(H323_G729Capability, H323AudioCapability)
-
-  public:
-    /// Specific G.729 encoding algorithm.
-    enum Mode {
-      e_Normal,
-      e_AnnexA,
-      e_AnnexB,
-      e_AnnexA_AnnexB
-    };
-
-  /**@name Construction */
-  //@{
-    /**Create a new G.729 capability.
-     */
-    H323_G729Capability(
-      Mode mode 
-    );
-  //@}
-
-  /**@name Overrides from class PObject */
-  //@{
-    /**Create a copy of the object.
-      */
-    virtual PObject * Clone() const;
-  //@}
-
-  /**@name Identification functions */
-  //@{
-    /**Get the sub-type of the capability. This is a code dependent on the
-       main type of the capability.
-     */
-    virtual unsigned GetSubType() const;
-
-    /**Get the name of the media data format this class represents.
-     */
-    virtual PString GetFormatName() const;
-  //@}
-
-  protected:
-    Mode mode;
-};
-
-
-/**This class describes the G.723.1 codec capability.
- */
-class H323_G7231Capability : public H323AudioCapability
-{
-  PCLASSINFO(H323_G7231Capability, H323AudioCapability)
-
-  public:
-  /**@name Construction */
-  //@{
-    /**Create a new G.723.1 capability.
-     */
-    H323_G7231Capability(
-      PBoolean allowSIDFrames = PTrue   ///<  Allow SID frames in data stream.
-    );
-  //@}
-
-  /**@name Overrides from class PObject */
-  //@{
-    /**Create a copy of the object.
-      */
-    virtual PObject * Clone() const;
-  //@}
-
-  /**@name Identification functions */
-  //@{
-    /**Get the sub-type of the capability. This is a code dependent on the
-       main type of the capability.
-     */
-    virtual unsigned GetSubType() const;
-
-    /**Get the name of the media data format this class represents.
-     */
-    virtual PString GetFormatName() const;
-  //@}
-
-  /**@name Protocol manipulation */
-  //@{
-    /**This function is called whenever and outgoing TerminalCapabilitySet
-       or OpenLogicalChannel PDU is being constructed for the control channel.
-       It allows the capability to set the PDU fields from information in
-       members specific to the class.
-
-       The default behaviour sets the data rate field in the PDU.
-     */
-    virtual PBoolean OnSendingPDU(
-      H245_AudioCapability & pdu,  ///<  PDU to set information on
-      unsigned packetSize          ///<  Packet size to use in capability
-    ) const;
-
-    /**This function is called whenever and incoming TerminalCapabilitySet
-       or OpenLogicalChannel PDU has been used to construct the control
-       channel. It allows the capability to set from the PDU fields,
-       information in members specific to the class.
-
-       The default behaviour gets the data rate field from the PDU.
-     */
-    virtual PBoolean OnReceivedPDU(
-      const H245_AudioCapability & pdu,  ///<  PDU to get information from
-      unsigned & packetSize              ///<  Packet size to use in capability
-    );
-  //@}
-
-  protected:
-    PBoolean allowSIDFrames;
-};
-
-
-/**This class describes the GSM 06.10 codec capability.
- */
-class H323_GSM0610Capability : public H323AudioCapability
-{
-  PCLASSINFO(H323_GSM0610Capability, H323AudioCapability)
-
-  public:
-  /**@name Construction */
-  //@{
-    /**Create a new GSM 06.10 capability.
-     */
-    H323_GSM0610Capability();
-  //@}
-
-  /**@name Overrides from class PObject */
-  //@{
-    /**Create a copy of the object.
-      */
-    virtual PObject * Clone() const;
-  //@}
-
-  /**@name Identification functions */
-  //@{
-    /**Get the sub-type of the capability. This is a code dependent on the
-       main type of the capability.
-
-       This returns one of the four possible combinations of mode and speed
-       using the enum values of the protocol ASN H245_AudioCapability class.
-     */
-    virtual unsigned GetSubType() const;
-
-    /**Get the name of the media data format this class represents.
-     */
-    virtual PString GetFormatName() const;
-
-    /**Set the maximum size (in frames) of data that will be transmitted in a
-       single PDU.
-
-       This will also be the desired number that will be sent by most codec
-       implemetations.
-
-       The default behaviour sets the txFramesInPacket variable.
-     */
-    virtual void SetTxFramesInPacket(
-      unsigned frames   ///<  Number of frames per packet
-    );
-  //@}
-
-  /**@name Protocol manipulation */
-  //@{
-    /**This function is called whenever and outgoing TerminalCapabilitySet
-       or OpenLogicalChannel PDU is being constructed for the control channel.
-       It allows the capability to set the PDU fields from information in
-       members specific to the class.
-
-       The default behaviour sets the data rate field in the PDU.
-     */
-    virtual PBoolean OnSendingPDU(
-      H245_AudioCapability & pdu,  ///<  PDU to set information on
-      unsigned packetSize          ///<  Packet size to use in capability
-    ) const;
-
-    /**This function is called whenever and incoming TerminalCapabilitySet
-       or OpenLogicalChannel PDU has been used to construct the control
-       channel. It allows the capability to set from the PDU fields,
-       information in members specific to the class.
-
-       The default behaviour gets the data rate field from the PDU.
-     */
-    virtual PBoolean OnReceivedPDU(
-      const H245_AudioCapability & pdu,  ///<  PDU to get information from
-      unsigned & packetSize              ///<  Packet size to use in capability
-    );
-  //@}
-};
-
-#endif
-
-
 ///////////////////////////////////////////////////////////////////////////////
 
 /**This class describes the UserInput psuedo-channel.
@@ -2328,6 +2385,8 @@ class H323Capabilities : public PObject
   //@}
 
   protected:
+    void SetMediaPacketizations(const PStringSet & packetizations) { mediaPacketizations = packetizations; }
+    
     H323CapabilitiesList table;
     H323CapabilitiesSet  set;
     PStringSet           mediaPacketizations;
@@ -2342,7 +2401,10 @@ typedef PFactory<H323Capability, std::string> H323CapabilityFactory;
 
 #define H323_REGISTER_CAPABILITY(cls, capName)   static H323CapabilityFactory::Worker<cls> cls##Factory(capName, true); \
 
-#endif // __OPAL_H323CAPS_H
+
+#endif // OPAL_H323
+
+#endif // OPAL_H323_H323CAPS_H
 
 
 /////////////////////////////////////////////////////////////////////////////
