@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23729 $
+ * $Revision: 23728 $
  * $Author: rjongbloed $
- * $Date: 2009-10-29 23:46:33 +0000 (Thu, 29 Oct 2009) $
+ * $Date: 2009-10-29 18:42:22 -0500 (Thu, 29 Oct 2009) $
  */
 
 #ifndef OPAL_H323_H323CON_H
@@ -85,6 +85,7 @@ class H245_FlowControlCommand;
 class H245_MiscellaneousCommand;
 class H245_MiscellaneousIndication;
 class H245_JitterIndication;
+class H245_ArrayOf_GenericParameter;
 
 class H323SignalPDU;
 class H323ControlPDU;
@@ -268,6 +269,21 @@ class H323Connection : public OpalRTPConnection
       */
     virtual OpalMediaFormatList GetMediaFormats() const;
 
+    /**Get next available session ID for the media type.
+      */
+    virtual unsigned GetNextSessionID(
+      const OpalMediaType & mediaType,   ///< Media type of stream being opened
+      bool isSource                      ///< Stream is a source/sink
+    );
+
+#if OPAL_FAX
+    /**Switch to/from FAX mode.
+      */
+    virtual bool SwitchFaxMediaStreams(
+      bool enableFax  ///< Enable FAX or return to audio mode
+    );
+#endif
+
     /**Open source or sink media stream for session.
       */
     virtual OpalMediaStreamPtr OpenMediaStream(
@@ -592,10 +608,15 @@ class H323Connection : public OpalRTPConnection
      */
     virtual bool RetrieveConnection();
 
-    /**Return PTrue if the current connection is on hold.
-     * Simply calls IsCallOnHold() which is kept for backward compatibility.
+    /**Return true if the current connection is on hold.
+       The bool parameter indicates if we are testing if the remote system
+       has us on hold, or we have them on hold.
+
+       Simply calls IsCallOnHold() which is kept for backward compatibility.
      */
-    virtual PBoolean IsConnectionOnHold();
+    virtual bool IsConnectionOnHold(
+      bool fromRemote  ///< Test if remote has us on hold, or we have them
+    );
 
 #if OPAL_H450
 
@@ -1123,6 +1144,65 @@ class H323Connection : public OpalRTPConnection
       const H245_JitterIndication & pdu  ///<  Received PDU
     );
 
+#if OPAL_H239
+    /**Handle a H.239 message from remote.
+     */
+    virtual bool OnH239Message(
+      unsigned subMessage,
+      const H245_ArrayOf_GenericParameter & params
+    );
+
+    /**Handle a H.239 flow control request.
+       Default behaviour simply sends an acknowedge response.
+      */
+    virtual bool OnH239FlowControlRequest(
+      unsigned logicalChannel,
+      unsigned bitRate
+    );
+
+    /**Handle a H.239 flow control ack/reject response.
+       Default behaviour does nothing
+      */
+    virtual bool OnH239FlowControlResponse(
+      unsigned logicalChannel,
+      bool rejected
+    );
+
+    /**Handle a H.239 presentation token request.
+       Default behaviour simply sends an acknowedge response.
+      */
+    virtual bool OnH239PresentationRequest(
+      unsigned logicalChannel,
+      unsigned symmetryBreaking,
+      unsigned terminalLabel
+    );
+
+    /**Handle a H.239 presentation token ack/reject response.
+       Default behaviour simply sends a release command.
+      */
+    virtual bool OnH239PresentationResponse(
+      unsigned logicalChannel,
+      unsigned terminalLabel,
+      bool rejected
+    );
+
+    /**Handle a H.239 presentation token release command.
+       Default behaviour does nothing.
+      */
+    virtual bool OnH239PresentationRelease(
+      unsigned logicalChannel,
+      unsigned terminalLabel
+    );
+
+    /**Handle a H.239 presentation token owner indication.
+       Default behaviour does nothing.
+      */
+    virtual bool OnH239PresentationIndication(
+      unsigned logicalChannel,
+      unsigned terminalLabel
+    );
+#endif
+
     /**Error discriminator for the OnControlProtocolError() function.
       */
     enum ControlProtocolErrors {
@@ -1751,7 +1831,7 @@ class H323Connection : public OpalRTPConnection
        OPAL media stream model. It is maintained for backward compatibility
        with older applications only.
       */
-    virtual PBoolean P_DEPRECATED RequestModeChangeT38(
+    virtual PBoolean RequestModeChangeT38(
       const char * capabilityNames = "T.38\nT38FaxUDP"
     );
 
@@ -1923,10 +2003,30 @@ class H323Connection : public OpalRTPConnection
       unsigned seconds  ///<  max duration of call in seconds
     );
   //@}
-	
+    
+#if OPAL_H239
+    /**Get the local H.239 control capability.
+     */
+    bool GetLocalH239Control() const { return m_h239Control; }
+
+    /**Set the local H.239 control capability.
+     */
+    void SetLocalH239Control(
+      bool on   ///< H.239 control capability is to be sent to remote
+    ) { m_h239Control = on; }
+
+    /**Get the remote H.239 control capability.
+     */
+    bool GetRemoteH239Control() const;
+
+    /**Get the remote H.239 options.
+     */
+    OpalMediaFormatList GetRemoteH239Formats() const;
+#endif
+
     virtual PBoolean OnSendFeatureSet(unsigned, H225_FeatureSet &) const;
-	
-	virtual void OnReceiveFeatureSet(unsigned, const H225_FeatureSet &) const;
+    
+    virtual void OnReceiveFeatureSet(unsigned, const H225_FeatureSet &) const;
 
 #if OPAL_H460
     /** Get the connection FeatureSet
@@ -1945,12 +2045,6 @@ class H323Connection : public OpalRTPConnection
 
     virtual void OnMediaCommand(OpalMediaCommand & note, INT extra);
     
-    // H.245 - compliant handling of session IDs
-    // if no internal session ID is found, zero is returned
-    unsigned GetExternalSessionID(unsigned internalSessionID) const;
-    unsigned GetExternalSessionID(unsigned internalSessionID, const OpalMediaType & mediaType);
-    unsigned GetInternalSessionID(unsigned externalSessionID, const OpalMediaType & mediaType);
-
   protected:
     /**Internal function to check if call established.
        This checks all the criteria for establishing a call an initiating the
@@ -1967,6 +2061,7 @@ class H323Connection : public OpalRTPConnection
 
     H323EndPoint & endpoint;
 
+    H323TransportAddress m_remoteConnectAddress;
     int                  remoteCallWaiting; // Number of call's waiting at the remote endpoint
     PBoolean                 gatekeeperRouted;
     unsigned             distinctiveRing;
@@ -2046,14 +2141,6 @@ class H323Connection : public OpalRTPConnection
     H323LogicalChannelList fastStartChannels;
     OpalMediaStreamPtr     fastStartMediaStream;
     
-    // H.245 - compliant handling of session IDs
-    unsigned nextSessionID;
-    typedef std::map<unsigned, unsigned> InternalToExternalSessionIDMap;
-    typedef std::map<OpalMediaType, unsigned> MediaTypeToInternalSessionIDMap;
-    InternalToExternalSessionIDMap internalToExternalSessionIDMap;
-    MediaTypeToInternalSessionIDMap mediaTypeToInternalSessionIDMap;
-    PMutex sessionIDHandlingMutex;
-
 #if PTRACING
     static const char * GetConnectionStatesName(ConnectionStates s);
     friend ostream & operator<<(ostream & o, ConnectionStates s) { return o << GetConnectionStatesName(s); }
@@ -2070,13 +2157,17 @@ class H323Connection : public OpalRTPConnection
     H245NegRequestMode               * requestModeProcedure;
     H245NegRoundTripDelay            * roundTripDelayProcedure;
 
+#if OPAL_H239
+    bool m_h239Control;
+#endif
+
 #if OPAL_H450
-    H450xDispatcher                  * h450dispatcher;
-    H4502Handler                     * h4502handler;
-    H4504Handler                     * h4504handler;
-    H4506Handler                     * h4506handler;
-    H4507Handler                     * h4507handler;
-    H45011Handler                    * h45011handler;
+    H450xDispatcher * h450dispatcher;
+    H4502Handler    * h4502handler;
+    H4504Handler    * h4504handler;
+    H4506Handler    * h4506handler;
+    H4507Handler    * h4507handler;
+    H45011Handler   * h45011handler;
 #endif
 
 #if OPAL_H460
