@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21283 $
+ * $Revision: 23923 $
  * $Author: rjongbloed $
- * $Date: 2008-10-11 07:10:58 +0000 (Sat, 11 Oct 2008) $
+ * $Date: 2010-01-11 19:49:45 -0600 (Mon, 11 Jan 2010) $
  */
 
 #ifndef OPAL_H323_H323PDU_H
@@ -244,14 +244,10 @@ class H323SignalPDU : public H225_H323_UserInformation
       */
     void SetQ931Fields(
       const H323Connection & connection,
-      PBoolean insertPartyNumbers = PFalse,
-      unsigned plan = 1,
-      unsigned type = 0,
-      int presentation = -1,
-      int screening = -1
+      bool insertPartyNumbers = false
     );
 
-#ifdef H323_H460
+#ifdef OPAL_H460
     /** When sending the H460 message in the Setup PDU you have to ensure
         the ARQ is received first then add the Fields to the Setup PDU
     	so we require a call back
@@ -384,6 +380,23 @@ class H323ControlPDU : public H245_MultimediaSystemControlMessage
       unsigned type
     );
 
+    H245_GenericMessage & BuildGenericRequest(
+      const PString & identifier,
+      unsigned subMsgId
+    );
+    H245_GenericMessage & BuildGenericResponse(
+      const PString & identifier,
+      unsigned subMsgId
+    );
+    H245_GenericMessage & BuildGenericCommand(
+      const PString & identifier,
+      unsigned subMsgId
+    );
+    H245_GenericMessage & BuildGenericIndication(
+      const PString & identifier,
+      unsigned subMsgId
+    );
+
     H245_FunctionNotUnderstood & BuildFunctionNotUnderstood(
       const H323ControlPDU & pdu
     );
@@ -471,20 +484,32 @@ PString H323GetAliasAddressE164(const H225_ArrayOf_AliasAddress & aliases);
 
 H323Connection::CallEndReason H323TranslateToCallEndReason(
   Q931::CauseValues cause,
-  const H225_ReleaseCompleteReason & reason
+  unsigned reason
 );
 Q931::CauseValues H323TranslateFromCallEndReason(
-  const H323Connection & connection,
-  H225_ReleaseCompleteReason & rcReason
+  H323Connection::CallEndReason callEndReason,
+  H225_ReleaseCompleteReason & releaseCompleteReason
 );
 
 void H323GetApplicationInfo(OpalProductInfo & info, const H225_VendorIdentifier & vendor);
 
+void H323SetRTPPacketization(
+  H245_ArrayOf_RTPPayloadType & rtpPacketizations,
+  PINDEX & packetizationsCount,
+  const OpalMediaFormat & mediaFormat,
+  RTP_DataFrame::PayloadTypes payloadType
+);
 bool H323SetRTPPacketization(
   H245_RTPPayloadType & rtpPacketization,
   const OpalMediaFormat & mediaFormat,
   RTP_DataFrame::PayloadTypes payloadType
 );
+bool H323SetRTPPacketization(
+  H245_RTPPayloadType & rtpPacketization,
+  const PString & mediaPacketizationString,
+  const OpalMediaFormat & mediaFormat,
+  RTP_DataFrame::PayloadTypes payloadType
+);
 PString H323GetRTPPacketization(
   const H245_RTPPayloadType & rtpPacketization
 );
@@ -493,6 +518,59 @@ bool H323GetRTPPacketization(
   const H245_RTPPayloadType & rtpPacketization
 );
 
+PString H323GetCapabilityIdentifier(
+  const H245_CapabilityIdentifier & capId
+);
+bool H323SetCapabilityIdentifier(
+  const PString & str,
+  H245_CapabilityIdentifier & capId
+);
+
+const H245_ParameterValue * H323GetGenericParameter(
+  const H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal
+);
+bool H323GetGenericParameterBoolean(
+  const H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal
+);
+unsigned H323GetGenericParameterInteger(
+  const H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  unsigned defValue = 0,
+  H245_ParameterValue::Choices subType = H245_ParameterValue::e_unsignedMin
+);
+H245_ParameterValue * H323AddGenericParameter(
+  H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  bool reorder = true
+);
+void H323AddGenericParameterBoolean(
+  H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  bool value = true,
+  bool reorder = true
+);
+void H323AddGenericParameterInteger(
+  H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  unsigned value,
+  H245_ParameterValue::Choices subType = H245_ParameterValue::e_unsignedMin,
+  bool reorder = true
+);
+void H323AddGenericParameterString(
+  H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  const PString & value,
+  bool reorder = true
+);
+void H323AddGenericParameterOctets(
+  H245_ArrayOf_GenericParameter & params,
+  unsigned ordinal,
+  const PBYTEArray & value,
+  bool reorder = true
+);
+
 #if PTRACING
 void H323TraceDumpPDU(
   const char * proto,
