@@ -23,13 +23,13 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21004 $
+ * $Revision: 23923 $
  * $Author: rjongbloed $
- * $Date: 2008-09-16 07:08:56 +0000 (Tue, 16 Sep 2008) $
+ * $Date: 2010-01-11 19:49:45 -0600 (Mon, 11 Jan 2010) $
  */
 
-#ifndef __OPAL_Q931_H
-#define __OPAL_Q931_H
+#ifndef OPAL_H323_Q931_H
+#define OPAL_H323_Q931_H
 
 #ifdef P_USE_PRAGMA
 #pragma interface
@@ -316,10 +316,10 @@ class Q931 : public PObject
 
     void SetCallingPartyNumber(
       const PString & number, ///<  Number string
-      unsigned plan = 1,      ///<  1 = ISDN/Telephony numbering system
+      unsigned plan = 1,      ///<  1 = ISDN/Telephony numbering system, see Q.931 Table 4-11 for more
       unsigned type = 0,      ///<  0 = Unknown number type
-      int presentation = -1,  ///<  0 = presentation allowed, -1 = no octet3a
-      int screening = -1      ///<   0 = user provided, not screened
+      int presentation = -1,  ///<  0 = presentation allowed, 1 = presentation restricted, -1 = no octet3a
+      int screening = -1      ///<  0 = user provided, not screened, -1 = no octet3a
     );
     PBoolean GetCallingPartyNumber(
       PString & number,               ///<  Number string
@@ -414,7 +414,7 @@ class Q931 : public PObject
 };
 
 
-#endif // __OPAL_Q931_H
+#endif // OPAL_H323_Q931_H
 
 
 /////////////////////////////////////////////////////////////////////////////
