@@ -27,9 +27,9 @@
  *     US Joint Forces Command Joint Concept Development & Experimentation (J9)
  *     http://www.jfcom.mil/about/abt_j9.htm
  *
- * $Revision: 23491 $
+ * $Revision: 23953 $
  * $Author: rjongbloed $
- * $Date: 2009-09-18 02:17:05 +0000 (Fri, 18 Sep 2009) $
+ * $Date: 2010-01-20 02:08:46 -0600 (Wed, 20 Jan 2010) $
  */
 
 #ifndef OPAL_OPAL_CONNECTION_H
@@ -48,7 +48,12 @@
 #include <ptclib/dtmf.h>
 #include <ptlib/safecoll.h>
 #include <rtp/rtp.h>
+
+#if OPAL_HAS_IM
 #include <im/t140.h>
+#include <im/rfc4103.h>
+#include <im/im.h>
+#endif
 
 class OpalEndPoint;
 class OpalCall;
@@ -56,6 +61,7 @@ class OpalSilenceDetector;
 class OpalEchoCanceler;
 class OpalRFC2833Proto;
 class OpalRFC2833Info;
+class PURL;
 
 
 #define OPAL_OPT_AUTO_START           "AutoStart"             ///< String option for auto-started media types
@@ -65,16 +71,21 @@ class OpalRFC2833Info;
 #define OPAL_OPT_CALLING_PARTY_NAME   "Calling-Party-Name"    ///< String option to set outgoing local name
 #define OPAL_OPT_CALLING_PARTY_DOMAIN "Calling-Party-Domain"  ///< String option to set outgoing local host/address/domain
 #define OPAL_OPT_CALLING_DISPLAY_NAME "Calling-Display-Name"  ///< String option to set outgoing display name
+#define OPAL_OPT_PRESENTATION_BLOCK   "Presentation-Block"    ///< String option to block outgoing calling number presentation
 #define OPAL_OPT_ORIGINATOR_ADDRESS   "Originator-Address"    ///< String option for originator address used by IVR
 #define OPAL_OPT_INTERFACE            "Interface"             ///< String option to set the interface used for outgoing call
 #define OPAL_OPT_ENABLE_INBAND_DTMF   "EnableInbandDTMF"      ///< String option to enable in band DTMF detection
+#define OPAL_OPT_ENABLE_INBAND_DTMF   "EnableInbandDTMF"      ///< String option to enable in band DTMF detection/send
+#define OPAL_OPT_DETECT_INBAND_DTMF   "DetectInBandDTMF"      ///< String option to enable in band DTMF detection
+#define OPAL_OPT_SEND_INBAND_DTMF     "SendInBandDTMF"        ///< String option to enable in band DTMF send
 #define OPAL_OPT_DTMF_MULT            "dtmfmult"
 #define OPAL_OPT_DTMF_DIV             "dtmfdiv"
 #define OPAL_OPT_DISABLE_JITTER       "Disable-Jitter"        ///< String option to disable jitter buffer if "true"
 #define OPAL_OPT_MAX_JITTER           "Max-Jitter"            ///< String option to set maximum jitter in milliseconds
 #define OPAL_OPT_MIN_JITTER           "Min-Jitter"            ///< String option to set minimum jitter in milliseconds
 #define OPAL_OPT_RECORD_AUDIO         "Record-Audio"          ///< String option to start recording to a file for call
-#define OPAL_OPT_ALERTING_TYPE        "Alerting-Type"         ///< String option to set teh alerting type string for call
+#define OPAL_OPT_ALERTING_TYPE        "Alerting-Type"         ///< String option to set the alerting type string for call
+#define OPAL_OPT_REMOVE_CODEC         "Remove-Codec"          ///< String option to remove codecs for this call
 
 
 /*! \page pageOpalConnections Connection handling in the OPAL library
@@ -155,7 +166,6 @@ The different phases are.
  
   \li OpalConnection::AlertingPhase - the OpalConnection instance is aware the phone is ringing.
 
-
   \li OpalConnection::ConnectedPhase - both endpoints have accepted the call, there may or may
                             not be agreement on media atthis time. Note that in many systems
                             this constitutes the start of charging for a call.
@@ -332,41 +342,54 @@ class OpalConnection : public PSafeObject
        NOTE: if anything is added to this, you also need to add the field to
        the tables in connection.cxx and h323pdu.cxx.
       */
-    enum CallEndReason {
-      EndedByLocalUser,         /// Local endpoint application cleared call
-      EndedByNoAccept,          /// Local endpoint did not accept call OnIncomingCall()=PFalse
-      EndedByAnswerDenied,      /// Local endpoint declined to answer call
-      EndedByRemoteUser,        /// Remote endpoint application cleared call
-      EndedByRefusal,           /// Remote endpoint refused call
-      EndedByNoAnswer,          /// Remote endpoint did not answer in required time
-      EndedByCallerAbort,       /// Remote endpoint stopped calling
-      EndedByTransportFail,     /// Transport error cleared call
-      EndedByConnectFail,       /// Transport connection failed to establish call
-      EndedByGatekeeper,        /// Gatekeeper has cleared call
-      EndedByNoUser,            /// Call failed as could not find user (in GK)
-      EndedByNoBandwidth,       /// Call failed as could not get enough bandwidth
-      EndedByCapabilityExchange,/// Could not find common capabilities
-      EndedByCallForwarded,     /// Call was forwarded using FACILITY message
-      EndedBySecurityDenial,    /// Call failed a security check and was ended
-      EndedByLocalBusy,         /// Local endpoint busy
-      EndedByLocalCongestion,   /// Local endpoint congested
-      EndedByRemoteBusy,        /// Remote endpoint busy
-      EndedByRemoteCongestion,  /// Remote endpoint congested
-      EndedByUnreachable,       /// Could not reach the remote party
-      EndedByNoEndPoint,        /// The remote party is not running an endpoint
-      EndedByHostOffline,       /// The remote party host off line
-      EndedByTemporaryFailure,  /// The remote failed temporarily app may retry
-      EndedByQ931Cause,         /// The remote ended the call with unmapped Q.931 cause code
-      EndedByDurationLimit,     /// Call cleared due to an enforced duration limit
-      EndedByInvalidConferenceID, /// Call cleared due to invalid conference ID
-      EndedByNoDialTone,        /// Call cleared due to missing dial tone
-      EndedByNoRingBackTone,    /// Call cleared due to missing ringback tone
-      EndedByOutOfService,      /// Call cleared because the line is out of service, 
+    enum CallEndReasonCodes {
+      EndedByLocalUser,            /// Local endpoint application cleared call
+      EndedByNoAccept,             /// Local endpoint did not accept call OnIncomingCall()=PFalse
+      EndedByAnswerDenied,         /// Local endpoint declined to answer call
+      EndedByRemoteUser,           /// Remote endpoint application cleared call
+      EndedByRefusal,              /// Remote endpoint refused call
+      EndedByNoAnswer,             /// Remote endpoint did not answer in required time
+      EndedByCallerAbort,          /// Remote endpoint stopped calling
+      EndedByTransportFail,        /// Transport error cleared call
+      EndedByConnectFail,          /// Transport connection failed to establish call
+      EndedByGatekeeper,           /// Gatekeeper has cleared call
+      EndedByNoUser,               /// Call failed as could not find user (in GK)
+      EndedByNoBandwidth,          /// Call failed as could not get enough bandwidth
+      EndedByCapabilityExchange,   /// Could not find common capabilities
+      EndedByCallForwarded,        /// Call was forwarded using FACILITY message
+      EndedBySecurityDenial,       /// Call failed a security check and was ended
+      EndedByLocalBusy,            /// Local endpoint busy
+      EndedByLocalCongestion,      /// Local endpoint congested
+      EndedByRemoteBusy,           /// Remote endpoint busy
+      EndedByRemoteCongestion,     /// Remote endpoint congested
+      EndedByUnreachable,          /// Could not reach the remote party
+      EndedByNoEndPoint,           /// The remote party is not running an endpoint
+      EndedByHostOffline,          /// The remote party host off line
+      EndedByTemporaryFailure,     /// The remote failed temporarily app may retry
+      EndedByQ931Cause,            /// The remote ended the call with unmapped Q.931 cause code
+      EndedByDurationLimit,        /// Call cleared due to an enforced duration limit
+      EndedByInvalidConferenceID,  /// Call cleared due to invalid conference ID
+      EndedByNoDialTone,           /// Call cleared due to missing dial tone
+      EndedByNoRingBackTone,       /// Call cleared due to missing ringback tone
+      EndedByOutOfService,         /// Call cleared because the line is out of service, 
       EndedByAcceptingCallWaiting, /// Call cleared because another call is answered
       EndedByGkAdmissionFailed,    /// Call cleared because gatekeeper admission request failed.
-      NumCallEndReasons,
+      NumCallEndReasons
+    };
 
-      EndedWithQ931Code = 0x100  /// Q931 code specified in MS byte
+    struct CallEndReason {
+      CallEndReason(
+        CallEndReasonCodes reason = NumCallEndReasons,
+        unsigned cause = 0
+      ) : code(reason), q931(cause) { }
+      explicit CallEndReason(
+        long reason
+      ) : code((CallEndReasonCodes)reason), q931(0) { }
+
+      operator CallEndReasonCodes() const { return code; }
+
+      CallEndReasonCodes code:8; // Normalised OPAL code
+      unsigned           q931:8; // PSTN Interworking code, actually Q.850
     };
 
 #if PTRACING
@@ -420,6 +443,46 @@ class OpalConnection : public PSafeObject
 
     class StringOptions : public PStringToString 
     {
+      public:
+        // Make sure the key set in a StringOptions is caseless
+        PBoolean SetAt(const char * key, const PString & data)
+        {
+          return PStringToString::SetAt(PCaselessString(key), data);
+        }
+        PBoolean SetAt(const PString & key, const PString & data)
+        {
+          return PStringToString::SetAt(PCaselessString(key), data);
+        }
+        PBoolean SetAt(const PCaselessString & key, const PString & data)
+        {
+          return PStringToString::SetAt(key, data);
+        }
+
+        bool GetBoolean(
+          const char * key,
+          bool dflt = false
+        ) const;
+        bool GetBoolean(
+          const PString & key,
+          bool dflt = false
+        ) const { return GetBoolean((const char *)key, dflt); }
+
+        long GetInteger(
+          const char * key,
+          long dflt = 0
+        ) const;
+        long GetInteger(
+          const PString & key,
+          long dflt = 0
+        ) const { return GetInteger((const char *)key, dflt); }
+
+        /** Extract the parameters that start with "OPAL-XXX" from the URL and
+            insert into the string options dictionary. The parameters are
+            removed from the URL.
+          */
+        void ExtractFromURL(
+          PURL & url
+        );
     };
 
   /**@name Construction */
@@ -478,6 +541,7 @@ class OpalConnection : public PSafeObject
       AlertingPhase,        //!< The remote says there is a phone ringing, somewhere
       ConnectedPhase,       //!< There is agreement on having a call, usually means billing will apply
       EstablishedPhase,     //!< Media is flowing, control streams  are all operational
+      ForwardingPhase,      //!< Connection is in the process of being forwarded
       ReleasingPhase,       //!< Hangup packet has been sent/received, media and control not yet stopped
       ReleasedPhase,        //!< Media and control streams have been terminated
       NumPhases             //!< Number of available phases. Can be used to indicate an unknown phase
@@ -489,7 +553,15 @@ class OpalConnection : public PSafeObject
       */
     inline Phases GetPhase() const { return phase; }
 
-    /**Get the call clearand reason for this connection shutting down.
+    /**Set the phase of the connection.
+       Note that this is primarily for internal use and calling from user code
+       is likely to have very strange results.
+      */
+    void SetPhase(
+      Phases phaseToSet  ///< phaseToSet the phase to set
+    );
+
+    /**Get the reason for this connection shutting down.
        Note that this function is only generally useful in the
        H323EndPoint::OnConnectionCleared() function. This is due to the
        connection not being cleared before that, and the object not even
@@ -499,6 +571,15 @@ class OpalConnection : public PSafeObject
       */
     CallEndReason GetCallEndReason() const { return callEndReason; }
 
+    /**Get the reason for this connection shutting down as text.
+      */
+    static PString GetCallEndReasonText(CallEndReason reason);
+    PString GetCallEndReasonText() const { return GetCallEndReasonText(callEndReason); }
+
+    /**Get the reason for this connection shutting down as text.
+      */
+    static void SetCallEndReasonText(CallEndReasonCodes reasonCode, const PString & newText);
+
     /**Set the call clearance reason.
        An application should have no cause to use this function. It is present
        for the H323EndPoint::ClearCall() function to set the clearance reason.
@@ -528,12 +609,12 @@ class OpalConnection : public PSafeObject
     /**Get the Q.931 cause code (Q.850) that terminated this call.
        See Q931::CauseValues for common values.
      */
-    unsigned GetQ931Cause() const { return q931Cause; }
+    unsigned GetQ931Cause() const { return callEndReason.q931; }
 
     /**Set the outgoing Q.931 cause code (Q.850) that is sent for this call
        See Q931::CauseValues for common values.
      */
-    void SetQ931Cause(unsigned v) { q931Cause = v; }
+    void SetQ931Cause(unsigned v) { callEndReason.q931 = v; }
 
     /**Initiate the transfer of an existing call (connection) to a new remote 
        party.
@@ -554,9 +635,13 @@ class OpalConnection : public PSafeObject
      */
     virtual bool RetrieveConnection();
 
-    /**Return PTrue if the current connection is on hold.
+    /**Return true if the current connection is on hold.
+       The bool parameter indicates if we are testing if the remote system
+       has us on hold, or we have them on hold.
      */
-    virtual PBoolean IsConnectionOnHold();
+    virtual bool IsConnectionOnHold(
+      bool fromRemote  ///< Test if remote has us on hold, or we have them
+    );
 
     /**Call back indicating result of last hold/retrieve operation.
        This also indicates if the local connection has been put on hold by the
@@ -596,8 +681,6 @@ class OpalConnection : public PSafeObject
        descendants of OpalConnection - code that uses existing descendants will be unaffected
      */
     virtual PBoolean OnIncomingConnection(unsigned int options, OpalConnection::StringOptions * stringOptions);
-    virtual PBoolean OnIncomingConnection(unsigned int options);
-    virtual PBoolean OnIncomingConnection();
 
     /**Start an outgoing connection.
        This function will initiate the connection to the remote entity, for
@@ -819,12 +902,62 @@ class OpalConnection : public PSafeObject
        that an application may remove or reorder the media formats before they
        are used to open media streams.
 
+       This function may also be executed to adjust the media format list for
+       other connections in the call. If this happens then the "otherConnection"
+       parameter will be non-NULL.
+
        The default behaviour calls the OpalEndPoint function of the same name.
       */
     virtual void AdjustMediaFormats(
-      OpalMediaFormatList & mediaFormats  ///<  Media formats to use
+      bool local,                          ///<  Media formats a local ones to be presented to remote
+      OpalMediaFormatList & mediaFormats,  ///<  Media formats to use
+      OpalConnection * otherConnection     ///<  Other connection we are adjusting media for
     ) const;
-    
+
+    /**Get next available session ID for the media type.
+
+       Default behaviour returns zero indicating that this connection type
+       does not care what the session ID is, and the other connection in the
+       call should be asked. If neither care, then teh default of the media
+       type is used.
+      */
+    virtual unsigned GetNextSessionID(
+      const OpalMediaType & mediaType,   ///< Media type of stream being opened
+      bool isSource                      ///< Stream is a source/sink
+    );
+
+    /** Indicate whether a particular media type can auto-start.
+        This is typically used for things like video or fax to contol if on
+        initial connection, that media type is opened straight away. Streams
+        of that type may be opened later, during the call, by using the
+        OpalCall::OpenSourceMediaStreams() function.
+      */
+    virtual OpalMediaType::AutoStartMode GetAutoStart(
+      const OpalMediaType & mediaType  ///< media type to check
+    ) const;
+
+    /**Open source media streams, if needed.
+     */
+    virtual void AutoStartMediaStreams(
+      bool force = false ///< Force re-open even if already open
+    );
+
+#if OPAL_FAX
+    /**Switch to/from FAX mode.
+      */
+    virtual bool SwitchFaxMediaStreams(
+      bool enableFax  ///< Enable FAX or return to audio mode
+    );
+
+    /**Indicate status of switch to/from FAX mode.
+
+       Default behaviour does nothing.
+      */
+    virtual void OnSwitchedFaxMediaStreams(
+      bool enabledFax  ///< Enabled FAX or audio mode
+    );
+#endif
+
     /**Open source or sink media stream for session.
       */
     virtual OpalMediaStreamPtr OpenMediaStream(
@@ -893,6 +1026,9 @@ class OpalConnection : public PSafeObject
 
     /**Get a media stream.
        Locates a stream given an identifier string.
+
+       If streamID is an empty string, then the first source/sink stream
+       of any type, session or id is returned.
       */
     OpalMediaStreamPtr GetMediaStream(
       const PString & streamID,  ///<  Stream ID to search for.
@@ -910,13 +1046,19 @@ class OpalConnection : public PSafeObject
     ) const;
 
     /**Get a media stream.
-       Locates a stream given a media type. Each session would usually
-       have two media streams associated with it, so the source flag
-       may be used to distinguish which channel to return.
+       Locates a stream given a media type. The source flag may be used to
+       distinguish which stream durection to return.
+
+       If mediaType is empty (i.e. OpalMediaType()), then the first
+       source/sink stream of any type, session or id is returned.
+
+       The previous parameter may be used to enumerate multiple stream of the
+       same type and direction. If NULL then the first stream is returned.
       */
     OpalMediaStreamPtr GetMediaStream(
-      const OpalMediaType & mediaType,  ///<  Media type to search for.
-      bool source          ///<  Indicates the direction of stream.
+      const OpalMediaType & mediaType,    ///<  Media type to search for.
+      bool source,                        ///<  Indicates the direction of stream.
+      OpalMediaStreamPtr previous = NULL  ///< Previous stream to start search from
     ) const;
 
     /**Call back when opening a media stream.
@@ -944,13 +1086,34 @@ class OpalConnection : public PSafeObject
     
     /**Call back when patching a media stream.
        This function is called when a connection has created a new media
-       patch between two streams.
+       patch between two streams. This is usually called twice per media patch,
+       once for the source stream and once for the sink stream.
+
+       Note this is not called within the context of the patch thread and is
+       called before that thread has started.
       */
     virtual void OnPatchMediaStream(
-      PBoolean isSource,
+      PBoolean isSource,        ///< Is source/sink call
       OpalMediaPatch & patch    ///<  New patch
     );
-	
+
+    /**Call back when media stream patch thread starts.
+      */
+    virtual void OnStartMediaPatch(
+      OpalMediaPatch & patch    ///< Patch being started
+    );
+
+    /**Call back when media stream patch thread stops.
+      */
+    virtual void OnStopMediaPatch(
+      OpalMediaPatch & patch    ///< Patch being stopped
+    );
+
+    /** Notifier function for OpalVideoUpdatePicture.
+        Calls the SendIntraFrameRequest on the rtp session
+      */
+    PDECLARE_NOTIFIER(OpalMediaCommand, OpalConnection, OnMediaCommand);
+
     /**Attaches the RFC 2833 handler to the media patch
        This method may be called from subclasses, e.g. within
        OnPatchMediaStream()
@@ -967,15 +1130,6 @@ class OpalConnection : public PSafeObject
     ) const;
 
 #if OPAL_VIDEO
-
-    /**Add video media formats available on a connection.
-
-       The default behaviour calls the OpalEndPoint function of the same name.
-      */
-    virtual void AddVideoMediaFormats(
-      OpalMediaFormatList & mediaFormats  ///<  Media formats to use
-    ) const;
-
     /**Create an PVideoInputDevice for a source media stream.
       */
     virtual PBoolean CreateVideoInputDevice(
@@ -1242,6 +1396,13 @@ class OpalConnection : public PSafeObject
       */
     void SetDisplayName(const PString & name) { displayName = name; }
 
+    /**Determine if remote presentation of Caller-ID is to be blocked.
+       Applies to an outgoing call to a "network" based endpoint type.
+       Corresponds to Q.931 Calling-Party-Number Information Elementin H.323.
+       Will remove display name for SIP.
+      */
+    virtual bool IsPresentationBlocked() const;
+
     /**Get the remote party display name.
       */
     const PString & GetRemotePartyName() const { return remotePartyName; }
@@ -1261,7 +1422,7 @@ class OpalConnection : public PSafeObject
     const PString & GetRemotePartyNumber() const { return remotePartyNumber; }
 
     /**Get the remote party address.
-       This is typically a URL like sip:user@hostname, though it may be just a host
+       This is typically a URL like sip:user\@hostname, though it may be just a host
        address. It should not be used as a "call back" address, use the
        GetRemotePartyURL() function for that purpose.
       */
@@ -1377,19 +1538,21 @@ class OpalConnection : public PSafeObject
      */
     OpalSilenceDetector * GetSilenceDetector() const { return silenceDetector; }
     
+#if OPAL_AEC
     /**Get the echo canceler active on connection.
     */
     OpalEchoCanceler * GetEchoCanceler() const { return echoCanceler; }
+#endif
 
     /**Get the protocol-specific unique identifier for this connection.
        Default behaviour just returns the connection token.
      */
     virtual PString GetIdentifier() const;
 
-    /**Get the maximum RTP payload size. This function allows a user to
-       override the value returned on a connection by connection basis, for
-       example knowing the connection is on a LAN with ethernet MTU the
-       payload size could be increased.
+    /**Get the maximum transmitted RTP payload size.
+       This function allows a user to override the value returned on a
+       connection by connection basis, for example knowing the connection is
+       on a LAN with ethernet MTU the payload size could be increased.
 
        Defaults to the value returned by the OpalManager function of the same
        name.
@@ -1422,59 +1585,70 @@ class OpalConnection : public PSafeObject
     virtual void ApplyStringOptions(OpalConnection::StringOptions & stringOptions);
     virtual void OnApplyStringOptions();
 
-    virtual void PreviewPeerMediaFormats(const OpalMediaFormatList & fmts);
+#if OPAL_HAS_MIXER
 
     virtual void EnableRecording();
     virtual void DisableRecording();
 
-    virtual void OnMediaPatchStart(
-      unsigned sessionId, 
-          bool isSource
-    );
-    virtual void OnMediaPatchStop(
-      unsigned sessionId, 
-          bool isSource
-    );
-
-    /** Notifier function for OpalVideoUpdatePicture.
-        Calls the SendIntraFrameRequest on the rtp session
-      */
-    PDECLARE_NOTIFIER(OpalMediaCommand, OpalConnection, OnMediaCommand);
-
-    /** Indicate whether a particular media type can auto-start.
-        This is typically used for things like video or fax to contol if on
-        initial connection, that media type is opened straight away. Streams
-        of that type may be opened later, during the call, by using the
-        OpalCall::OpenSourceMediaStreams() function.
-      */
-    virtual OpalMediaType::AutoStartMode GetAutoStart(
-      const OpalMediaType & mediaType  ///< media type to check
-    ) const;
+#endif
 
 #if OPAL_HAS_IM
-    struct IMInfo : public PObject {
-      unsigned sessionId;
-      OpalMediaFormat mediaFormat;
-      T140String body;
-    };
-
-    virtual bool SendIM(const OpalMediaFormat & format, const T140String & body);
-    virtual bool SendIM(const OpalMediaFormat & format, const RTP_DataFrame & body);
-    virtual void OnReceiveIM(const IMInfo & im);
-
-    void AddIMListener(
-      const PNotifier & listener
+    /**
+      * Called to transmit an IM to the other connection in the call
+      */
+    virtual bool TransmitInternalIM(
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & body
     );
+
+    /**
+      * Called when this connection receives an IM from the other connection in the call
+      */
+    virtual void OnReceiveInternalIM(
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & body
+    );
+
+    /**
+      * Called to transmit an IM to the other end of the connection
+      */
+    virtual bool TransmitExternalIM(
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & body
+    );
+
+    /**
+      * Called when this connection receives an IM from the other end of the connection
+      */
+    virtual bool OnReceiveExternalIM(
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & body
+    );
+
+    /**  Used for creating IM messages. 
+      *  0 = used for sending internal
+      *  1 = used for sending external
+      */
+    RFC4103Context & GetRFC4103Context(PINDEX i) { return m_rfc4103Context[i]; };
+
+  protected:
+    RFC4103Context m_rfc4103Context[2];
+
 #endif
 
   protected:
     void OnConnectedInternal();
 
-#if OPAL_PTLIB_DTMF
-    PDECLARE_NOTIFIER(RTP_DataFrame, OpalConnection, OnUserInputInBandDTMF);
-#endif
     PDECLARE_NOTIFIER(PThread, OpalConnection, OnReleaseThreadMain);
+
+#if OPAL_HAS_MIXER
     PDECLARE_NOTIFIER(RTP_DataFrame, OpalConnection, OnRecordAudio);
+#if OPAL_VIDEO
+    PDECLARE_NOTIFIER(RTP_DataFrame, OpalConnection, OnRecordVideo);
+#endif
+    void OnStartRecording(OpalMediaPatch * patch);
+    void OnStopRecording(OpalMediaPatch * patch);
+#endif
 
   // Member variables
     OpalCall             & ownerCall;
@@ -1506,13 +1680,11 @@ class OpalConnection : public PSafeObject
     SendUserInputModes    sendUserInputMode;
     PString               userInputString;
     PSyncPoint            userInputAvailable;
-    PBoolean              detectInBandDTMF;
-    unsigned              dtmfScaleMultiplier;
-    unsigned              dtmfScaleDivisor;
-    unsigned              q931Cause;
 
     OpalSilenceDetector * silenceDetector;
+#if OPAL_AEC
     OpalEchoCanceler    * echoCanceler;
+#endif
 
     MediaAddressesDict         mediaTransportAddresses;
     PSafeList<OpalMediaStream> mediaStreams;
@@ -1524,21 +1696,34 @@ class OpalConnection : public PSafeObject
     // The In-Band DTMF detector. This is used inside an audio filter which is
     // added to the audio channel.
 #if OPAL_PTLIB_DTMF
-    PDTMFDecoder        dtmfDecoder;
-#endif
+    PDTMFDecoder m_dtmfDecoder;
+    bool         m_detectInBandDTMF;
+    unsigned     m_dtmfScaleMultiplier;
+    unsigned     m_dtmfScaleDivisor;
+    PNotifier    m_dtmfNotifier;
+    PDECLARE_NOTIFIER(RTP_DataFrame, OpalConnection, OnDetectInBandDTMF);
 
-    /**Set the phase of the connection.
-       @param phaseToSet the phase to set
-      */
-    void SetPhase(Phases phaseToSet);
+    bool         m_sendInBandDTMF;
+    bool         m_installedInBandDTMF;
+    PDTMFEncoder m_inBandDTMF;
+    PINDEX       m_emittedInBandDTMF;
+    PMutex       m_inBandMutex;
+    PDECLARE_NOTIFIER(RTP_DataFrame, OpalConnection, OnSendInBandDTMF);
+#endif
 
 #if PTRACING
     friend ostream & operator<<(ostream & o, Phases p);
 #endif
 
     StringOptions m_connStringOptions;
-    PString recordAudioFilename;
-    PNotifier recordNotifier;
+
+#if OPAL_HAS_MIXER
+    PString       m_recordingFilename;
+    PNotifier     m_recordAudioNotifier;
+#if OPAL_VIDEO
+    PNotifier     m_recordVideoNotifier;
+#endif
+#endif
 
 #if OPAL_STATISTICS
     unsigned m_VideoUpdateRequestsSent;
@@ -1564,10 +1749,23 @@ class OpalConnection : public PSafeObject
     };
     AutoStartMap m_autoStartInfo;
 
-#if OPAL_HAS_IM
-    mutable PList<PNotifier> m_imListeners;
+#if OPAL_FAX
+    enum {
+      e_NotSwitchingFaxMediaStreams,
+      e_SwitchingToFaxMediaStreams,
+      e_SwitchingFromFaxMediaStreams
+    } m_faxMediaStreamsSwitchState;
 #endif
 
+  private:
+    P_REMOVE_VIRTUAL(PBoolean, OnIncomingConnection(unsigned int), false);
+    P_REMOVE_VIRTUAL(PBoolean, OnIncomingConnection(), false);
+    P_REMOVE_VIRTUAL(PBoolean, IsConnectionOnHold(), false);
+    P_REMOVE_VIRTUAL_VOID(OnMediaPatchStart(unsigned, bool));
+    P_REMOVE_VIRTUAL_VOID(OnMediaPatchStop(unsigned, bool));
+    P_REMOVE_VIRTUAL_VOID(AdjustMediaFormats(OpalMediaFormatList &) const);
+    P_REMOVE_VIRTUAL_VOID(AdjustMediaFormats(OpalMediaFormatList &, OpalConnection *) const);
+    P_REMOVE_VIRTUAL_VOID(PreviewPeerMediaFormats(const OpalMediaFormatList &));
 };
 
 #endif // OPAL_OPAL_CONNECTION_H
