@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22680 $
+ * $Revision: 24156 $
  * $Author: rjongbloed $
- * $Date: 2009-05-21 00:48:57 +0000 (Thu, 21 May 2009) $
+ * $Date: 2010-03-23 23:12:09 -0500 (Tue, 23 Mar 2010) $
  */
 
 #ifndef OPAL_OPAL_MEDIAFMT_H
@@ -45,9 +45,12 @@
 #include <opal/buildopts.h>
 
 #include <opal/mediatype.h>
-
 #include <rtp/rtp.h>
 
+#if OPAL_VIDEO
+#include <ptlib/videoio.h>
+#endif
+
 #include <limits>
 
 #ifdef min
@@ -117,15 +120,17 @@ class OpalMediaFormatList : public OpalMediaFormatBaseList
       const OpalMediaFormatList & formats    ///<  Formats to remove
     );
 
-    /**Get a format position in the list matching the payload type.
+    /**Get a format iterator in the list matching the payload type.
 
-       Returns P_MAX_INDEX if not in list.
+
+       Returns end() if not in list.
       */
     const_iterator FindFormat(
       RTP_DataFrame::PayloadTypes rtpPayloadType, ///<  RTP payload type code
       const unsigned clockRate,                   ///<  clock rate
       const char * rtpEncodingName = NULL,        ///<  RTP payload type name
-      const char * protocol = NULL                ///<  protocol to be valid for (if NULL, then all)
+      const char * protocol = NULL,               ///<  protocol to be valid for (if NULL, then all)
+      const_iterator start = const_iterator()     ///<  location to start search
     ) const;
 
     /**Get a position in the list of the first entry matching the wildcard.
@@ -258,6 +263,11 @@ class OpalMediaOption : public PObject
     virtual bool Merge(
       const OpalMediaOption & option
     );
+
+    virtual bool ValidateMerge(
+      const OpalMediaOption & option
+    ) const;
+
     virtual Comparison CompareValue(
       const OpalMediaOption & option
     ) const = 0;
@@ -353,7 +363,7 @@ class OpalMediaOptionValue : public OpalMediaOption
 
     virtual void ReadFrom(istream & strm)
     {
-      T temp;
+      T temp = 0;
       strm >> temp;
       if (temp >= m_minimum && temp <= m_maximum)
         m_value = temp;
@@ -446,6 +456,7 @@ class OpalMediaOptionEnum : public OpalMediaOption
     PINDEX GetValue() const { return m_value; }
     void SetValue(PINDEX value);
 
+    const PStringArray & GetEnumerations() const { return m_enumerations; }
     void SetEnumerations(const PStringArray & e)
     {
       m_enumerations = e;
@@ -550,6 +561,8 @@ class OpalMediaFormatInternal : public PObject
       time_t timeStamp
     );
 
+    const PCaselessString & GetName() const { return formatName; }
+
     virtual PObject * Clone() const;
     virtual void PrintOn(ostream & strm) const;
 
@@ -578,6 +591,9 @@ class OpalMediaFormatInternal : public PObject
     virtual bool ToNormalisedOptions();
     virtual bool ToCustomisedOptions();
     virtual bool Merge(const OpalMediaFormatInternal & mediaFormat);
+
+    virtual bool ValidateMerge(const OpalMediaFormatInternal & mediaFormat) const;
+
     virtual bool IsValidForProtocol(const PString & protocol) const;
 
   protected:
@@ -745,6 +761,12 @@ class OpalMediaFormat : public PContainer
       */
     bool ToCustomisedOptions();
 
+    /**Update media format information.
+      */
+    bool Update(
+      const OpalMediaFormat & mediaFormat
+    );
+
     /**Merge with another media format. This will alter and validate
        the options for this media format according to the merge rule for
        each option. The parameter is typically a "capability" while the
@@ -758,47 +780,52 @@ class OpalMediaFormat : public PContainer
     bool Merge(
       const OpalMediaFormat & mediaFormat
     );
+    /**Return true if the given mediaFormat will successfully merge.
+      */
+    bool ValidateMerge(
+      const OpalMediaFormat & mediaFormat
+    ) const;
 
     /**Get the name of the format
       */
-    PString GetName() const { PWaitAndSignal m(_mutex); return m_info == NULL ? "" : m_info->formatName; }
+    PString GetName() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? "" : m_info->formatName; }
 
     /**Return PTrue if media format info is valid. This may be used if the
        single string constructor is used to check that it matched something
        in the registered media formats database.
       */
-    PBoolean IsValid() const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->IsValid(); }
+    PBoolean IsValid() const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->IsValid(); }
 
     /**Return PTrue if media format info may be sent via RTP. Some formats are internal
        use only and are never transported "over the wire".
       */
-    PBoolean IsTransportable() const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->IsTransportable(); }
+    PBoolean IsTransportable() const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->IsTransportable(); }
 
     /**Get the RTP payload type that is to be used for this media format.
        This will either be an intrinsic one for the media format eg GSM or it
        will be automatically calculated as a dynamic media format that will be
        uniqueue amongst the registered media formats.
       */
-    RTP_DataFrame::PayloadTypes GetPayloadType() const { PWaitAndSignal m(_mutex); return m_info == NULL ? RTP_DataFrame::IllegalPayloadType : m_info->rtpPayloadType; }
-    void SetPayloadType(RTP_DataFrame::PayloadTypes type) { PWaitAndSignal m(_mutex); MakeUnique(); if (m_info != NULL) m_info->rtpPayloadType = type; }
+    RTP_DataFrame::PayloadTypes GetPayloadType() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? RTP_DataFrame::IllegalPayloadType : m_info->rtpPayloadType; }
+    void SetPayloadType(RTP_DataFrame::PayloadTypes type) { PWaitAndSignal m(m_mutex); MakeUnique(); if (m_info != NULL) m_info->rtpPayloadType = type; }
 
     /**Get the RTP encoding name that is to be used for this media format.
       */
-    const char * GetEncodingName() const { PWaitAndSignal m(_mutex); return m_info == NULL ? "" : m_info->rtpEncodingName.GetPointer(); }
+    const char * GetEncodingName() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? "" : m_info->rtpEncodingName.GetPointer(); }
 
     /** Get the media type for this format
       */
-    OpalMediaType GetMediaType() const { PWaitAndSignal m(_mutex); return m_info == NULL ? OpalMediaType() : m_info->mediaType; }
+    OpalMediaType GetMediaType() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? OpalMediaType() : m_info->mediaType; }
 
     /**Determine if the media format requires a jitter buffer. As a rule an
        audio codec needs a jitter buffer and all others do not.
       */
-    bool NeedsJitterBuffer() const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->GetOptionBoolean(NeedsJitterOption(), PFalse); }
+    bool NeedsJitterBuffer() const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->GetOptionBoolean(NeedsJitterOption(), PFalse); }
     static const PString & NeedsJitterOption();
 
     /**Get the average bandwidth used in bits/second.
       */
-    unsigned GetBandwidth() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(MaxBitRateOption(), 0); }
+    unsigned GetBandwidth() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(MaxBitRateOption(), 0); }
     static const PString & MaxBitRateOption();
     static const PString & TargetBitRateOption();
 
@@ -806,13 +833,13 @@ class OpalMediaFormat : public PContainer
        media format has no intrinsic maximum frame size, eg a video format
        would return zero but G.723.1 would return 24.
       */
-    PINDEX GetFrameSize() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(MaxFrameSizeOption(), 0); }
+    PINDEX GetFrameSize() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(MaxFrameSizeOption(), 0); }
     static const PString & MaxFrameSizeOption();
 
     /**Get the frame time in RTP timestamp units. If this returns zero then
        the media format is not real time and has no intrinsic timing eg T.120
       */
-    unsigned GetFrameTime() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(FrameTimeOption(), 0); }
+    unsigned GetFrameTime() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(FrameTimeOption(), 0); }
     static const PString & FrameTimeOption();
 
     /**Get the number of RTP timestamp units per millisecond.
@@ -826,7 +853,7 @@ class OpalMediaFormat : public PContainer
 
     /**Get the clock rate in Hz for this format.
       */
-    unsigned GetClockRate() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(ClockRateOption(), 1000); }
+    unsigned GetClockRate() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->GetOptionInteger(ClockRateOption(), 1000); }
     static const PString & ClockRateOption();
 
     /**Get the name of the OpalMediaOption indicating the protocol the format is being used on.
@@ -840,18 +867,18 @@ class OpalMediaFormat : public PContainer
     /**Get all of the option values of the format as a dictionary.
        Each entry is a name value pair.
       */
-    PStringToString GetOptions() const { PWaitAndSignal m(_mutex); return m_info == NULL ? PStringToString() : m_info->GetOptions(); }
+    PStringToString GetOptions() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? PStringToString() : m_info->GetOptions(); }
 
     /**Get the number of options this media format has.
       */
-    PINDEX GetOptionCount() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->options.GetSize(); }
+    PINDEX GetOptionCount() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->options.GetSize(); }
 
     /**Get the option instance at the specified index. This contains the
        description and value for the option.
       */
     const OpalMediaOption & GetOption(
       PINDEX index   ///<  Index of option in list to get
-    ) const { PWaitAndSignal m(_mutex); return m_info->options[index]; }
+    ) const { PWaitAndSignal m(m_mutex); return m_info->options[index]; }
 
     /**Get the option value of the specified name as a string.
 
@@ -860,7 +887,7 @@ class OpalMediaFormat : public PContainer
     bool GetOptionValue(
       const PString & name,   ///<  Option name
       PString & value         ///<  String to receive option value
-    ) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->GetOptionValue(name, value); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->GetOptionValue(name, value); }
 
     /**Set the option value of the specified name as a string.
        Note the option will not be added if it does not exist, the option
@@ -871,7 +898,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionValue(
       const PString & name,   ///<  Option name
       const PString & value   ///<  New option value as string
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionValue(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionValue(name, value); }
 
     /**Get the option value of the specified name as a boolean. The default
        value is returned if the option is not present.
@@ -879,7 +906,7 @@ class OpalMediaFormat : public PContainer
     bool GetOptionBoolean(
       const PString & name,   ///<  Option name
       bool dflt = PFalse       ///<  Default value if option not present
-    ) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->GetOptionBoolean(name, dflt); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->GetOptionBoolean(name, dflt); }
 
     /**Set the option value of the specified name as a boolean.
        Note the option will not be added if it does not exist, the option
@@ -890,7 +917,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionBoolean(
       const PString & name,   ///<  Option name
       bool value              ///<  New value for option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionBoolean(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionBoolean(name, value); }
 
     /**Get the option value of the specified name as an integer. The default
        value is returned if the option is not present.
@@ -898,7 +925,7 @@ class OpalMediaFormat : public PContainer
     int GetOptionInteger(
       const PString & name,   ///<  Option name
       int dflt = 0            ///<  Default value if option not present
-    ) const { PWaitAndSignal m(_mutex); return m_info == NULL ? dflt : m_info->GetOptionInteger(name, dflt); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info == NULL ? dflt : m_info->GetOptionInteger(name, dflt); }
 
     /**Set the option value of the specified name as an integer.
        Note the option will not be added if it does not exist, the option
@@ -910,7 +937,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionInteger(
       const PString & name,   ///<  Option name
       int value               ///<  New value for option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionInteger(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionInteger(name, value); }
 
     /**Get the option value of the specified name as a real. The default
        value is returned if the option is not present.
@@ -918,7 +945,7 @@ class OpalMediaFormat : public PContainer
     double GetOptionReal(
       const PString & name,   ///<  Option name
       double dflt = 0         ///<  Default value if option not present
-    ) const { PWaitAndSignal m(_mutex); return m_info == NULL ? dflt : m_info->GetOptionReal(name, dflt); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info == NULL ? dflt : m_info->GetOptionReal(name, dflt); }
 
     /**Set the option value of the specified name as a real.
        Note the option will not be added if it does not exist, the option
@@ -929,7 +956,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionReal(
       const PString & name,   ///<  Option name
       double value            ///<  New value for option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionReal(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionReal(name, value); }
 
     /**Get the option value of the specified name as an index into an
        enumeration list. The default value is returned if the option is not
@@ -938,7 +965,7 @@ class OpalMediaFormat : public PContainer
     PINDEX GetOptionEnum(
       const PString & name,   ///<  Option name
       PINDEX dflt = 0         ///<  Default value if option not present
-    ) const { PWaitAndSignal m(_mutex); return m_info == NULL ? dflt : m_info->GetOptionEnum(name, dflt); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info == NULL ? dflt : m_info->GetOptionEnum(name, dflt); }
 
     /**Set the option value of the specified name as an index into an enumeration.
        Note the option will not be added if it does not exist, the option
@@ -949,7 +976,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionEnum(
       const PString & name,   ///<  Option name
       PINDEX value            ///<  New value for option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionEnum(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionEnum(name, value); }
 
     /**Get the option value of the specified name as a string. The default
        value is returned if the option is not present.
@@ -957,7 +984,7 @@ class OpalMediaFormat : public PContainer
     PString GetOptionString(
       const PString & name,                   ///<  Option name
       const PString & dflt = PString::Empty() ///<  Default value if option not present
-    ) const { PWaitAndSignal m(_mutex); return m_info == NULL ? dflt : m_info->GetOptionString(name, dflt); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info == NULL ? dflt : m_info->GetOptionString(name, dflt); }
 
     /**Set the option value of the specified name as a string.
        Note the option will not be added if it does not exist, the option
@@ -968,7 +995,7 @@ class OpalMediaFormat : public PContainer
     bool SetOptionString(
       const PString & name,   ///<  Option name
       const PString & value   ///<  New value for option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionString(name, value); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionString(name, value); }
 
     /**Get the option value of the specified name as an octet array.
        Returns PFalse if not present.
@@ -976,7 +1003,7 @@ class OpalMediaFormat : public PContainer
     bool GetOptionOctets(
       const PString & name, ///<  Option name
       PBYTEArray & octets   ///<  Octets in option
-    ) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->GetOptionOctets(name, octets); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->GetOptionOctets(name, octets); }
 
     /**Set the option value of the specified name as an octet array.
        Note the option will not be added if it does not exist, the option
@@ -987,12 +1014,22 @@ class OpalMediaFormat : public PContainer
     bool SetOptionOctets(
       const PString & name,       ///<  Option name
       const PBYTEArray & octets   ///<  Octets in option
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionOctets(name, octets); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionOctets(name, octets); }
     bool SetOptionOctets(
       const PString & name,       ///<  Option name
       const BYTE * data,          ///<  Octets in option
       PINDEX length               ///<  Number of octets
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionOctets(name, data, length); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->SetOptionOctets(name, data, length); }
+
+    /**Create a media format list containing all of the registered media 
+	   formats matching the encoding name and clock rate, or the payload type.
+      */
+    static OpalMediaFormatList GetMatchingRegisteredMediaFormats(
+      RTP_DataFrame::PayloadTypes rtpPayloadType, ///<  RTP payload type code
+      const unsigned clockRate,                   ///<  clock rate
+      const char * rtpEncodingName = NULL,        ///<  RTP payload type name
+      const char * protocol = NULL                ///<  protocol to be valid for (if NULL, then all)
+    );
 
     /**Get a copy of the list of media formats that have been registered.
       */
@@ -1014,12 +1051,12 @@ class OpalMediaFormat : public PContainer
     bool AddOption(
       OpalMediaOption * option,
       PBoolean overwrite = PFalse
-    ) { PWaitAndSignal m(_mutex); MakeUnique(); return m_info != NULL && m_info->AddOption(option, overwrite); }
+    ) { PWaitAndSignal m(m_mutex); MakeUnique(); return m_info != NULL && m_info->AddOption(option, overwrite); }
 
     /**
       * Determine if media format has the specified option.
       */
-    bool HasOption(const PString & name) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->FindOption(name) != NULL; }
+    bool HasOption(const PString & name) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->FindOption(name) != NULL; }
 
     /**
       * Get a pointer to the specified media format option.
@@ -1027,35 +1064,43 @@ class OpalMediaFormat : public PContainer
       */
     OpalMediaOption * FindOption(
       const PString & name
-    ) const { PWaitAndSignal m(_mutex); return m_info == NULL ? NULL : m_info->FindOption(name); }
+    ) const { PWaitAndSignal m(m_mutex); return m_info == NULL ? NULL : m_info->FindOption(name); }
 
     /** Returns PTrue if the media format is valid for the protocol specified
         This allow plugin codecs to customise which protocols they are valid for
         The default implementation returns true unless the protocol is H.323
         and the rtpEncodingName is NULL
       */
-    bool IsValidForProtocol(const PString & protocol) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->IsValidForProtocol(protocol); }
+    bool IsValidForProtocol(const PString & protocol) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->IsValidForProtocol(protocol); }
 
-    time_t GetCodecVersionTime() const { PWaitAndSignal m(_mutex); return m_info == NULL ? 0 : m_info->codecVersionTime; }
+    time_t GetCodecVersionTime() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? 0 : m_info->codecVersionTime; }
 
     ostream & PrintOptions(ostream & strm) const
     {
-      PWaitAndSignal m(_mutex);
+      PWaitAndSignal m(m_mutex);
       if (m_info != NULL)
         strm << setw(-1) << *m_info;
       return strm;
     }
 
+#if OPAL_VIDEO
+    /** Adjust the video device arguments from the media format.
+     */
+    void AdjustVideoArgs(
+      PVideoDevice::OpenArgs & args  /// Videoo device arguments to adjust
+    ) const;
+#endif
+
     // Backward compatibility
-    virtual PBoolean IsEmpty() const { PWaitAndSignal m(_mutex); return m_info == NULL || !m_info->IsValid(); }
-    operator PString() const { PWaitAndSignal m(_mutex); return m_info == NULL ? "" : m_info->formatName; }
-    operator const char *() const { PWaitAndSignal m(_mutex); return m_info == NULL ? "" : m_info->formatName; }
-    bool operator==(const char * other) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->formatName == other; }
-    bool operator!=(const char * other) const { PWaitAndSignal m(_mutex); return m_info == NULL || m_info->formatName != other; }
-    bool operator==(const PString & other) const { PWaitAndSignal m(_mutex); return m_info != NULL && m_info->formatName == other; }
-    bool operator!=(const PString & other) const { PWaitAndSignal m(_mutex); return m_info == NULL || m_info->formatName != other; }
-    bool operator==(const OpalMediaFormat & other) const { PWaitAndSignal m(_mutex); return Compare(other) == EqualTo; }
-    bool operator!=(const OpalMediaFormat & other) const { PWaitAndSignal m(_mutex); return Compare(other) != EqualTo; }
+    virtual PBoolean IsEmpty() const { PWaitAndSignal m(m_mutex); return m_info == NULL || !m_info->IsValid(); }
+    operator PString() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? "" : m_info->formatName; }
+    operator const char *() const { PWaitAndSignal m(m_mutex); return m_info == NULL ? "" : m_info->formatName; }
+    bool operator==(const char * other) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->formatName == other; }
+    bool operator!=(const char * other) const { PWaitAndSignal m(m_mutex); return m_info == NULL || m_info->formatName != other; }
+    bool operator==(const PString & other) const { PWaitAndSignal m(m_mutex); return m_info != NULL && m_info->formatName == other; }
+    bool operator!=(const PString & other) const { PWaitAndSignal m(m_mutex); return m_info == NULL || m_info->formatName != other; }
+    bool operator==(const OpalMediaFormat & other) const { PWaitAndSignal m(m_mutex); return Compare(other) == EqualTo; }
+    bool operator!=(const OpalMediaFormat & other) const { PWaitAndSignal m(m_mutex); return Compare(other) != EqualTo; }
     friend bool operator==(const char * other, const OpalMediaFormat & fmt) { return fmt.m_info != NULL && fmt.m_info->formatName == other; }
     friend bool operator!=(const char * other, const OpalMediaFormat & fmt) { return fmt.m_info == NULL || fmt.m_info->formatName != other; }
     friend bool operator==(const PString & other, const OpalMediaFormat & fmt) { return fmt.m_info != NULL && fmt.m_info->formatName == other; }
@@ -1070,9 +1115,10 @@ class OpalMediaFormat : public PContainer
     PBoolean SetSize(PINDEX) { return PTrue; }
 
   protected:
-    PMutex _mutex;
     void Construct(OpalMediaFormatInternal * info);
+
     OpalMediaFormatInternal * m_info;
+    PMutex                    m_mutex;
 
   friend class OpalMediaFormatInternal;
   friend class OpalMediaFormatList;
@@ -1102,6 +1148,9 @@ class OpalAudioFormat : public OpalMediaFormat
 {
     PCLASSINFO(OpalAudioFormat, OpalMediaFormat);
   public:
+    OpalAudioFormat(
+      OpalMediaFormatInternal * info = NULL
+    ) : OpalMediaFormat(info) { }
     OpalAudioFormat(
       const char * fullName,    ///<  Full name of media format
       RTP_DataFrame::PayloadTypes rtpPayloadType, ///<  RTP payload type code
@@ -1144,6 +1193,9 @@ class OpalVideoFormat : public OpalMediaFormat
 {
     PCLASSINFO(OpalVideoFormat, OpalMediaFormat);
   public:
+    OpalVideoFormat(
+      OpalMediaFormatInternal * info = NULL
+    ) : OpalMediaFormat(info) { }
     OpalVideoFormat(
       const char * fullName,    ///<  Full name of media format
       RTP_DataFrame::PayloadTypes rtpPayloadType, ///<  RTP payload type code
@@ -1166,6 +1218,29 @@ class OpalVideoFormat : public OpalMediaFormat
     static const PString & RateControlEnableOption();
     static const PString & RateControllerOption();
     static const PString & RateControllerBitRateScalerOption();
+
+    /**The "role" of the content in the video stream based on this media
+       format. This is based on RFC4796 and H.239 semantics and is an
+       enumeration consisting of:
+          Value          H.239                RFC4796
+          NoRole         Not used             (not populated)
+          Presentation   Presentation (0x01)  slides
+          Main           Live (0x02)          main
+          Speaker        Live (0x02)          speaker
+          SignLanguage   Live (0x02)          sl
+      */
+    enum ContentRole {
+      eNoRole,
+      ePresentation,
+      eMainRole,
+      eSpeaker,
+      eSignLanguage,
+      eNumRoles
+    };
+    enum { ContentRoleMask = 15 };
+    __inline static unsigned ContentRoleBit(ContentRole contentRole) { return contentRole != eNoRole ? (1<<(contentRole-1)) : 0; }
+    static const PString & ContentRoleOption();
+    static const PString & ContentRoleMaskOption();
 };
 #endif
 
@@ -1203,7 +1278,8 @@ class OpalVideoFormat : public OpalMediaFormat
 #define OPAL_iLBC           "iLBC"
 #define OPAL_H261           "H.261"
 #define OPAL_H263           "H.263"
-#define OPAL_H264           "H.264"
+#define OPAL_H264_MODE0     "H.264-0"
+#define OPAL_H264_MODE1     "H.264-1"
 #define OPAL_MPEG4          "MPEG4"
 #define OPAL_RFC2833        "UserInput/RFC2833"
 #define OPAL_CISCONSE       "NamedSignalEvent"
@@ -1222,7 +1298,7 @@ extern const OpalAudioFormat & GetOpalG711_ULAW_64K();
 extern const OpalAudioFormat & GetOpalG711_ALAW_64K();
 extern const OpalAudioFormat & GetOpalG722();
 extern const OpalAudioFormat & GetOpalG7221();
-extern const OpalMediaFormat & GetOpalG7222();
+extern const OpalAudioFormat & GetOpalG7222();
 extern const OpalAudioFormat & GetOpalG726_40K();
 extern const OpalAudioFormat & GetOpalG726_32K();
 extern const OpalAudioFormat & GetOpalG726_24K();
@@ -1237,8 +1313,8 @@ extern const OpalAudioFormat & GetOpalG7231_5k3();
 extern const OpalAudioFormat & GetOpalG7231A_6k3();
 extern const OpalAudioFormat & GetOpalG7231A_5k3();
 extern const OpalAudioFormat & GetOpalGSM0610();
-extern const OpalMediaFormat & GetOpalGSMAMR();
-extern const OpalMediaFormat & GetOpaliLBC();
+extern const OpalAudioFormat & GetOpalGSMAMR();
+extern const OpalAudioFormat & GetOpaliLBC();
 
 extern const OpalMediaFormat & GetOpalRFC2833();
 
