@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23028 $
+ * $Revision: 23772 $
  * $Author: rjongbloed $
- * $Date: 2009-07-02 10:39:41 +0000 (Thu, 02 Jul 2009) $
+ * $Date: 2009-11-16 20:39:18 -0600 (Mon, 16 Nov 2009) $
  */
 
 #ifndef OPAL_OPAL_MEDIASTRM_H
@@ -161,12 +161,14 @@ class OpalMediaStream : public PSafeObject
     /**Callback that is called on the source stream once the media patch has started.
        The default behaviour calls OpalConnection::OnMediaPatchStart()
       */
-    virtual void OnPatchStart();
+    virtual void OnStartMediaPatch();
 
     /**Callback that is called on the source stream once the media patch has started.
        The default behaviour calls OpalConnection::OnMediaPatchStop()
       */
-    virtual void OnPatchStop();
+    virtual void OnStopMediaPatch(
+      OpalMediaPatch & patch    ///< Media patch that is stopping
+    );
 
     /**Write a list of RTP frames of data to the sink media stream.
        The default behaviour simply calls WritePacket() on each of the
@@ -229,7 +231,7 @@ class OpalMediaStream : public PSafeObject
       */
     virtual PBoolean SetDataSize(
       PINDEX dataSize,  ///< New data size (in total)
-      PINDEX frameSize  ///< Individual frame size (if applicable)
+      PINDEX frameTime  ///< Individual frame time (if applicable)
     );
 
     /**Get the data size in bytes that is expected to be used. Some media
@@ -246,16 +248,16 @@ class OpalMediaStream : public PSafeObject
     virtual PBoolean IsSynchronous() const = 0;
 
     /**Indicate if the media stream requires a OpalMediaPatch thread (active patch).
-       This is called on the source stream and is passed the sink stream that the
-       patch will initially be using. The function could conditionally require
-       the patch thread to execute a thread reading and writing data, or prevent
-       it from doing so as it can do so in hardware in some way, e.g. if both
-       streams where on the same OpalLineInterfaceDevice.
+       This is called on the source/sink stream and is passed the sink/source
+       stream that the patch will initially be using. The function could
+       conditionally require the patch thread to execute a thread reading and
+       writing data, or prevent  it from doing so as it can do so in hardware
+       in some way, e.g. both streams are on the same OpalLineInterfaceDevice.
 
        The default behaviour simply returns true.
       */
     virtual PBoolean RequiresPatchThread(
-      OpalMediaStream * sinkStream  ///< Sink stream for this source
+      OpalMediaStream * stream  ///< Other stream in patch
     ) const;
     virtual PBoolean RequiresPatchThread() const; // For backward compatibility
 
@@ -284,6 +286,10 @@ class OpalMediaStream : public PSafeObject
      */
     unsigned GetSessionID() const { return sessionID; }
 
+    /**Get the session number of the stream.
+     */
+    void SetSessionID(unsigned id) { sessionID = id; }
+
     /**  Get the ID associated with this stream. Used for detecting two 
       *  the streams associated with a bidirectional media channel
       */
@@ -366,6 +372,11 @@ class OpalMediaStream : public PSafeObject
 
     OpalMediaPatch * mediaPatch;
     PNotifier        commandNotifier;
+
+  private:
+    P_REMOVE_VIRTUAL_VOID(OnPatchStart());
+    P_REMOVE_VIRTUAL_VOID(OnPatchStop());
+    P_REMOVE_VIRTUAL_VOID(OnStopMediaPatch());
 };
 
 typedef PSafePtr<OpalMediaStream> OpalMediaStreamPtr;
@@ -413,6 +424,14 @@ class OpalNullMediaStream : public OpalMediaStream, public OpalMediaStreamPacing
       bool isSource,                       ///<  Is a source stream
       bool isSynchronous = false           ///<  Can accept data and block accordingly
     );
+    OpalNullMediaStream(
+      OpalConnection & conn,
+      const OpalMediaFormat & mediaFormat, ///<  Media format for stream
+      unsigned sessionID,                  ///<  Session number for stream
+      bool isSource,                       ///<  Is a source stream
+      bool usePacingDelay,                 ///<  Use delay to pace stream I/O
+      bool requiresPatchThread             ///<  Requires a patch thread to execute
+    );
   //@}
 
   /**@name Overrides of OpalMediaStream class */
@@ -436,18 +455,19 @@ class OpalNullMediaStream : public OpalMediaStream, public OpalMediaStreamPacing
     );
 	
     /**Indicate if the media stream requires a OpalMediaPatch thread (active patch).
-       The default behaviour returns false.
+       The default behaviour returns the value of m_isSynchronous.
       */
     virtual PBoolean RequiresPatchThread() const;
 
     /**Indicate if the media stream is synchronous.
-       Returns false.
+       Returns m_isSynchronous.
       */
     virtual PBoolean IsSynchronous() const;
   //@}
 
   protected:
-    bool           m_isSynchronous;
+    bool m_isSynchronous;
+    bool m_requiresPatchThread;
 };
 
 
@@ -492,6 +512,13 @@ class OpalRTPMediaStream : public OpalMediaStream
       */
     virtual PBoolean Close();
 
+    /**Set the paused state for stream.
+       This will stop reading/writing data from the stream.
+      */
+    virtual void SetPaused(
+      bool pause    ///< Indicate that the stream should be paused
+    );
+
     /**Read an RTP frame of data from the source media stream.
        The new behaviour simply calls RTP_Session::ReadData().
       */
@@ -510,7 +537,7 @@ class OpalRTPMediaStream : public OpalMediaStream
       */
     virtual PBoolean SetDataSize(
       PINDEX dataSize,  ///< New data size (in total)
-      PINDEX frameSize  ///< Individual frame size (if applicable)
+      PINDEX frameTime  ///< Individual frame time (if applicable)
     );
 
     /**Indicate if the media stream is synchronous.
@@ -518,12 +545,27 @@ class OpalRTPMediaStream : public OpalMediaStream
       */
     virtual PBoolean IsSynchronous() const;
 
+    /**Indicate if the media stream requires a OpalMediaPatch thread (active patch).
+       The default behaviour dermines if the media will be flowing between two
+       RTP sessions within the same process. If so the
+       OpalRTPConnection::OnLocalRTP() is called, and if it returns true
+       indicating local handling then this function returns faklse to disable
+       the patch thread.
+      */
+    virtual PBoolean RequiresPatchThread() const;
+
     /**Enable jitter buffer for the media stream.
 
        The default behaviour does nothing.
       */
     virtual void EnableJitterBuffer() const;
 
+    /**Set the patch thread that is using this stream.
+      */
+    virtual PBoolean SetPatch(
+      OpalMediaPatch * patch  ///<  Media patch thread
+    );
+
     /** Return current RTP session
       */
     virtual RTP_Session & GetRtpSession() const
@@ -704,6 +746,7 @@ class OpalAudioMediaStream : public OpalRawMediaStream
       unsigned sessionID,                  ///<  Session number for stream
       bool isSource,                       ///<  Is a source stream
       PINDEX buffers,                      ///<  Number of buffers on sound channel
+      unsigned bufferTime,                 ///<  Buffering time on sound channel (milliseconds)
       PSoundChannel * channel,             ///<  Audio device to stream to/from
       bool autoDelete = true               ///<  Automatically delete PSoundChannel
     );
@@ -716,6 +759,7 @@ class OpalAudioMediaStream : public OpalRawMediaStream
       unsigned sessionID,                  ///<  Session number for stream
       bool isSource,                       ///<  Is a source stream
       PINDEX buffers,                      ///<  Number of buffers on sound channel
+      unsigned bufferTime,                 ///<  Buffering time on sound channel (milliseconds)
       const PString & deviceName           ///<  Name of audio device to stream to/from
     );
   //@}
@@ -729,7 +773,7 @@ class OpalAudioMediaStream : public OpalRawMediaStream
       */
     virtual PBoolean SetDataSize(
       PINDEX dataSize,  ///< New data size (in total)
-      PINDEX frameSize  ///< Individual frame size (if applicable)
+      PINDEX frameTime  ///< Individual frame time (if applicable)
     );
 
     /**Indicate if the media stream is synchronous.
@@ -739,7 +783,8 @@ class OpalAudioMediaStream : public OpalRawMediaStream
   //@}
 
   protected:
-    PINDEX soundChannelBuffers;
+    PINDEX   m_soundChannelBuffers;
+    unsigned m_soundChannelBufferTime;
 };
 
 #endif // OPAL_PTLIB_AUDIO
@@ -820,7 +865,7 @@ class OpalVideoMediaStream : public OpalMediaStream
       */
     virtual PBoolean SetDataSize(
       PINDEX dataSize,  ///< New data size (in total)
-      PINDEX frameSize  ///< Individual frame size (if applicable)
+      PINDEX frameTime  ///< Individual frame time (if applicable)
     );
 
     /** Get the input device (e.g. for statistics)
@@ -867,6 +912,8 @@ class OpalUDPMediaStream : public OpalMediaStream
     );
   //@}
 
+    ~OpalUDPMediaStream();
+
   /**@name Overrides of OpalMediaStream class */
   //@{
 
