@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21936 $
+ * $Revision: 23741 $
  * $Author: rjongbloed $
- * $Date: 2009-01-17 08:02:24 +0000 (Sat, 17 Jan 2009) $
+ * $Date: 2009-11-01 20:27:28 -0600 (Sun, 01 Nov 2009) $
  */
 
 #ifndef OPAL_OPAL_PCSS_H
@@ -39,25 +39,22 @@
 
 #include <opal/buildopts.h>
 
-#ifndef OPAL_PTLIB_AUDIO
-#ifdef _MSC_VER
-#pragma message("PTLib soundcard support not available")
-#else
-#warning "PTLib soundcard support not available"
-#endif
-#else
+#if OPAL_HAS_PCSS
+
+#if OPAL_PTLIB_AUDIO
 
 #include <ptlib/sound.h>
-#include <opal/endpoint.h>
+#include <opal/localep.h>
+
 
 class OpalPCSSConnection;
 
 
 /** PC Sound System endpoint.
  */
-class OpalPCSSEndPoint : public OpalEndPoint
+class OpalPCSSEndPoint : public OpalLocalEndPoint
 {
-    PCLASSINFO(OpalPCSSEndPoint, OpalEndPoint);
+    PCLASSINFO(OpalPCSSEndPoint, OpalLocalEndPoint);
   public:
   /**@name Construction */
   //@{
@@ -104,7 +101,7 @@ class OpalPCSSEndPoint : public OpalEndPoint
 
        The default behaviour is pure.
      */
-    virtual PBoolean MakeConnection(
+    virtual PSafePtr<OpalConnection> MakeConnection(
       OpalCall & call,           ///<  Owner of connection
       const PString & party,     ///<  Remote party to call
       void * userData = NULL,    ///<  Arbitrary data to pass to connection
@@ -113,17 +110,42 @@ class OpalPCSSEndPoint : public OpalEndPoint
     );
   //@}
 
+  /**@name Overrides from OpalLocalEndPoint */
+  //@{
+    /**Call back to indicate that remote is ringing.
+       If false is returned the call is aborted.
+
+       The default implementation does nothing.
+      */
+    virtual bool OnOutgoingCall(
+      const OpalLocalConnection & connection ///<  Connection having event
+    );
+
+    /**Call back to indicate that remote is ringing.
+       If false is returned the call is aborted.
+
+       The default implementation answers the call immediately.
+      */
+    virtual bool OnIncomingCall(
+      OpalLocalConnection & connection ///<  Connection having event
+    );
+
+    /**Call back to indicate that the remote user has indicated something.
+       If false is returned the call is aborted.
+
+       The default implementation does nothing.
+      */
+    virtual bool OnUserInput(
+      const OpalLocalConnection & connection, ///<  Connection having event
+      const PString & indication
+    );
+  //@}
+
   /**@name Customisation call backs */
   //@{
     /**Create a connection for the PCSS endpoint.
        The default implementation is to create a OpalPCSSConnection.
       */
-    virtual OpalPCSSConnection * CreateConnection(
-      OpalCall & call,    ///<  Owner of connection
-      const PString & playDevice, ///<  Sound channel play device
-      const PString & recordDevice, ///<  Sound channel record device
-      void * userData     ///<  Arbitrary data to pass to connection
-    );
     virtual OpalPCSSConnection * CreateConnection(
       OpalCall & call,    ///<  Owner of connection
       const PString & playDevice, ///<  Sound channel play device
@@ -198,17 +220,6 @@ class OpalPCSSEndPoint : public OpalEndPoint
       const OpalPCSSConnection & connection, ///<  Connection having event
       const PString & indication
     );
-
-    
-    /**Call back when patching a media stream.
-       This function is called when a connection has created a new media
-       patch between two streams.
-      */
-    virtual void OnPatchMediaStream(
-      const OpalPCSSConnection & connection, ///<  Connection having new patch
-      PBoolean isSource,                         ///<  Source patch
-      OpalMediaPatch & patch                 ///<  New patch
-    );
   //@}
 
   /**@name Member variable access */
@@ -244,30 +255,50 @@ class OpalPCSSEndPoint : public OpalEndPoint
     const PString & GetSoundChannelRecordDevice() const { return soundChannelRecordDevice; }
 
     /**Get default the sound channel buffer depth.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
       */
     unsigned GetSoundChannelBufferDepth() const { return soundChannelBuffers; }
 
     /**Set the default sound channel buffer depth.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
       */
     void SetSoundChannelBufferDepth(
       unsigned depth    ///<  New depth
     );
-  //@}
 
-    OpalMediaFormatList GetMediaFormats() const;
+    /**Get default the sound channel buffer time in milliseconds.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
+      */
+    unsigned GetSoundChannelBufferTime() const { return m_soundChannelBufferTime; }
+
+    /**Set the default sound channel buffer time in milliseconds.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
+      */
+    void SetSoundChannelBufferTime(
+      unsigned depth    ///<  New depth in milliseconds
+    );
+  //@}
 
   protected:
     PString  soundChannelPlayDevice;
     PString  soundChannelRecordDevice;
     unsigned soundChannelBuffers;
+    unsigned m_soundChannelBufferTime;
+
+  private:
+    P_REMOVE_VIRTUAL(OpalPCSSConnection *, CreateConnection(OpalCall &, const PString &, const PString &, void *), 0)
 };
 
 
 /** PC Sound System connection.
  */
-class OpalPCSSConnection : public OpalConnection
+class OpalPCSSConnection : public OpalLocalConnection
 {
-    PCLASSINFO(OpalPCSSConnection, OpalConnection);
+    PCLASSINFO(OpalPCSSConnection, OpalLocalConnection);
   public:
   /**@name Construction */
   //@{
@@ -289,39 +320,6 @@ class OpalPCSSConnection : public OpalConnection
 
   /**@name Overrides from OpalConnection */
   //@{
-    /**Get indication of connection being to a "network".
-       This indicates the if the connection may be regarded as a "network"
-       connection. The distinction is about if there is a concept of a "remote"
-       party being connected to and is best described by example: sip, h323,
-       iax and pstn are all "network" connections as they connect to something
-       "remote". While pc, pots and ivr are not as the entity being connected
-       to is intrinsically local.
-      */
-    virtual bool IsNetworkConnection() const { return false; }
-
-    /**Start an outgoing connection.
-       This function will initiate the connection to the remote entity, for
-       example in H.323 it sends a SETUP, in SIP it sends an INVITE etc.
-
-       The default behaviour does.
-      */
-    virtual PBoolean SetUpConnection();
-
-    /**Indicate to remote endpoint an alert is in progress.
-       If this is an incoming connection and the AnswerCallResponse is in a
-       AnswerCallDeferred or AnswerCallPending state, then this function is
-       used to indicate to that endpoint that an alert is in progress. This is
-       usually due to another connection which is in the call (the B party)
-       has received an OnAlerting() indicating that its remote endpoint is
-       "ringing".
-
-       The default behaviour does nothing.
-      */
-    virtual PBoolean SetAlerting(
-      const PString & calleeName,   ///<  Name of endpoint being alerted.
-      PBoolean withMedia                ///<  Open media with alerting
-    );
-
     /**Initiate the transfer of an existing call (connection) to a new remote 
        party.
 
@@ -352,26 +350,6 @@ class OpalPCSSConnection : public OpalConnection
       PBoolean isSource                        ///<  Is a source stream
     );
 
-    /**Call back when patching a media stream.
-       This function is called when a connection has created a new media
-       patch between two streams.
-       Add the echo canceler patch and call the endpoint function of
-       the same name.
-       Add a PCM silence detector filter.
-      */
-    virtual void OnPatchMediaStream(
-      PBoolean isSource,
-      OpalMediaPatch & patch    ///<  New patch
-    );
-
-    /**Open source or sink media stream for session.
-      */
-    virtual OpalMediaStreamPtr OpenMediaStream(
-      const OpalMediaFormat & mediaFormat, ///<  Media format to open
-      unsigned sessionID,                  ///<  Session to start stream on
-      bool isSource                        ///< Stream is a source/sink
-    );
-
     /**Set  the volume (gain) for the audio media channel to the specified percentage.
       */
     virtual PBoolean SetAudioVolume(
@@ -385,25 +363,10 @@ class OpalPCSSConnection : public OpalConnection
     virtual unsigned GetAudioSignalLevel(
       PBoolean source                   ///< true for source (microphone), false for sink (speaker)
     );
-
-    /**Send a user input indication to the remote endpoint.
-       This sends an arbitrary string as a user indication. If DTMF tones in
-       particular are required to be sent then the SendIndicationTone()
-       function should be used.
-
-       The default behaviour plays the DTMF tones on the line.
-      */
-    virtual PBoolean SendUserInputString(
-      const PString & value                   ///<  String value of indication
-    );
   //@}
 
   /**@name New operations */
   //@{
-    /**Accept the incoming connection.
-      */
-    virtual void AcceptIncoming();
-
     /**Create an PSoundChannel based media stream.
       */
     virtual PSoundChannel * CreateSoundChannel(
@@ -427,8 +390,16 @@ class OpalPCSSConnection : public OpalConnection
     const PString & GetSoundChannelRecordDevice() const { return soundChannelRecordDevice; }
 
     /**Get default the sound channel buffer depth.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
       */
     unsigned GetSoundChannelBufferDepth() const { return soundChannelBuffers; }
+
+    /**Get default the sound channel buffer time in milliseconds.
+       Note the largest of the depth in frames and the depth in milliseconds
+       as returned by GetSoundBufferTime() is used.
+      */
+    unsigned GetSoundChannelBufferTime() const { return m_soundChannelBufferTime; }
   //@}
 
 
@@ -437,10 +408,21 @@ class OpalPCSSConnection : public OpalConnection
     PString            soundChannelPlayDevice;
     PString            soundChannelRecordDevice;
     unsigned           soundChannelBuffers;
+    unsigned           m_soundChannelBufferTime;
 };
 
+#else
+
+#ifdef _MSC_VER
+#pragma message("PTLib soundcard support not available")
+#else
+#warning "PTLib soundcard support not available"
+#endif
+
 #endif // OPAL_PTLIB_AUDIO
 
+#endif // OPAL_HAS_PCSS
+
 #endif // OPAL_OPAL_PCSS_H
 
 
