@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21811 $
+ * $Revision: 24148 $
  * $Author: rjongbloed $
- * $Date: 2008-12-16 10:06:55 +0000 (Tue, 16 Dec 2008) $
+ * $Date: 2010-03-23 03:29:03 -0500 (Tue, 23 Mar 2010) $
  */
 
 #ifndef OPAL_OPAL_RTPCONN_H
@@ -113,7 +113,12 @@ class OpalRTPMediaSession : public OpalMediaSession
 {
   PCLASSINFO(OpalRTPMediaSession, OpalMediaSession);
   public:
-    OpalRTPMediaSession(OpalConnection & conn, const OpalMediaType & mediaType, unsigned sessionId);
+    OpalRTPMediaSession(
+      OpalConnection & conn,
+      const OpalMediaType & mediaType,
+      unsigned sessionId,
+      RTP_Session * rtpSession
+    );
     OpalRTPMediaSession(const OpalRTPMediaSession & obj);
 
     PObject * Clone() const { return new OpalRTPMediaSession(*this); }
@@ -124,7 +129,7 @@ class OpalRTPMediaSession : public OpalMediaSession
 
     virtual bool IsRTP() const { return true; }
 
-    virtual bool HasFailed() const { return rtpSession != NULL && rtpSession->HasFailed(); }
+    virtual bool HasFailed() const { return (rtpSession != NULL) && (rtpSession->HasFailed() || (rtpSession->GetPacketsReceived() == 0)); }
 
     virtual OpalTransportAddress GetLocalMediaAddress() const;
 
@@ -161,6 +166,10 @@ class OpalRTPSessionManager : public PObject
 
   /**@name Operations */
   //@{
+    /**Get next available session ID for the media type.
+      */
+    unsigned GetNextSessionID();
+
     /**Add an RTP session for the specified ID.
 
        This function MUST be called only after the UseSession() function has
@@ -191,6 +200,16 @@ class OpalRTPSessionManager : public PObject
     OpalMediaSession * GetMediaSession(
       unsigned sessionID
     ) const;
+
+    /**Change the sessionID for an existing session.
+       This will adjust the RTP session and media streams.
+
+       Return false if no such session exists.
+      */
+    bool ChangeSessionID(
+      unsigned fromSessionID,   ///< Session ID to search for
+      unsigned toSessionID      ///< Session ID to change to
+    );
   //@}
 
     PMutex & GetMutex() { return m_mutex; }
@@ -233,6 +252,13 @@ class OpalRTPConnection : public OpalConnection
 
   /**@name RTP Session Management */
   //@{
+    /**Get next available session ID for the media type.
+      */
+    virtual unsigned GetNextSessionID(
+      const OpalMediaType & mediaType,   ///< Media type of stream being opened
+      bool isSource                      ///< Stream is a source/sink
+    );
+
     /**Get an RTP session for the specified ID.
        If there is no session of the specified ID, NULL is returned.
       */
@@ -272,11 +298,30 @@ class OpalRTPConnection : public OpalConnection
     virtual RTP_Session * CreateSession(
       const OpalTransport & transport,
       unsigned sessionID,
+      const OpalMediaType & mediaType,
       RTP_QOS * rtpqos
     );
 
+    /** Create a new underlying RTP session instance.
+      */
+    virtual RTP_UDP * CreateRTPSession(
+      unsigned sessionId,
+      const OpalMediaType & mediaType,
+      bool remoteIsNat
+    );
+
+    /**Change the sessionID for an existing session.
+       This will adjust the RTP session and media streams.
+
+       Return false if no such session exists.
+      */
+    virtual bool ChangeSessionID(
+      unsigned fromSessionID,   ///< Session ID to search for
+      unsigned toSessionID      ///< Session ID to change to
+    );
   //@}
 
+  /**@name NAT Management */
   //@{
     /** Return PTrue if the remote appears to be behind a NAT firewall
     */
@@ -332,7 +377,10 @@ class OpalRTPConnection : public OpalConnection
       RTP_DataFrame::PayloadTypes rfc2833; ///<  Payload type for RFC2833
       RTP_DataFrame::PayloadTypes ciscoNSE; ///<  Payload type for RFC2833
     };
+  //@}
 
+  /**@name Overrides from OpalConnection */
+  //@{
     /**Get information on the media channel for the connection.
        The default behaviour checked the mediaTransportAddresses dictionary
        for the session ID and returns information based on that. It also uses
@@ -372,21 +420,54 @@ class OpalRTPConnection : public OpalConnection
       PBoolean isSource                        ///<  Is a source stream
     );
 
-    /**Overrides from OpalConnection
+    /**Adjust media formats available on a connection.
+       This is called by a connection after it has called
+       OpalCall::GetMediaFormats() to get all media formats that it can use so
+       that an application may remove or reorder the media formats before they
+       are used to open media streams.
+
+       This function may also be executed to adjust the media format list for
+       other connections in the call. If this happens then the "otherConnection"
+       parameter will be non-NULL.
+
+       The default behaviour calls the OpalEndPoint function of the same name.
       */
-    virtual void OnPatchMediaStream(PBoolean isSource, OpalMediaPatch & patch);
+    virtual void AdjustMediaFormats(
+      bool local,                          ///<  Media formats a local ones to be presented to remote
+      OpalMediaFormatList & mediaFormats,  ///<  Media formats to use
+      OpalConnection * otherConnection     ///<  Other connection we are adjusting media for
+    ) const;
 
+    /**Call back when patching a media stream.
+       This function is called when a connection has created a new media
+       patch between two streams. This is usually called twice per media patch,
+       once for the source stream and once for the sink stream.
+
+       Note this is not called within the context of the patch thread and is
+       called before that thread has started.
+      */
+    virtual void OnPatchMediaStream(
+      PBoolean isSource,        ///< Is source/sink call
+      OpalMediaPatch & patch    ///<  New patch
+    );
+
+    /** Notifier function for OpalVideoUpdatePicture.
+        Calls the SendIntraFrameRequest on the rtp session
+      */
     void OnMediaCommand(OpalMediaCommand & command, INT extra);
-
-    PDECLARE_NOTIFIER(OpalRFC2833Info, OpalRTPConnection, OnUserInputInlineRFC2833);
-    PDECLARE_NOTIFIER(OpalRFC2833Info, OpalRTPConnection, OnUserInputInlineCiscoNSE);
+  //@}
 
     virtual void SessionFailing(RTP_Session & session);
 
   protected:
+    PDECLARE_NOTIFIER(OpalRFC2833Info, OpalRTPConnection, OnUserInputInlineRFC2833);
+    PDECLARE_NOTIFIER(OpalRFC2833Info, OpalRTPConnection, OnUserInputInlineCiscoNSE);
+
     OpalRTPSessionManager m_rtpSessions;
     OpalRFC2833Proto * rfc2833Handler;
+#if OPAL_T38_CAPABILITY
     OpalRFC2833Proto * ciscoNSEHandler;
+#endif
 
     PBoolean remoteIsNAT;
     PBoolean useRTPAggregation;
