@@ -21,13 +21,13 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 20384 $
+ * $Revision: 23281 $
  * $Author: rjongbloed $
- * $Date: 2008-06-04 10:14:22 +0000 (Wed, 04 Jun 2008) $
+ * $Date: 2009-08-27 06:43:30 -0500 (Thu, 27 Aug 2009) $
  */
 
-#ifndef __OPAL_RTPEP_H
-#define __OPAL_RTPEP_H
+#ifndef OPAL_OPAL_RTPEP_H
+#define OPAL_OPAL_RTPEP_H
 
 #ifdef P_USE_PRAGMA
 #pragma interface
@@ -54,33 +54,33 @@ class OpalRTPEndPoint : public OpalEndPoint
     /**Destroy the endpoint.
      */
     ~OpalRTPEndPoint();
+  //@}
 
-#if OPAL_RTP_AGGREGATE
+  /**@name Overrides from OpalEndPoint */
+  //@{
+    /**Get the data formats this endpoint is capable of operating.
+       This provides a list of media data format names that may be used by an
+       OpalMediaStream may be created by a connection from this endpoint.
 
-    virtual PBoolean UseRTPAggregation() const;
+       Note that a specific connection may not actually support all of the
+       media formats returned here, but should return no more.
 
-    /**Set the RTP aggregation size
+       The default behaviour is pure.
       */
-    void SetRTPAggregationSize(
-      PINDEX size            ///< max connections per aggregation thread. Value of 1 or zero disables aggregation
+    virtual OpalMediaFormatList GetMediaFormats() const;
+
+    /**Call back for closed a media stream.
+
+       The default behaviour checks for local RTP session then calls the
+       OpalManager function of the same name.
+      */
+    virtual void OnClosedMediaStream(
+      const OpalMediaStream & stream     ///<  Media stream being closed
     );
+  //@}
 
-    /**Get the RTP aggregation size
-      */
-    PINDEX GetRTPAggregationSize() const;
-
-    /** Get the aggregator used for RTP channels
-      */
-    PHandleAggregator * GetRTPAggregator();
-
-#endif
-	
-    /**Callback to allow interface adjustments before connecting to the remote party
-       The default implementation does nothing and returns PTrue
-      */
-    virtual PBoolean AdjustInterfaceTable(PIPSocket::Address & remoteAddress,
-                                      PIPSocket::InterfaceTable & interfaceTable);
-
+  /**@name RTP handling */
+  //@{
     /**Determine if the RTP session needs to accommodate a NAT router.
        For endpoints that do not use STUN or something similar to set up all the
        correct protocol embeddded addresses correctly when a NAT router is between
@@ -106,36 +106,47 @@ class OpalRTPEndPoint : public OpalEndPoint
       PBoolean incoming                       ///< Incoming/outgoing connection
     );
 
-    virtual void SetDefaultSecurityMode(const PString & v)
-    { defaultSecurityMode = v; }
-
-    virtual PString GetDefaultSecurityMode() const 
-    { return defaultSecurityMode; }
-
-    /**Get the data formats this endpoint is capable of operating.
-       This provides a list of media data format names that may be used by an
-       OpalMediaStream may be created by a connection from this endpoint.
-
-       Note that a specific connection may not actually support all of the
-       media formats returned here, but should return no more.
-
-       The default behaviour is pure.
-      */
-    virtual OpalMediaFormatList GetMediaFormats() const;
-  //@}
-    
-  protected:
-    PString defaultSecurityMode; 
-
-#if OPAL_RTP_AGGREGATE
-    PMutex rtpAggregationMutex;
-    PBoolean useRTPAggregation; 
-    PINDEX rtpAggregationSize;
-    PHandleAggregator * rtpAggregator;
+#ifdef OPAL_ZRTP
+    virtual bool GetZRTPEnabled() const;
 #endif
 
+    /**Indicate is a local RTP connection.
+       This is called when a new media stream has been created and it has been
+       detected that media will be flowing between two RTP sessions within the
+       same process. An application could take advantage of this by optimising
+       the transfer in some way, rather than the full media path of codecs and
+       sockets whcih might not be necessary.
+
+       The return value is true if the application is going to execute some
+       form of bypass, and the media patch threads should not be started.
+
+       The default behaviour calls the OpanManager function of the same name.
+      */
+    virtual bool OnLocalRTP(
+      OpalConnection & connection1, ///< First connection
+      OpalConnection & connection2, ///< Second connection
+      unsigned sessionID,           ///< Session ID of RTP session
+      bool opened                   ///< Media streams are opened/closed
+    ) const;
+
+    // Check for local RTP connection. Internal function.
+    bool CheckForLocalRTP(const OpalRTPMediaStream & stream);
   //@}
+
+  protected:
+#ifdef OPAL_ZRTP
+    bool zrtpEnabled;
+#endif
+
+    struct LocalRtpInfo {
+      LocalRtpInfo(OpalConnection & connection) : m_connection(connection), m_previousResult(-1) { }
+
+      OpalConnection & m_connection;
+      int              m_previousResult;
+    };
+    typedef std::map<WORD, LocalRtpInfo> LocalRtpInfoMap;
+    LocalRtpInfoMap m_connectionsByRtpLocalPort;
 };
 
 
-#endif // __OPAL_RTPEP_H
+#endif // OPAL_OPAL_RTPEP_H
