@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23625 $
+ * $Revision: 24146 $
  * $Author: rjongbloed $
- * $Date: 2009-10-09 01:44:30 +0000 (Fri, 09 Oct 2009) $
+ * $Date: 2010-03-22 21:58:12 -0500 (Mon, 22 Mar 2010) $
  */
 
 #ifndef OPAL_SIP_SIPCON_H
@@ -44,11 +44,16 @@
 #include <opal/rtpconn.h>
 #include <sip/sippdu.h>
 #include <sip/handlers.h>
+
 #if OPAL_VIDEO
 #include <opal/pcss.h>                  // for OpalPCSSConnection
 #include <codec/vidcodec.h>             // for OpalVideoUpdatePicture command
 #endif
 
+#if OPAL_HAS_IM
+#include <im/rfc4103.h>
+#endif
+
 class OpalCall;
 class SIPEndPoint;
 
@@ -60,7 +65,7 @@ class SIPEndPoint;
 
 /**Session Initiation Protocol connection.
  */
-//class SIPConnection : public OpalConnection
+
 class SIPConnection : public OpalRTPConnection
 {
   PCLASSINFO(SIPConnection, OpalRTPConnection);
@@ -187,9 +192,13 @@ class SIPConnection : public OpalRTPConnection
      */
     virtual bool RetrieveConnection();
 
-    /**Return PTrue if the current connection is on hold.
+    /**Return true if the current connection is on hold.
+       The bool parameter indicates if we are testing if the remote system
+       has us on hold, or we have them on hold.
      */
-    virtual PBoolean IsConnectionOnHold();
+    virtual bool IsConnectionOnHold(
+      bool fromRemote  ///< Test if remote has us on hold, or we have them
+    );
 
     /**Indicate to remote endpoint an alert is in progress.
        If this is an incoming connection and the AnswerCallResponse is in a
@@ -333,6 +342,13 @@ class SIPConnection : public OpalRTPConnection
       */
     virtual void OnReceivedNOTIFY(SIP_PDU & pdu);
 
+    /**Callback function on receipt of an allowed NOTIFY message.
+       Allowed events are determined by the m_allowedEvents member variable.
+      */
+    virtual void OnAllowedEventNotify(
+      const PString & eventName  ///< Name of event
+    );
+
     /**Handle an incoming REFER PDU
       */
     virtual void OnReceivedREFER(SIP_PDU & pdu);
@@ -437,6 +453,7 @@ class SIPConnection : public OpalRTPConnection
     OpalTransportAddress GetDefaultSDPConnectAddress(WORD port = 0) const;
 
     OpalTransport & GetTransport() const { return *transport; }
+    bool SetTransport(OpalTransport * transport);
 
     SIPEndPoint & GetEndPoint() const { return endpoint; }
     SIPDialogContext & GetDialog() { return m_dialog; }
@@ -457,11 +474,21 @@ class SIPConnection : public OpalRTPConnection
     virtual void OnStartTransaction(SIPTransaction & transaction);
 
     virtual void OnReceivedMESSAGE(SIP_PDU & pdu);
-    virtual void OnMessageReceived(const SIPURL & from, const SIP_PDU & pdu);
+
+    P_REMOVE_VIRTUAL_VOID(OnMessageReceived(const SIPURL & /*from*/, const SIP_PDU & /*pdu*/));
+    P_REMOVE_VIRTUAL_VOID(OnMessageReceived(const SIP_PDU & /*pdu*/));
+
+#if OPAL_HAS_IM
+    virtual bool TransmitExternalIM(
+      const OpalMediaFormat & format, 
+      RTP_IMFrame & body
+    );
+#endif
 
     PString GetLocalPartyURL() const;
 
   protected:
+    PDECLARE_NOTIFIER(PTimer, SIPConnection, OnSessionTimeout);
     PDECLARE_NOTIFIER(PTimer, SIPConnection, OnInviteResponseRetry);
     PDECLARE_NOTIFIER(PTimer, SIPConnection, OnAckTimeout);
 
@@ -474,7 +501,8 @@ class SIPConnection : public OpalRTPConnection
       const OpalMediaType & mediaType,
       unsigned sessionID,
       OpalRTPSessionManager & rtpSessions,
-      SDPSessionDescription & sdpOut
+      SDPSessionDescription & sdpOut,
+      bool isReINVITE
     );
     virtual bool AnswerSDPMediaDescription(
       const SDPSessionDescription & sdpIn,
@@ -503,9 +531,7 @@ class SIPConnection : public OpalRTPConnection
 
     friend class SIPInvite;
     static PBoolean WriteINVITE(OpalTransport & transport, void * param);
-    bool WriteINVITE(OpalTransport & transport);
-
-    OpalTransport * CreateTransport(const OpalTransportAddress & address, PBoolean isLocalAddress = PFalse);
+    bool WriteINVITE();
 
     void UpdateRemoteAddresses();
 
@@ -520,6 +546,7 @@ class SIPConnection : public OpalRTPConnection
     SIPEndPoint         & endpoint;
     OpalTransport       * transport;
     bool                  deleteTransport;
+    PStringList           m_allowedEvents;
 
     enum HoldState {
       eHoldOff,
@@ -537,14 +564,16 @@ class SIPConnection : public OpalRTPConnection
     PTime                 originalInviteTime;
     time_t                m_sdpSessionId;
     unsigned              m_sdpVersion; // Really a sequence number
-    bool                  needReINVITE;
+    bool                  m_needReINVITE;
     bool                  m_handlingINVITE;
+    bool                  m_symmetricOpenStream;
     SIPDialogContext      m_dialog;
     OpalGloballyUniqueID  m_dialogNotifyId;
     int                   m_appearanceCode;
     PString               m_alertInfo;
     SIPAuthentication   * m_authentication;
     unsigned              m_authenticatedCseq;
+    PTimer                sessionTimer;
 
     std::map<SIP_PDU::Methods, unsigned> m_lastRxCSeq;
 
@@ -552,9 +581,14 @@ class SIPConnection : public OpalRTPConnection
     PTimer                    ackRetry;
     SIP_PDU                   ackPacket;
     bool                      ackReceived;
-    PSafePtr<SIPTransaction>  referTransaction;
+    bool                      m_referInProgress;
     PSafeList<SIPTransaction> forkedInvitations; // Not for re-INVITE
     PSafeList<SIPTransaction> pendingInvitations; // For re-INVITE
+    PSafeList<SIPTransaction> m_pendingTransactions;
+
+#if OPAL_FAX
+    bool m_switchedToFaxMode;
+#endif
 
     enum {
       ReleaseWithBYE,
@@ -563,16 +597,15 @@ class SIPConnection : public OpalRTPConnection
       ReleaseWithNothing,
     } releaseMethod;
 
-    OpalMediaFormatList remoteFormatList;
-
-  protected:
-    PTimer sessionTimer;
-  public:
-    PDECLARE_NOTIFIER(PTimer, SIPConnection, OnSessionTimeout);
+    OpalMediaFormatList m_remoteFormatList;
+    OpalMediaFormatList m_answerFormatList;
+    void SetRemoteMediaFormats(SDPSessionDescription * sdp);
 
   private:
     P_REMOVE_VIRTUAL_VOID(OnCreatingINVITE(SIP_PDU&));
     P_REMOVE_VIRTUAL_VOID(OnReceivedTrying(SIP_PDU &));
+
+  friend class SIPTransaction;
 };
 
 
