@@ -46,18 +46,21 @@
 
  */
 
+#ifndef PLUGIN_CODEC_DLL_EXPORTS
+#include "plugin-config.h"
+#endif
 
 #define _CRT_SECURE_NO_DEPRECATE
 
-#include "plugin-config.h"
-
 #include "h263-1998.h"
-#include <limits.h>
+#include <limits>
+#include <stdio.h>
 #include <math.h>
-#include "trace.h"
 #include <string.h>
-#include "dyna.h"
-#include "mpi.h"
+
+#include "../common/mpi.h"
+#include "../common/ffmpeg.h"
+#include "../common/trace.h"
 
 #include "tracer.h"
 
@@ -99,9 +102,10 @@ static struct StdSizes {
   { CIF16_WIDTH, CIF16_HEIGHT, PLUGINCODEC_CIF16_MPI },
 };
 
-static FFMPEGLibrary FFMPEGLibraryInstance(CODEC_ID_H263P);
+//static FFMPEGLibrary FFMPEGLibraryInstance(CODEC_ID_H263P);
 
-static void logCallbackFFMPEG (void* v, int level, const char* fmt , va_list arg) {
+static void logCallbackFFMPEG (void* v, int level, const char* fmt , va_list arg)
+{
   char buffer[512];
   int severity = 0;
   if (v) {
@@ -208,33 +212,33 @@ H263_Base_EncoderContext::H263_Base_EncoderContext(const char * _prefix)
 #if TRACE_FILE
   , tracer(_prefix, true)
 #endif
+  , _context(NULL)
 { 
+  _inputFrameBuffer = NULL;
 }
 
 H263_Base_EncoderContext::~H263_Base_EncoderContext()
 {
+  free(_inputFrameBuffer);
 }
 
 bool H263_Base_EncoderContext::Open(CodecID codecId)
 {
   TRACE_AND_LOG(tracer, 1, "Opening encoder");
 
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return false;
-
-  _codec = FFMPEGLibraryInstance.AvcodecFindEncoder(codecId);
+  _codec = avcodec_find_encoder(codecId);
   if (_codec == NULL) {
     TRACE_AND_LOG(tracer, 1, "Codec not found for encoder");
     return false;
   }
 
-  _context = FFMPEGLibraryInstance.AvcodecAllocContext();
+  _context = avcodec_alloc_context();
   if (_context == NULL) {
     TRACE_AND_LOG(tracer, 1, "Failed to allocate context for encoder");
     return false;
   }
 
-  _inputFrame = FFMPEGLibraryInstance.AvcodecAllocFrame();
+  _inputFrame = avcodec_alloc_frame();
   if (_inputFrame == NULL) {
     TRACE_AND_LOG(tracer, 1, "Failed to allocate frame for encoder");
     return false;
@@ -251,20 +255,20 @@ bool H263_Base_EncoderContext::Open(CodecID codecId)
 
   _context->max_b_frames = 0;
   _context->pix_fmt = PIX_FMT_YUV420P;
-  //_context->rtp_mode = 1;
-  //_context->rtp_payload_size = 500;
 
   // X-Lite does not like Custom Picture frequency clocks...
   _context->time_base.num = 100; 
   _context->time_base.den = 2997;
   _context->gop_size      = 125;
 
-//  _context->flags = 0;
   // avoid copying input/output
   _context->flags |= CODEC_FLAG_INPUT_PRESERVED; // we guarantee to preserve input for max_b_frames+1 frames
   _context->flags |= CODEC_FLAG_EMU_EDGE;        // don't draw edges
   _context->flags |= CODEC_FLAG_PASS1;
 
+  _context->error_concealment = 3;
+  _context->error_recognition = 5;
+
   // debugging flags
   if (Trace::CanTraceUserPlane(4)) {
     _context->debug |= FF_DEBUG_RC;
@@ -277,7 +281,7 @@ bool H263_Base_EncoderContext::Open(CodecID codecId)
   SetFrameWidth(_height);
   SetFrameHeight(_width);
   SetTargetBitrate(256000);
-  SetTSTO(31);
+  SetTSTO(0);
   DisableAnnex(D);
   DisableAnnex(F);
   DisableAnnex(I);
@@ -299,27 +303,28 @@ void H263_Base_EncoderContext::SetMaxKeyFramePeriod (unsigned period)
 
 void H263_Base_EncoderContext::SetTargetBitrate (unsigned rate)
 {
-  _context->bit_rate = (rate * 3) >> 2;        // average bit rate
-  _context->bit_rate_tolerance = rate >> 1;
+  m_targetBitRate = rate;
+  CODEC_TRACER(tracer, "target bit rate set to " << m_targetBitRate);
+
+  _context->bit_rate = (m_targetBitRate * 3) >> 2;        // average bit rate
+  _context->bit_rate_tolerance = m_targetBitRate >> 1;
   _context->rc_min_rate = 0;                   // minimum bitrate
-  _context->rc_max_rate = rate;                // maximum bitrate
+  _context->rc_max_rate = m_targetBitRate;                // maximum bitrate
   _context->rc_buffer_size = 224;
 
   /* ratecontrol qmin qmax limiting method
      0-> clipping, 1-> use a nice continous function to limit qscale wthin qmin/qmax.
   */
   
-  _context->rc_qsquish = 0;                    // limit q by clipping 
+  _context->rc_qsquish = 0;            // limit q by clipping 
   _context->rc_eq = (char*) "1";       // rate control equation
   _context->rc_buffer_size = rate * 64;
-
-  CODEC_TRACER(tracer, "target bit rate set to " << rate);
 }
 
 void H263_Base_EncoderContext::SetFrameWidth (unsigned width)
 {
   _width = width;
-  FFMPEGLibraryInstance.AvSetDimensions(_context, _width, _height);
+  avcodec_set_dimensions(_context, _width, _height);
 
   _inputFrame->linesize[0] = width;
   _inputFrame->linesize[1] = width / 2;
@@ -331,7 +336,7 @@ void H263_Base_EncoderContext::SetFrameWidth (unsigned width)
 void H263_Base_EncoderContext::SetFrameHeight (unsigned height)
 {
   _height = height;
-  FFMPEGLibraryInstance.AvSetDimensions(_context, _width, _height);
+  avcodec_set_dimensions(_context, _width, _height);
   CODEC_TRACER(tracer, "frame height set to " << height);
 }
 
@@ -347,7 +352,7 @@ void H263_Base_EncoderContext::SetTSTO (unsigned tsto)
 
   _context->qmin = H263P_MIN_QUANT;
   _context->qmax = round ( (31.0 - H263P_MIN_QUANT) / 31.0 * tsto + H263P_MIN_QUANT);
-  _context->qmax = std::min( _context->qmax, 31);
+  _context->qmax = min( _context->qmax, 31);
 
   _context->mb_qmin = _context->qmin;
   _context->mb_qmax = _context->qmax;
@@ -371,7 +376,8 @@ void H263_Base_EncoderContext::EnableAnnex (Annex annex)
     case F:
       // Annex F: Advanced Prediction Mode
       // does not work with eyeBeam
-      _context->flags |= CODEC_FLAG_OBMC; 
+      // DO NOT ENABLE THIS FLAG. FFMPEG IS NOT THREAD_SAFE WHEN THIS FLAG IS SET
+      //_context->flags |= CODEC_FLAG_OBMC; 
       break;
     case I:
       // Annex I: Advanced Intra Coding
@@ -460,6 +466,13 @@ bool H263_Base_EncoderContext::OpenCodec()
   CODEC_TRACER(tracer, "Size is " << _width << "x" << _height);
   CODEC_TRACER(tracer, "rc_max_rate is " <<  _context->rc_max_rate);
   CODEC_TRACER(tracer, "GOP is " << _context->gop_size);
+  CODEC_TRACER(tracer, "qmin set to " << _context->qmin);
+  CODEC_TRACER(tracer, "qmax set to " << _context->qmax);
+  CODEC_TRACER(tracer, "mb_qmin set to " << _context->mb_qmin);
+  CODEC_TRACER(tracer, "mb_qmax set to " << _context->mb_qmax);
+  CODEC_TRACER(tracer, "bit_rate set to " << _context->bit_rate);
+  CODEC_TRACER(tracer, "bit_rate_tolerance set to " <<_context->bit_rate_tolerance);
+  CODEC_TRACER(tracer, "rc_min_rate set to " << _context->rc_min_rate);
   CODEC_TRACER_FLAG(tracer, CODEC_FLAG_H263P_UMV);
   CODEC_TRACER_FLAG(tracer, CODEC_FLAG_OBMC);
   CODEC_TRACER_FLAG(tracer, CODEC_FLAG_AC_PRED);
@@ -467,14 +480,14 @@ bool H263_Base_EncoderContext::OpenCodec()
   CODEC_TRACER_FLAG(tracer, CODEC_FLAG_LOOP_FILTER);
   CODEC_TRACER_FLAG(tracer, CODEC_FLAG_H263P_AIV);
 
-  return FFMPEGLibraryInstance.AvcodecOpen(_context, _codec) == 0;
+  return avcodec_open(_context, _codec) == 0;
 }
 
 void H263_Base_EncoderContext::CloseCodec()
 {
   if (_context != NULL) {
     if (_context->codec != NULL) {
-      FFMPEGLibraryInstance.AvcodecClose(_context);
+      avcodec_close(_context);
     }
   }
 }
@@ -500,13 +513,11 @@ H263_RFC2190_EncoderContext::~H263_RFC2190_EncoderContext()
 {
   WaitAndSignal m(_mutex);
 
-  if (FFMPEGLibraryInstance.IsLoaded()) {
+  CloseCodec();
 
-    CloseCodec();
+  av_free(_context);
+  av_free(_inputFrame);
 
-    FFMPEGLibraryInstance.AvcodecFree(_context);
-    FFMPEGLibraryInstance.AvcodecFree(_inputFrame);
-  }  
   TRACE_AND_LOG(tracer, 3, "encoder closed");
 }
 
@@ -518,11 +529,11 @@ static void rtp_callback(struct AVCodecContext *avctx, void * _data, int size, i
   context->RTPCallBack(avctx, _data, size, mb_nb);
 }
 
-void H263_RFC2190_EncoderContext::RTPCallBack(struct AVCodecContext *avctx, void * _data, int size, int mbCount)
+void H263_RFC2190_EncoderContext::RTPCallBack(struct AVCodecContext * /*avctx*/, void * _data, int size, int mbCount)
 {
   // sometimes, FFmpeg encodes the same frame multiple times
   // we need to detect this in order to avoid duplicating the encoded data
-  if ((_data == &packetizer.buffer[0]) && (packetizer.fragments.size() != 0)) {
+  if ((_data == packetizer.m_buffer) && (packetizer.fragments.size() != 0)) {
     packetizer.fragments.resize(0);
     currentMb = 0;
     currentBytes = 0;
@@ -569,7 +580,7 @@ bool H263_RFC2190_EncoderContext::InitContext()
   return true;
 }
 
-void H263_RFC2190_EncoderContext::SetMaxRTPFrameSize (unsigned size)
+void H263_RFC2190_EncoderContext::SetMaxRTPFrameSize (unsigned /*size*/)
 {
   //_context->rtp_payload_size = (size * 6 / 7);
   
@@ -584,9 +595,6 @@ int H263_RFC2190_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
 {
   WaitAndSignal m(_mutex);
 
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return 0;
-
   if (_codec == NULL) {
     TRACE_AND_LOG(tracer, 1, "Encoder\tCodec not initialized");
     return 0;
@@ -602,7 +610,17 @@ int H263_RFC2190_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
   // if still running out packets from previous frame, then return it
   if (packetizer.GetPacket(dstRTP, flags) != 0) {
     CODEC_TRACER_RTP(tracer, "Tx frame:", dstRTP, RFC2190Dump);
-    dstLen = dstRTP.GetFrameLen();
+    dstLen = dstRTP.GetHeaderSize() + dstRTP.GetPayloadSize();
+    return 1;
+  }
+
+  // zero payload means do nothing
+  if (srcRTP.GetPayloadSize() == 0) {
+    TRACE_AND_LOG(tracer, 1, "Zero payload passed");
+    dstLen = dstRTP.GetHeaderSize();
+    dstRTP.SetPayloadSize(0);
+    dstRTP.SetMarker(true);
+    flags |= 1;
     return 1;
   }
 
@@ -630,6 +648,18 @@ int H263_RFC2190_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
       TRACE_AND_LOG(tracer, 1, "Reopening codec failed");
       return 0;
     }
+
+    if (_inputFrameBuffer != NULL)
+      free(_inputFrameBuffer);
+#if HAVE_POSIX_MEMALIGN
+    if (posix_memalign((void **)&_inputFrameBuffer, 64, header->width*header->height*3/2 + (FF_INPUT_BUFFER_PADDING_SIZE*2)) != 0) 
+#else
+    if ((_inputFrameBuffer = (BYTE *)malloc(header->width*header->height*3/2 + (FF_INPUT_BUFFER_PADDING_SIZE*2))) == NULL) 
+#endif
+    {
+      TRACE_AND_LOG(tracer, 1, "Unable to allocate memory for frame buffer");
+      return 0;
+    }
   }
 
   CODEC_TRACER(tracer, "Input:seq=" << _frameCount
@@ -642,30 +672,62 @@ int H263_RFC2190_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
   int frameSize = (size * 3) >> 1;
  
   // we need FF_INPUT_BUFFER_PADDING_SIZE allocated bytes after the YVU420P image for the encoder
-  memset (_inputFrameBuffer, 0 , FF_INPUT_BUFFER_PADDING_SIZE);
-  memcpy (_inputFrameBuffer + FF_INPUT_BUFFER_PADDING_SIZE, OPAL_VIDEO_FRAME_DATA_PTR(header), frameSize);
-  memset (_inputFrameBuffer + FF_INPUT_BUFFER_PADDING_SIZE + frameSize, 0 , FF_INPUT_BUFFER_PADDING_SIZE);
+  memcpy (_inputFrameBuffer, OPAL_VIDEO_FRAME_DATA_PTR(header), frameSize);
+  memset (_inputFrameBuffer + frameSize, 0 , FF_INPUT_BUFFER_PADDING_SIZE);
+  _inputFrame->data[0] = _inputFrameBuffer;
 
-  _inputFrame->data[0] = _inputFrameBuffer + FF_INPUT_BUFFER_PADDING_SIZE;
   _inputFrame->data[1] = _inputFrame->data[0] + size;
   _inputFrame->data[2] = _inputFrame->data[1] + (size / 4);
   _inputFrame->pict_type = (flags && forceIFrame) ? FF_I_TYPE : 0;
 
   currentMb = 0;
   currentBytes = 0;
+
   packetizer.fragments.resize(0);
-  packetizer.buffer.resize(frameSize);
 
-  CODEC_TRACER(tracer, "Encoder called with " << frameSize << " bytes and frame type " << _inputFrame->pict_type << " at " << header->width << "x" << header->height);
+  size_t newOutputSize = 100000;
 
-  int encodedLen = FFMPEGLibraryInstance.AvcodecEncodeVideo(_context, &packetizer.buffer[0], frameSize, _inputFrame);  
+  if (packetizer.m_buffer != NULL) {
+    if (packetizer.m_bufferSize < newOutputSize) {
+      free(packetizer.m_buffer);
+      packetizer.m_buffer = NULL;
+    }
+  }
+  if (packetizer.m_buffer == NULL) {
+    packetizer.m_bufferSize = newOutputSize;
+#if HAVE_POSIX_MEMALIGN
+    if (posix_memalign((void **)&packetizer.m_buffer, 64, packetizer.m_bufferSize) != 0) 
+#else
+    if ((packetizer.m_buffer = (BYTE *)malloc(packetizer.m_bufferSize)) == NULL) 
+#endif
+    {
+      TRACE_AND_LOG(tracer, 1, "Unable to allocate memory for packet buffer");
+      return 0;
+    }
+  }
 
-  packetizer.buffer.resize(encodedLen);
+  //CODEC_TRACER(tracer, "Encoder called with " << frameSize << " bytes and frame type " << _inputFrame->pict_type << " at " << header->width << "x" << header->height);
+
+  int encodedLen = avcodec_encode_video(_context, packetizer.m_buffer, packetizer.m_bufferSize, _inputFrame);  
+
+  if (encodedLen < 0) {
+    TRACE_AND_LOG(tracer, 1, "Encoder failed");
+    return 0;
+  }
+  if (encodedLen == 0) {
+    TRACE_AND_LOG(tracer, 1, "Encoder returned empty frame");
+    dstRTP.SetPayloadSize(0);
+    dstLen = dstRTP.GetHeaderSize();
+    flags |= 1;
+    return 1;
+  }
+
+  packetizer.m_bufferLen = encodedLen;
 
   // push the encoded frame through the packetizer
 #if TRACE_FILE
-  if (encodedLen > 0) {
-    const unsigned char * p =  &packetizer.buffer[0];
+  {
+    const unsigned char * p =  packetizer.m_buffer;
     CODEC_TRACER(tracer, "Raw data: " << hex << setfill('0') << setprecision(2)
                          << (int)p[0] << ' ' << (int)p[1] << ' ' << (int)p[2] << ' ' << (int)p[3] << ' ' << (int)p[4]
                          << setfill(' ') << dec);
@@ -683,7 +745,7 @@ int H263_RFC2190_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
   // return the first encoded block of data
   if (packetizer.GetPacket(dstRTP, flags)) {
     CODEC_TRACER_RTP(tracer, "Tx frame:", dstRTP, RFC2190Dump);
-    dstLen = dstRTP.GetFrameLen();
+    dstLen = dstRTP.GetHeaderSize() + dstRTP.GetPayloadSize();
   }
 
   return 1;
@@ -701,16 +763,14 @@ H263_RFC2429_EncoderContext::~H263_RFC2429_EncoderContext()
 {
   WaitAndSignal m(_mutex);
 
-  if (FFMPEGLibraryInstance.IsLoaded()) {
+  CloseCodec();
 
-    CloseCodec();
+  if (_txH263PFrame)
+    delete _txH263PFrame;
 
-    if (_txH263PFrame)
-      delete _txH263PFrame;
+  av_free(_context);
+  av_free(_inputFrame);
 
-    FFMPEGLibraryInstance.AvcodecFree(_context);
-    FFMPEGLibraryInstance.AvcodecFree(_inputFrame);
-  }  
   TRACE_AND_LOG(tracer, 3, "encoder closed");
 }
 
@@ -739,7 +799,7 @@ void H263_RFC2429_EncoderContext::SetMaxRTPFrameSize (unsigned size)
     else  
     _context->rtp_payload_size = size;
 
-  _txH263PFrame->SetMaxPayloadSize(size);
+  _txH263PFrame->SetMaxPayloadSize((uint16_t)size);
 }
 
 
@@ -747,9 +807,6 @@ int H263_RFC2429_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
 {
   WaitAndSignal m(_mutex);
 
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return 0;
-
   if (_codec == NULL) {
     TRACE_AND_LOG(tracer, 1, "Codec not initialized");
     return 0;
@@ -795,6 +852,16 @@ int H263_RFC2429_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
       TRACE_AND_LOG(tracer, 1, "Reopening codec failed");
       return 0;
     }
+    if (_inputFrameBuffer != NULL)
+      free(_inputFrameBuffer);
+#if HAVE_POSIX_MEMALIGN
+    if (posix_memalign((void **)&_inputFrameBuffer, 64, header->width*header->height*3/2 + (FF_INPUT_BUFFER_PADDING_SIZE*2)) != 0) {
+#else
+    if ((_inputFrameBuffer = (BYTE *)malloc(header->width*header->height*3/2 + (FF_INPUT_BUFFER_PADDING_SIZE*2))) != NULL) {
+#endif
+      TRACE_AND_LOG(tracer, 1, "Unable to allocate memory for frame buffer");
+      return 0;
+    }
   }
 
   CODEC_TRACER(tracer, "Input:seq=" << _frameCount
@@ -816,7 +883,7 @@ int H263_RFC2429_EncoderContext::EncodeFrames(const BYTE * src, unsigned & srcLe
  
   _txH263PFrame->BeginNewFrame();
   _txH263PFrame->SetTimestamp(srcRTP.GetTimestamp());
-  _txH263PFrame->SetFrameSize (FFMPEGLibraryInstance.AvcodecEncodeVideo(_context, _txH263PFrame->GetFramePtr(), frameSize, _inputFrame));  
+  _txH263PFrame->SetFrameSize (avcodec_encode_video(_context, _txH263PFrame->GetFramePtr(), frameSize, _inputFrame));  
   _frameCount++; 
 
   if (_txH263PFrame->GetFrameSize() == 0) {
@@ -844,21 +911,18 @@ H263_Base_DecoderContext::H263_Base_DecoderContext(const char * _prefix)
   , tracer(_prefix, false)
 #endif
 {
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return;
-
-  if ((_codec = FFMPEGLibraryInstance.AvcodecFindDecoder(CODEC_ID_H263)) == NULL) {
+  if ((_codec = avcodec_find_decoder(CODEC_ID_H263)) == NULL) {
     TRACE_AND_LOG(tracer, 1, "Codec not found for decoder");
     return;
   }
 
-  _context = FFMPEGLibraryInstance.AvcodecAllocContext();
+  _context = avcodec_alloc_context();
   if (_context == NULL) {
     TRACE_AND_LOG(tracer, 1, "Failed to allocate context for decoder");
     return;
   }
 
-  _outputFrame = FFMPEGLibraryInstance.AvcodecAllocFrame();
+  _outputFrame = avcodec_alloc_frame();
   if (_outputFrame == NULL) {
     TRACE_AND_LOG(tracer, 1, "Failed to allocate frame for decoder");
     return;
@@ -883,12 +947,10 @@ H263_Base_DecoderContext::H263_Base_DecoderContext(const char * _prefix)
 
 H263_Base_DecoderContext::~H263_Base_DecoderContext()
 {
-  if (FFMPEGLibraryInstance.IsLoaded()) {
-    CloseCodec();
+  CloseCodec();
 
-    FFMPEGLibraryInstance.AvcodecFree(_context);
-    FFMPEGLibraryInstance.AvcodecFree(_outputFrame);
-  }
+  av_free(_context);
+  av_free(_outputFrame);
 }
 
 bool H263_Base_DecoderContext::OpenCodec()
@@ -898,7 +960,7 @@ bool H263_Base_DecoderContext::OpenCodec()
     return 0;
   }
 
-  if (FFMPEGLibraryInstance.AvcodecOpen(_context, _codec) < 0) {
+  if (avcodec_open(_context, _codec) < 0) {
     TRACE_AND_LOG(tracer, 1, "Failed to open H.263 decoder");
     return false;
   }
@@ -912,7 +974,7 @@ void H263_Base_DecoderContext::CloseCodec()
 {
   if (_context != NULL) {
     if (_context->codec != NULL) {
-      FFMPEGLibraryInstance.AvcodecClose(_context);
+      avcodec_close(_context);
       TRACE_AND_LOG(tracer, 4, "Closed H.263 decoder" );
     }
   }
@@ -937,9 +999,6 @@ H263_RFC2429_DecoderContext::~H263_RFC2429_DecoderContext()
 
 bool H263_RFC2429_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcLen, BYTE * dst, unsigned & dstLen, unsigned int & flags)
 {
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return 0;
-
   TRACE_AND_LOG(tracer, 4, "Codec opened");
 
   // create RTP frame from source buffer
@@ -955,7 +1014,7 @@ bool H263_RFC2429_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcL
     _rxH263PFrame->BeginNewFrame();
     flags = (_gotAGoodFrame ? PluginCodec_ReturnCoderRequestIFrame : 0);
     _gotAGoodFrame = false;
-    return 1;
+    return true;
   }
   
   if (srcRTP.GetMarker()==0)
@@ -996,7 +1055,7 @@ bool H263_RFC2429_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcL
   int gotPicture = 0;
 
   TRACE_AND_LOG(tracer, 4, "Decoding " << _rxH263PFrame->GetFrameSize()  << " bytes");
-  int bytesDecoded = FFMPEGLibraryInstance.AvcodecDecodeVideo(_context, _outputFrame, &gotPicture, _rxH263PFrame->GetFramePtr(), _rxH263PFrame->GetFrameSize());
+  int bytesDecoded = avcodec_decode_video(_context, _outputFrame, &gotPicture, _rxH263PFrame->GetFramePtr(), _rxH263PFrame->GetFrameSize());
 
   _rxH263PFrame->BeginNewFrame();
 
@@ -1081,62 +1140,121 @@ H263_RFC2190_DecoderContext::~H263_RFC2190_DecoderContext()
 {
 }
 
+static bool ReturnEmptyFrame(RTPFrame & dstRTP, unsigned & dstLen, unsigned int & flags)
+{
+  flags |= PluginCodec_ReturnCoderLastFrame;
+  dstRTP.SetPayloadSize(0);
+  dstLen = 0;
+  return true;
+}
+
 bool H263_RFC2190_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcLen, BYTE * dst, unsigned & dstLen, unsigned int & flags)
 {
-  if (!FFMPEGLibraryInstance.IsLoaded())
-    return 0;
-
   // create RTP frame from source buffer
   RTPFrame srcRTP(src, srcLen);
 
+  // create RTP frame from destination
+  RTPFrame dstRTP(dst, dstLen, 0);
+  dstRTP.SetTimestamp(srcRTP.GetTimestamp());
+
+  if (dstLen < (12 + sizeof(PluginCodec_Video_FrameHeader))) {
+    flags = 0;
+    TRACE_AND_LOG(tracer, 1, "Destination buffer " << dstLen << " insufficient for video header");
+    ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
+
+  dstLen = 0;
+
   CODEC_TRACER_RTP(tracer, "Rx frame:", srcRTP, RFC2190Dump);
 
   // push new frame through the depacketiser
   bool requestIFrame, isIFrame;
   int code = depacketizer.SetPacket(srcRTP, requestIFrame, isIFrame);
-  if (code <= 0)
-    return true;
+  if (code <= 0) {
+    flags = requestIFrame ? PluginCodec_ReturnCoderRequestIFrame : 0;
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
+
+  if ((depacketizer.frame.size() < 3)  ||
+      (depacketizer.frame[0] != 0x00) ||
+      (depacketizer.frame[1] != 0x00) ||
+      (depacketizer.frame[2] & 0x80) != 0x80) {
+    TRACE_AND_LOG(tracer, 1, "Frame does not start with correct code");
+    flags = PluginCodec_ReturnCoderRequestIFrame;
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
 
   TRACE_AND_LOG(tracer, 4, "Decoder called with " << depacketizer.frame.size()  << " bytes");
 
+#if FFMPEG_HAS_DECODE_ERROR_COUNT
+  unsigned error_before = _context->decode_error_count;
+#endif
+
   int gotPicture = 0;
-  int bytesDecoded = FFMPEGLibraryInstance.AvcodecDecodeVideo(_context, _outputFrame, &gotPicture, &depacketizer.frame[0], depacketizer.frame.size());
+  int bytesDecoded = avcodec_decode_video(_context, _outputFrame, &gotPicture, &depacketizer.frame[0], depacketizer.frame.size());
 
   depacketizer.NewFrame();
 
   if (!gotPicture) {
-    TRACE_AND_LOG(tracer, 1, "Decoded "<< bytesDecoded << " bytes without getting a Picture"); 
     flags = PluginCodec_ReturnCoderRequestIFrame;
-    return 1;
+    TRACE_AND_LOG(tracer, 1, "Decoded "<< bytesDecoded << " bytes without getting a Picture"); 
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
   }
 
   TRACE_AND_LOG(tracer, 4, "Decoder processed " << bytesDecoded << " bytes, creating frame at " << _context->width << "x" << _context->height);
 
   // if error occurred, tell the other end to send another I-frame and hopefully we can resync
-  if (bytesDecoded < 0) {
+  if (bytesDecoded < 0
+#if FFMPEG_HAS_DECODE_ERROR_COUNT
+      || error_before != _context->decode_error_count
+#endif
+  ) {
     flags = PluginCodec_ReturnCoderRequestIFrame;
-    return 1;
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
+
+  if (bytesDecoded == 0) {
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
   }
 
   // if decoded frame size is not legal, request an I-Frame
-  if (_context->width == 0 || _context->height == 0) {
+  if ((_context->width <= 0) ||
+      (_context->height <= 0) ||
+      (_context->width > CIF4_WIDTH) ||
+      (_context->height > CIF4_HEIGHT) ||
+      (_context->height * _context->width > (CIF4_WIDTH * CIF4_HEIGHT))) {
     TRACE_AND_LOG(tracer, 1, "Received frame with invalid size");
     flags = PluginCodec_ReturnCoderRequestIFrame;
-    return 1;
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
   }
 
   // create RTP frame from destination buffer
-  RTPFrame dstRTP(dst, dstLen, 0);
-  dstLen = 0;
+  unsigned frameBytes = (_context->width * _context->height * 12) / 8;
+  if (dstRTP.GetPayloadSize() - sizeof(PluginCodec_Video_FrameHeader) < frameBytes) {
+    TRACE_AND_LOG(tracer, 1, "Destination buffer size " << dstRTP.GetPayloadSize() << " too small for frame of size " << _context->width  << "x" <<  _context->height);
+    flags = PluginCodec_ReturnCoderRequestIFrame;
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
 
-  int frameBytes = (_context->width * _context->height * 12) / 8;
   PluginCodec_Video_FrameHeader * header = (PluginCodec_Video_FrameHeader *)dstRTP.GetPayloadPtr();
-  header->x = header->y = 0;
-  header->width = _context->width;
+  header->x      = header->y = 0;
+  header->width  = _context->width;
   header->height = _context->height;
   int size = _context->width * _context->height;
-  if (_outputFrame->data[1] == _outputFrame->data[0] + size
-      && _outputFrame->data[2] == _outputFrame->data[1] + (size >> 2)) {
+
+  if ((unsigned)dstRTP.GetFrameLen() < (frameBytes + sizeof(PluginCodec_Video_FrameHeader))) {
+    flags = PluginCodec_ReturnCoderRequestIFrame;
+    TRACE_AND_LOG(tracer, 1, "Destination buffer " << dstLen << " insufficient for decoded data size " << header->width << "x" << header->height);
+    return ReturnEmptyFrame(dstRTP, dstLen, flags);
+  }
+
+  dstRTP.SetPayloadSize(sizeof(PluginCodec_Video_FrameHeader) + frameBytes);
+  dstLen = dstRTP.GetHeaderSize() + dstRTP.GetPayloadSize();
+
+  if (
+       (_outputFrame->data[1] == (_outputFrame->data[0] + size)) &&
+       (_outputFrame->data[2] == (_outputFrame->data[1] + (size >> 2)))
+     ) {
     memcpy(OPAL_VIDEO_FRAME_DATA_PTR(header), _outputFrame->data[0], frameBytes);
   } else {
     unsigned char *dst = OPAL_VIDEO_FRAME_DATA_PTR(header);
@@ -1150,7 +1268,7 @@ bool H263_RFC2190_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcL
         memcpy(dst, src, dst_stride*h);
         dst += dst_stride*h;
       } else {
-        while (h--) {
+        while (h-- > 0) {
           memcpy(dst, src, dst_stride);
           dst += dst_stride;
           src += src_stride;
@@ -1159,12 +1277,9 @@ bool H263_RFC2190_DecoderContext::DecodeFrames(const BYTE * src, unsigned & srcL
     }
   }
 
-  dstRTP.SetPayloadSize(sizeof(PluginCodec_Video_FrameHeader) + frameBytes);
   dstRTP.SetTimestamp(srcRTP.GetTimestamp());
   dstRTP.SetMarker(true);
 
-  dstLen = dstRTP.GetFrameLen();
-
   flags = PluginCodec_ReturnCoderLastFrame |
           (isIFrame ? PluginCodec_ReturnCoderIFrame : 0) |
           (requestIFrame ? PluginCodec_ReturnCoderRequestIFrame : 0);
@@ -1256,10 +1371,11 @@ static void FindBoundingBox(const char * const * * parm,
                                              int & maxWidth,
                                              int & maxHeight,
                                              int & frameTime,
-                                             int & bitRate)
-{ 
+                                             int & targetBitRate,
+                                             int & maxBitRate)
+{
   // initialise the MPI values to disabled
-       int i;
+  int i;
   for (i = 0; i < 5; i++)
     mpi[i] = PLUGINCODEC_MPI_DISABLED;
 
@@ -1275,8 +1391,8 @@ static void FindBoundingBox(const char * const * * parm,
   int frameRate     = 10;      // 10 fps
   int origFrameTime = 900;     // 10 fps in video RTP timestamps
   int maxBR = 0;
-  int maxBitRate = 0;
-  int targetBitRate = 0;
+  maxBitRate = 0;
+  targetBitRate = 0;
 
   // extract the MPI values set in the custom options, and find the min/max of them
   frameTime = 0;
@@ -1374,13 +1490,19 @@ static void FindBoundingBox(const char * const * * parm,
   }
 
   // find an appropriate max bit rate
-  bitRate = 0;
-  if (maxBR == 0)
-    bitRate = maxBitRate;
-  else if (maxBitRate == 0)
-    bitRate = maxBR * 100;
-  else
-    bitRate = PMIN(maxBR * 100, maxBitRate);
+  if (maxBitRate == 0) {
+    if (maxBR != 0)
+      maxBitRate = maxBR * 100;
+    else if (targetBitRate != 0)
+      maxBitRate = targetBitRate;
+    else
+      maxBitRate = 327000;
+  }
+  else if (maxBR > 0)
+    maxBitRate = PMIN(maxBR * 100, maxBitRate);
+
+  if (targetBitRate == 0)
+    targetBitRate = 327000;
 }
 
 static int to_normalised_options(const struct PluginCodec_Definition *, void *, const char *, void * parm, unsigned * parmLen)
@@ -1390,8 +1512,8 @@ static int to_normalised_options(const struct PluginCodec_Definition *, void *,
 
   // find bounding box enclosing all MPI values
   int mpi[5];
-  int minWidth, minHeight, maxHeight, maxWidth, frameTime, bitRate;
-  FindBoundingBox((const char * const * *)parm, mpi, minWidth, minHeight, maxWidth, maxHeight, frameTime, bitRate);
+  int minWidth, minHeight, maxHeight, maxWidth, frameTime, targetBitRate, maxBitRate;
+  FindBoundingBox((const char * const * *)parm, mpi, minWidth, minHeight, maxWidth, maxHeight, frameTime, targetBitRate, maxBitRate);
 
   char ** options = (char **)calloc(16+(5*2)+2, sizeof(char *));
   *(char ***)parm = options;
@@ -1409,11 +1531,11 @@ static int to_normalised_options(const struct PluginCodec_Definition *, void *,
   options[ 8] = strdup(PLUGINCODEC_OPTION_FRAME_TIME);
   options[ 9] = num2str(frameTime);
   options[10] = strdup(PLUGINCODEC_OPTION_MAX_BIT_RATE);
-  options[11] = num2str(bitRate);
+  options[11] = num2str(maxBitRate);
   options[12] = strdup(PLUGINCODEC_OPTION_TARGET_BIT_RATE);
-  options[13] = num2str(bitRate);
+  options[13] = num2str(targetBitRate);
   options[14] = strdup("MaxBR");
-  options[15] = num2str((bitRate+50)/100);
+  options[15] = num2str((maxBitRate+50)/100);
   for (int i = 0; i < 5; i++) {
     options[16+i*2] = strdup(StandardVideoSizes[i].optionName);
     options[16+i*2+1] = num2str(mpi[i]);
@@ -1429,8 +1551,8 @@ static int to_customised_options(const struct PluginCodec_Definition *, void *,
 
   // find bounding box enclosing all MPI values
   int mpi[5];
-  int minWidth, minHeight, maxHeight, maxWidth, frameTime, bitRate;
-  FindBoundingBox((const char * const * *)parm, mpi, minWidth, minHeight, maxWidth, maxHeight, frameTime, bitRate);
+  int minWidth, minHeight, maxHeight, maxWidth, frameTime, targetBitRate, maxBitRate;
+  FindBoundingBox((const char * const * *)parm, mpi, minWidth, minHeight, maxWidth, maxHeight, frameTime, targetBitRate, maxBitRate);
 
   char ** options = (char **)calloc(14+5*2+2, sizeof(char *));
   *(char ***)parm = options;
@@ -1446,11 +1568,11 @@ static int to_customised_options(const struct PluginCodec_Definition *, void *,
   options[ 6] = strdup(PLUGINCODEC_OPTION_MAX_RX_FRAME_HEIGHT);
   options[ 7] = num2str(maxHeight);
   options[ 8] = strdup(PLUGINCODEC_OPTION_MAX_BIT_RATE);
-  options[ 9] = num2str(bitRate);
+  options[ 9] = num2str(maxBitRate);
   options[10] = strdup(PLUGINCODEC_OPTION_TARGET_BIT_RATE);
-  options[11] = num2str(bitRate);
+  options[11] = num2str(targetBitRate);
   options[12] = strdup("MaxBR");
-  options[13] = num2str((bitRate+50)/100);
+  options[13] = num2str((maxBitRate+50)/100);
   for (int i = 0; i < 5; i++) {
     options[14+i*2] = strdup(StandardVideoSizes[i].optionName);
     options[14+i*2+1] = num2str(mpi[i]);
@@ -1554,7 +1676,7 @@ static int codec_decoder(const struct PluginCodec_Definition *,
                                    unsigned int * flag)
 {
   H263_Base_DecoderContext * context = (H263_Base_DecoderContext *)_context;
-  return context->DecodeFrames((const BYTE *)from, *fromLen, (BYTE *)to, *toLen, *flag);
+  return context->DecodeFrames((const BYTE *)from, *fromLen, (BYTE *)to, *toLen, *flag) ? 1 : 0;
 }
 
 static int decoder_get_output_data_size(const PluginCodec_Definition * codec, void *, const char *, void *, unsigned *)
@@ -1943,14 +2065,11 @@ extern "C" {
       Trace::SetLevelUserPlane(0);
     }
 
-    if (!FFMPEGLibraryInstance.Load(1)) {
-      *count = 0;
-      TRACE(1, "H.263\tCodec\tDisabled");
-      return NULL;
-    }
+    avcodec_init();
+    avcodec_register_all();
 
-    FFMPEGLibraryInstance.AvLogSetLevel(AV_LOG_FATAL);
-    FFMPEGLibraryInstance.AvLogSetCallback(&logCallbackFFMPEG);
+    av_log_set_level(AV_LOG_FATAL);
+    av_log_set_callback(&logCallbackFFMPEG);
 
     if (version < PLUGIN_CODEC_VERSION_OPTIONS) {
       *count = 0;
