@@ -23,16 +23,27 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22072 $
+ * $Revision: 23520 $
  * $Author: rjongbloed $
- * $Date: 2009-02-16 11:18:30 +0000 (Mon, 16 Feb 2009) $
+ * $Date: 2009-09-23 23:57:23 -0500 (Wed, 23 Sep 2009) $
  */
 
 #include "precompile.h"
 #include "main.h"
 
+#include <codec/ratectl.h>
 #include <opal/patch.h>
 
+#include <ptclib/random.h>
+
+#define OUTPUT_BPS(strm, rate) \
+  if (rate < 10000ULL) strm << rate << ' '; \
+  else if (rate < 10000000ULL) strm << rate/1000.0 << " k"; \
+  else if (rate < 10000000000ULL) strm << rate/1000000.0 << " M"; \
+  strm << "bps"; \
+
+PMutex coutMutex;
+static unsigned g_infoCount = 0;
 
 PCREATE_PROCESS(CodecTest);
 
@@ -53,7 +64,8 @@ void CodecTest::Main()
 
   args.Parse("b-bit-rate:"
              "c-crop."
-             "C-rate-control."
+             "C-rate-control:"
+             "d-drop:"
              "D-display-device:"
              "-display-driver:"
              "F-audio-frames:"
@@ -62,6 +74,7 @@ void CodecTest::Main()
              "-grab-format:"
              "-grab-channel:"
              "h-help."
+             "i-info."
              "m-suppress-marker."
              "M-force-marker."
              "O-option:"
@@ -73,10 +86,13 @@ void CodecTest::Main()
              "R-record-device:"
              "-record-driver:"
              "s-frame-size:"
+             "-single-step."
              "S-simultaneous:"
+             "T-statistics."
              "-count:"
              "-noprompt."
              "-snr."
+             "-list."
 #if PTRACING
              "o-output:"             "-no-output."
              "t-trace."              "-no-trace."
@@ -89,6 +105,43 @@ void CodecTest::Main()
          PTrace::Blocks | PTrace::Timestamp | PTrace::Thread | PTrace::FileAndLine);
 #endif
 
+  if (args.HasOption("list")) {
+    OpalPluginCodecManager * codecMgr = PFactory<PPluginModuleManager>::CreateInstanceAs<OpalPluginCodecManager>("OpalPluginCodecManager");
+    if (codecMgr == NULL) {
+      cerr << "error: Cannot get codec manager" << endl;
+      return;
+    }
+    PPluginModuleManager::PluginListType pluginList = codecMgr->GetPluginList();
+
+    cout << "Plugin codecs:" << endl;
+    for (int i = 0; i < pluginList.GetSize(); i++) {
+      PDynaLink & plugin =  pluginList.GetDataAt(i);
+
+      PStringList codecNames;
+      PluginCodec_GetCodecFunction getCodecs;
+      if (plugin.GetFunction(PLUGIN_CODEC_GET_CODEC_FN_STR, (PDynaLink::Function &)getCodecs)) {
+        unsigned count = 0;
+        PluginCodec_Definition * codecs = (*getCodecs)(&count, PLUGIN_CODEC_VERSION_OPTIONS);
+        while (count > 0) {
+          PString name(codecs->descr);
+          if (codecNames.GetStringsIndex(name) == P_MAX_INDEX)
+            codecNames.AppendString(name);
+          --count;
+          ++codecs;
+        }
+      }
+
+      cout << "   " << plugin.GetName(true) << endl;
+      if (codecNames.GetSize() == 0) 
+        cout << "      No codecs defined" << endl;
+      else
+        cout << "      " << setfill(',') << codecNames << setfill(' ') << endl;
+      
+    }
+    cout << "\n\n";
+    return;
+  }
+
   if (args.HasOption('h') || args.GetCount() == 0) {
     PError << "usage: " << GetFile().GetTitle() << " [ options ] fmtname [ fmtname ]\n"
               "  where fmtname is the Media Format Name for the codec(s) to test, up to two\n"
@@ -112,7 +165,7 @@ void CodecTest::Main()
               "  -r --frame-rate size    : video frame rate (frames/second)\n"
               "  -b --bit-rate size      : video bit rate (bits/second)\n"
               "  -O --option opt=val     : set media format option to value\n"
-              "  -S --single-step        : video single frame at a time mode\n"
+              "  --single-step           : video single frame at a time mode\n"
               "  -c --crop               : crop rather than scale if resizing\n"
               "  -m --suppress-marker    : suppress marker bits to decoder\n"
               "  -M --force-marker       : force marker bits to decoder\n"
@@ -120,9 +173,12 @@ void CodecTest::Main()
               "  -S --simultanoues n     : Number of simultaneous encode/decode threads\n"
               "  -T --statistics         : output statistics files\n"
               "  -C --rate-control       : enable rate control\n"
+              "  -d --drop N             : randomly drop N% of encoded packets\n"
               "  --count n               : set number of frames to transcode\n"
               "  --noprompt              : do not prompt for commands, i.e. exit when input closes\n"
               "  --snr                   : calculate signal-to-noise ratio between input and output\n"
+              "  -i --info               : display per-frame info (use multiple times for more info)\n"
+              "  --list                  : list all available plugin codecs\n"
 #if PTRACING
               "  -o or --output file     : file name for output of log messages\n"       
               "  -t or --trace           : degree of verbosity in error log (more times for more detail)\n"     
@@ -132,40 +188,46 @@ void CodecTest::Main()
     return;
   }
 
+  g_infoCount = args.GetOptionCount('i');
+
   unsigned threadCount = args.GetOptionString('S').AsInteger();
   if (threadCount > 0) {
+    PList<TestThreadInfo> infos;
     unsigned i;
-    TestThreadInfo ** infos = (TestThreadInfo **)malloc(threadCount * sizeof(TestThreadInfo *));
     for (i = 0; i < threadCount; ++i) {
-      cout << "------------------------------" << endl
-           << "Starting thread " << i+1 << endl;
+      cout << "Creating thread " << i+1 << endl;
       TestThreadInfo * info = new TestThreadInfo(i+1);
-      if (!info->audio.Initialise(args) || !info->video.Initialise(args)) {
-        cout << "failed" << endl;
+      if (!info->Initialise(args)) {
+        cout << "Failed to create thread, aborting." << endl;
         return;
       }
-      info->audio.Resume();
-      info->video.Resume();
-      infos[i] = info;
+      infos.Append(info);
     }
-    cout << endl << "Tests running" << endl;
 
-    cout << "press return to stop" << endl;
+    cout << "Starting threads." << endl;
+    for (i = 0; i < threadCount; ++i)
+      infos[i].Start();
+
+    coutMutex.Wait();
+    cout << "\n\n\nTests running, press return to stop." << endl;
+    coutMutex.Signal();
+
     PCaselessString cmd;
     cin >> cmd;
 
-    for (i = 0; i < threadCount; ++i) {
-      infos[i]->audio.Stop();
-      infos[i]->video.Stop();
-    }
+    coutMutex.Wait();
+    cout << "\n\n\nStopping tests." << endl;
+    coutMutex.Signal();
+
+    for (i = 0; i < threadCount; ++i)
+      infos[i].Stop();
 
     return;
   }
 
-  AudioThread audio(0);
-  VideoThread video(0);
+  TestThreadInfo test(0);
 
-  if (!audio.Initialise(args) || !video.Initialise(args))
+  if (!test.Initialise(args))
     return;
 
 #if 0
@@ -178,14 +240,12 @@ void CodecTest::Main()
   }
 #endif
 
-  audio.Resume();
-  video.Resume();
+  test.Start();
 
   // command line
   if (args.HasOption("noprompt")) {
     cout << "Waiting for finish" << endl;
-    audio.WaitForTermination();
-    video.WaitForTermination();
+    test.Wait();
   }
   else {
     for (;;) {
@@ -201,47 +261,47 @@ void CodecTest::Main()
       if (cmd.NumCompare("n") == EqualTo) {
         int steps = cmd.Mid(1).AsUnsigned();
         do {
-          video.frameWait.Signal();
+          test.video.frameWait.Signal();
         } while (--steps > 0);
         continue;
       }
 
       if (cmd == "vfu") {
-        if (video.encoder == NULL)
+        if (test.video.encoder == NULL)
           cout << "\nNo video encoder running!" << endl;
         else
-          video.encoder->ExecuteCommand(OpalVideoUpdatePicture());
+          test.video.encoder->ExecuteCommand(OpalVideoUpdatePicture());
         continue;
       }
 
       if (cmd == "fg") {
-        if (video.grabber == NULL)
+        if (test.video.grabber == NULL)
           cout << "\nNo video grabber running!" << endl;
-        else if (!video.grabber->SetVFlipState(!video.grabber->GetVFlipState()))
+        else if (!test.video.grabber->SetVFlipState(!test.video.grabber->GetVFlipState()))
           cout << "\nCould not toggle Vflip state of video grabber device" << endl;
         continue;
       }
 
       if (cmd == "fd") {
-        if (video.display == NULL)
+        if (test.video.display == NULL)
           cout << "\nNo video display running!" << endl;
-        else if (!video.display->SetVFlipState(!video.display->GetVFlipState()))
+        else if (!test.video.display->SetVFlipState(!test.video.display->GetVFlipState()))
           cout << "\nCould not toggle Vflip state of video display device" << endl;
         continue;
       }
 
       unsigned width, height;
       if (PVideoFrameInfo::ParseSize(cmd, width, height)) {
-        video.pause.Signal();
-        if (video.grabber == NULL)
+        test.video.pause.Signal();
+        if (test.video.grabber == NULL)
           cout << "\nNo video grabber running!" << endl;
-        else if (!video.grabber->SetFrameSizeConverter(width, height))
+        else if (!test.video.grabber->SetFrameSizeConverter(width, height))
           cout << "Video grabber device could not be set to size " << width << 'x' << height << endl;
-        if (video.display == NULL)
+        if (test.video.display == NULL)
           cout << "\nNo video display running!" << endl;
-        else if (!video.display->SetFrameSizeConverter(width, height))
+        else if (!test.video.display->SetFrameSizeConverter(width, height))
           cout << "Video display device could not be set to size " << width << 'x' << height << endl;
-        video.resume.Signal();
+        test.video.resume.Signal();
         continue;
       }
 
@@ -257,8 +317,7 @@ void CodecTest::Main()
     } // end for
 
     cout << "Exiting." << endl;
-    audio.Stop();
-    video.Stop();
+    test.Stop();
   }
 }
 
@@ -456,6 +515,7 @@ bool AudioThread::Initialise(PArgList & args)
   }
 
   cout << "opened and initialised." << endl;
+  running = true;
 
   return true;
 }
@@ -650,32 +710,38 @@ bool VideoThread::Initialise(PArgList & args)
   }
   cout << "Target bit rate set to " << mediaFormat.GetOptionInteger(OpalVideoFormat::TargetBitRateOption()) << " bps" << endl;
 
-  if (args.HasOption('T')) {
-    frameFn  = "frame_stats.csv";
-    packetFn = "packet_stats.csv";
-  }
-
-  if (args.HasOption('C')) {
-    rcEnable = true;
-    unsigned rate = mediaFormat.GetOptionInteger(OpalVideoFormat::TargetBitRateOption());
-    rateController.Open(rate, 90000 / 10); // frameRate);
-    cout << "Video rate controller enabled for bit rate " << rate << " bps and frame rate " << frameRate << endl;
-  }
-
   SetOptions(args, mediaFormat);
 
-  if (encoder == NULL) 
-    frameTime = mediaFormat.GetFrameTime();
+  unsigned bitRate = mediaFormat.GetOptionInteger(OpalVideoFormat::TargetBitRateOption());
+  PString rc = args.GetOptionString('C');
+  if (rc.IsEmpty()) 
+    rateController = NULL;
   else {
+    rateController = PFactory<OpalVideoRateController>::CreateInstance(rc);
+    if (rateController != NULL) {
+      rateController->Open(mediaFormat);
+      cout << "Video rate controller enabled for bit rate " << bitRate << " bps and frame rate " << frameRate << endl;
+    }
+  }
+
+  if (args.HasOption('T'))
+    frameFn = "frame_stats.csv";
+
+  frameTime = mediaFormat.GetFrameTime();
+  if (encoder != NULL) {
     encoder->UpdateMediaFormats(OpalMediaFormat(), mediaFormat);
     if (args.HasOption('p'))
       encoder->SetMaxOutputSize(args.GetOptionString('p').AsUnsigned());
   }
 
-  singleStep = args.HasOption('S');
+  singleStep = args.HasOption("single-step");
 
   sumYSNR = sumCbSNR = sumCrSNR = 0.0;
   snrCount = 0;
+  running = true;
+
+  m_dropPercent = args.GetOptionString('d').AsInteger();
+  cout << "Dropping " << m_dropPercent << "% of encoded frames" << endl;
 
   return true;
 }
@@ -704,19 +770,37 @@ void VideoThread::Main()
 
   //decoder->SetCommandNotifier(PCREATE_NOTIFIER(OnTranscoderCommand));
 
-  forceIFrame = false;
+  m_forceIFrame = false;
   TranscoderThread::Main();
 }
 
 void TranscoderThread::OnTranscoderCommand(OpalMediaCommand & cmd, INT)
 {
   if (PIsDescendant(&cmd, OpalVideoUpdatePicture)) {
+    coutMutex.Wait();
     cout << "decoder lost sync" << endl;
-    forceIFrame = true;
+    coutMutex.Signal();
+    m_forceIFrame = true;
     ((OpalVideoTranscoder *)encoder)->ForceIFrame();
   }
   else {
+    coutMutex.Wait();
     cout << "unknown decoder command " << cmd.GetName() << endl;
+    coutMutex.Signal();
+  }
+}
+
+void VideoThread::CalcSNR(const RTP_DataFrame & dst)
+{
+  while (snrSourceFrames.size() > 0) {
+    const RTP_DataFrame * src = snrSourceFrames.front();
+    snrSourceFrames.pop();
+    if (src->GetTimestamp() == dst.GetTimestamp()) {
+      CalcSNR(*src, dst);
+      delete src;
+      return;
+    }
+    delete src;
   }
 }
 
@@ -768,86 +852,169 @@ void VideoThread::CalcSNR(const RTP_DataFrame & src, const RTP_DataFrame & dst)
 
 void VideoThread::ReportSNR()
 {
+  coutMutex.Wait();
+
   /* The PSNR is the mean of the sum of squares of the differences,
      normalized to the range 0..1
   */
   double const yPsnr = sumYSNR / snrCount;
 
+  cout << "Y  color component ";
   if (yPsnr <= 1e-9)
-    cout << "Y  color component identical" << endl;
+    cout << "identical";
   else
-    printf("Y  color component avg SNR : %.2f dB\n", 10 * log10(1/yPsnr));
+    cout << setprecision(2) << (10 * log10(1/yPsnr)) << " dB";
+  cout << endl;
 
   double const cbPsnr = sumCbSNR / snrCount;
 
+  cout << "Cb color component ";
   if (cbPsnr <= 1e-9)
-    cout << "Cb color component identical" << endl;
+    cout << "identical";
   else
-    printf("Cb color component avg SNR : %.2f dB\n", 10 * log10(1/cbPsnr));
+    cout << setprecision(2) << (10 * log10(1/cbPsnr)) << " dB";
+  cout << endl;
 
   double const crPsnr = sumCrSNR / snrCount;
 
+  cout << "Cr color component ";
   if (crPsnr <= 1e-9)
-    cout << "Cr color component identical" << endl;
+    cout << "identical";
   else
-    printf("Cr color component avg SNR : %.2f dB\n", 10 * log10(1/crPsnr));
+    cout << setprecision(2) << (10 * log10(1/crPsnr)) << " dB";
+  cout << endl;
+
+  coutMutex.Signal();
 }
 
+struct FrameHistoryEntry {
+  FrameHistoryEntry(unsigned size_, PInt64 timeStamp_, bool iFrame_)
+    : size(size_), timeStamp(timeStamp_), iFrame(iFrame_)
+  { }
+
+  unsigned size;
+  PInt64 timeStamp;
+  bool iFrame;
+};
 
 void TranscoderThread::Main()
 {
-  PUInt64 byteCount = 0;
-  PUInt64 frameCount = 0;
-  PUInt64 packetCount = 0;
-  WORD sequenceNumber = 0;
+  PUInt64 maximumBitRate = 0;
+  PUInt64 totalInputFrameCount = 0;
+  PUInt64 totalEncodedPacketCount = 0;
+  PUInt64 totalOutputFrameCount = 0;
+  PUInt64 totalEncodedByteCount = 0;
+  PUInt64 skippedFrames = 0;
+  PUInt64 totalDroppedPacketCount = 0;
+
   bool oldSrcState = true;
   bool oldOutState = true;
   bool oldEncState = true;
   bool oldDecState = true;
 
-  PTimeInterval startTick = PTimer::Tick();
-  while (running && framesToTranscode < 0 || (framesToTranscode-- > 0)) {
+  WORD sequenceNumber = 0;
 
-    RTP_DataFrame srcFrame(0);
-    bool state = Read(srcFrame);
-    if (oldSrcState != state) {
-      oldSrcState = state;
-      cerr << "Source " << (state ? "restor" : "fail") << "ed at frame " << frameCount << endl;
+  bool isVideo = PIsDescendant(encoder, OpalVideoTranscoder);
+
+  PTimeInterval startTick = PTimer::Tick();
+
+  if (isVideo) {
+    ((VideoThread *)this)->m_bitRateCalc.SetQuanta(frameTime/90);
+  }
+
+  //////////////////////////////////////////////
+  //
+  // main loop
+  //
+
+  RTP_DataFrame * srcFrame_ = NULL;
+
+  while ((running && framesToTranscode < 0) || (framesToTranscode-- > 0)) {
+
+    //////////////////////////////////////////////
+    //
+    //  acquire and format source frame
+    //
+    if (srcFrame_ != NULL)
+      delete srcFrame_;
+    srcFrame_ = new RTP_DataFrame(0);
+    RTP_DataFrame & srcFrame = *srcFrame_;
+
+    {
+      bool state = Read(srcFrame);
+      if (oldSrcState != state) {
+        oldSrcState = state;
+        cerr << "Source " << (state ? "restor" : "fail") << "ed at frame " << totalInputFrameCount << endl;
+        break;
+      }
     }
 
     srcFrame.SetTimestamp(timestamp);
     timestamp += frameTime;
+    ++totalInputFrameCount;
 
-    if (rcEnable && rateController.SkipFrame()) {
-      cerr << "Rate controller forced frame skip" << endl;
-      continue;
+    //////////////////////////////////////////////
+    //
+    //  allow the rate controller to skip frames
+    //
+    bool rateControlForceIFrame = false;
+    if (rateController != NULL && rateController->SkipFrame(rateControlForceIFrame)) {
+      coutMutex.Wait();
+      cout << "Packet pacer forced skip of input frame " << totalInputFrameCount-1 << endl;
+      coutMutex.Signal();
+      ++skippedFrames;
     }
 
-    RTP_DataFrameList encFrames;
-    if (encoder == NULL) 
-      encFrames.Append(new RTP_DataFrame(srcFrame)); 
     else {
-      state = encoder->ConvertFrames(srcFrame, encFrames);
-      if (oldEncState != state) {
-        oldEncState = state;
-        cerr << "Encoder " << (state ? "restor" : "fail") << "ed at frame " << frameCount << endl;
-        continue;
-      }
-      if (PIsDescendant(encoder, OpalVideoTranscoder) && ((OpalVideoTranscoder *)encoder)->WasLastFrameIFrame())
-        cerr << "Encoder returned I-Frame at frame " << frameCount << endl;
-    }
 
-    for (PINDEX i = 0; i < encFrames.GetSize(); i++)
-      encFrames[i].SetSequenceNumber(++sequenceNumber);
-
-    unsigned long frameSize = 0;
-    unsigned long framePacketCount = 0;
-    bool isIFrame = false;
-    for (PINDEX i = 0; i < encFrames.GetSize(); i++) {
-      RTP_DataFrameList outFrames;
-      if (encoder == NULL)
-        outFrames = encFrames;
+      //////////////////////////////////////////////
+      //
+      //  push frames through encoder
+      //
+      RTP_DataFrameList encFrames;
+      if (encoder == NULL) 
+        encFrames.Append(new RTP_DataFrame(srcFrame)); 
       else {
+        if (isVideo) {
+          if (m_forceIFrame) {
+            coutMutex.Wait();
+            cout << "Decoder forced I-frame at input frame " << totalInputFrameCount-1 << endl;
+            coutMutex.Signal();
+          }
+          if (rateControlForceIFrame) {
+            coutMutex.Wait();
+            cout << "Rate controller forced I-frame at input frame " << totalInputFrameCount-1 << endl;
+            coutMutex.Signal();
+          }
+          if (m_forceIFrame || rateControlForceIFrame)
+            ((OpalVideoTranscoder *)encoder)->ForceIFrame();
+        }
+
+        bool state = encoder->ConvertFrames(srcFrame, encFrames);
+        if (oldEncState != state) {
+          oldEncState = state;
+          cerr << "Encoder " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount-1 << endl;
+          continue;
+        }
+        if (isVideo && ((OpalVideoTranscoder *)encoder)->WasLastFrameIFrame()) {
+          coutMutex.Wait();
+          cout << "Encoder returned I-Frame at input frame " << totalInputFrameCount-1 << endl;
+          coutMutex.Signal();
+        }
+      }
+
+      //////////////////////////////////////////////
+      //
+      //  re-format encoded frames
+      //
+      unsigned long encodedPayloadSize = 0;
+      unsigned long encodedPacketCount = 0;
+      unsigned long encodedDataSize    = 0;
+      for (PINDEX i = 0; i < encFrames.GetSize(); i++) {
+        encFrames[i].SetSequenceNumber(++sequenceNumber);
+        ++encodedPacketCount;
+        encodedPayloadSize += encFrames[i].GetPayloadSize();
+        encodedDataSize    += encFrames[i].GetPayloadSize() + encFrames[i].GetHeaderSize();
         switch (markerHandling) {
           case SuppressMarkers :
             encFrames[i].SetMarker(false);
@@ -859,72 +1026,228 @@ void TranscoderThread::Main()
             break;
         }
 
-        state = decoder->ConvertFrames(encFrames[i], outFrames);
-
-        if (PIsDescendant(decoder, OpalVideoTranscoder) && ((OpalVideoTranscoder *)decoder)->WasLastFrameIFrame())
-          isIFrame = true;
-
-        if (oldDecState != state) {
-          oldDecState = state;
-          cerr << "Decoder " << (state ? "restor" : "fail") << "ed at packet " << packetCount << endl;
-          continue;
-        }
-        UpdateStats(encFrames[i]);
-      }
-
-      for (PINDEX j = 0; j < outFrames.GetSize(); j++) {
-
-        if (calcSNR) 
-          CalcSNR(srcFrame, outFrames[j]);
-
-        state = Write(outFrames[j]);
-        if (oldOutState != state) {
-          oldOutState = state;
-          cerr << "Frame display " << (state ? "restor" : "fail") << "ed at packet " << packetCount << endl;
+        if (g_infoCount > 0) {
+          RTP_DataFrame & rtp = encFrames[i];
+          coutMutex.Wait();
+          if (g_infoCount > 1) 
+            cout << "Inframe=" << totalInputFrameCount << ",outframe=#" << i << ":pt=" << rtp.GetPayloadType() << ",psz=" << rtp.GetPayloadSize() << ",m=" << (rtp.GetMarker() ? "1" : "0") << ",";
+          cout << "ssrc=" << hex << rtp.GetSyncSource() << dec << ",ts=" << rtp.GetTimestamp() << ",seq = " << rtp.GetSequenceNumber();
+          if (g_infoCount > 2) {
+            cout << "\n   data=";
+            cout << hex << setfill('0') << ::setw(2);
+            for (PINDEX i = 0; i < PMIN(10, rtp.GetPayloadSize()); ++i)
+              cout << (int)rtp.GetPayloadPtr()[i] << ' ';
+            cout << dec << setfill(' ') << ::setw(0);
+          }
+          cout << endl;
+          coutMutex.Signal();
         }
       }
-      byteCount += encFrames[i].GetPayloadSize();
-      frameSize += encFrames[i].GetPayloadSize() + encFrames[i].GetHeaderSize();
-      framePacketCount++;
-      packetCount++;
+
+      totalEncodedPacketCount += encFrames.GetSize();
+
+      if (isVideo && calcSNR) {
+        ((VideoThread *)this)->SaveSNRFrame(srcFrame_);
+        srcFrame_ = NULL;
+      }
+
+      //////////////////////////////////////////////
+      //
+      //  drop encoded frames if required
+      //
+      if (m_dropPercent > 0) {
+        PINDEX i = 0;
+        while (i < encFrames.GetSize()) {
+          int n = PRandom::Number() % 100; 
+          if (n >= m_dropPercent) 
+            i++;
+          else {
+            ++totalDroppedPacketCount;
+            encFrames.RemoveAt(i);
+          }
+        }
+      }
+
+      //////////////////////////////////////////////
+      //
+      //  push audio/video frames through NULL decoder
+      //
+      if (encoder == NULL) {
+        totalEncodedByteCount += encodedPayloadSize;
+        RTP_DataFrameList outFrames;
+        outFrames = encFrames;
+        if (outFrames.GetSize() != 1)
+          cerr << "NULL decoder returned != 1 output frame for input frame " << totalInputFrameCount-1 << endl;
+        else {
+          bool state = Write(outFrames[0]);
+          if (oldOutState != state) {
+            oldOutState = state;
+            cerr << "Output write " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount << endl;
+          }
+          totalOutputFrameCount++;
+        }
+      }
+
+      //////////////////////////////////////////////
+      //
+      //  push audio/video frames through explicit decoder
+      //
+      else if (rateController == NULL) {
+        totalEncodedByteCount += encodedPayloadSize;
+        if (isVideo)
+          ((VideoThread *)this)->CalcVideoPacketStats(encFrames, ((OpalVideoTranscoder *)decoder)->WasLastFrameIFrame());
+        RTP_DataFrameList outFrames;
+        for (PINDEX i = 0; i < encFrames.GetSize(); i++) {
+          bool state = decoder->ConvertFrames(encFrames[i], outFrames);
+          if (oldDecState != state) {
+            oldDecState = state;
+            cerr << "Decoder " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount-1 << endl;
+            continue;
+          }
+          if (outFrames.GetSize() > 1) 
+            cerr << "Non rate controlled video decoder returned != 1 output frame for input frame " << totalInputFrameCount-1 << endl;
+          else if (outFrames.GetSize() == 1) {
+            if (isVideo) {
+              if (((OpalVideoTranscoder *)decoder)->WasLastFrameIFrame()) {
+                coutMutex.Wait();
+                cout << "Decoder returned I-Frame at output frame " << totalOutputFrameCount << endl;
+                coutMutex.Signal();
+              } 
+              if (g_infoCount > 1) {
+                coutMutex.Wait();
+                cout << "Decoder generated payload of size " << outFrames[0].GetPayloadSize() << endl;
+                coutMutex.Signal();
+              }
+            }
+            bool state = Write(outFrames[0]);
+            if (oldOutState != state) {
+              oldOutState = state;
+              cerr << "Output write " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount << endl;
+            }
+            if (isVideo && calcSNR) 
+              ((VideoThread *)this)->CalcSNR(outFrames[0]);
+            totalOutputFrameCount++;
+          }
+        }
+      }
+
+      //////////////////////////////////////////////
+      //
+      //  push video frames into rate controller 
+      //
+      else 
+        rateController->Push(encFrames, ((OpalVideoTranscoder *)encoder)->WasLastFrameIFrame());
     }
 
-    if (isIFrame)
-      cerr << "Decoder returned I-Frame at frame " << frameCount << endl;
+    //////////////////////////////////////////////
+    //
+    //  pop video frames from rate controller and explicit decoder
+    //
+    if (rateController != NULL) {
+      for (;;) {
+        bool outIFrame;
+        RTP_DataFrameList pacedFrames;
+        if (!rateController->Pop(pacedFrames, outIFrame, false))
+          break;
+        ((VideoThread *)this)->CalcVideoPacketStats(pacedFrames, outIFrame);
+        RTP_DataFrameList outFrames;
+        for (PINDEX i = 0; i < pacedFrames.GetSize(); i++) {
+          totalEncodedByteCount += pacedFrames[i].GetPayloadSize();
+          bool state = decoder->ConvertFrames(pacedFrames[i], outFrames);
+          if (oldDecState != state) {
+            oldDecState = state;
+            cerr << "Decoder " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount-1 << endl;
+            continue;
+          }
+          if (outFrames.GetSize() > 1) 
+            cerr << "Rate controlled video decoder returned > 1 output frame for input frame " << totalInputFrameCount-1 << endl;
+          else if (outFrames.GetSize() == 1) {
+            if (((OpalVideoTranscoder *)decoder)->WasLastFrameIFrame()) {
+              coutMutex.Wait();
+              cout << "Decoder returned I-Frame at output frame " << totalOutputFrameCount << endl;
+              coutMutex.Signal();
+            }
+            bool state = Write(outFrames[0]);
+            if (oldOutState != state) {
+              oldOutState = state;
+              cerr << "Output write " << (state ? "restor" : "fail") << "ed at input frame " << totalInputFrameCount-1 << endl;
+            }
+            if (isVideo && calcSNR) 
+              ((VideoThread *)this)->CalcSNR(outFrames[0]);
+            totalOutputFrameCount++;
+          }
+        }
+      }
+    }
 
-    if (rcEnable)
-      rateController.AddFrame(frameSize, framePacketCount);
+    {
+      PTimeInterval duration = PTimer::Tick() - startTick;
+      PInt64 msecs = duration.GetMilliSeconds();
+      if (msecs < frameTime / 90)
+        msecs = frameTime / 90;
+      PUInt64 bitRate = totalEncodedByteCount*8*1000/msecs;
+      if (bitRate > maximumBitRate)
+        maximumBitRate = bitRate;
+    }
 
     if (pause.Wait(0)) {
       pause.Acknowledge();
       resume.Wait();
     }
-
-    frameCount++;
   }
 
+  //
+  //  end of main loop
+  //
+  ///////////////////////////////////////////////////////
+
   PTimeInterval duration = PTimer::Tick() - startTick;
 
-  cout << fixed << setprecision(1);
-  if (byteCount < 10000ULL)
-    cout << byteCount << ' ';
-  else if (byteCount < 10000000ULL)
-    cout << byteCount/1000.0 << " k";
-  else if (byteCount < 10000000000ULL)
-    cout << byteCount/1000000.0 << " M";
-  cout << "bytes, "
-       << packetCount << " packets, "
-       << frameCount << " frames over " << duration << " seconds at "
-       << (frameCount*1000.0/duration.GetMilliSeconds()) << " f/s and ";
+  PProcess::Times cpuTimes;
+  GetTimes(cpuTimes);
 
-  PUInt64 bitRate = byteCount*8/duration.GetSeconds();
-  if (bitRate < 10000ULL)
-    cout << bitRate << ' ';
-  else if (bitRate < 10000000ULL)
-    cout << bitRate/1000.0 << " k";
-  else if (bitRate < 10000000000ULL)
-    cout << bitRate/1000000.0 << " M";
-  cout << "bits/s." << endl;
+  coutMutex.Wait();
+
+  cout << fixed << setprecision(1);
+  if (totalEncodedByteCount < 10000ULL)
+    cout << totalEncodedByteCount << ' ';
+  else if (totalEncodedByteCount < 10000000ULL)
+    cout << totalEncodedByteCount/1000.0 << " k";
+  else if (totalEncodedByteCount < 10000000000ULL)
+    cout << totalEncodedByteCount /1000000.0 << " M";
+  cout << "bytes,"
+       << totalEncodedPacketCount << " packets,"
+       << totalInputFrameCount << " frames over " << duration << " seconds at "
+       << (totalInputFrameCount*1000.0/duration.GetMilliSeconds()) << " fps" << endl;
+
+  if (m_dropPercent > 0) {
+    cout << totalDroppedPacketCount << " dropped frames(" << totalDroppedPacketCount*100.0/totalEncodedPacketCount << "%)" << endl;
+  }
+
+  cout << "Average bit rate = ";
+  {
+    PInt64 msecs = duration.GetMilliSeconds();
+    if (msecs == 0) 
+      cout << "N/A";
+    else {
+      PUInt64 bitRate = totalEncodedByteCount*8*1000/msecs;
+      OUTPUT_BPS(cout, bitRate);
+    }
+  }
+
+  cout << ",max bit rate = ";
+  if (maximumBitRate == 0)
+    cout << "N/A";
+  else {
+    OUTPUT_BPS(cout, maximumBitRate);
+  }
+
+  cout << "," << skippedFrames << " skipped frames (" << (skippedFrames * 100.0)/totalInputFrameCount << "%)";
+  cout << endl;
+
+  cout << "CPU used: " << cpuTimes << endl;
+
+  coutMutex.Signal();
 
   if (calcSNR) 
     ReportSNR();
@@ -946,8 +1269,10 @@ bool AudioThread::Write(const RTP_DataFrame & frame)
 
 void AudioThread::Stop()
 {
-  running = false;
-  WaitForTermination();
+  if (running) {
+    running = false;
+    WaitForTermination();
+  }
 }
 
 
@@ -965,6 +1290,7 @@ bool VideoThread::Read(RTP_DataFrame & data)
   frame->x = frame->y = 0;
   frame->width = width;
   frame->height = height;
+  data.SetPayloadSize(sizeof(OpalVideoTranscoder::FrameHeader) + width*height*3/2);
 
   return grabber->GetFrameData(OPAL_VIDEO_FRAME_DATA_PTR(frame));
 }
@@ -982,37 +1308,57 @@ bool VideoThread::Write(const RTP_DataFrame & data)
 
 void VideoThread::Stop()
 {
-  running = false;
-  frameWait.Signal();
-  WaitForTermination();
+  if (running) {
+    running = false;
+    frameWait.Signal();
+    WaitForTermination();
+  }
 }
 
-void VideoThread::InitStats()
+
+void VideoThread::CalcVideoPacketStats(const RTP_DataFrameList & packets, bool isIFrame)
 {
-  packetCount= 0;
-  frameCount = 0;
-  frameBytes = 0;
-  totalFrameBytes = 0;
+  static unsigned maximumBitRate = 0;
+  static unsigned maximumAvgBitRate = 0;
+
+  for (PINDEX i = 0; i < packets.GetSize(); ++i) {
+    
+    //m_bitRateCalc.AddPacket(packets[i].GetPayloadSize());
+
+    if (rateController != NULL) {
+      OpalBitRateCalculator & m_bitRateCalc = rateController->m_bitRateCalc;
+
+      unsigned r = m_bitRateCalc.GetBitRate();
+      if (r > maximumBitRate)
+        maximumBitRate = r;
+
+      unsigned a = m_bitRateCalc.GetAverageBitRate();
+      if (a > maximumAvgBitRate)
+        maximumAvgBitRate = a;
+
+      PStringStream str;
+      str << "index=" << (packets[i].GetTimestamp() / 3600)
+          << ",ps=" << packets[i].GetPayloadSize()
+          << ",rate=" << r
+          << ",avg=" << a 
+          << ",maxrate=" << maximumBitRate
+          << ",maxavg=" << maximumAvgBitRate 
+          << ",f=" << (isIFrame?"I":"") << (packets[i].GetMarker()?"M":"");
+      ((VideoThread *)this)->WriteFrameStats(str);
+    }
+  }
 }
 
-void VideoThread::UpdateStats(const RTP_DataFrame & frame)
+
+void VideoThread::WriteFrameStats(const PString & str)
 {
-  if (!packetFn.IsEmpty()) {
-    if (packetStatFile.Open(packetFn, PFile::WriteOnly))
-      packetStatFile << "PacketNum,PacketSize\n";
-    packetFn.MakeEmpty();
+  if (!frameFn.IsEmpty()) {
+    frameStatFile.Open(frameFn, PFile::WriteOnly);
+    frameFn.MakeEmpty();
   }
 
-  packetCount++;
-  frameBytes      += frame.GetPayloadSize();
-  totalFrameBytes += frame.GetPayloadSize();
-
-  if (packetStatFile.IsOpen()) 
-    packetStatFile << packetCount << ','
-                   << frame.GetPayloadSize() << endl;
-
-  if (frame.GetMarker()) 
-    UpdateFrameStats();
+  if (frameStatFile.IsOpen()) 
+    frameStatFile << str << endl;
 }
 
 PInt64 BpsTokbps(PInt64 Bps)
@@ -1020,24 +1366,4 @@ PInt64 BpsTokbps(PInt64 Bps)
   return ((Bps * 8) + 500) / 1000;
 }
 
-void VideoThread::UpdateFrameStats()
-{
-  if (!frameFn.IsEmpty()) {
-    if (frameStatFile.Open(frameFn, PFile::WriteOnly))
-      frameStatFile << "FrameNum,FrameBits,FrameKBPS,TotalKBPS\n";
-    frameFn.MakeEmpty();
-  }
-
-  frameCount++;
-
-  if (frameStatFile.IsOpen()) {
-    frameStatFile << frameCount << ','
-                  << (frameBytes * 8) << ','
-                  << BpsTokbps(frameBytes * frameRate) << ','
-                  << BpsTokbps((totalFrameBytes * frameRate) / frameCount) << endl;
-  }
-
-  frameBytes = 0;
-}
-
 // End of File ///////////////////////////////////////////////////////////////
