@@ -21,9 +21,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21163 $
+ * $Revision: 22254 $
  * $Author: rjongbloed $
- * $Date: 2008-09-25 06:36:06 +0000 (Thu, 25 Sep 2008) $
+ * $Date: 2009-03-23 19:40:36 -0500 (Mon, 23 Mar 2009) $
  */
 
 #ifndef _IvrOPAL_MAIN_H
@@ -41,6 +41,18 @@ class MyManager : public OpalManager
 };
 
 
+class MySIPEndPoint : public SIPEndPoint
+{
+    PCLASSINFO(MySIPEndPoint, SIPEndPoint)
+
+  public:
+    MySIPEndPoint(OpalManager & manager) : SIPEndPoint(manager) { }
+    void OnRegistrationStatus(const RegistrationStatus & status);
+
+    PSyncPoint m_completed;
+};
+
+
 class IvrOPAL : public PProcess
 {
     PCLASSINFO(IvrOPAL, PProcess)
@@ -50,6 +62,7 @@ class IvrOPAL : public PProcess
     ~IvrOPAL();
 
     virtual void Main();
+    virtual bool OnInterrupt(bool);
 
   private:
     MyManager * m_manager;
