@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 20741 $
+ * $Revision: 22444 $
  * $Author: rjongbloed $
- * $Date: 2008-08-27 04:18:35 +0000 (Wed, 27 Aug 2008) $
+ * $Date: 2009-04-20 18:49:06 -0500 (Mon, 20 Apr 2009) $
  */
 
 #include <ptlib.h>
@@ -43,9 +43,6 @@
 #include <codec/rfc4175.h>
 #include <codec/opalplugin.h>
 
-namespace PWLibStupidLinkerHacks {
-  int rfc4175Loader;
-};
 
 #define   FRAME_WIDTH   1920
 #define   FRAME_HEIGHT  1080
@@ -70,7 +67,6 @@ const OpalVideoFormat & GetOpalRFC4175_YCbCr420()
   return RFC4175_YCbCr420;
 }
 
-OPAL_REGISTER_RFC4175_VIDEO(YUV420P, YCbCr420)
 
 const OpalVideoFormat & GetOpalRFC4175_RGB()
 {
@@ -78,7 +74,6 @@ const OpalVideoFormat & GetOpalRFC4175_RGB()
   return RFC4175_RGB;
 }
 
-OPAL_REGISTER_RFC4175_VIDEO(RGB24, RGB)
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -266,11 +261,10 @@ void OpalRFC4175Encoder::AddNewDstFrame()
   FinishOutputFrame();
 
   // allocate a new output frame
-  RTP_DataFrame * frame = new RTP_DataFrame;
+  RTP_DataFrame * frame = new RTP_DataFrame(maximumPacketSize - RTP_DataFrame::MinHeaderSize);
   dstFrames->Append(frame);
 
   // initialise payload size for maximum size
-  frame->SetPayloadSize(maximumPacketSize - frame->GetHeaderSize());
   frame->SetPayloadType(outputMediaFormat.GetPayloadType());
   // initialise current output scanline count;
   dstScanLineCount = 0;
@@ -503,11 +497,10 @@ PBoolean Opal_RFC4175YCbCr420_to_YUV420P::DecodeFrames(RTP_DataFrameList & outpu
   PTRACE(4, "RFC4175\tDecoding output from " << inputFrames.GetSize() << " input frames");
 
   // allocate destination frame
-  output.Append(new RTP_DataFrame());
+  output.Append(new RTP_DataFrame(sizeof(PluginCodec_Video_FrameHeader) + PixelsToBytes(frameWidth*frameHeight)));
   RTP_DataFrame & outputFrame = output.back();
   outputFrame.SetMarker(PTrue);
   outputFrame.SetPayloadType(outputMediaFormat.GetPayloadType());
-  outputFrame.SetPayloadSize(sizeof(PluginCodec_Video_FrameHeader) + PixelsToBytes(frameWidth*frameHeight));
 
   // get pointer to header and payload
   PluginCodec_Video_FrameHeader * hdr = (PluginCodec_Video_FrameHeader *)outputFrame.GetPayloadPtr();
@@ -622,10 +615,9 @@ PBoolean Opal_RFC4175RGB_to_RGB24::DecodeFrames(RTP_DataFrameList & output)
   PTRACE(4, "RFC4175\tDecoding output from " << inputFrames.GetSize() << " input frames");
 
   // allocate destination frame
-  output.Append(new RTP_DataFrame());
+  output.Append(new RTP_DataFrame(sizeof(PluginCodec_Video_FrameHeader) + PixelsToBytes(frameWidth*frameHeight)));
   RTP_DataFrame & outputFrame = output.back();
   outputFrame.SetMarker(PTrue);
-  outputFrame.SetPayloadSize(sizeof(PluginCodec_Video_FrameHeader) + PixelsToBytes(frameWidth*frameHeight));
 
   // get pointer to header and payload
   PluginCodec_Video_FrameHeader * hdr = (PluginCodec_Video_FrameHeader *)outputFrame.GetPayloadPtr();
