@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21543 $
+ * $Revision: 23566 $
  * $Author: rjongbloed $
- * $Date: 2008-11-18 03:30:26 +0000 (Tue, 18 Nov 2008) $
+ * $Date: 2009-10-01 01:06:00 -0500 (Thu, 01 Oct 2009) $
  */
 
 #include <ptlib.h>
@@ -46,6 +46,7 @@
 #include <h323/h323con.h>
 #include <h323/h323pdu.h>
 #include <h323/transaddr.h>
+#include <t38/h323t38.h>
 
 
 #define DEFINE_G711_CAPABILITY(cls, code, capName) \
@@ -55,11 +56,15 @@ class cls : public H323_G711Capability { \
 }; \
 H323_REGISTER_CAPABILITY(cls, capName) \
 
-#ifndef NO_H323_AUDIO_CODECS
 
+#ifndef NO_H323_AUDIO_CODECS
 DEFINE_G711_CAPABILITY(H323_G711ALaw64Capability, H323_G711Capability::ALaw, OPAL_G711_ALAW_64K)
 DEFINE_G711_CAPABILITY(H323_G711uLaw64Capability, H323_G711Capability::muLaw, OPAL_G711_ULAW_64K)
+#endif
 
+
+#if OPAL_T38_CAPABILITY
+H323_REGISTER_CAPABILITY(H323_T38Capability, OPAL_T38);
 #endif
 
 
@@ -164,7 +169,8 @@ PBoolean H323Capability::IsMatch(const PASN_Choice & subTypePDU) const
 
 PBoolean H323Capability::OnSendingPDU(H245_DataType & /*pdu*/) const
 {
-  return mediaFormat.ToCustomisedOptions();
+  GetWritableMediaFormat().SetOptionString(OpalMediaFormat::ProtocolOption(), "H.323");
+  return m_mediaFormat.ToCustomisedOptions();
 }
 
 
@@ -200,13 +206,15 @@ PBoolean H323Capability::OnReceivedPDU(const H245_Capability & cap)
       capabilityDirection = e_NoDirection;
   }
 
-  return PTrue;
+  GetWritableMediaFormat().SetOptionString(OpalMediaFormat::ProtocolOption(), "H.323");
+  return m_mediaFormat.ToNormalisedOptions();
 }
 
 
 PBoolean H323Capability::OnReceivedPDU(const H245_DataType & /*pdu*/, PBoolean /*receiver*/)
 {
-  return mediaFormat.ToNormalisedOptions();
+  GetWritableMediaFormat().SetOptionString(OpalMediaFormat::ProtocolOption(), "H.323");
+  return m_mediaFormat.ToNormalisedOptions();
 }
 
 
@@ -218,35 +226,26 @@ PBoolean H323Capability::IsUsable(const H323Connection &) const
 
 OpalMediaFormat H323Capability::GetMediaFormat() const
 {
-  return mediaFormat.IsValid() ? mediaFormat : OpalMediaFormat(GetFormatName());
+  return m_mediaFormat.IsValid() ? m_mediaFormat : OpalMediaFormat(GetFormatName());
 }
 
 
-bool H323Capability::SetMediaFormatOptions(const OpalMediaFormat &format)
+bool H323Capability::UpdateMediaFormat(const OpalMediaFormat &format)
 {
+  OpalMediaFormat & mediaFormat = GetWritableMediaFormat();
   if (mediaFormat != format)
     return false;
 
-  for (PINDEX i = 0; i < format.GetOptionCount(); i++) {
-    PString name = format.GetOption(i).GetName();
-    //  Find the src format's option in the dst format's option list
-    OpalMediaOption *option = mediaFormat.FindOption(name);
-    if (option != NULL) {   //  Option not found in destination format
-      PString value;
-      format.GetOptionValue(name, value);
-      mediaFormat.SetOptionValue(name, value);
-    }
-  }
-
+  mediaFormat = format;
   return true;
 }
 
 
-OpalMediaFormat & H323Capability::GetWritableMediaFormat()
+OpalMediaFormat & H323Capability::GetWritableMediaFormat() const
 {
-  if (!mediaFormat.IsValid())
-    mediaFormat = GetFormatName();
-  return mediaFormat;
+  if (!m_mediaFormat.IsValid())
+    m_mediaFormat = GetFormatName();
+  return m_mediaFormat;
 }
 
 
@@ -497,25 +496,10 @@ PObject::Comparison H323NonStandardCapabilityInfo::CompareData(const PBYTEArray
 
 /////////////////////////////////////////////////////////////////////////////
 
-H323GenericCapabilityInfo::H323GenericCapabilityInfo(const PString & standardId, PINDEX maxBitRate)
-  : maxBitRate(maxBitRate)
+H323GenericCapabilityInfo::H323GenericCapabilityInfo(const PString & standardId, unsigned bitRate)
+  : m_identifier(standardId)
+  , maxBitRate(bitRate)
 {
-  identifier = new H245_CapabilityIdentifier(H245_CapabilityIdentifier::e_standard);
-  PASN_ObjectId & object_id = *identifier;
-  object_id = standardId;
-}
-
-
-H323GenericCapabilityInfo::H323GenericCapabilityInfo(const H323GenericCapabilityInfo & obj)
-  : maxBitRate(obj.maxBitRate)
-{
-  identifier = new H245_CapabilityIdentifier(*obj.identifier);
-}
-
-
-H323GenericCapabilityInfo::~H323GenericCapabilityInfo()
-{
-  delete identifier;
 }
 
 
@@ -523,7 +507,7 @@ PBoolean H323GenericCapabilityInfo::OnSendingGenericPDU(H245_GenericCapability &
                                                     const OpalMediaFormat & mediaFormat,
                                                     H323Capability::CommandType type) const
 {
-  pdu.m_capabilityIdentifier = *identifier;
+  H323SetCapabilityIdentifier(m_identifier, pdu.m_capabilityIdentifier);
 
   unsigned bitRate = maxBitRate != 0 ? maxBitRate : ((mediaFormat.GetBandwidth()+99)/100);
   if (bitRate != 0) {
@@ -551,20 +535,13 @@ PBoolean H323GenericCapabilityInfo::OnSendingGenericPDU(H245_GenericCapability &
         break;
     }
 
-    H245_GenericParameter * param = new H245_GenericParameter;
+    H245_ArrayOf_GenericParameter & params =
+            genericInfo.mode == OpalMediaOption::H245GenericInfo::Collapsing ? pdu.m_collapsing : pdu.m_nonCollapsing;
 
-    param->m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
-    (PASN_Integer &)param->m_parameterIdentifier = genericInfo.ordinal;
-
-    if (PIsDescendant(&option, OpalMediaOptionBoolean)) {
-      if (!((const OpalMediaOptionBoolean &)option).GetValue()) {
-        delete param;
-        continue; // Do not include a logical at all if it is false
-      }
-      param->m_parameterValue.SetTag(H245_ParameterValue::e_logical);
-    }
+    if (PIsDescendant(&option, OpalMediaOptionBoolean))
+      H323AddGenericParameterBoolean(params, genericInfo.ordinal, ((const OpalMediaOptionBoolean &)option).GetValue());
     else if (PIsDescendant(&option, OpalMediaOptionUnsigned)) {
-      unsigned tag;
+      H245_ParameterValue::Choices tag;
       switch (genericInfo.integerType) {
         default :
         case OpalMediaOption::H245GenericInfo::UnsignedInt :
@@ -580,28 +557,20 @@ PBoolean H323GenericCapabilityInfo::OnSendingGenericPDU(H245_GenericCapability &
           break;
       }
 
-      param->m_parameterValue.SetTag(tag);
-      (PASN_Integer &)param->m_parameterValue = ((const OpalMediaOptionUnsigned &)option).GetValue();
-    }
-    else {
-      param->m_parameterValue.SetTag(H245_ParameterValue::e_octetString);
-      PASN_OctetString & octetString = param->m_parameterValue;
-      if (PIsDescendant(&option, OpalMediaOptionOctets))
-        octetString = ((const OpalMediaOptionOctets &)option).GetValue();
-      else
-        octetString = option.AsString();
-    }
-
-    if (genericInfo.mode == OpalMediaOption::H245GenericInfo::Collapsing) {
-      pdu.IncludeOptionalField(H245_GenericCapability::e_collapsing);
-      pdu.m_collapsing.Append(param);
-    }
-    else {
-      pdu.IncludeOptionalField(H245_GenericCapability::e_nonCollapsing);
-      pdu.m_nonCollapsing.Append(param);
+      H323AddGenericParameterInteger(params, genericInfo.ordinal, ((const OpalMediaOptionUnsigned &)option).GetValue(), tag);
     }
+    else if (PIsDescendant(&option, OpalMediaOptionOctets))
+      H323AddGenericParameterOctets(params, genericInfo.ordinal, ((const OpalMediaOptionOctets &)option).GetValue());
+    else
+      H323AddGenericParameterString(params, genericInfo.ordinal, option.AsString());
   }
 
+  if (pdu.m_collapsing.GetSize() > 0)
+    pdu.IncludeOptionalField(H245_GenericCapability::e_collapsing);
+
+  if (pdu.m_nonCollapsing.GetSize() > 0)
+    pdu.IncludeOptionalField(H245_GenericCapability::e_nonCollapsing);
+
   return PTrue;
 }
 
@@ -609,8 +578,8 @@ PBoolean H323GenericCapabilityInfo::OnReceivedGenericPDU(OpalMediaFormat & media
                                                      const H245_GenericCapability & pdu,
                                                      H323Capability::CommandType type)
 {
-  if (pdu.m_capabilityIdentifier != *identifier)
-    return PFalse;
+  if (H323GetCapabilityIdentifier(pdu.m_capabilityIdentifier) != m_identifier)
+    return false;
 
   if (pdu.HasOptionalField(H245_GenericCapability::e_maxBitRate)) {
     maxBitRate = pdu.m_maxBitRate;
@@ -637,68 +606,65 @@ PBoolean H323GenericCapabilityInfo::OnReceivedGenericPDU(OpalMediaFormat & media
         break;
     }
 
-    const H245_ArrayOf_GenericParameter * params;
+    const H245_ParameterValue * param;
     if (genericInfo.mode == OpalMediaOption::H245GenericInfo::Collapsing) {
       if (!pdu.HasOptionalField(H245_GenericCapability::e_collapsing))
         continue;
-      params = &pdu.m_collapsing;
+      param = H323GetGenericParameter(pdu.m_collapsing, genericInfo.ordinal);
     }
     else {
       if (!pdu.HasOptionalField(H245_GenericCapability::e_nonCollapsing))
         continue;
-      params = &pdu.m_nonCollapsing;
+      param = H323GetGenericParameter(pdu.m_nonCollapsing, genericInfo.ordinal);
     }
 
     if (PIsDescendant(&option, OpalMediaOptionBoolean))
       ((OpalMediaOptionBoolean &)option).SetValue(false);
 
-    for (PINDEX j = 0; j < params->GetSize(); j++) {
-      const H245_GenericParameter & param = (*params)[j];
-      if (param.m_parameterIdentifier.GetTag() == H245_ParameterIdentifier::e_standard &&
-                         (const PASN_Integer &)param.m_parameterIdentifier == genericInfo.ordinal) {
-        if (PIsDescendant(&option, OpalMediaOptionBoolean)) {
-          if (param.m_parameterValue.GetTag() == H245_ParameterValue::e_logical) {
-            ((OpalMediaOptionBoolean &)option).SetValue(true);
-            break;
-          }
-        }
-        else if (PIsDescendant(&option, OpalMediaOptionUnsigned)) {
-          unsigned tag;
-          switch (genericInfo.integerType) {
-            default :
-            case OpalMediaOption::H245GenericInfo::UnsignedInt :
-              tag = option.GetMerge() == OpalMediaOption::MinMerge ? H245_ParameterValue::e_unsignedMin : H245_ParameterValue::e_unsignedMax;
-              break;
- 
-            case OpalMediaOption::H245GenericInfo::Unsigned32 :
-              tag = option.GetMerge() == OpalMediaOption::MinMerge ? H245_ParameterValue::e_unsigned32Min : H245_ParameterValue::e_unsigned32Max;
-              break;
- 
-            case OpalMediaOption::H245GenericInfo::BooleanArray :
-              tag = H245_ParameterValue::e_booleanArray;
-              break;
-          }
- 
-          if (param.m_parameterValue.GetTag() == tag) {
-            ((OpalMediaOptionUnsigned &)option).SetValue((const PASN_Integer &)param.m_parameterValue);
-            break;
-          }
-        }
-        else {
-          if (param.m_parameterValue.GetTag() == H245_ParameterValue::e_octetString) {
-            const PASN_OctetString & octetString = param.m_parameterValue;
-            if (PIsDescendant(&option, OpalMediaOptionOctets))
-              ((OpalMediaOptionOctets &)option).SetValue(octetString);
-            else
-              ((OpalMediaOption &)option).FromString(octetString.AsString());
-            break;
-          }
-        }
+    if (param == NULL)
+      continue;
 
-        PTRACE(2, "Invalid generic parameter type (" << param.m_parameterValue.GetTagName()
-               << ") for option \"" << option.GetName() << "\" (" << option.GetClass() << ')');
+    if (PIsDescendant(&option, OpalMediaOptionBoolean)) {
+      if (param->GetTag() == H245_ParameterValue::e_logical) {
+        ((OpalMediaOptionBoolean &)option).SetValue(true);
+        break;
       }
     }
+    else if (PIsDescendant(&option, OpalMediaOptionUnsigned)) {
+      unsigned tag;
+      switch (genericInfo.integerType) {
+        default :
+        case OpalMediaOption::H245GenericInfo::UnsignedInt :
+          tag = option.GetMerge() == OpalMediaOption::MinMerge ? H245_ParameterValue::e_unsignedMin : H245_ParameterValue::e_unsignedMax;
+          break;
+
+        case OpalMediaOption::H245GenericInfo::Unsigned32 :
+          tag = option.GetMerge() == OpalMediaOption::MinMerge ? H245_ParameterValue::e_unsigned32Min : H245_ParameterValue::e_unsigned32Max;
+          break;
+
+        case OpalMediaOption::H245GenericInfo::BooleanArray :
+          tag = H245_ParameterValue::e_booleanArray;
+          break;
+      }
+
+      if (param->GetTag() == tag) {
+        ((OpalMediaOptionUnsigned &)option).SetValue((const PASN_Integer &)*param);
+        break;
+      }
+    }
+    else {
+      if (param->GetTag() == H245_ParameterValue::e_octetString) {
+        const PASN_OctetString & octetString = *param;
+        if (PIsDescendant(&option, OpalMediaOptionOctets))
+          ((OpalMediaOptionOctets &)option).SetValue(octetString);
+        else
+          ((OpalMediaOption &)option).FromString(octetString.AsString());
+        break;
+      }
+    }
+
+    PTRACE(2, "H323\tInvalid generic parameter type (" << param->GetTagName()
+           << ") for option \"" << option.GetName() << "\" (" << option.GetClass() << ')');
   }
 
   return PTrue;
@@ -706,12 +672,12 @@ PBoolean H323GenericCapabilityInfo::OnReceivedGenericPDU(OpalMediaFormat & media
 
 PBoolean H323GenericCapabilityInfo::IsMatch(const H245_GenericCapability & param) const
 {
-  return param.m_capabilityIdentifier == *identifier;
+  return H323GetCapabilityIdentifier(param.m_capabilityIdentifier) == m_identifier;
 }
 
 PObject::Comparison H323GenericCapabilityInfo::CompareInfo(const H323GenericCapabilityInfo & obj) const
 {
-  return identifier->Compare(*obj.identifier);
+  return m_identifier.Compare(obj.m_identifier);
 }
 
 
@@ -832,8 +798,6 @@ PBoolean H323AudioCapability::OnSendingPDU(H245_AudioMode & pdu) const
 
 PBoolean H323AudioCapability::OnReceivedPDU(const H245_Capability & cap)
 {
-  H323Capability::OnReceivedPDU(cap);
-
   if (cap.GetTag() != H245_Capability::e_receiveAudioCapability &&
       cap.GetTag() != H245_Capability::e_receiveAndTransmitAudioCapability)
     return PFalse;
@@ -854,7 +818,7 @@ PBoolean H323AudioCapability::OnReceivedPDU(const H245_Capability & cap)
            << txFramesInPacket << " as remote allows " << packetSize);
   }
 
-  return PTrue;
+  return H323Capability::OnReceivedPDU(cap);
 }
 
 
@@ -1075,9 +1039,18 @@ PBoolean H323VideoCapability::OnSendingPDU(H245_VideoCapability & /*pdu*/) const
 }
 
 
-PBoolean H323VideoCapability::OnSendingPDU(H245_VideoCapability & pdu, CommandType) const
+PBoolean H323VideoCapability::OnSendingPDU(H245_VideoCapability & pdu, CommandType type) const
 {
+#if OPAL_H239
+  PINDEX role = 0;
+  if (type != e_OLC || (role = GetMediaFormat().GetOptionEnum(OpalVideoFormat::ContentRoleOption())) == 0)
+    return OnSendingPDU(pdu);
+
+  H323H239VideoCapability h239(GetMediaFormat());
+  return h239.OnSendingPDU(pdu, type);
+#else
   return OnSendingPDU(pdu);
+#endif
 }
 
 
@@ -1090,20 +1063,18 @@ PBoolean H323VideoCapability::OnSendingPDU(H245_ModeElement & mode) const
 
 PBoolean H323VideoCapability::OnReceivedPDU(const H245_Capability & cap)
 {
-  H323Capability::OnReceivedPDU(cap);
-
   if (cap.GetTag() != H245_Capability::e_receiveVideoCapability &&
       cap.GetTag() != H245_Capability::e_receiveAndTransmitVideoCapability)
     return PFalse;
 
-  return OnReceivedPDU((const H245_VideoCapability &)cap, e_TCS);
+  return OnReceivedPDU((const H245_VideoCapability &)cap, e_TCS) && H323Capability::OnReceivedPDU(cap);
 }
 
 
 PBoolean H323VideoCapability::OnReceivedPDU(const H245_DataType & dataType, PBoolean receiver)
 {
   if (dataType.GetTag() != H245_DataType::e_videoData) {
-    PTRACE(5, "dataType.GetTag() " << dataType.GetTag() << " != H245_DataType::e_videoData");
+    PTRACE(5, "H323\tdataType.GetTag() " << dataType.GetTag() << " != H245_DataType::e_videoData");
     return PFalse;
   }
 
@@ -1260,6 +1231,312 @@ PBoolean H323GenericVideoCapability::IsMatch(const PASN_Choice & subTypePDU) con
 }
 
 
+/////////////////////////////////////////////////////////////////////////////
+
+#if OPAL_H239
+
+H323ExtendedVideoCapability::H323ExtendedVideoCapability(const PString & identifier)
+  : H323GenericVideoCapability(identifier)
+{
+}
+
+
+unsigned H323ExtendedVideoCapability::GetSubType() const
+{
+  return H245_VideoCapability::e_extendedVideoCapability;
+}
+
+
+PBoolean H323ExtendedVideoCapability::OnSendingPDU(H245_VideoCapability & pdu, CommandType type) const
+{
+  pdu.SetTag(H245_VideoCapability::e_extendedVideoCapability);
+  H245_ExtendedVideoCapability & extcap = pdu;
+
+  unsigned roleMask = UINT_MAX;
+
+  for (OpalMediaFormatList::const_iterator videoFormat = m_videoFormats.begin(); videoFormat != m_videoFormats.end(); ++videoFormat) {
+    if (videoFormat->GetMediaType() == OpalMediaType::Video()) {
+      H323Capability * capability = H323Capability::Create(videoFormat->GetName());
+      if (capability != NULL) {
+        capability->UpdateMediaFormat(*videoFormat);
+        H245_Capability h245Cap;
+        if (capability->OnSendingPDU(h245Cap)) {
+          PINDEX size = extcap.m_videoCapability.GetSize();
+          extcap.m_videoCapability.SetSize(size+1);
+          extcap.m_videoCapability[size] = h245Cap;
+          roleMask &= videoFormat->GetOptionInteger(OpalVideoFormat::ContentRoleMaskOption());
+        }
+        delete capability;
+      }
+    }
+  }
+  if (extcap.m_videoCapability.GetSize() == 0) {
+    PTRACE(2, "H323\tCannot encode H.239 video capability, no extended video codecs available");
+    return false;
+  }
+
+  OpalMediaFormat videoCapExtMediaFormat(GetFormatName());
+  if ((roleMask&0xfffc) != 0)
+    roleMask = (roleMask&3)|2;
+  videoCapExtMediaFormat.SetOptionInteger(OpalVideoFormat::ContentRoleMaskOption(), roleMask);
+
+  extcap.IncludeOptionalField(H245_ExtendedVideoCapability::e_videoCapabilityExtension);
+  extcap.m_videoCapabilityExtension.SetSize(1);
+  return OnSendingGenericPDU(extcap.m_videoCapabilityExtension[0], OpalMediaFormat(GetFormatName()), type);
+}
+
+
+PBoolean H323ExtendedVideoCapability::OnSendingPDU(H245_VideoMode &) const
+{
+  return false;
+}
+
+
+PBoolean H323ExtendedVideoCapability::OnReceivedPDU(const H245_VideoCapability & pdu, CommandType type)
+{
+  if (pdu.GetTag() != H245_VideoCapability::e_extendedVideoCapability)
+    return false;
+
+  const H245_ExtendedVideoCapability & extcap = pdu;
+  if (!extcap.HasOptionalField(H245_ExtendedVideoCapability::e_videoCapabilityExtension)) {
+    PTRACE(2, "H323\tNo H.239 video capability extension");
+    return false;
+  }
+
+  PINDEX i = 0;
+  for (;;) {
+    if (i >= extcap.m_videoCapabilityExtension.GetSize()) {
+      PTRACE(2, "H323\tNo H.239 video capability extension for " << m_identifier);
+      return false;
+    }
+
+    if (H323GenericCapabilityInfo::IsMatch(extcap.m_videoCapabilityExtension[i]))
+      break;
+
+    ++i;
+  }
+
+  OpalMediaFormat videoCapExtMediaFormat(GetFormatName());
+  if (!OnReceivedGenericPDU(videoCapExtMediaFormat, extcap.m_videoCapabilityExtension[i], type))
+    return false;
+
+  unsigned roleMask = videoCapExtMediaFormat.GetOptionInteger(OpalVideoFormat::ContentRoleMaskOption());
+
+  H323Capabilities allVideoCapabilities;
+  H323CapabilityFactory::KeyList_T stdCaps = H323CapabilityFactory::GetKeyList();
+
+  for (H323CapabilityFactory::KeyList_T::const_iterator iterCap = stdCaps.begin(); iterCap != stdCaps.end(); ++iterCap) {
+    H323Capability * capability = H323Capability::Create(*iterCap);
+    if (capability->GetMainType() == H323Capability::e_Video)
+      allVideoCapabilities.Add(capability);
+    else
+      delete capability;
+  }
+
+  m_videoFormats.RemoveAll();
+
+  for (i = 0; i < extcap.m_videoCapability.GetSize(); ++i) {
+    const H245_VideoCapability & vidCap = extcap.m_videoCapability[i];
+    for (PINDEX c = 0; c < allVideoCapabilities.GetSize(); c++) {
+      H323VideoCapability & capability = (H323VideoCapability &)allVideoCapabilities[c];
+      if (capability.IsMatch(vidCap) && capability.OnReceivedPDU(vidCap, type)) {
+        OpalMediaFormat mediaFormat = capability.GetMediaFormat();
+        mediaFormat.SetOptionInteger(OpalVideoFormat::ContentRoleMaskOption(), roleMask);
+        if (mediaFormat.ToNormalisedOptions())
+          m_videoFormats += mediaFormat;
+      }
+    }
+  }
+
+  PTRACE(4, "H323\tExtended video: " << setfill(',') << m_videoFormats);
+  return !m_videoFormats.IsEmpty();
+}
+
+
+PBoolean H323ExtendedVideoCapability::IsMatch(const PASN_Choice & subTypePDU) const
+{
+  if (!H323Capability::IsMatch(subTypePDU))
+    return false;
+
+  const H245_ExtendedVideoCapability & extcap = (const H245_ExtendedVideoCapability &)subTypePDU.GetObject();
+  if (!extcap.HasOptionalField(H245_ExtendedVideoCapability::e_videoCapabilityExtension))
+    return false;
+
+  for (PINDEX i = 0; i < extcap.m_videoCapabilityExtension.GetSize(); ++i) {
+    if (H323GenericCapabilityInfo::IsMatch(extcap.m_videoCapabilityExtension[i]))
+      return true;
+  }
+
+  return false;
+}
+
+
+/////////////////////////////////////////////////////////////////////////////
+
+H323GenericControlCapability::H323GenericControlCapability(const PString & identifier)
+  : H323GenericCapabilityInfo(identifier)
+{
+}
+
+
+H323Capability::MainTypes H323GenericControlCapability::GetMainType() const
+{
+  return e_GenericControl;
+}
+
+
+unsigned H323GenericControlCapability::GetSubType() const
+{
+  return 0; // Unused
+}
+
+
+PBoolean H323GenericControlCapability::OnSendingPDU(H245_Capability & pdu) const
+{
+  pdu.SetTag(H245_Capability::e_genericControlCapability);
+  return OnSendingGenericPDU(pdu, GetMediaFormat(), e_OLC);
+}
+
+
+PBoolean H323GenericControlCapability::OnSendingPDU(H245_ModeElement &) const
+{
+  PTRACE(1, "H323\tCannot create GenericControlCapability mode");
+  return false; // Can't be in mode request.
+}
+
+
+PBoolean H323GenericControlCapability::OnReceivedPDU(const H245_Capability & pdu)
+{
+  if (pdu.GetTag() != H245_Capability::e_genericControlCapability)
+    return false;
+  return OnReceivedGenericPDU(GetWritableMediaFormat(), pdu, e_OLC);
+}
+
+
+PBoolean H323GenericControlCapability::IsMatch(const PASN_Choice & subTypePDU) const
+{
+  return H323GenericCapabilityInfo::IsMatch((const H245_GenericCapability &)subTypePDU.GetObject());
+}
+
+
+H323Channel * H323GenericControlCapability::CreateChannel(H323Connection &,
+                                                          H323Channel::Directions,
+                                                          unsigned,
+                                                          const H245_H2250LogicalChannelParameters *) const
+{
+  PTRACE(1, "H323\tCannot create GenericControlCapability channel");
+  return NULL;
+}
+
+
+/////////////////////////////////////////////////////////////////////////////
+
+const char H239MediaType[] = "H.239";
+
+class OpalH239MediaType : public OpalMediaTypeDefinition 
+{
+  public:
+    OpalH239MediaType()
+      : OpalMediaTypeDefinition(H239MediaType, NULL)
+    { }
+    virtual PString GetRTPEncoding() const { return PString::Empty(); }
+#if OPAL_SIP
+    virtual SDPMediaDescription * CreateSDPMediaDescription(const OpalTransportAddress &) { return NULL; }
+#endif
+};
+
+OPAL_INSTANTIATE_MEDIATYPE2(H239, H239MediaType, OpalH239MediaType);
+
+
+
+H323H239VideoCapability::H323H239VideoCapability(const OpalMediaFormat & mediaFormat)
+  : H323ExtendedVideoCapability("0.0.8.239.1.2")
+{
+  GetWritableMediaFormat() = mediaFormat;
+}
+
+
+PObject::Comparison H323H239VideoCapability::Compare(const PObject & obj) const
+{
+  Comparison comparison = H323ExtendedVideoCapability::Compare(obj);
+  if (comparison != EqualTo)
+    return comparison;
+
+  return GetMediaFormat().Compare(((H323Capability &)obj).GetMediaFormat());
+}
+
+
+PObject * H323H239VideoCapability::Clone() const
+{
+  return new H323H239VideoCapability(*this);
+}
+
+
+PString H323H239VideoCapability::GetFormatName() const
+{
+  static class H239VideoMediaFormat : public OpalMediaFormat { 
+    public: 
+      H239VideoMediaFormat() 
+        : OpalMediaFormat("H.239-Video", H239MediaType, RTP_DataFrame::MaxPayloadType, NULL, false, 0, 0, 0, 0)
+      {
+        OpalMediaOption * option = new OpalMediaOptionUnsigned(OpalVideoFormat::ContentRoleMaskOption(), true, OpalMediaOption::AndMerge, 1, 1, 3);
+
+        OpalMediaOption::H245GenericInfo genericInfo;
+        genericInfo.ordinal = 1;
+        genericInfo.mode = OpalMediaOption::H245GenericInfo::Collapsing;
+        genericInfo.integerType = OpalMediaOption::H245GenericInfo::BooleanArray;
+        genericInfo.excludeTCS = false;
+        genericInfo.excludeOLC = false;
+        genericInfo.excludeReqMode = true;
+        option->SetH245Generic(genericInfo);
+
+        AddOption(option);
+      } 
+  } const name;
+  return name;
+}
+
+
+PBoolean H323H239VideoCapability::OnSendingPDU(H245_VideoCapability & pdu, CommandType type) const
+{
+  const_cast<H323H239VideoCapability*>(this)->m_videoFormats += GetMediaFormat();
+  return H323ExtendedVideoCapability::OnSendingPDU(pdu, type);
+}
+
+
+PBoolean H323H239VideoCapability::OnReceivedPDU(const H245_VideoCapability & pdu, CommandType type)
+{
+  if (!H323ExtendedVideoCapability::OnReceivedPDU(pdu, type))
+    return false;
+
+  GetWritableMediaFormat() = m_videoFormats[0];
+  return true;
+}
+
+
+/////////////////////////////////////////////////////////////////////////////
+
+H323H239ControlCapability::H323H239ControlCapability()
+  : H323GenericControlCapability("0.0.8.239.1.1")
+{
+}
+
+
+PObject * H323H239ControlCapability::Clone() const
+{
+  return new H323H239ControlCapability(*this);
+}
+
+
+PString H323H239ControlCapability::GetFormatName() const
+{
+  static const OpalMediaFormat name("H.239-Control", H239MediaType, RTP_DataFrame::IllegalPayloadType, NULL, false, 0, 0, 0, 0);
+  return name;
+}
+
+
+#endif  // OPAL_H239
+
 #endif // OPAL_VIDEO
 
 /////////////////////////////////////////////////////////////////////////////
@@ -1324,15 +1601,14 @@ PBoolean H323DataCapability::OnSendingPDU(H245_ModeElement & mode) const
 
 PBoolean H323DataCapability::OnReceivedPDU(const H245_Capability & cap)
 {
-  H323Capability::OnReceivedPDU(cap);
-
   if (cap.GetTag() != H245_Capability::e_receiveDataApplicationCapability &&
       cap.GetTag() != H245_Capability::e_receiveAndTransmitDataApplicationCapability)
     return PFalse;
 
   const H245_DataApplicationCapability & app = cap;
   maxBitRate = app.m_maxBitRate;
-  return OnReceivedPDU(app, e_TCS);
+
+  return OnReceivedPDU(app, e_TCS) && H323Capability::OnReceivedPDU(cap);
 }
 
 
@@ -1475,180 +1751,6 @@ PString H323_G711Capability::GetFormatName() const
 }
 
 
-/////////////////////////////////////////////////////////////////////////////
-
-#if 0
-
-H323_G728Capability::H323_G728Capability()
-  : H323AudioCapability(240, 30) // 240ms max, 30ms desired
-{
-}
-
-
-PObject * H323_G728Capability::Clone() const
-{
-  return new H323_G728Capability(*this);
-}
-
-
-unsigned H323_G728Capability::GetSubType() const
-{
-  return H245_AudioCapability::e_g728;
-}
-
-
-PString H323_G728Capability::GetFormatName() const
-{
-  return OpalG728;
-}
-
-
-/////////////////////////////////////////////////////////////////////////////
-
-static const char * const G729Name[4] = {
-  OPAL_G729,
-  OPAL_G729A,
-  OPAL_G729B,
-  OPAL_G729AB
-};
-
-H323_G729Capability::H323_G729Capability(Mode m)
-{
-  mode = m;
-}
-
-
-PObject * H323_G729Capability::Clone() const
-{
-  return new H323_G729Capability(*this);
-}
-
-
-unsigned H323_G729Capability::GetSubType() const
-{
-  static const unsigned G729SubType[4] = {
-    H245_AudioCapability::e_g729,
-    H245_AudioCapability::e_g729AnnexA,
-    H245_AudioCapability::e_g729wAnnexB,
-    H245_AudioCapability::e_g729AnnexAwAnnexB
-  };
-  return G729SubType[mode];
-}
-
-
-PString H323_G729Capability::GetFormatName() const
-{
-  return G729Name[mode];
-}
-
-
-/////////////////////////////////////////////////////////////////////////////
-
-H323_G7231Capability::H323_G7231Capability(PBoolean sid)
-{
-  allowSIDFrames = sid;
-}
-
-
-PObject * H323_G7231Capability::Clone() const
-{
-  return new H323_G7231Capability(*this);
-}
-
-
-unsigned H323_G7231Capability::GetSubType() const
-{
-  return H245_AudioCapability::e_g7231;
-}
-
-
-PString H323_G7231Capability::GetFormatName() const
-{
-  return OpalG7231_6k3;
-}
-
-
-PBoolean H323_G7231Capability::OnSendingPDU(H245_AudioCapability & pdu,
-                                        unsigned packetSize) const
-{
-  pdu.SetTag(H245_AudioCapability::e_g7231);
-  H245_AudioCapability_g7231 & g7231 = pdu;
-  g7231.m_maxAl_sduAudioFrames = packetSize;
-  g7231.m_silenceSuppression = allowSIDFrames;
-  return PTrue;
-}
-
-
-PBoolean H323_G7231Capability::OnReceivedPDU(const H245_AudioCapability & pdu,
-                                         unsigned & packetSize)
-{
-  if (pdu.GetTag() != H245_AudioCapability::e_g7231)
-    return PFalse;
-
-  const H245_AudioCapability_g7231 & g7231 = pdu;
-  packetSize = g7231.m_maxAl_sduAudioFrames;
-  allowSIDFrames = g7231.m_silenceSuppression;
-  return PTrue;
-}
-
-
-///////////////////////////////////////////////////////////////////////////////
-
-H323_GSM0610Capability::H323_GSM0610Capability()
-{
-}
-
-
-PObject * H323_GSM0610Capability::Clone() const
-{
-  return new H323_GSM0610Capability(*this);
-}
-
-
-unsigned H323_GSM0610Capability::GetSubType() const
-{
-  return H245_AudioCapability::e_gsmFullRate;
-}
-
-
-PString H323_GSM0610Capability::GetFormatName() const
-{
-  return OpalGSM0610;
-}
-
-
-void H323_GSM0610Capability::SetTxFramesInPacket(unsigned frames)
-{
-  H323AudioCapability::SetTxFramesInPacket(frames > 7 ? 7 : frames);
-}
-
-
-PBoolean H323_GSM0610Capability::OnSendingPDU(H245_AudioCapability & cap,
-                                          unsigned packetSize) const
-{
-  cap.SetTag(H245_AudioCapability::e_gsmFullRate);
-
-  H245_GSMAudioCapability & gsm = cap;
-  gsm.m_audioUnitSize = packetSize*33;
-  return PTrue;
-}
-
-
-PBoolean H323_GSM0610Capability::OnReceivedPDU(const H245_AudioCapability & cap,
-                                           unsigned & packetSize)
-{
-  if (cap.GetTag() != H245_AudioCapability::e_gsmFullRate)
-    return PFalse;
-
-  const H245_GSMAudioCapability & gsm = cap;
-  packetSize = gsm.m_audioUnitSize/33;
-  if (packetSize == 0)
-    packetSize = 1;
-  return PTrue;
-}
-
-#endif   // 0
-
 /////////////////////////////////////////////////////////////////////////////
 
 static const char * const UIISubTypeNames[H323_UserInputCapability::NumSubTypes] = {
@@ -1677,7 +1779,7 @@ class H323_UserInputCapability_##type : public H323_UserInputCapability \
   DECLARE_USER_INPUT_CLASS(type) \
   const OpalMediaFormat UserInput_##type( \
     UIISubTypeNames[H323_UserInputCapability::type], \
-    "userinput", RTP_DataFrame::IllegalPayloadType, NULL, PFalse, 1, 0, 0, 0 \
+    OpalMediaType::UserInput(), RTP_DataFrame::IllegalPayloadType, NULL, PFalse, 1, 0, 0, 0 \
   ); \
   H323_REGISTER_CAPABILITY(H323_UserInputCapability_##type, UIISubTypeNames[H323_UserInputCapability::type]) \
 
@@ -1737,7 +1839,7 @@ H323Channel * H323_UserInputCapability::CreateChannel(H323Connection &,
                                                       unsigned,
                                                       const H245_H2250LogicalChannelParameters *) const
 {
-  PTRACE(1, "Codec\tCannot create UserInputCapability channel");
+  PTRACE(1, "H323\tCannot create UserInputCapability channel");
   return NULL;
 }
 
@@ -1761,28 +1863,26 @@ PBoolean H323_UserInputCapability::OnSendingPDU(H245_Capability & pdu) const
 
 PBoolean H323_UserInputCapability::OnSendingPDU(H245_DataType &) const
 {
-  PTRACE(1, "Codec\tCannot have UserInputCapability in DataType");
+  PTRACE(1, "H323\tCannot have UserInputCapability in DataType");
   return PFalse;
 }
 
 
 PBoolean H323_UserInputCapability::OnSendingPDU(H245_ModeElement &) const
 {
-  PTRACE(1, "Codec\tCannot have UserInputCapability in ModeElement");
+  PTRACE(1, "H323\tCannot have UserInputCapability in ModeElement");
   return PFalse;
 }
 
 
 PBoolean H323_UserInputCapability::OnReceivedPDU(const H245_Capability & pdu)
 {
-  H323Capability::OnReceivedPDU(pdu);
-
   if (pdu.GetTag() == H245_Capability::e_receiveRTPAudioTelephonyEventCapability) {
     subType = SignalToneRFC2833;
     const H245_AudioTelephonyEventCapability & atec = pdu;
     rtpPayloadType = (RTP_DataFrame::PayloadTypes)(int)atec.m_dynamicRTPPayloadType;
     // Really should verify atec.m_audioTelephoneEvent here
-    return PTrue;
+    return H323Capability::OnReceivedPDU(pdu);
   }
 
   if (pdu.GetTag() != H245_Capability::e_receiveUserInputCapability &&
@@ -1790,13 +1890,13 @@ PBoolean H323_UserInputCapability::OnReceivedPDU(const H245_Capability & pdu)
     return PFalse;
 
   const H245_UserInputCapability & ui = pdu;
-  return ui.GetTag() == UserInputCapabilitySubTypeCodes[subType];
+  return ui.GetTag() == UserInputCapabilitySubTypeCodes[subType] && H323Capability::OnReceivedPDU(pdu);
 }
 
 
 PBoolean H323_UserInputCapability::OnReceivedPDU(const H245_DataType &, PBoolean)
 {
-  PTRACE(1, "Codec\tCannot have UserInputCapability in DataType");
+  PTRACE(1, "H323\tCannot have UserInputCapability in DataType");
   return PFalse;
 }
 
@@ -1908,13 +2008,15 @@ H323Capabilities::H323Capabilities(const H323Connection & connection,
 
   // Decode out of the PDU, the list of known codecs.
   if (pdu.HasOptionalField(H245_TerminalCapabilitySet::e_capabilityTable)) {
-    H323Capabilities allCapabilities;
-    const H323Capabilities & localCapabilities = connection.GetLocalCapabilities();
-    for (PINDEX c = 0; c < localCapabilities.GetSize(); c++)
-      allCapabilities.Add(allCapabilities.Copy(localCapabilities[c]));
-    allCapabilities.AddAllCapabilities(connection.GetEndPoint(), 0, 0, "*");
+    H323Capabilities allCapabilities(connection.GetLocalCapabilities());
+    allCapabilities.AddAllCapabilities(0, 0, "*");
     H323_UserInputCapability::AddAllCapabilities(allCapabilities, P_MAX_INDEX, P_MAX_INDEX);
+#if OPAL_H239
+    allCapabilities.Add(new H323H239VideoCapability(OpalMediaFormat()));
+    allCapabilities.Add(new H323H239ControlCapability());
+#endif
     allCapabilities.SetMediaPacketizations(mediaPacketizations);
+    PTRACE(4, "H323\tParsing remote capabilities");
 
     for (PINDEX i = 0; i < pdu.m_capabilityTable.GetSize(); i++) {
       if (pdu.m_capabilityTable[i].HasOptionalField(H245_CapabilityTableEntry::e_capability)) {
@@ -1949,8 +2051,7 @@ H323Capabilities::H323Capabilities(const H323Connection & connection,
               }
             }
           }
-          if (copy->OnReceivedPDU(pdu.m_capabilityTable[i].m_capability) &&
-              copy->GetWritableMediaFormat().ToNormalisedOptions())
+          if (copy->OnReceivedPDU(pdu.m_capabilityTable[i].m_capability))
             table.Append(copy);
           else
             delete copy;
@@ -2345,6 +2446,11 @@ H323Capability * H323Capabilities::FindCapability(const H245_Capability & cap) c
 
       case H245_Capability::e_receiveRTPAudioTelephonyEventCapability :
         return FindCapability(H323Capability::e_UserInput, SignalToneRFC2833_SubType);
+
+      case H245_Capability::e_genericControlCapability :
+        if (capability.GetMainType() == H323Capability::e_GenericControl)
+          found = capability.IsMatch(cap);
+        break;
     }
 
     if (found) {
@@ -2372,6 +2478,7 @@ H323Capability * H323Capabilities::FindCapability(const H245_Capability & cap) c
       }
 
       PTRACE(4, "H323\tUnsupported media packetization " << packetizationString << ", not using capability " << cap);
+      return NULL;
     }
   }
 
@@ -2585,10 +2692,11 @@ void H323Capabilities::BuildPDU(const H323Connection & connection,
       entry.m_capabilityTableEntryNumber = capability.GetCapabilityNumber();
       entry.IncludeOptionalField(H245_CapabilityTableEntry::e_capability);
       capability.GetWritableMediaFormat().ToCustomisedOptions();
-      capability.OnSendingPDU(entry.m_capability);
-
-      H323SetRTPPacketization(h225_0.m_mediaPacketizationCapability.m_rtpPayloadType, rtpPacketizationCount,
-                              capability.GetMediaFormat(), RTP_DataFrame::IllegalPayloadType);
+      if (capability.OnSendingPDU(entry.m_capability))
+        H323SetRTPPacketization(h225_0.m_mediaPacketizationCapability.m_rtpPayloadType, rtpPacketizationCount,
+                                capability.GetMediaFormat(), RTP_DataFrame::IllegalPayloadType);
+      else
+        pdu.m_capabilityTable.SetSize(count);
     }
   }
 
@@ -2627,7 +2735,7 @@ void H323Capabilities::BuildPDU(const H323Connection & connection,
 
 PBoolean H323Capabilities::Merge(const H323Capabilities & newCaps)
 {
-  PTRACE_IF(4, !table.IsEmpty(), "H245\tCapability merge of:\n" << newCaps << "\nInto:\n" << *this);
+  PTRACE_IF(4, !table.IsEmpty(), "H323\tCapability merge of:\n" << newCaps << "\nInto:\n" << *this);
 
   // Add any new capabilities not already in set.
   PINDEX i;
@@ -2654,8 +2762,8 @@ PBoolean H323Capabilities::Merge(const H323Capabilities & newCaps)
     }
   }
 
-  PTRACE_IF(4, !table.IsEmpty(), "H245\tCapability merge result:\n" << *this);
-  PTRACE(3, "H245\tReceived capability set, is " << (table.IsEmpty() ? "rejected" : "accepted"));
+  PTRACE_IF(4, !table.IsEmpty(), "H323\tCapability merge result:\n" << *this);
+  PTRACE(3, "H323\tReceived capability set, is " << (table.IsEmpty() ? "rejected" : "accepted"));
   return !table.IsEmpty();
 }
 
