@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21715 $
- * $Author: hfriederich $
- * $Date: 2008-12-08 23:17:16 +0000 (Mon, 08 Dec 2008) $
+ * $Revision: 23567 $
+ * $Author: rjongbloed $
+ * $Date: 2009-10-01 01:20:36 -0500 (Thu, 01 Oct 2009) $
  */
 
 #include <ptlib.h>
@@ -117,7 +117,7 @@ void H245NegMasterSlaveDetermination::Stop()
   if (state == e_Idle)
     return;
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Idle;
 }
 
@@ -129,7 +129,7 @@ PBoolean H245NegMasterSlaveDetermination::HandleIncoming(const H245_MasterSlaveD
   PTRACE(3, "H245\tReceived MasterSlaveDetermination: state=" << state);
 
   if (state == e_Incoming) {
-    replyTimer.Stop();
+    replyTimer.Stop(false);
     state = e_Idle;
     return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                              "Duplicate MasterSlaveDetermination");
@@ -167,7 +167,7 @@ PBoolean H245NegMasterSlaveDetermination::HandleIncoming(const H245_MasterSlaveD
     if (retryCount < endpoint.GetMasterSlaveDeterminationRetries())
       return Restart(); // Try again
 
-    replyTimer.Stop();
+    replyTimer.Stop(false);
     state = e_Idle;
     return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
                                              "Retries exceeded");
@@ -207,7 +207,7 @@ PBoolean H245NegMasterSlaveDetermination::HandleAck(const H245_MasterSlaveDeterm
       return PFalse;
   }
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Idle;
 
   if (status != newStatus)
@@ -239,7 +239,7 @@ PBoolean H245NegMasterSlaveDetermination::HandleReject(const H245_MasterSlaveDet
       break;
   }
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Idle;
 
   return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
@@ -256,7 +256,7 @@ PBoolean H245NegMasterSlaveDetermination::HandleRelease(const H245_MasterSlaveDe
   if (state == e_Idle)
     return PTrue;
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Idle;
 
   return connection.OnControlProtocolError(H323Connection::e_MasterSlaveDetermination,
@@ -268,6 +268,9 @@ void H245NegMasterSlaveDetermination::HandleTimeout(PTimer &, INT)
 {
   PWaitAndSignal wait(mutex);
 
+  if (state == e_Idle)
+    return;
+
   PTRACE(3, "H245\tTimeout on MasterSlaveDetermination: state=" << state);
 
   if (state == e_Outgoing) {
@@ -352,7 +355,7 @@ void H245NegTerminalCapabilitySet::Stop(PBoolean dec)
   if (state == e_Idle)
     return;
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Idle;
   receivedCapabilites = PFalse;
 
@@ -421,7 +424,7 @@ PBoolean H245NegTerminalCapabilitySet::HandleAck(const H245_TerminalCapabilitySe
   if (pdu.m_sequenceNumber != outSequenceNumber)
     return PTrue;
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   state = e_Confirmed;
   PTRACE(3, "H245\tTerminalCapabilitySet Sent.");
   return PTrue;
@@ -444,7 +447,7 @@ PBoolean H245NegTerminalCapabilitySet::HandleReject(const H245_TerminalCapabilit
     return PTrue;
 
   state = e_Idle;
-  replyTimer.Stop();
+  replyTimer.Stop(false);
   return connection.OnControlProtocolError(H323Connection::e_CapabilityExchange,
                                            "Rejected");
 }
@@ -466,6 +469,9 @@ void H245NegTerminalCapabilitySet::HandleTimeout(PTimer &, INT)
 {
   PWaitAndSignal wait(mutex);
 
+  if (state == e_Idle)
+    return;
+
   PTRACE(3, "H245\tTimeout on TerminalCapabilitySet: state=" << state);
 
   H323ControlPDU reply;
@@ -573,15 +579,6 @@ PBoolean H245NegLogicalChannel::OpenWhileLocked(const H323Capability & capabilit
            << ", channel->OnSendingPDU() failed");
     return PFalse;
   }
-  
-  // ensure the correct (external) session ID is used
-  if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
-    H245_H2250LogicalChannelParameters & param = open.m_reverseLogicalChannelParameters.m_multiplexParameters;
-    param.m_sessionID = connection.GetExternalSessionID(sessionID, capability.GetMediaFormat().GetMediaType());
-  } else {
-    H245_H2250LogicalChannelParameters & param = open.m_forwardLogicalChannelParameters.m_multiplexParameters;
-    param.m_sessionID = connection.GetExternalSessionID(sessionID, capability.GetMediaFormat().GetMediaType());
-  }
 
   if (replacementFor > 0) {
     if (open.HasOptionalField(H245_OpenLogicalChannel::e_reverseLogicalChannelParameters)) {
@@ -633,6 +630,8 @@ PBoolean H245NegLogicalChannel::CloseWhileLocked()
   else {
     reply.BuildCloseLogicalChannel(channelNumber);
     state = e_AwaitingRelease;
+    if (channel != NULL)
+      channel->Close();
   }
 
   return connection.WriteControlPDU(reply);
@@ -711,7 +710,7 @@ PBoolean H245NegLogicalChannel::HandleOpenAck(const H245_OpenLogicalChannelAck &
                                                "Ack unknown channel");
     case e_AwaitingEstablishment :
       state = e_Established;
-      replyTimer.Stop();
+      replyTimer.Stop(false);
 
       if (!channel->OnReceivedAckPDU(pdu)) {
         if (connection.GetRemoteProductInfo().name != "Cisco IOS")
@@ -754,7 +753,7 @@ PBoolean H245NegLogicalChannel::HandleOpenConfirm(const H245_OpenLogicalChannelC
       return connection.OnControlProtocolError(H323Connection::e_LogicalChannel,
                                                "Confirm established channel");
     case e_AwaitingConfirmation :
-      replyTimer.Stop();
+      replyTimer.Stop(false);
       state = e_Established;
       // Channel was already opened when OLC sent, if have error here it is
       // somthing other than an asymmetric codec conflict, so close it down.
@@ -958,7 +957,7 @@ void H245NegLogicalChannel::Release()
   channel = NULL;
   mutex.Signal();
 
-  replyTimer.Stop();
+  replyTimer.Stop(false);
 
   if (chan != NULL) {
     chan->Close();
@@ -1215,10 +1214,13 @@ H323Channel * H245NegLogicalChannels::FindChannelBySession(unsigned rtpSessionId
   PINDEX i;
   H323Channel::Directions desiredDirection = fromRemote ? H323Channel::IsReceiver : H323Channel::IsTransmitter;
   for (i = 0; i < GetSize(); i++) {
-    H323Channel * channel = channels.GetDataAt(i).GetChannel();
-    if (channel != NULL && channel->GetSessionID() == rtpSessionId &&
-                           channel->GetDirection() == desiredDirection)
-      return channel;
+    H245NegLogicalChannel & logChan = channels.GetDataAt(i);
+    if (logChan.IsAwaitingEstablishment() || logChan.IsEstablished()) {
+      H323Channel * channel = logChan.GetChannel();
+      if (channel != NULL && channel->GetSessionID() == rtpSessionId &&
+                             channel->GetDirection() == desiredDirection)
+        return channel;
+    }
   }
 
   return NULL;
@@ -1256,8 +1258,10 @@ H245NegRequestMode::H245NegRequestMode(H323EndPoint & end, H323Connection & conn
 PBoolean H245NegRequestMode::StartRequest(const PString & newModes)
 {
   PStringArray modes = newModes.Lines();
-  if (modes.IsEmpty())
-    return PFalse;
+  if (modes.IsEmpty()) {
+    PTRACE(2, "H245\tNo new mode to request");
+    return false;
+  }
 
   H245_ArrayOf_ModeDescription descriptions;
   PINDEX modeCount = 0;
@@ -1285,8 +1289,10 @@ PBoolean H245NegRequestMode::StartRequest(const PString & newModes)
     }
   }
 
-  if (modeCount == 0)
-    return PFalse;
+  if (modeCount == 0) {
+    PTRACE(2, "H245\tUnsupported new mode to request");
+    return false;
+  }
 
   return StartRequest(descriptions);
 }
@@ -1297,8 +1303,10 @@ PBoolean H245NegRequestMode::StartRequest(const H245_ArrayOf_ModeDescription & n
   PTRACE(3, "H245\tStarted request mode: outSeq=" << outSequenceNumber
          << (awaitingResponse ? " awaitingResponse" : " idle"));
 
-  if (awaitingResponse)
-    return PFalse;
+  if (awaitingResponse) {
+    PTRACE(2, "H245\tAwaiting response to previous mode request");
+    return false;
+  }
 
   // Initiate a mode request
   outSequenceNumber = (outSequenceNumber+1)%256;
@@ -1350,7 +1358,7 @@ PBoolean H245NegRequestMode::HandleAck(const H245_RequestModeAck & pdu)
 
   if (awaitingResponse && pdu.m_sequenceNumber == outSequenceNumber) {
     awaitingResponse = PFalse;
-    replyTimer.Stop();
+    replyTimer.Stop(false);
     connection.OnAcceptModeChange(pdu);
   }
 
@@ -1364,7 +1372,7 @@ PBoolean H245NegRequestMode::HandleReject(const H245_RequestModeReject & pdu)
 
   if (awaitingResponse && pdu.m_sequenceNumber == outSequenceNumber) {
     awaitingResponse = PFalse;
-    replyTimer.Stop();
+    replyTimer.Stop(false);
     connection.OnRefusedModeChange(&pdu);
   }
 
@@ -1390,9 +1398,8 @@ void H245NegRequestMode::HandleTimeout(PTimer &, INT)
     pdu.Build(H245_IndicationMessage::e_requestModeRelease);
     connection.WriteControlPDU(pdu);
     connection.OnRefusedModeChange(NULL);
+    connection.OnControlProtocolError(H323Connection::e_ModeRequest, "Timeout");
   }
-
-  connection.OnControlProtocolError(H323Connection::e_ModeRequest, "Timeout");
 }
 
 
@@ -1454,7 +1461,7 @@ PBoolean H245NegRoundTripDelay::HandleResponse(const H245_RoundTripDelayResponse
          << (awaitingResponse ? " awaitingResponse" : " idle"));
 
   if (awaitingResponse && pdu.m_sequenceNumber == sequenceNumber) {
-    replyTimer.Stop();
+    replyTimer.Stop(false);
     awaitingResponse = PFalse;
     roundTripTime = tripEndTime - tripStartTime;
     retryCount = 3;
