@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22300 $
+ * $Revision: 23923 $
  * $Author: rjongbloed $
- * $Date: 2009-03-26 03:40:31 +0000 (Thu, 26 Mar 2009) $
+ * $Date: 2010-01-11 19:49:45 -0600 (Mon, 11 Jan 2010) $
  */
 
 #include <ptlib.h>
@@ -478,6 +478,186 @@ bool H323GetRTPPacketization(OpalMediaFormat & mediaFormat, const H245_RTPPayloa
 }
 
 
+PString H323GetCapabilityIdentifier(const H245_CapabilityIdentifier & capId)
+{
+  switch (capId.GetTag()) {
+    case H245_CapabilityIdentifier::e_standard :
+      return ((const PASN_ObjectId &)capId).AsString();
+
+    case H245_CapabilityIdentifier::e_h221NonStandard :
+    {
+      PString str;
+      const H245_NonStandardParameter & nonStd = capId;
+      if (nonStd.m_nonStandardIdentifier.GetTag() == H245_NonStandardIdentifier::e_object)
+        str = ((const PASN_ObjectId &)nonStd.m_nonStandardIdentifier).AsString();
+      else {
+        const H245_NonStandardIdentifier_h221NonStandard & h221 = nonStd.m_nonStandardIdentifier;
+        str.sprintf("c=%u,cx=%u,o=%u",
+                    h221.m_t35CountryCode.GetValue(),
+                    h221.m_t35Extension.GetValue(),
+                    h221.m_manufacturerCode.GetValue());
+      }
+      if (nonStd.m_data.GetSize() > 0)
+        str += ':' + nonStd.m_data.AsString();
+      return str;
+    }
+  }
+
+  return PString::Empty();
+}
+
+
+static int ExtractVar(const PString & str, const PString & var)
+{
+  PRegularExpression regex("(^|[ \t\n,]+)" + var + "[ \t\n]*=[ \t\n]*[0-9]",
+                           PRegularExpression::IgnoreCase|PRegularExpression::Extended);
+
+  PINDEX pos, len;
+  if (!str.FindRegEx(regex, pos, len))
+    return -1;
+
+  return str.Mid(pos+len-1).AsUnsigned();
+}
+
+
+bool H323SetCapabilityIdentifier(const PString & str, H245_CapabilityIdentifier & capId)
+{
+  PASN_ObjectId oid;
+  oid.SetValue(str);
+  if (oid.AsString() == str) {
+    capId.SetTag(H245_CapabilityIdentifier::e_standard);
+    ((PASN_ObjectId &)capId) = oid;
+    return true;
+  }
+
+  PINDEX colon = str.Find(':');
+  if (colon == 0)
+    return false;
+
+  if (colon != P_MAX_INDEX && oid.AsString() == str.Left(colon)) {
+    capId.SetTag(H245_CapabilityIdentifier::e_h221NonStandard);
+    H245_NonStandardParameter & nonStd = capId;
+    nonStd.m_nonStandardIdentifier.SetTag(H245_NonStandardIdentifier::e_object);
+    ((PASN_ObjectId &)nonStd.m_nonStandardIdentifier) = oid;
+    nonStd.m_data = str.Mid(colon+1);
+    return true;
+  }
+
+  int country = ExtractVar(str, 'c');
+  int manufacturer = ExtractVar(str, 'o');
+  if (country >= 0 && manufacturer >= 0) {
+    capId.SetTag(H245_CapabilityIdentifier::e_h221NonStandard);
+    H245_NonStandardParameter & nonStd = capId;
+    nonStd.m_nonStandardIdentifier.SetTag(H245_NonStandardIdentifier::e_h221NonStandard);
+    H245_NonStandardIdentifier_h221NonStandard & h221 = nonStd.m_nonStandardIdentifier;
+
+    h221.m_t35CountryCode = country;
+    h221.m_manufacturerCode = manufacturer;
+
+    int extension = ExtractVar(str, "cx");
+    if (extension >= 0)
+      h221.m_t35Extension = extension;
+
+    if (colon != P_MAX_INDEX)
+      nonStd.m_data = str.Mid(colon+1);
+    return true;
+  }
+
+  return false;
+}
+
+
+const H245_ParameterValue * H323GetGenericParameter(const H245_ArrayOf_GenericParameter & params, unsigned ordinal)
+{
+  for (PINDEX i = 0; i < params.GetSize(); i++) {
+    const H245_GenericParameter & param = params[i];
+    if (param.m_parameterIdentifier.GetTag() == H245_ParameterIdentifier::e_standard &&
+                         ((const PASN_Integer &)param.m_parameterIdentifier) == ordinal)
+      return &param.m_parameterValue;
+  }
+
+  return NULL;
+}
+
+
+bool H323GetGenericParameterBoolean(const H245_ArrayOf_GenericParameter & params, unsigned ordinal)
+{
+  const H245_ParameterValue * param = H323GetGenericParameter(params, ordinal);
+  return param != NULL && param->GetTag() == H245_ParameterValue::e_logical;
+}
+
+
+unsigned H323GetGenericParameterInteger(const H245_ArrayOf_GenericParameter & params,
+                                        unsigned ordinal,
+                                        unsigned defValue,
+                                        H245_ParameterValue::Choices subType)
+{
+  const H245_ParameterValue * param = H323GetGenericParameter(params, ordinal);
+  if (param == NULL || param->GetTag() != (unsigned)subType)
+    return defValue;
+  return (const PASN_Integer &)*param;
+}
+
+
+H245_ParameterValue * H323AddGenericParameter(H245_ArrayOf_GenericParameter & params, unsigned ordinal, bool reorder)
+{
+  PINDEX size = params.GetSize();
+  params.SetSize(size+1);
+
+  PINDEX pos = size;
+  if (reorder) {
+    for (pos = size; pos > 0; pos--) {
+      const H245_GenericParameter & param = params[pos-1];
+      if (param.m_parameterIdentifier.GetTag() == H245_ParameterIdentifier::e_standard &&
+                            ((const PASN_Integer &)param.m_parameterIdentifier) < ordinal)
+        break;
+      params[pos] = param;
+    }
+  }
+
+  H245_GenericParameter & param = params[pos];
+  param.m_parameterIdentifier.SetTag(H245_ParameterIdentifier::e_standard);
+  (PASN_Integer &)param.m_parameterIdentifier = ordinal;
+  return &param.m_parameterValue;
+}
+
+
+void H323AddGenericParameterBoolean(H245_ArrayOf_GenericParameter & params, unsigned ordinal, bool value, bool reorder)
+{
+  // Do not include a logical at all if it is false
+  if (value)
+    H323AddGenericParameter(params, ordinal, reorder)->SetTag(H245_ParameterValue::e_logical);
+}
+
+
+void H323AddGenericParameterInteger(H245_ArrayOf_GenericParameter & params,
+                                    unsigned ordinal,
+                                    unsigned value,
+                                    H245_ParameterValue::Choices subType,
+                                    bool reorder)
+{
+  H245_ParameterValue * param = H323AddGenericParameter(params, ordinal, reorder);
+  param->SetTag(subType);
+  (PASN_Integer &)*param = value;
+}
+
+
+void H323AddGenericParameterString(H245_ArrayOf_GenericParameter & params, unsigned ordinal, const PString & value, bool reorder)
+{
+  H245_ParameterValue * param = H323AddGenericParameter(params, ordinal, reorder);
+  param->SetTag(H245_ParameterValue::e_octetString);
+  (PASN_OctetString &)*param = value;
+}
+
+
+void H323AddGenericParameterOctets(H245_ArrayOf_GenericParameter & params, unsigned ordinal, const PBYTEArray & value, bool reorder)
+{
+  H245_ParameterValue * param = H323AddGenericParameter(params, ordinal, reorder);
+  param->SetTag(H245_ParameterValue::e_octetString);
+  (PASN_OctetString &)*param = value;
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 #if OPAL_H460
@@ -729,12 +909,11 @@ H225_Information_UUIE & H323SignalPDU::BuildInformation(const H323Connection & c
 }
 
 
-H323Connection::CallEndReason H323TranslateToCallEndReason(Q931::CauseValues cause,
-                                                           unsigned reason)
+static OpalConnection::CallEndReason H323TranslateToCallEndReasonCode(Q931::CauseValues cause, unsigned reasonTag)
 {
   switch (cause) {
     case Q931::ErrorInCauseIE :
-      switch (reason) {
+      switch (reasonTag) {
         case H225_ReleaseCompleteReason::e_noBandwidth :
           return H323Connection::EndedByNoBandwidth;
 
@@ -775,6 +954,9 @@ H323Connection::CallEndReason H323TranslateToCallEndReason(Q931::CauseValues cau
     case Q931::ResourceUnavailable :
       return H323Connection::EndedByRemoteCongestion;
 
+    case Q931::CallRejected :
+      return H323Connection::EndedByRefusal;
+
     case Q931::NoResponse :
     case Q931::NoAnswer :
       return H323Connection::EndedByNoAnswer;
@@ -802,6 +984,11 @@ H323Connection::CallEndReason H323TranslateToCallEndReason(Q931::CauseValues cau
   }
 }
 
+H323Connection::CallEndReason H323TranslateToCallEndReason(Q931::CauseValues cause, unsigned reasonTag)
+{
+  return OpalConnection::CallEndReason(H323TranslateToCallEndReasonCode(cause, reasonTag), cause);
+}
+
 
 Q931::CauseValues H323TranslateFromCallEndReason(H323Connection::CallEndReason callEndReason,
                                                  H225_ReleaseCompleteReason & releaseCompleteReason)
@@ -832,11 +1019,18 @@ Q931::CauseValues H323TranslateFromCallEndReason(H323Connection::CallEndReason c
     Q931::TemporaryFailure,                                 /// EndedByTemporaryFailure   The remote failed temporarily app may retry
     Q931::UnknownCauseIE,                                   /// EndedByQ931Cause,         The remote ended the call with unmapped Q.931 cause code
     Q931::NormalUnspecified,                                /// EndedByDurationLimit,     Call cleared due to an enforced duration limit
+    Q931::InvalidCallReference,                             /// EndedByInvalidConferenceID,  Call cleared due to invalid conference ID
+    Q931::NoResponse,                                       /// EndedByNoDialTone,           Call cleared due to missing dial tone
+    Q931::NoResponse,                                       /// EndedByNoRingBackTone,       Call cleared due to missing ringback tone
+    Q931::DestinationOutOfOrder,                            /// EndedByOutOfService,         Call cleared because the line is out of service, 
+    Q931::Redirection,                                      /// EndedByAcceptingCallWaiting, Call cleared because another call is answered
+    Q931::ExchangeRoutingError,                             /// EndedByGkAdmissionFailed,    Call cleared because gatekeeper admission request failed.
   };
 
+  if (callEndReason.q931 != Q931::UnknownCauseIE)
+    return (Q931::CauseValues)callEndReason.q931;
+
   int code = ReasonCodes[callEndReason];
-  if (code == Q931::UnknownCauseIE)
-    return (Q931::CauseValues)callEndReason;
   if (code >= 0)
     return (Q931::CauseValues)code;
 
@@ -857,14 +1051,12 @@ H225_ReleaseComplete_UUIE &
   SetH225Version(connection, release.m_protocolIdentifier);
   release.m_callIdentifier.m_guid = connection.GetCallIdentifier();
 
-  Q931::CauseValues cause = (Q931::CauseValues)connection.GetQ931Cause();
-  if (cause == Q931::ErrorInCauseIE)
-    cause = H323TranslateFromCallEndReason(connection.GetCallEndReason(), release.m_reason);
+  Q931::CauseValues cause = H323TranslateFromCallEndReason(connection.GetCallEndReason(), release.m_reason);
   if (cause != Q931::ErrorInCauseIE)
     q931pdu.SetCause(cause);
   else
     release.IncludeOptionalField(H225_ReleaseComplete_UUIE::e_reason);
-  
+
 #if OPAL_H460
   SendFeatureSet<H225_ReleaseComplete_UUIE>(&connection, H460_MessageType::e_releaseComplete, m_h323_uu_pdu, release);
 #endif
@@ -1215,12 +1407,7 @@ unsigned H323SignalPDU::GetDistinctiveRing() const
 }
 
 
-void H323SignalPDU::SetQ931Fields(const H323Connection & connection,
-                                  PBoolean insertPartyNumbers,
-                                  unsigned plan,
-                                  unsigned type,
-                                  int presentation,
-                                  int screening)
+void H323SignalPDU::SetQ931Fields(const H323Connection & connection, bool insertPartyNumbers)
 {
   const PStringList & aliases = connection.GetLocalAliasNames();
 
@@ -1230,23 +1417,8 @@ void H323SignalPDU::SetQ931Fields(const H323Connection & connection,
 
   {
     const OpalConnection::StringOptions & stringOptions = connection.GetStringOptions();
-    PString strLocal, strDisplay;
-    strLocal   = stringOptions(OPAL_OPT_CALLING_PARTY_NAME);
-    strDisplay = stringOptions(OPAL_OPT_CALLING_DISPLAY_NAME);
-    if (strDisplay.IsEmpty())
-      strDisplay = strLocal;
-
-    if (!strLocal.IsEmpty())
-      localName = strLocal;
-    else
-      localName = connection.GetLocalPartyName();
-
-    if (!strDisplay.IsEmpty())
-      displayName = strDisplay;
-    else if (!strLocal.IsEmpty())
-      displayName = localName;
-    else
-      displayName = connection.GetDisplayName();
+    localName   = stringOptions(OPAL_OPT_CALLING_PARTY_NAME, connection.GetLocalPartyName());
+    displayName = stringOptions(OPAL_OPT_CALLING_DISPLAY_NAME, connection.GetDisplayName());
   }
 
   if (IsE164(localName)) {
@@ -1261,7 +1433,7 @@ void H323SignalPDU::SetQ931Fields(const H323Connection & connection,
     }
   }
   else {
-    if (!localName && displayName.IsEmpty())
+    if (displayName.IsEmpty())
       displayName = localName;
     for (PStringList::const_iterator alias = aliases.begin(); alias != aliases.end(); ++alias) {
       if (IsE164(*alias)) {
@@ -1271,8 +1443,6 @@ void H323SignalPDU::SetQ931Fields(const H323Connection & connection,
     }
   }
 
-  if (displayName.IsEmpty())
-    displayName = number;
   q931pdu.SetDisplayName(displayName);
 
   if (insertPartyNumbers) {
@@ -1283,6 +1453,11 @@ void H323SignalPDU::SetQ931Fields(const H323Connection & connection,
         otherNumber = otherName;
     }
 
+    int plan = 1; // ISDN/Telephony numbering system
+    int type = 0; // Unknown number type
+    int presentation = connection.IsPresentationBlocked() ? 1 : -1; // Presentation Blocked, or no octet3a
+    int screening = presentation == -1 ? -1 : 0; // User-provided, not screened
+
     if (connection.HadAnsweredCall()) {
       if (!number)
         q931pdu.SetCalledPartyNumber(number, plan, type);
@@ -1647,6 +1822,53 @@ H245_MiscellaneousIndication & H323ControlPDU::BuildMiscellaneousIndication(unsi
 }
 
 
+H245_GenericMessage & H323ControlPDU::BuildGenericRequest(const PString & identifier, unsigned subMsgId)
+{
+  H245_GenericMessage & msg = Build(H245_RequestMessage::e_genericRequest);
+  H323SetCapabilityIdentifier(identifier, msg.m_messageIdentifier);
+  msg.m_subMessageIdentifier = subMsgId;
+  return msg;
+}
+
+
+H245_GenericMessage & H323ControlPDU::BuildGenericResponse(const PString & identifier, unsigned subMsgId)
+{
+  H245_GenericMessage & msg = Build(H245_ResponseMessage::e_genericResponse);
+  H323SetCapabilityIdentifier(identifier, msg.m_messageIdentifier);
+
+  msg.IncludeOptionalField(H245_GenericMessage::e_subMessageIdentifier);
+  msg.m_subMessageIdentifier = subMsgId;
+
+  // Assume always have a content
+  msg.IncludeOptionalField(H245_GenericMessage::e_messageContent);
+  return msg;
+}
+
+
+H245_GenericMessage & H323ControlPDU::BuildGenericCommand(const PString & identifier, unsigned subMsgId)
+{
+  H245_GenericMessage & msg = Build(H245_CommandMessage::e_genericCommand);
+  H323SetCapabilityIdentifier(identifier, msg.m_messageIdentifier);
+
+  msg.IncludeOptionalField(H245_GenericMessage::e_subMessageIdentifier);
+  msg.m_subMessageIdentifier = subMsgId;
+
+  return msg;
+}
+
+
+H245_GenericMessage & H323ControlPDU::BuildGenericIndication(const PString & identifier, unsigned subMsgId)
+{
+  H245_GenericMessage & msg = Build(H245_IndicationMessage::e_genericIndication);
+  H323SetCapabilityIdentifier(identifier, msg.m_messageIdentifier);
+
+  msg.IncludeOptionalField(H245_GenericMessage::e_subMessageIdentifier);
+  msg.m_subMessageIdentifier = subMsgId;
+
+  return msg;
+}
+
+
 H245_FunctionNotUnderstood &
       H323ControlPDU::BuildFunctionNotUnderstood(const H323ControlPDU & pdu)
 {
