@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22096 $
+ * $Revision: 24148 $
  * $Author: rjongbloed $
- * $Date: 2009-02-21 03:25:23 +0000 (Sat, 21 Feb 2009) $
+ * $Date: 2010-03-23 03:29:03 -0500 (Tue, 23 Mar 2010) $
  */
 
 #include <ptlib.h>
@@ -56,11 +56,14 @@
 
 OpalCall::OpalCall(OpalManager & mgr)
   : manager(mgr)
-  , myToken(mgr.GetNextCallToken())
+  , myToken(mgr.GetNextToken('C'))
   , isEstablished(PFalse)
   , isClearing(PFalse)
   , callEndReason(OpalConnection::NumCallEndReasons)
   , endCallSyncPoint(NULL)
+#if OPAL_HAS_MIXER
+  , m_recordManager(NULL)
+#endif
 {
   manager.activeCalls.SetAt(myToken, this);
 
@@ -76,6 +79,10 @@ OpalCall::OpalCall(OpalManager & mgr)
 
 OpalCall::~OpalCall()
 {
+#if OPAL_HAS_MIXER
+  delete m_recordManager;
+#endif
+
   PTRACE(3, "Call\t" << *this << " destroyed.");
 }
 
@@ -122,6 +129,8 @@ void OpalCall::Clear(OpalConnection::CallEndReason reason, PSyncPoint * sync)
 
   UnlockReadWrite();
 
+  InternalOnClear();
+
   PSafePtr<OpalConnection> connection;
   while (EnumerateConnections(connection, PSafeReadWrite))
     connection->Release(reason);
@@ -131,7 +140,10 @@ void OpalCall::Clear(OpalConnection::CallEndReason reason, PSyncPoint * sync)
 void OpalCall::OnCleared()
 {
   manager.OnClearedCall(*this);
-  manager.GetRecordManager().Close(myToken);
+
+#if OPAL_HAS_MIXER
+  StopRecording();
+#endif
 
   if (!LockReadWrite())
     return;
@@ -205,7 +217,7 @@ OpalConnection::AnswerCallResponse OpalCall::OnAnswerCall(OpalConnection & PTRAC
                                                           const PString & PTRACE_PARAM(caller))
 {
   PTRACE(3, "Call\tOnAnswerCall " << connection << " caller \"" << caller << '"');
-  return OpalConnection::AnswerCallPending;
+  return OpalConnection::AnswerCallDeferred;
 }
 
 
@@ -216,17 +228,20 @@ PBoolean OpalCall::OnConnected(OpalConnection & connection)
   if (isClearing || !LockReadOnly())
     return false;
 
-  bool ok = connectionsActive.GetSize() == 1 && !m_partyB.IsEmpty();
+  bool havePartyB = connectionsActive.GetSize() == 1 && !m_partyB.IsEmpty();
 
   UnlockReadOnly();
 
-  if (ok) {
-    if (!manager.MakeConnection(*this, m_partyB, NULL, 0,
-                                const_cast<OpalConnection::StringOptions *>(&connection.GetStringOptions())))
-      connection.Release(OpalConnection::EndedByNoUser);
-    return OnSetUp(connection);
+  if (havePartyB) {
+    if (manager.MakeConnection(*this, m_partyB, NULL, 0,
+                                const_cast<OpalConnection::StringOptions *>(&connection.GetStringOptions())) != NULL)
+      return OnSetUp(connection);
+
+    connection.Release(OpalConnection::EndedByNoUser);
+    return false;
   }
 
+  bool ok = false;
   PSafePtr<OpalConnection> otherConnection;
   while (EnumerateConnections(otherConnection, PSafeReadWrite, &connection)) {
     if (otherConnection->GetPhase() >= OpalConnection::ConnectedPhase ||
@@ -314,7 +329,7 @@ bool OpalCall::IsOnHold() const
 {
   PSafePtr<OpalConnection> connection;
   while (EnumerateConnections(connection, PSafeReadOnly)) {
-    if (connection->IsConnectionOnHold())
+    if (connection->IsConnectionOnHold(false) || connection->IsConnectionOnHold(true))
       return true;
   }
 
@@ -322,24 +337,50 @@ bool OpalCall::IsOnHold() const
 }
 
 
-bool OpalCall::Transfer(OpalConnection & connection, const PString & newAddress)
+bool OpalCall::Transfer(const PString & newAddress, OpalConnection * connection)
 {
-  if (newAddress.NumCompare(connection.GetPrefixName()+':') == EqualTo)
-    return connection.TransferConnection(newAddress);
+  PCaselessString prefix = newAddress.Left(newAddress.Find(':'));
 
-  PSafePtr<OpalConnection> connectionToKeep;
-  EnumerateConnections(connectionToKeep, PSafeReference, &connection);
+  if (connection == NULL) {
+    for (PSafePtr<OpalConnection> conn = GetConnection(0); conn != NULL; ++conn) {
+      if (prefix == conn->GetPrefixName())
+        return conn->TransferConnection(newAddress);
+    }
 
-  if (!manager.MakeConnection(*this, newAddress))
+    PTRACE(2, "Call\tUnable to resolve transfer to \"" << newAddress << '"');
+    return false;
+  }
+
+  if (prefix == "*")
+    return connection->TransferConnection(connection->GetPrefixName() + newAddress.Mid(1));
+
+  if (prefix == connection->GetPrefixName() || manager.HasCall(newAddress))
+    return connection->TransferConnection(newAddress);
+
+  PTRACE(3, "Call\tTransferring " << *connection << " to \"" << newAddress << '"');
+
+  PSafePtr<OpalConnection> connectionToKeep = GetOtherPartyConnection(*connection);
+  if (connectionToKeep == NULL)
     return false;
 
-  connection.Release(OpalConnection::EndedByCallForwarded);
+  PSafePtr<OpalConnection> newConnection = manager.MakeConnection(*this, newAddress);
+  if (newConnection == NULL)
+    return false;
 
-  // Close streams, but as we Released above should not do re-INVITE/OLC
-  connection.CloseMediaStreams();
+  OpalConnection::Phases oldPhase = connection->GetPhase();
+  connection->SetPhase(OpalConnection::ForwardingPhase);
 
   // Restart with new connection
-  return OnSetUp(*connectionToKeep);
+  if (newConnection->SetUpConnection() && newConnection->OnSetUpConnection()) {
+    connectionToKeep->AutoStartMediaStreams(true);
+    connection->Release(OpalConnection::EndedByCallForwarded);
+    newConnection->StartMediaStreams();
+    return true;
+  }
+
+  newConnection->Release(OpalConnection::EndedByTemporaryFailure);
+  connection->SetPhase(oldPhase);
+  return false;
 }
 
 
@@ -353,6 +394,7 @@ OpalMediaFormatList OpalCall::GetMediaFormats(const OpalConnection & connection,
   PSafePtr<OpalConnection> otherConnection;
   while (EnumerateConnections(otherConnection, PSafeReadOnly, includeSpecifiedConnection ? NULL : &connection)) {
     OpalMediaFormatList possibleFormats = OpalTranscoder::GetPossibleFormats(otherConnection->GetMediaFormats());
+    connection.AdjustMediaFormats(true, possibleFormats, otherConnection);
     if (first) {
       commonFormats = possibleFormats;
       first = PFalse;
@@ -368,7 +410,7 @@ OpalMediaFormatList OpalCall::GetMediaFormats(const OpalConnection & connection,
     }
   }
 
-  connection.AdjustMediaFormats(commonFormats);
+  connection.AdjustMediaFormats(true, commonFormats, NULL);
 
   PTRACE(4, "Call\tGetMediaFormats for " << connection << '\n'
          << setfill('\n') << commonFormats << setfill(' '));
@@ -380,12 +422,31 @@ OpalMediaFormatList OpalCall::GetMediaFormats(const OpalConnection & connection,
 PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection, 
                                      const OpalMediaType & mediaType,
                                                   unsigned sessionID, 
-                                   const OpalMediaFormat & preselectedFormat)
+                                   const OpalMediaFormat & preselectedFormat
+#if OPAL_VIDEO
+                            , OpalVideoFormat::ContentRole contentRole
+#endif
+                                   )
 {
   PSafeLockReadOnly lock(*this);
   if (isClearing || !lock.IsLocked())
     return false;
 
+#if PTRACING
+  PStringStream traceText;
+  if (PTrace::CanTrace(2)) {
+    traceText << " for " << mediaType << " session " << sessionID;
+    if (preselectedFormat.IsValid())
+      traceText << " (" << preselectedFormat << ')';
+    traceText << " on " << connection;
+  }
+#endif
+
+  if (IsOnHold()) {
+    PTRACE(3, "Call\tOpenSourceMediaStreams (call on hold)" << traceText);
+    return false;
+  }
+
   // Check if already done
   OpalMediaStreamPtr sinkStream;
   OpalMediaStreamPtr sourceStream = connection.GetMediaStream(sessionID, true);
@@ -395,17 +456,19 @@ PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection,
       sinkStream = patch->GetSink();
     if (sourceStream->GetMediaFormat() == preselectedFormat ||
         (sinkStream != NULL && sinkStream->GetMediaFormat() == preselectedFormat)) {
-      PTRACE(3, "Call\tOpenSourceMediaStreams (already opened) for session " << sessionID << " on " << connection);
+      PTRACE(3, "Call\tOpenSourceMediaStreams (already opened)" << traceText);
       return true;
     }
-    if (sourceStream->IsOpen()) {
-      PTRACE(3, "Call\tOpenSourceMediaStreams (is opening) for session " << sessionID << " on " << connection);
+    if (sinkStream == NULL && sourceStream->IsOpen()) {
+      PTRACE(3, "Call\tOpenSourceMediaStreams (is opening)" << traceText);
       return true;
     }
   }
 
-  PTRACE(3, "Call\tOpenSourceMediaStreams " << (sourceStream != NULL ? "replacing" : "opening")
-         << ' ' << mediaType << " session " << sessionID << " on " << connection);
+  if (sessionID == 0)
+    sessionID = connection.GetNextSessionID(mediaType, true);
+
+  PTRACE(3, "Call\tOpenSourceMediaStreams " << (sourceStream != NULL ? "replace" : "open") << traceText);
   sourceStream.SetNULL();
 
   // Create the sinks and patch if needed
@@ -419,45 +482,90 @@ PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection,
 
   PSafePtr<OpalConnection> otherConnection;
   while (EnumerateConnections(otherConnection, PSafeReadWrite, &connection)) {
+    PStringArray order;
+    if (sinkFormat.IsValid())
+      order += sinkFormat.GetName(); // Preferential treatment to format already in use
+    order += '@' + mediaType;        // And media of the same type
+
     OpalMediaFormatList sinkMediaFormats = otherConnection->GetMediaFormats();
-    if (preselectedFormat.IsValid()  && sinkMediaFormats.HasFormat(preselectedFormat))
+    connection.AdjustMediaFormats(false, sinkMediaFormats, otherConnection);
+    if (sinkMediaFormats.IsEmpty()) {
+      PTRACE(2, "Call\tOpenSourceMediaStreams failed with no sink formats" << traceText);
+      return false;
+    }
+
+    if (preselectedFormat.IsValid() && sinkMediaFormats.HasFormat(preselectedFormat))
       sinkMediaFormats = preselectedFormat;
     else
-      sinkMediaFormats.Reorder(sinkFormat.GetName()); // Preferential treatment to format already in use
+      sinkMediaFormats.Reorder(order);
 
     OpalMediaFormatList sourceMediaFormats;
     if (sourceFormat.IsValid())
       sourceMediaFormats = sourceFormat; // Use the source format already established
     else {
-      OpalMediaFormatList connectionMediaFormats = connection.GetMediaFormats();
-      if (preselectedFormat.IsValid()  && connectionMediaFormats.HasFormat(preselectedFormat))
-        sourceMediaFormats = preselectedFormat;
-      else {
-        // Only include media formats that are of the same media type
-        for (OpalMediaFormatList::iterator i = connectionMediaFormats.begin(); i != connectionMediaFormats.end(); ++i) {
-          if (i->GetMediaType() == mediaType)
-            sourceMediaFormats += *i;
-        }
+      sourceMediaFormats = connection.GetMediaFormats();
+      otherConnection->AdjustMediaFormats(false, sourceMediaFormats, &connection);
+      if (sourceMediaFormats.IsEmpty()) {
+        PTRACE(2, "Call\tOpenSourceMediaStreams failed with no source formats" << traceText);
+        return false;
       }
+
+      if (preselectedFormat.IsValid() && sourceMediaFormats.HasFormat(preselectedFormat))
+        sourceMediaFormats = preselectedFormat;
+      else
+        sourceMediaFormats.Reorder(order);
     }
 
     // Get other media directions format so we give preference to symmetric codecs
-    OpalMediaStreamPtr otherDirection = otherConnection->GetMediaStream(sessionID, true);
+    OpalMediaStreamPtr otherDirection = sessionID != 0 ? otherConnection->GetMediaStream(sessionID, true)
+                                                       : otherConnection->GetMediaStream(mediaType, true);
     if (otherDirection != NULL) {
       PString priorityFormat = otherDirection->GetMediaFormat().GetName();
       sourceMediaFormats.Reorder(priorityFormat);
       sinkMediaFormats.Reorder(priorityFormat);
     }
 
-    if (!SelectMediaFormats(sourceMediaFormats,
+#if OPAL_VIDEO
+    if (contentRole != OpalVideoFormat::eNoRole) {
+      // Remove all media formats no supporting the role
+      OpalMediaFormatList * lists[2] = { &sourceMediaFormats, &sinkMediaFormats };
+      for (PINDEX i = 0; i < 2; i++) {
+        OpalMediaFormatList::iterator format = lists[i]->begin();
+        while (format != lists[i]->end()) {
+          if ( format->GetMediaType() != mediaType ||
+              (format->IsTransportable() &&
+              (format->GetOptionInteger(OpalVideoFormat::ContentRoleMaskOption())&OpalVideoFormat::ContentRoleBit(contentRole)) == 0))
+            *lists[i] -= *format++;
+          else
+            ++format;
+        }
+        if (lists[i]->IsEmpty()) {
+          PTRACE(2, "Call\tUnsupported Content Role " << contentRole << traceText);
+          return false;
+        }
+      }
+    }
+#endif
+
+    OpalMediaFormatList localMediaFormats = OpalMediaFormat::GetAllRegisteredMediaFormats();
+    otherConnection->AdjustMediaFormats(true, localMediaFormats, &connection);
+    if (!SelectMediaFormats(mediaType,
+                            sourceMediaFormats,
                             sinkMediaFormats,
-                            connection.GetLocalMediaFormats(),
+                            localMediaFormats,
                             sourceFormat,
                             sinkFormat))
       return false;
 
-    if (sessionID == 0)
-      sessionID = mediaType.GetDefinition()->GetDefaultSessionId();
+#if OPAL_VIDEO
+    sourceFormat.SetOptionEnum(OpalVideoFormat::ContentRoleOption(), contentRole);
+    sinkFormat.SetOptionEnum(OpalVideoFormat::ContentRoleOption(), contentRole);
+#endif
+
+    if (sessionID == 0) {
+      sessionID = otherConnection->GetNextSessionID(mediaType, false);
+      PTRACE(3, "Call\tOpenSourceMediaStreams using session " << sessionID << " on " << connection);
+    }
 
     // Finally have the negotiated formats, open the streams
     if (sourceStream == NULL) {
@@ -495,18 +603,19 @@ PBoolean OpalCall::OpenSourceMediaStreams(OpalConnection & connection,
 }
 
 
-bool OpalCall::SelectMediaFormats(const OpalMediaFormatList & srcFormats,
+bool OpalCall::SelectMediaFormats(const OpalMediaType & mediaType,
+                                  const OpalMediaFormatList & srcFormats,
                                   const OpalMediaFormatList & dstFormats,
                                   const OpalMediaFormatList & allFormats,
                                   OpalMediaFormat & srcFormat,
                                   OpalMediaFormat & dstFormat) const
 {
-  if (OpalTranscoder::SelectFormats(srcFormats, dstFormats, allFormats, srcFormat, dstFormat)) {
+  if (OpalTranscoder::SelectFormats(mediaType, srcFormats, dstFormats, allFormats, srcFormat, dstFormat)) {
     PTRACE(3, "Call\tSelected media formats " << srcFormat << " -> " << dstFormat);
     return true;
   }
 
-  PTRACE(2, "Call\tSelectMediaFormats could not find compatible media format:\n"
+  PTRACE(2, "Call\tSelectMediaFormats could not find compatible " << mediaType << " format:\n"
             "  source formats=" << setfill(',') << srcFormats << "\n"
             "   sink  formats=" << dstFormats << setfill(' '));
   return false;
@@ -582,6 +691,13 @@ void OpalCall::OnReleased(OpalConnection & connection)
     if (last != NULL)
       last->Release(connection.GetCallEndReason());
   }
+
+  InternalOnClear();
+}
+
+
+void OpalCall::InternalOnClear()
+{
   if (connectionsActive.IsEmpty() && manager.activeCalls.Contains(GetToken())) {
     OnCleared();
     manager.activeCalls.RemoveAt(GetToken());
@@ -595,51 +711,94 @@ void OpalCall::OnHold(OpalConnection & /*connection*/,
 {
 }
 
+#if OPAL_HAS_MIXER
 
-PBoolean OpalCall::StartRecording(const PFilePath & fn, bool mono)
+bool OpalCall::StartRecording(const PFilePath & fn, const OpalRecordManager::Options & options)
 {
+  StopRecording();
+
+  OpalRecordManager * newManager = OpalRecordManager::Factory::CreateInstance(fn.GetType());
+  if (newManager == NULL) {
+    PTRACE(2, "OPAL\tCannot record to file type " << fn);
+    return false;
+  }
+
   // create the mixer entry
-  if (!manager.GetRecordManager().Open(myToken, fn, mono))
-    return PFalse;
+  if (!newManager->Open(fn, options)) {
+    delete newManager;
+    return false;
+  }
+
+  PSafeLockReadWrite lock(*this);
+  if (!lock.IsLocked())
+    return false;
+
+  m_recordManager = newManager;
 
   // tell each connection to start sending data
   PSafePtr<OpalConnection> connection;
   while (EnumerateConnections(connection, PSafeReadWrite))
     connection->EnableRecording();
 
-  return PTrue;
+  return true;
 }
 
-
 bool OpalCall::IsRecording() const
 {
-  return manager.GetRecordManager().IsOpen(myToken);
+  PSafeLockReadOnly lock(*this);
+  return m_recordManager != NULL && m_recordManager->IsOpen();
 }
 
 
 void OpalCall::StopRecording()
 {
+  PSafeLockReadWrite lock(*this);
+  if (!lock.IsLocked() || m_recordManager == NULL)
+    return;
+
   // tell each connection to stop sending data
   PSafePtr<OpalConnection> connection;
   while (EnumerateConnections(connection, PSafeReadWrite))
     connection->DisableRecording();
 
-  manager.GetRecordManager().Close(myToken);
+  m_recordManager->Close();
+  delete m_recordManager;
+  m_recordManager = NULL;
+}
+
+
+bool OpalCall::OnStartRecording(const PString & streamId, const OpalMediaFormat & format)
+{
+  return m_recordManager != NULL && m_recordManager->OpenStream(streamId, format);
+}
+
+
+void OpalCall::OnStopRecording(const PString & streamId)
+{
+  if (m_recordManager != NULL)
+    m_recordManager->CloseStream(streamId);
 }
 
 
 void OpalCall::OnRecordAudio(const PString & streamId, const RTP_DataFrame & frame)
 {
-  if (!manager.GetRecordManager().WriteAudio(myToken, streamId, frame))
-    StopRecording();
+  if (m_recordManager != NULL && !m_recordManager->WriteAudio(streamId, frame))
+    m_recordManager->CloseStream(streamId);
 }
 
 
-void OpalCall::OnStopRecordAudio(const PString & streamId)
+#if OPAL_VIDEO
+
+void OpalCall::OnRecordVideo(const PString & streamId, const RTP_DataFrame & frame)
 {
-  manager.GetRecordManager().CloseStream(myToken, streamId);
+  if (m_recordManager != NULL && !m_recordManager->WriteVideo(streamId, frame))
+    m_recordManager->CloseStream(streamId);
 }
 
+#endif
+
+#endif
+
 
 bool OpalCall::IsNetworkOriginated() const
 {
@@ -697,7 +856,7 @@ bool OpalCall::EnumerateConnections(PSafePtr<OpalConnection> & connection,
 
   while (connection != NULL) {
     if (connection != skipConnection &&
-        connection->GetPhase() < OpalConnection::ReleasingPhase &&
+        connection->GetPhase() <= OpalConnection::EstablishedPhase &&
         connection.SetSafetyMode(mode))
       return true;
     ++connection;
