@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23491 $
+ * $Revision: 23954 $
  * $Author: rjongbloed $
- * $Date: 2009-09-17 21:17:05 -0500 (Thu, 17 Sep 2009) $
+ * $Date: 2010-01-20 17:11:47 -0600 (Wed, 20 Jan 2010) $
  */
 
 #include <ptlib.h>
@@ -47,14 +47,18 @@
 #include <codec/silencedetect.h>
 #include <codec/echocancel.h>
 #include <codec/rfc2833.h>
+#include <codec/vidcodec.h>
 #include <rtp/rtp.h>
 #include <t120/t120proto.h>
 #include <t38/t38proto.h>
 #include <ptclib/url.h>
 
 #if OPAL_HAS_IM
-#include <im/sipim.h>
 #include <im/rfc4103.h>
+#if OPAL_SIP
+#include <im/sipim.h>
+#include <sip/sipcon.h>
+#endif
 #endif
 
 #define new PNEW
@@ -70,6 +74,7 @@ ostream & operator<<(ostream & out, OpalConnection::Phases phase)
     "AlertingPhase",
     "ConnectedPhase",
     "EstablishedPhase",
+    "ForwardingPhase",
     "ReleasingPhase",
     "ReleasedPhase"
   };
@@ -159,6 +164,44 @@ ostream & operator<<(ostream & o, OpalConnection::SendUserInputModes m)
 
 #endif
 
+
+static POrdinalToString::Initialiser const CallEndReasonStringsInitialiser[] = {
+  { OpalConnection::EndedByLocalUser,            "Local party cleared call" },
+  { OpalConnection::EndedByNoAccept,             "Local party did not accept call" },
+  { OpalConnection::EndedByAnswerDenied,         "Local party declined to answer call" },
+  { OpalConnection::EndedByRemoteUser,           "Remote party cleared call" },
+  { OpalConnection::EndedByRefusal,              "Remote party refused call" },
+  { OpalConnection::EndedByNoAnswer,             "Remote party did not answer in required time" },
+  { OpalConnection::EndedByCallerAbort,          "Remote party stopped calling" },
+  { OpalConnection::EndedByTransportFail,        "Call failed due to a transport error" },
+  { OpalConnection::EndedByConnectFail,          "Connection to remote failed" },
+  { OpalConnection::EndedByGatekeeper,           "Gatekeeper has cleared call" },
+  { OpalConnection::EndedByNoUser,               "Call failed as could not find user" },
+  { OpalConnection::EndedByNoBandwidth,          "Call failed due to insufficient bandwidth" },
+  { OpalConnection::EndedByCapabilityExchange,   "Call failed as could not find common media capabilities" },
+  { OpalConnection::EndedByCallForwarded,        "Call was forwarded" },
+  { OpalConnection::EndedBySecurityDenial,       "Call failed security check" },
+  { OpalConnection::EndedByLocalBusy,            "Local party busy" },
+  { OpalConnection::EndedByLocalCongestion,      "Local party congested" },
+  { OpalConnection::EndedByRemoteBusy,           "Remote party busy" },
+  { OpalConnection::EndedByRemoteCongestion,     "Remote switch congested" },
+  { OpalConnection::EndedByUnreachable,          "Remote party could not be reached" },
+  { OpalConnection::EndedByNoEndPoint,           "Remote party application is not running" },
+  { OpalConnection::EndedByHostOffline,          "Remote party host is off line" },
+  { OpalConnection::EndedByTemporaryFailure,     "Remote system failed temporarily" },
+  { OpalConnection::EndedByQ931Cause,            "Call cleared with Q.931 cause code %u" },
+  { OpalConnection::EndedByDurationLimit,        "Call cleared due to an enforced duration limit" },
+  { OpalConnection::EndedByInvalidConferenceID,  "Call cleared due to invalid conference ID" },
+  { OpalConnection::EndedByNoDialTone,           "Call cleared due to missing dial tone" },
+  { OpalConnection::EndedByNoRingBackTone,       "Call cleared due to missing ringback tone" },
+  { OpalConnection::EndedByOutOfService,         "Call cleared because the line is out of service" },
+  { OpalConnection::EndedByAcceptingCallWaiting, "Call cleared because another call is answered" },
+  { OpalConnection::EndedByGkAdmissionFailed,    "Call cleared because gatekeeper admission request failed." },
+};
+
+static POrdinalToString CallEndReasonStrings(PARRAYSIZE(CallEndReasonStringsInitialiser), CallEndReasonStringsInitialiser);
+
+
 /////////////////////////////////////////////////////////////////////////////
 
 OpalConnection::OpalConnection(OpalCall & call,
@@ -181,19 +224,42 @@ OpalConnection::OpalConnection(OpalCall & call,
   , remotePartyName(token)
   , callEndReason(NumCallEndReasons)
   , synchronousOnRelease(true)
-  , q931Cause(0x100)
   , silenceDetector(NULL)
+#if OPAL_AEC
   , echoCanceler(NULL)
+#endif
+#if OPAL_PTLIB_DTMF
+  , m_dtmfScaleMultiplier(1)
+  , m_dtmfScaleDivisor(1)
 #ifdef _MSC_VER
 #pragma warning(disable:4355)
 #endif
-  , recordNotifier(PCREATE_NOTIFIER(OnRecordAudio))
+  , m_dtmfNotifier(PCREATE_NOTIFIER(OnDetectInBandDTMF))
+#ifdef _MSC_VER
+#pragma warning(default:4355)
+#endif
+  , m_sendInBandDTMF(true)
+  , m_installedInBandDTMF(false)
+  , m_emittedInBandDTMF(0)
+#endif
+#ifdef _MSC_VER
+#pragma warning(disable:4355)
+#endif
+#if OPAL_HAS_MIXER
+  , m_recordAudioNotifier(PCREATE_NOTIFIER(OnRecordAudio))
+#if OPAL_VIDEO
+  , m_recordVideoNotifier(PCREATE_NOTIFIER(OnRecordVideo))
+#endif
+#endif
 #ifdef _MSC_VER
 #pragma warning(default:4355)
 #endif
 #if OPAL_STATISTICS
   , m_VideoUpdateRequestsSent(0)
 #endif
+#if OPAL_FAX
+  , m_faxMediaStreamsSwitchState(e_NotSwitchingFaxMediaStreams)
+#endif
 {
   PTRACE(3, "OpalCon\tCreated connection " << *this);
 
@@ -204,12 +270,25 @@ OpalConnection::OpalConnection(OpalCall & call,
   if (stringOptions != NULL)
     m_connStringOptions = *stringOptions;
 
-  detectInBandDTMF = !endpoint.GetManager().DetectInBandDTMFDisabled();
   minAudioJitterDelay = endpoint.GetManager().GetMinAudioJitterDelay();
   maxAudioJitterDelay = endpoint.GetManager().GetMaxAudioJitterDelay();
   bandwidthAvailable = endpoint.GetInitialBandwidth();
 
-  dtmfScaleMultiplier = dtmfScaleDivisor = 1;
+#if OPAL_PTLIB_DTMF
+  switch (options&DetectInBandDTMFOptionMask) {
+    case DetectInBandDTMFOptionDisable :
+      m_detectInBandDTMF = false;
+      break;
+
+    case DetectInBandDTMFOptionEnable :
+      m_detectInBandDTMF = true;
+      break;
+
+    default :
+      m_detectInBandDTMF = !endpoint.GetManager().DetectInBandDTMFDisabled();
+      break;
+  }
+#endif
 
   switch (options & SendDTMFMask) {
     case SendDTMFAsString:
@@ -226,6 +305,11 @@ OpalConnection::OpalConnection(OpalCall & call,
       sendUserInputMode = ep.GetSendUserInputMode();
       break;
   }
+
+#if OPAL_HAS_IM
+  m_rfc4103Context[0].SetMediaFormat(OpalT140);
+  m_rfc4103Context[1].SetMediaFormat(OpalT140);
+#endif
 }
 
 OpalConnection::~OpalConnection()
@@ -233,7 +317,9 @@ OpalConnection::~OpalConnection()
   mediaStreams.RemoveAll();
 
   delete silenceDetector;
+#if OPAL_AEC
   delete echoCanceler;
+#endif
 #if OPAL_T120DATA
   delete t120handler;
 #endif
@@ -276,9 +362,9 @@ bool OpalConnection::RetrieveConnection()
 }
 
 
-PBoolean OpalConnection::IsConnectionOnHold()
+PBoolean OpalConnection::IsConnectionOnHold(bool /*fromRemote*/)
 {
-  return PFalse;
+  return false;
 }
 
 
@@ -289,14 +375,22 @@ void OpalConnection::OnHold(bool fromRemote, bool onHold)
 }
 
 
+PString OpalConnection::GetCallEndReasonText(CallEndReason reason)
+{
+  return psprintf(CallEndReasonStrings(reason.code), reason.q931);
+}
+
+
+void OpalConnection::SetCallEndReasonText(CallEndReasonCodes reasonCode, const PString & newText)
+{
+  CallEndReasonStrings.SetAt(reasonCode, newText);
+}
+
+
 void OpalConnection::SetCallEndReason(CallEndReason reason)
 {
   // Only set reason if not already set to something
   if (callEndReason == NumCallEndReasons) {
-    if ((reason & EndedWithQ931Code) != 0) {
-      SetQ931Cause((int)reason >> 24);
-      reason = (CallEndReason)(reason & 0xff);
-    }
     PTRACE(3, "OpalCon\tCall end reason for " << *this << " set to " << reason);
     callEndReason = reason;
     ownerCall.SetCallEndReason(reason);
@@ -387,23 +481,9 @@ void OpalConnection::OnReleased()
 }
 
 
-PBoolean OpalConnection::OnIncomingConnection()
-{
-  return PTrue;
-}
-
-
-PBoolean OpalConnection::OnIncomingConnection(unsigned int /*options*/)
-{
-  return PTrue;
-}
-
-
 PBoolean OpalConnection::OnIncomingConnection(unsigned options, OpalConnection::StringOptions * stringOptions)
 {
-  return OnIncomingConnection() &&
-         OnIncomingConnection(options) &&
-         endpoint.OnIncomingConnection(*this, options, stringOptions);
+  return endpoint.OnIncomingConnection(*this, options, stringOptions);
 }
 
 
@@ -523,12 +603,71 @@ void OpalConnection::OnEstablished()
 }
 
 
-void OpalConnection::AdjustMediaFormats(OpalMediaFormatList & mediaFormats) const
+void OpalConnection::AdjustMediaFormats(bool local, OpalMediaFormatList & mediaFormats, OpalConnection * otherConnection) const
 {
-  endpoint.AdjustMediaFormats(*this, mediaFormats);
+  if (otherConnection != NULL && otherConnection != this)
+    otherConnection->AdjustMediaFormats(local, mediaFormats, NULL);
+  else {
+    mediaFormats.Remove(m_connStringOptions(OPAL_OPT_REMOVE_CODEC).Lines());
+    endpoint.AdjustMediaFormats(local, *this, mediaFormats);
+  }
 }
 
 
+unsigned OpalConnection::GetNextSessionID(const OpalMediaType & /*mediaType*/, bool /*isSource*/)
+{
+  return 0;
+}
+
+
+void OpalConnection::AutoStartMediaStreams(bool force)
+{
+  OpalMediaTypeFactory::KeyList_T mediaTypes = OpalMediaType::GetList();
+  for (OpalMediaTypeFactory::KeyList_T::iterator iter = mediaTypes.begin(); iter != mediaTypes.end(); ++iter) {
+    OpalMediaType mediaType = *iter;
+    if ((GetAutoStart(mediaType)&OpalMediaType::Transmit) != 0 && (force || GetMediaStream(mediaType, true) == NULL))
+      ownerCall.OpenSourceMediaStreams(*this, mediaType, mediaType.GetDefinition()->GetDefaultSessionId());
+  }
+  StartMediaStreams();
+}
+
+
+#if OPAL_FAX
+bool OpalConnection::SwitchFaxMediaStreams(bool enableFax)
+{
+  if (m_faxMediaStreamsSwitchState != e_NotSwitchingFaxMediaStreams) {
+    PTRACE(2, "OpalCon\tNested call to SwitchFaxMediaStreams on " << *this);
+    return false;
+  }
+
+  PTRACE(3, "OpalCon\tSwitchFaxMediaStreams to " << (enableFax ? "fax" : "audio") << " on " << *this);
+  OpalMediaFormat format = enableFax ? OpalT38 : OpalG711uLaw;
+  if (!ownerCall.OpenSourceMediaStreams(*this, format.GetMediaType(), 1, format))
+    return false;
+
+  m_faxMediaStreamsSwitchState = enableFax ? e_SwitchingToFaxMediaStreams : e_SwitchingFromFaxMediaStreams;
+  return true;
+}
+
+
+void OpalConnection::OnSwitchedFaxMediaStreams(bool enabledFax)
+{
+  if (m_faxMediaStreamsSwitchState != e_NotSwitchingFaxMediaStreams) {
+    PTRACE(3, "OpalCon\tSwitch of media streams to "
+           << (m_faxMediaStreamsSwitchState == e_SwitchingToFaxMediaStreams ? "fax" : "audio") << ' '
+           << (enabledFax != (m_faxMediaStreamsSwitchState == e_SwitchingToFaxMediaStreams) ? "failed" : "succeeded")
+           << " on " << *this);
+
+    m_faxMediaStreamsSwitchState = e_NotSwitchingFaxMediaStreams;
+
+    PSafePtr<OpalConnection> other = GetOtherPartyConnection();
+    if (other != NULL)
+      other->OnSwitchedFaxMediaStreams(enabledFax);
+  }
+}
+#endif
+
+
 OpalMediaStreamPtr OpalConnection::OpenMediaStream(const OpalMediaFormat & mediaFormat, unsigned sessionID, bool isSource)
 {
   PSafeLockReadWrite safeLock(*this);
@@ -700,78 +839,148 @@ PBoolean OpalConnection::OnOpenMediaStream(OpalMediaStream & stream)
 }
 
 
+void OpalConnection::OnClosedMediaStream(const OpalMediaStream & stream)
+{
+#if OPAL_HAS_MIXER
+  OnStopRecording(stream.GetPatch());
+#endif
+  endpoint.OnClosedMediaStream(stream);
+}
+
+#if OPAL_HAS_MIXER
+
 static PString MakeRecordingKey(const OpalMediaPatch & patch)
 {
   return psprintf("%08x", &patch);
 }
 
 
-void OpalConnection::OnClosedMediaStream(const OpalMediaStream & stream)
+void OpalConnection::OnStartRecording(OpalMediaPatch * patch)
 {
-  OpalMediaPatch * patch = stream.GetPatch();
-  if (patch != NULL)
-    GetCall().OnStopRecordAudio(MakeRecordingKey(*patch));
+  if (patch == NULL)
+    return;
 
-  endpoint.OnClosedMediaStream(stream);
+  if (!ownerCall.OnStartRecording(MakeRecordingKey(*patch), patch->GetSource().GetMediaFormat())) {
+    PTRACE(4, "OpalCon\tNo record filter added on connection " << *this << ", patch " << *patch);
+    return;
+  }
+
+  patch->AddFilter(m_recordAudioNotifier, OPAL_PCM16);
+#if OPAL_VIDEO
+  patch->AddFilter(m_recordVideoNotifier, OPAL_YUV420P);
+#endif
+
+  PTRACE(4, "OpalCon\tAdded record filter on connection " << *this << ", patch " << *patch);
 }
 
 
-void OpalConnection::OnPatchMediaStream(PBoolean PTRACE_PARAM(isSource), OpalMediaPatch & patch)
+void OpalConnection::OnStopRecording(OpalMediaPatch * patch)
 {
-  if (!recordAudioFilename.IsEmpty())
-    GetCall().StartRecording(recordAudioFilename);
-  else if (GetCall().IsRecording()) {
-    patch.AddFilter(recordNotifier, OPAL_PCM16);
-    PTRACE(4, "OpalCon\tAdded record filter on connection " << *this << ", patch " << patch);
+  if (patch == NULL)
+    return;
+
+  ownerCall.OnStopRecording(MakeRecordingKey(*patch));
+
+  patch->RemoveFilter(m_recordAudioNotifier, OPAL_PCM16);
+#if OPAL_VIDEO
+  patch->RemoveFilter(m_recordVideoNotifier, OPAL_YUV420P);
+#endif
+
+  PTRACE(4, "OpalCon\tRemoved record filter on " << *patch);
+}
+
+#endif
+
+void OpalConnection::OnPatchMediaStream(PBoolean isSource, OpalMediaPatch & patch)
+{
+  patch.SetCommandNotifier(PCREATE_NOTIFIER(OnMediaCommand), !isSource);
+
+  OpalMediaFormat mediaFormat = patch.GetSource().GetMediaFormat();
+  if (mediaFormat.GetMediaType() == OpalMediaType::Audio()) {
+    PTRACE(3, "OpalCon\tAdding audio filters to patch " << patch);
+
+    if (isSource && silenceDetector != NULL) {
+      silenceDetector->SetParameters(endpoint.GetManager().GetSilenceDetectParams(), mediaFormat.GetClockRate());
+      patch.AddFilter(silenceDetector->GetReceiveHandler(), mediaFormat);
+    }
+#if OPAL_AEC
+    if (echoCanceler) {
+      echoCanceler->SetParameters(endpoint.GetManager().GetEchoCancelParams());
+      echoCanceler->SetClockRate(mediaFormat.GetClockRate());
+      patch.AddFilter(isSource ? echoCanceler->GetReceiveHandler() : echoCanceler->GetSendHandler(), OpalPCM16);
+    }
+#endif
+
+#if OPAL_PTLIB_DTMF
+    if (m_detectInBandDTMF && isSource) {
+      patch.AddFilter(m_dtmfNotifier, OPAL_PCM16);
+      PTRACE(4, "OpalCon\tAdded detect DTMF filter on connection " << *this << ", patch " << patch);
+    }
+
+    if (m_sendInBandDTMF && !isSource) {
+      m_installedInBandDTMF = true;
+      patch.AddFilter(PCREATE_NOTIFIER(OnSendInBandDTMF), OPAL_PCM16);
+      PTRACE(4, "OpalCon\tAdded send DTMF filter on connection " << *this << ", patch " << patch);
+    }
+#endif
   }
 
+#if OPAL_HAS_MIXER
+  if (!m_recordingFilename.IsEmpty())
+    ownerCall.StartRecording(m_recordingFilename);
+  else if (ownerCall.IsRecording())
+    OnStartRecording(&patch);
+#endif
+
   PTRACE(3, "OpalCon\t" << (isSource ? "Source" : "Sink") << " stream of connection " << *this << " uses patch " << patch);
 }
 
 
+#if OPAL_HAS_MIXER
 void OpalConnection::EnableRecording()
 {
-  if (!LockReadWrite())
-    return;
-
   OpalMediaStreamPtr stream = GetMediaStream(OpalMediaType::Audio(), true);
-  if (stream != NULL) {
-    OpalMediaPatch * patch = stream->GetPatch();
-    if (patch != NULL) {
-      patch->AddFilter(recordNotifier, OPAL_PCM16);
-      PTRACE(4, "OpalCon\tAdded record filter on " << MakeRecordingKey(*patch));
-    }
-  }
-  
-  UnlockReadWrite();
+  if (stream != NULL)
+    OnStartRecording(stream->GetPatch());
+
+  stream = GetMediaStream(OpalMediaType::Video(), true);
+  if (stream != NULL)
+    OnStartRecording(stream->GetPatch());
 }
 
 
 void OpalConnection::DisableRecording()
 {
-  if (!LockReadWrite())
-    return;
-
   OpalMediaStreamPtr stream = GetMediaStream(OpalMediaType::Audio(), true);
-  if (stream != NULL) {
-    OpalMediaPatch * patch = stream->GetPatch();
-    if (patch != NULL) {
-      patch->RemoveFilter(recordNotifier, OPAL_PCM16);
-      PTRACE(4, "OpalCon\tRemoved record filter on " << MakeRecordingKey(*patch));
-    }
-  }
-  
-  UnlockReadWrite();
+  if (stream != NULL)
+    OnStopRecording(stream->GetPatch());
+
+  stream = GetMediaStream(OpalMediaType::Video(), true);
+  if (stream != NULL)
+    OnStopRecording(stream->GetPatch());
 }
 
 
 void OpalConnection::OnRecordAudio(RTP_DataFrame & frame, INT param)
 {
   const OpalMediaPatch * patch = (const OpalMediaPatch *)param;
-  GetCall().OnRecordAudio(MakeRecordingKey(*patch), frame);
+  ownerCall.OnRecordAudio(MakeRecordingKey(*patch), frame);
 }
 
 
+#if OPAL_VIDEO
+
+void OpalConnection::OnRecordVideo(RTP_DataFrame & frame, INT param)
+{
+  const OpalMediaPatch * patch = (const OpalMediaPatch *)param;
+  ownerCall.OnRecordVideo(MakeRecordingKey(*patch), frame);
+}
+
+#endif
+
+#endif
+
+
 void OpalConnection::AttachRFC2833HandlerToPatch(PBoolean /*isSource*/, OpalMediaPatch & /*patch*/)
 {
 }
@@ -799,11 +1008,19 @@ OpalMediaStreamPtr OpalConnection::GetMediaStream(unsigned sessionId, bool sourc
 }
 
 
-OpalMediaStreamPtr OpalConnection::GetMediaStream(const OpalMediaType & mediaType, bool source) const
+OpalMediaStreamPtr OpalConnection::GetMediaStream(const OpalMediaType & mediaType,
+                                                  bool source,
+                                                  OpalMediaStreamPtr mediaStream) const
 {
-  for (OpalMediaStreamPtr mediaStream(mediaStreams, PSafeReference); mediaStream != NULL; ++mediaStream) {
-    if (mediaStream->GetMediaFormat().GetMediaType() == mediaType && mediaStream->IsSource() == source)
+  if (mediaStream == NULL)
+    mediaStream = OpalMediaStreamPtr(mediaStreams, PSafeReference);
+  else
+    ++mediaStream;
+
+  while (mediaStream != NULL) {
+    if ((mediaType.empty() || mediaStream->GetMediaFormat().GetMediaType() == mediaType) && mediaStream->IsSource() == source)
       return mediaStream;
+    ++mediaStream;
   }
 
   return NULL;
@@ -816,14 +1033,9 @@ PBoolean OpalConnection::IsMediaBypassPossible(unsigned /*sessionID*/) const
   return false;
 }
 
+
 #if OPAL_VIDEO
 
-void OpalConnection::AddVideoMediaFormats(OpalMediaFormatList & mediaFormats) const
-{
-  endpoint.AddVideoMediaFormats(mediaFormats, this);
-}
-
-
 PBoolean OpalConnection::CreateVideoInputDevice(const OpalMediaFormat & mediaFormat,
                                             PVideoInputDevice * & device,
                                             PBoolean & autoDelete)
@@ -941,10 +1153,28 @@ PBoolean OpalConnection::SendUserInputString(const PString & value)
 }
 
 
+#if OPAL_PTLIB_DTMF
+PBoolean OpalConnection::SendUserInputTone(char tone, unsigned duration)
+{
+  if (!m_installedInBandDTMF)
+    return false;
+
+  if (duration <= 0)
+    duration = PDTMFEncoder::DefaultToneLen;
+
+  PTRACE(3, "OPAL\tSending in-band DTMF tone '" << tone << "', duration=" << duration);
+  m_inBandMutex.Wait();
+  m_inBandDTMF.AddTone(tone, duration);
+  m_inBandMutex.Signal();
+
+  return true;
+}
+#else
 PBoolean OpalConnection::SendUserInputTone(char /*tone*/, unsigned /*duration*/)
 {
   return false;
 }
+#endif
 
 
 void OpalConnection::OnUserInputString(const PString & value)
@@ -998,22 +1228,47 @@ PBoolean OpalConnection::PromptUserInput(PBoolean /*play*/)
 
 
 #if OPAL_PTLIB_DTMF
-void OpalConnection::OnUserInputInBandDTMF(RTP_DataFrame & frame, INT)
+void OpalConnection::OnDetectInBandDTMF(RTP_DataFrame & frame, INT)
 {
   // This function is set up as an 'audio filter'.
   // This allows us to access the 16 bit PCM audio (at 8Khz sample rate)
   // before the audio is passed on to the sound card (or other output device)
 
   // Pass the 16 bit PCM audio through the DTMF decoder   
-  PString tones = dtmfDecoder.Decode((const short *)frame.GetPayloadPtr(), frame.GetPayloadSize()/sizeof(short), dtmfScaleMultiplier, dtmfScaleDivisor);
+  PString tones = m_dtmfDecoder.Decode((const short *)frame.GetPayloadPtr(),
+                                       frame.GetPayloadSize()/sizeof(short),
+                                       m_dtmfScaleMultiplier,
+                                       m_dtmfScaleDivisor);
   if (!tones.IsEmpty()) {
-    PTRACE(3, "OPAL\tDTMF detected. " << tones);
-    PINDEX i;
-    for (i = 0; i < tones.GetLength(); i++) {
-      OnUserInputTone(tones[i], 0);
-    }
+    PTRACE(3, "OPAL\tDTMF detected: \"" << tones << '"');
+    for (PINDEX i = 0; i < tones.GetLength(); i++)
+      OnUserInputTone(tones[i], PDTMFDecoder::DetectTime);
   }
 }
+
+void OpalConnection::OnSendInBandDTMF(RTP_DataFrame & frame, INT)
+{
+  if (m_inBandDTMF.IsEmpty())
+    return;
+
+  // Can't do the usual LockReadWrite() as deadlocks
+  m_inBandMutex.Wait();
+
+  PINDEX bytes = (m_inBandDTMF.GetSize() - m_emittedInBandDTMF)*sizeof(short);
+  if (bytes > frame.GetPayloadSize())
+    bytes = frame.GetPayloadSize();
+  memcpy(frame.GetPayloadPtr(), &m_inBandDTMF[m_emittedInBandDTMF], bytes);
+
+  m_emittedInBandDTMF += bytes/sizeof(short);
+
+  if (m_emittedInBandDTMF >= m_inBandDTMF.GetSize()) {
+    PTRACE(4, "OPAL\tSent in-band DTMF tone, " << m_inBandDTMF.GetSize() << " samples");
+    m_inBandDTMF.SetSize(0);
+    m_emittedInBandDTMF = 0;
+  }
+
+  m_inBandMutex.Signal();
+}
 #endif
 
 
@@ -1035,6 +1290,12 @@ PString OpalConnection::GetLocalPartyURL() const
 }
 
 
+bool OpalConnection::IsPresentationBlocked() const
+{
+  return m_connStringOptions.GetBoolean(OPAL_OPT_PRESENTATION_BLOCK);
+}
+
+
 static PString MakeURL(const PString & prefix, const PString & partyName)
 {
   if (partyName.IsEmpty())
@@ -1144,33 +1405,32 @@ void OpalConnection::ApplyStringOptions(OpalConnection::StringOptions & stringOp
   PTRACE(4, "OpalCon\tApplying string options:\n" << stringOptions);
 
   if (LockReadWrite()) {
+    PCaselessString str;
 
     m_connStringOptions = stringOptions;
-  
-    PCaselessString str = stringOptions(OPAL_OPT_ENABLE_INBAND_DTMF);
-    if (!str.IsEmpty())
-      detectInBandDTMF = str *= "true";
-    str = stringOptions(OPAL_OPT_DTMF_MULT);
-    if (!str.IsEmpty()) {
-      dtmfScaleMultiplier = str.AsInteger();
-      dtmfScaleDivisor    = 1;
-    }
-    str = stringOptions(OPAL_OPT_DTMF_DIV);
-    if (!str.IsEmpty())
-      dtmfScaleDivisor = str.AsInteger();
+
+#if OPAL_PTLIB_DTMF
+    m_sendInBandDTMF   = stringOptions.GetBoolean(OPAL_OPT_ENABLE_INBAND_DTMF, m_sendInBandDTMF);
+    m_detectInBandDTMF = stringOptions.GetBoolean(OPAL_OPT_DETECT_INBAND_DTMF, m_detectInBandDTMF);
+    m_sendInBandDTMF   = stringOptions.GetBoolean(OPAL_OPT_SEND_INBAND_DTMF,   m_sendInBandDTMF);
+
+    m_dtmfScaleMultiplier = stringOptions.GetInteger(OPAL_OPT_DTMF_MULT, m_dtmfScaleMultiplier);
+    m_dtmfScaleDivisor    = stringOptions.GetInteger(OPAL_OPT_DTMF_DIV,  m_dtmfScaleDivisor);
+#endif
 
     m_autoStartInfo.Initialise(stringOptions);
 
-    if (stringOptions.Contains(OPAL_OPT_DISABLE_JITTER))
+    if (stringOptions.GetBoolean(OPAL_OPT_DISABLE_JITTER))
       maxAudioJitterDelay = minAudioJitterDelay = 0;
-    str = stringOptions(OPAL_OPT_MAX_JITTER);
-    if (!str.IsEmpty())
-      maxAudioJitterDelay = str.AsUnsigned();
-    str = stringOptions(OPAL_OPT_MIN_JITTER);
-    if (!str.IsEmpty())
-      minAudioJitterDelay = str.AsUnsigned();
+    else {
+      maxAudioJitterDelay = stringOptions.GetInteger(OPAL_OPT_MAX_JITTER, maxAudioJitterDelay);
+      minAudioJitterDelay = stringOptions.GetInteger(OPAL_OPT_MIN_JITTER, minAudioJitterDelay);
+    }
+
+#if OPAL_HAS_MIXER
     if (stringOptions.Contains(OPAL_OPT_RECORD_AUDIO))
-      recordAudioFilename = m_connStringOptions(OPAL_OPT_RECORD_AUDIO);
+      m_recordingFilename = m_connStringOptions(OPAL_OPT_RECORD_AUDIO);
+#endif
 
     str = stringOptions(OPAL_OPT_ALERTING_TYPE);
     if (!str.IsEmpty())
@@ -1180,10 +1440,6 @@ void OpalConnection::ApplyStringOptions(OpalConnection::StringOptions & stringOp
   }
 }
 
-void OpalConnection::PreviewPeerMediaFormats(const OpalMediaFormatList & /*fmts*/)
-{
-}
-
 
 OpalMediaFormatList OpalConnection::GetMediaFormats() const
 {
@@ -1196,11 +1452,18 @@ OpalMediaFormatList OpalConnection::GetLocalMediaFormats()
   return ownerCall.GetMediaFormats(*this, FALSE);
 }
 
-void OpalConnection::OnMediaPatchStart(unsigned, bool)
-{ }
 
-void OpalConnection::OnMediaPatchStop(unsigned,  bool )
-{ }
+void OpalConnection::OnStartMediaPatch(OpalMediaPatch & patch)
+{
+  GetEndPoint().GetManager().OnStartMediaPatch(*this, patch);
+}
+
+
+void OpalConnection::OnStopMediaPatch(OpalMediaPatch & patch)
+{
+  GetEndPoint().GetManager().OnStopMediaPatch(*this, patch);
+}
+
 
 void OpalConnection::OnMediaCommand(OpalMediaCommand & /*command*/, INT /*extra*/)
 {
@@ -1308,60 +1571,77 @@ OpalMediaType::AutoStartMode OpalConnection::AutoStartMap::GetAutoStart(const Op
 
 #if OPAL_HAS_IM
 
-template <class DataType>
-bool SendIMHelper(OpalConnection & conn, const OpalMediaFormat & format, const DataType & data)
+bool OpalConnection::TransmitInternalIM(const OpalMediaFormat & format, RTP_IMFrame & frame)
 {
-  if (!conn.LockReadWrite())
-    return false;
-
-  bool stat = true;
-
-  OpalMediaStreamPtr strm = conn.GetMediaStream(format.GetMediaType(), true);
-  if (strm == NULL) 
-    stat = false;
-  else {
-    OpalIMMediaStream * imStream = dynamic_cast<OpalIMMediaStream *>(&*strm);
-    if (imStream != NULL) 
-      imStream->PushIM(data);
-  }
-  
-  conn.UnlockReadWrite();
-
-  return stat;
-}
-
-bool OpalConnection::SendIM(const OpalMediaFormat & format, const T140String & text)
-{
-  return SendIMHelper<T140String>(*this, format, text);
-}
-
-bool OpalConnection::SendIM(const OpalMediaFormat & format, const RTP_DataFrame & frame)
-{
-  return SendIMHelper<RTP_DataFrame>(*this, format, frame);
-}
-
-void OpalConnection::OnReceiveIM(const IMInfo & im)
-{
-  if (!LockReadWrite())
-    return;
-
-  for (PList<PNotifier>::iterator f = m_imListeners.begin(); f != m_imListeners.end(); ++f) {
-    (*f)((IMInfo &)im, (INT)this);
+  // if the call contains any compatible IM stream, use it
+  for (OpalMediaStreamPtr mediaStream(mediaStreams, PSafeReference); mediaStream != NULL; ++mediaStream) {
+    if (mediaStream->IsSource() && (mediaStream->GetMediaFormat() == format)) {
+      PTRACE(3, "OpalCon\tSending " << format << " IM message within call");
+      return mediaStream->GetPatch()->PushFrame(frame);
+    }
   }
 
-  UnlockReadWrite();
+  // otherwise push the frame directly to the other connection
+  PSafePtr<OpalConnection> other = GetOtherPartyConnection();
+  if (other != NULL) {
+    PTRACE(3, "OpalCon\tSending " << format << " IM message directly to other connection");
+    other->OnReceiveInternalIM(format, frame);
+  }
+  return true;
 }
 
-void OpalConnection::AddIMListener(const PNotifier & listener)
+
+void OpalConnection::OnReceiveInternalIM(const OpalMediaFormat & format, RTP_IMFrame & rtp)
 {
-  if (!LockReadWrite())
-    return;
+  TransmitExternalIM(format, rtp);
+}
 
-  m_imListeners.Append(new PNotifier(listener));
 
-  UnlockReadWrite();
+bool OpalConnection::TransmitExternalIM(const OpalMediaFormat & format, RTP_IMFrame & frame)
+{
+  return endpoint.TransmitExternalIM(*this, format, frame);
+}
+
+bool OpalConnection::OnReceiveExternalIM(const OpalMediaFormat & format, RTP_IMFrame & body)
+{
+  return TransmitInternalIM(format, body);
 }
 
 #endif
 
+
+bool OpalConnection::StringOptions::GetBoolean(const char * key, bool dflt) const
+{
+  PString * value = GetAt(key);
+  if (value == NULL)
+    return dflt;
+
+  return value->IsEmpty() || (*value *= "true") || (*value *= "yes") || value->AsInteger() != 0;
+}
+
+
+long OpalConnection::StringOptions::GetInteger(const char * key, long dflt) const
+{
+  PString * value = GetAt(key);
+  if (value == NULL)
+    return dflt;
+
+  return value->AsInteger();
+}
+
+
+void OpalConnection::StringOptions::ExtractFromURL(PURL & url)
+{
+  PStringToString params = url.GetParamVars();
+  params.MakeUnique();
+  for (PINDEX i = 0; i < params.GetSize(); ++i) {
+    PCaselessString key = params.GetKeyAt(i);
+    if (key.NumCompare("OPAL-") == EqualTo) {
+      SetAt(key.Mid(5), params.GetDataAt(i));
+      url.SetParamVar(key, PString::Empty());
+    }
+  }
+}
+
+
 /////////////////////////////////////////////////////////////////////////////
