@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23601 $
+ * $Revision: 23926 $
  * $Author: rjongbloed $
- * $Date: 2009-10-07 07:53:21 +0000 (Wed, 07 Oct 2009) $
+ * $Date: 2010-01-12 18:44:33 -0600 (Tue, 12 Jan 2010) $
  */
 
 #include <ptlib.h>
@@ -50,19 +50,6 @@
 
 #define new PNEW
 
-namespace PWLibStupidLinkerHacks {
-extern int opalLoader;
-
-static class InstantiateMe
-{
-  public:
-    InstantiateMe()
-    {
-      opalLoader = 1;
-    }
-} instance;
-
-}; // namespace PWLibStupidLinkerHacks
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -392,10 +379,10 @@ PBoolean OpalEndPoint::HasConnection(const PString & token)
 }
 
 
-PBoolean OpalEndPoint::AddConnection(OpalConnection * connection)
+OpalConnection * OpalEndPoint::AddConnection(OpalConnection * connection)
 {
   if (connection == NULL)
-    return PFalse;
+    return NULL;
 
   connection->SetStringOptions(m_defaultStringOptions, false);
 
@@ -403,7 +390,7 @@ PBoolean OpalEndPoint::AddConnection(OpalConnection * connection)
 
   connectionsActive.SetAt(connection->GetToken(), connection);
 
-  return PTrue;
+  return connection;
 }
 
 
@@ -420,23 +407,9 @@ PBoolean OpalEndPoint::OnSetUpConnection(OpalConnection & PTRACE_PARAM(connectio
 }
 
 
-PBoolean OpalEndPoint::OnIncomingConnection(OpalConnection & /*connection*/)
-{
-  return PTrue;
-}
-
-
-PBoolean OpalEndPoint::OnIncomingConnection(OpalConnection & /*connection*/, unsigned /*options*/)
-{
-  return PTrue;
-}
-
-
 PBoolean OpalEndPoint::OnIncomingConnection(OpalConnection & connection, unsigned options, OpalConnection::StringOptions * stringOptions)
 {
-  return OnIncomingConnection(connection) &&
-         OnIncomingConnection(connection, options) &&
-         manager.OnIncomingConnection(connection, options, stringOptions);
+  return manager.OnIncomingConnection(connection, options, stringOptions);
 }
 
 
@@ -530,31 +503,11 @@ void OpalEndPoint::ClearAllCalls(OpalConnection::CallEndReason reason, PBoolean
 }
 
 
-OpalMediaFormatList OpalEndPoint::GetMediaFormats() const
-{
-  OpalMediaFormatList formats;
-
-  // Sound cards can only do 16 bit PCM, but at various sample rates
-  // The following will be in order of preference, so lets do wideband first
-  formats += OpalPCM16_48KHZ;
-  formats += OpalPCM16_32KHZ;
-  formats += OpalPCM16_16KHZ;
-  formats += OpalPCM16;
-
-#if OPAL_VIDEO
-  AddVideoMediaFormats(formats);
-#endif
-
-  AddIMMediaFormats(formats);
-
-  return formats;
-}
-
-
-void OpalEndPoint::AdjustMediaFormats(const OpalConnection & connection,
+void OpalEndPoint::AdjustMediaFormats(bool local,
+                                      const OpalConnection & connection,
                                       OpalMediaFormatList & mediaFormats) const
 {
-  manager.AdjustMediaFormats(connection, mediaFormats);
+  manager.AdjustMediaFormats(local, connection, mediaFormats);
 }
 
 
@@ -571,13 +524,6 @@ void OpalEndPoint::OnClosedMediaStream(const OpalMediaStream & stream)
 }
 
 #if OPAL_VIDEO
-void OpalEndPoint::AddVideoMediaFormats(OpalMediaFormatList & mediaFormats,
-                                        const OpalConnection * connection) const
-{
-  manager.AddVideoMediaFormats(mediaFormats, connection);
-}
-
-
 PBoolean OpalEndPoint::CreateVideoInputDevice(const OpalConnection & connection,
                                           const OpalMediaFormat & mediaFormat,
                                           PVideoInputDevice * & device,
@@ -652,10 +598,45 @@ PString OpalEndPoint::GetSSLCertificate() const
 }
 #endif
 
-void OpalEndPoint::AddIMMediaFormats(OpalMediaFormatList & mediaFormats, const OpalConnection * connection) const
+
+bool OpalEndPoint::Message(const PString & to, const PString & body)
 {
-  manager.AddIMMediaFormats(mediaFormats, connection);
+  PURL from;
+  PString conversationId;
+  return Message(to, "text/plain", body, from, conversationId);
 }
 
 
+PBoolean OpalEndPoint::Message(
+  const PURL & /*to*/, 
+  const PString & /*type*/,
+  const PString & /*body*/,
+  PURL & /*from*/, 
+  PString & /*conversationId*/
+)
+{
+  return false;
+}
+
+void OpalEndPoint::OnMessageReceived(
+  const PURL & from, 
+  const PString & fromName,
+  const PURL & to, 
+  const PString & type,
+  const PString & body,
+  const PString & conversationId
+)
+{
+  manager.OnMessageReceived(from, fromName, to, type, body, conversationId);
+}
+
+#if OPAL_HAS_IM
+
+bool OpalEndPoint::TransmitExternalIM(OpalConnection & conn, const OpalMediaFormat & format, RTP_IMFrame & frame)
+{
+  return manager.TransmitExternalIM(conn, format, frame);
+}
+
+#endif
+
 /////////////////////////////////////////////////////////////////////////////
