@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22630 $
+ * $Revision: 23978 $
  * $Author: rjongbloed $
- * $Date: 2009-05-18 06:12:59 +0000 (Mon, 18 May 2009) $
+ * $Date: 2010-01-24 19:28:06 -0600 (Sun, 24 Jan 2010) $
  */
 
 #include <ptlib.h>
@@ -38,6 +38,7 @@
 
 #include <opal/ivr.h>
 #include <opal/call.h>
+#include <opal/patch.h>
 
 
 #define new PNEW
@@ -48,20 +49,22 @@
 /////////////////////////////////////////////////////////////////////////////
 
 OpalIVREndPoint::OpalIVREndPoint(OpalManager & mgr, const char * prefix)
-  : OpalEndPoint(mgr, prefix, CanTerminateCall),
+  : OpalLocalEndPoint(mgr, prefix),
     defaultVXML("<?xml version=\"1.0\"?>\n"
                 "<vxml version=\"1.0\">\n"
                 "  <form id=\"root\">\n"
-                "    <audio src=\"welcome.wav\">\n"
+                "    <audio src=\"file:welcome.wav\">\n"
                 "      This is the OPAL, V X M L test program, please speak after the tone.\n"
                 "    </audio>\n"
-                "    <record name=\"msg\" beep=\"true\" dtmfterm=\"true\" dest=\"recording.wav\" maxtime=\"10s\"/>\n"
+                "    <record name=\"msg\" beep=\"true\" dtmfterm=\"true\" dest=\"file:recording.wav\" maxtime=\"10s\"/>\n"
                 "  </form>\n"
                 "</vxml>\n")
 {
-  nextTokenNumber = 1;
-
   defaultMediaFormats += OpalPCM16;
+  defaultMediaFormats += OpalRFC2833;
+#if OPAL_T38_CAPABILITY
+  defaultMediaFormats += OpalCiscoNSE;
+#endif
 
   PTRACE(4, "IVR\tCreated endpoint.");
 }
@@ -73,11 +76,11 @@ OpalIVREndPoint::~OpalIVREndPoint()
 }
 
 
-PBoolean OpalIVREndPoint::MakeConnection(OpalCall & call,
-                                     const PString & remoteParty,
-                                     void * userData,
-                               unsigned int /*options*/,
-                               OpalConnection::StringOptions * stringOptions)
+PSafePtr<OpalConnection> OpalIVREndPoint::MakeConnection(OpalCall & call,
+                                                    const PString & remoteParty,
+                                                             void * userData,
+                                                       unsigned int options,
+                                    OpalConnection::StringOptions * stringOptions)
 {
   PString ivrString = remoteParty;
 
@@ -92,7 +95,7 @@ PBoolean OpalIVREndPoint::MakeConnection(OpalCall & call,
   if (vxml.IsEmpty() || vxml == "*")
     vxml = defaultVXML;
 
-  return AddConnection(CreateConnection(call, CreateConnectionToken(), userData, vxml, stringOptions));
+  return AddConnection(CreateConnection(call, userData, vxml, options, stringOptions));
 }
 
 
@@ -104,21 +107,12 @@ OpalMediaFormatList OpalIVREndPoint::GetMediaFormats() const
 
 
 OpalIVRConnection * OpalIVREndPoint::CreateConnection(OpalCall & call,
-                                                      const PString & token,
                                                       void * userData,
                                                       const PString & vxml,
+                                                      unsigned int options,
                                                       OpalConnection::StringOptions * stringOptions)
 {
-  return new OpalIVRConnection(call, *this, token, userData, vxml, stringOptions);
-}
-
-
-PString OpalIVREndPoint::CreateConnectionToken()
-{
-  inUseFlag.Wait();
-  unsigned tokenNumber = nextTokenNumber++;
-  inUseFlag.Signal();
-  return psprintf("IVR/%u", tokenNumber);
+  return new OpalIVRConnection(call, *this, userData, vxml, options, stringOptions);
 }
 
 
@@ -147,11 +141,11 @@ PBoolean OpalIVREndPoint::StartVXML()
 
 OpalIVRConnection::OpalIVRConnection(OpalCall & call,
                                      OpalIVREndPoint & ep,
-                                     const PString & token,
-                                     void * /*userData*/,
+                                     void * userData,
                                      const PString & vxml,
+                                     unsigned int options,
                                      OpalConnection::StringOptions * stringOptions)
-  : OpalConnection(call, ep, token, 0, stringOptions),
+  : OpalLocalConnection(call, ep, userData, options, stringOptions, 'I'),
     endpoint(ep),
     vxmlToLoad(vxml),
     vxmlMediaFormats(ep.GetMediaFormats()),
@@ -180,24 +174,6 @@ PString OpalIVRConnection::GetLocalPartyURL() const
 }
 
 
-PBoolean OpalIVRConnection::SetUpConnection()
-{
-  originating = false;
-
-  OnApplyStringOptions();
-
-  PTRACE(3, "IVR\tSetUpConnection(" << remotePartyName << ')');
-
-  SetPhase(AlertingPhase);
-  OnAlerting();
-
-  OnConnectedInternal();
-
-  StartMediaStreams();
-
-  return PTrue;
-}
-
 void OpalIVRConnection::OnEstablished()
 {
   OpalConnection::OnEstablished();
@@ -230,14 +206,22 @@ PBoolean OpalIVRConnection::StartVXML()
   if (vxmlToLoad.IsEmpty())
     return endpoint.StartVXML();
 
-  if (vxmlToLoad.Find("<?xml") == 0) {
+  PCaselessString vxmlHead = vxmlToLoad.LeftTrim().Left(5);
+  if (vxmlHead == "<?xml" || vxmlHead == "<vxml") {
     PTRACE(4, "IVR\tStartVXML:\n" << vxmlToLoad);
     return vxmlSession.LoadVXML(vxmlToLoad);
   }
 
-  PFilePath vxmlFile = PURL(vxmlToLoad).AsFilePath();
-  if (vxmlFile.GetType() *= ".vxml")
-    return vxmlSession.LoadFile(vxmlFile);
+  PURL vxmlURL(vxmlToLoad, NULL);
+  if (vxmlURL.IsEmpty()) {
+    PFilePath vxmlFile = vxmlToLoad;
+    if (vxmlFile.GetType() *= ".vxml")
+      return vxmlSession.LoadFile(vxmlFile);
+  }
+  else {
+    if (vxmlURL.GetScheme() != "file" || (vxmlURL.AsFilePath().GetType() *= ".vxml"))
+      return vxmlSession.LoadURL(vxmlURL);
+  }
 
   ////////////////////////////////
 
@@ -266,7 +250,7 @@ PBoolean OpalIVRConnection::StartVXML()
       if (!voice.IsEmpty())
         fn = fn.GetDirectory() + voice + PDIR_SEPARATOR + fn.GetFileName();
 
-      PTRACE(3, "IVR\tPlaying file " << fn);
+      PTRACE(3, "IVR\tPlaying file " << fn << ' ' << repeat << " times, " << delay << "ms");
       vxmlSession.PlayFile(fn, repeat, delay);
       continue;
     }
@@ -325,47 +309,6 @@ PBoolean OpalIVRConnection::StartVXML()
   return PTrue;
 }
 
-PBoolean OpalIVRConnection::SetAlerting(const PString & calleeName, PBoolean)
-{
-  PTRACE(3, "IVR\tSetAlerting(" << calleeName << ')');
-
-  if (!LockReadWrite())
-    return PFalse;
-
-  SetPhase(AlertingPhase);
-  remotePartyName = calleeName;
-  UnlockReadWrite();
-
-  return PTrue;
-}
-
-
-PBoolean OpalIVRConnection::SetConnected()
-{
-  PTRACE(3, "IVR\tSetConnected()");
-
-  PSafeLockReadWrite safeLock(*this);
-  if (!safeLock.IsLocked())
-    return PFalse;
-
-  if (!StartVXML()) {
-    PTRACE(1, "IVR\tVXML session not loaded, aborting.");
-    Release(EndedByLocalUser);
-    return PFalse;
-  }
-
-  // if no media streams, try and start them
-  if (mediaStreams.IsEmpty()) {
-    ownerCall.OpenSourceMediaStreams(*this, OpalMediaType::Audio(), 1);
-    PSafePtr<OpalConnection> otherParty = GetOtherPartyConnection();
-    if (otherParty != NULL)
-      ownerCall.OpenSourceMediaStreams(*otherParty, OpalMediaType::Audio(), 1);
-  }
-
-  // if we have media streams, move to Established straight away
-  return OpalConnection::SetConnected();
-}
-
 
 OpalMediaFormatList OpalIVRConnection::GetMediaFormats() const
 {
@@ -391,11 +334,10 @@ PBoolean OpalIVRConnection::SendUserInputString(const PString & value)
   return PTrue;
 }
 
-void OpalIVRConnection::OnMediaPatchStop(unsigned sessionId, bool)
+void OpalIVRConnection::OnStopMediaPatch(OpalMediaPatch & patch)
 {
   // lose the audio, then lose the call
-  OpalMediaStreamPtr stream = GetMediaStream(OpalMediaType::Audio(), true);
-  if (stream == NULL || !stream->IsOpen() || stream->GetSessionID() == sessionId) {
+  if (patch.GetSource().GetMediaFormat().GetMediaType() == OpalMediaType::Audio()) {
     synchronousOnRelease = false; // Deadlocks if try to do it synchronously ...
     Release();
   }
