@@ -24,9 +24,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 23028 $
+ * $Revision: 23973 $
  * $Author: rjongbloed $
- * $Date: 2009-07-02 10:39:41 +0000 (Thu, 02 Jul 2009) $
+ * $Date: 2010-01-24 18:33:51 -0600 (Sun, 24 Jan 2010) $
  */
 
 #include <ptlib.h>
@@ -52,10 +52,16 @@
 /////////////////////////////////////////////////////////////////////////////
 
 OpalMediaPatch::OpalMediaPatch(OpalMediaStream & src)
-: source(src)
+  : source(src)
+#if OPAL_VIDEO
+  , m_videoDecoder(src.GetMediaFormat().IsTransportable() && src.GetMediaFormat().GetMediaType() == OpalMediaType::Video())
+#endif
+  , m_bypassActive(false)
+  , m_bypassToPatch(NULL)
+  , m_bypassFromPatch(NULL)
+  , patchThread(NULL)
 {
   src.SetPatch(this);
-  patchThread = NULL;
   PTRACE(5, "Patch\tCreated media patch " << this);
 }
 
@@ -75,7 +81,7 @@ OpalMediaPatch::~OpalMediaPatch()
 
 void OpalMediaPatch::PrintOn(ostream & strm) const
 {
-  strm << "Patch " << source;
+  strm << "Patch[" << this << "] " << source;
 
   inUse.StartRead();
 
@@ -115,8 +121,15 @@ void OpalMediaPatch::Close()
   PTRACE(3, "Patch\tClosing media patch " << *this);
 
   inUse.StartWrite();
+
+  if (m_bypassFromPatch != NULL)
+    m_bypassFromPatch->SetBypassPatch(NULL);
+  else
+    SetBypassPatch(NULL);
+
   filters.RemoveAll();
-  source.Close();
+  if (source.GetPatch() == this)
+    source.Close();
 
   while (sinks.GetSize() > 0) {
     OpalMediaStreamPtr stream = sinks.front().stream;
@@ -139,7 +152,7 @@ void OpalMediaPatch::Close()
       return;
     }
   }
-  
+
   inUse.EndWrite();
 }
 
@@ -170,6 +183,12 @@ PBoolean OpalMediaPatch::AddSink(const OpalMediaStreamPtr & sinkStream)
             "Destination format:\n" << setw(-1) << destinationFormat);
 
   if (sourceFormat == destinationFormat) {
+    PINDEX framesPerPacket = destinationFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(),
+                                  sourceFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), 1));
+    PINDEX packetSize = sourceFormat.GetFrameSize()*framesPerPacket;
+    PINDEX packetTime = sourceFormat.GetFrameTime()*framesPerPacket;
+    source.SetDataSize(packetSize, packetTime);
+    sinkStream->SetDataSize(packetSize, packetTime);
     PTRACE(3, "Patch\tAdded direct media stream sink " << *sinkStream);
     return true;
   }
@@ -179,7 +198,7 @@ PBoolean OpalMediaPatch::AddSink(const OpalMediaStreamPtr & sinkStream)
   if (sink->primaryCodec != NULL) {
     PTRACE(4, "Patch\tCreated primary codec " << sourceFormat << "->" << destinationFormat << " with ID " << id);
 
-    if (!sinkStream->SetDataSize(sink->primaryCodec->GetOptimalDataFrameSize(false), destinationFormat.GetFrameSize())) {
+    if (!sinkStream->SetDataSize(sink->primaryCodec->GetOptimalDataFrameSize(false), sourceFormat.GetFrameTime())) {
       PTRACE(1, "Patch\tSink stream " << *sinkStream << " cannot support data size "
               << sink->primaryCodec->GetOptimalDataFrameSize(PFalse));
       return false;
@@ -197,12 +216,31 @@ PBoolean OpalMediaPatch::AddSink(const OpalMediaStreamPtr & sinkStream)
       return false;
     }
 
+    if (intermediateFormat.GetMediaType() == OpalMediaType::Audio()) {
+      // try prepare intermediateFormat for correct frames to frames transcoding
+      // so we need make sure that tx frames time of destinationFormat be equal 
+      // to tx frames time of intermediateFormat (all this does not produce during
+      // Merge phase in FindIntermediateFormat)
+      int destinationPacketTime = destinationFormat.GetFrameTime()*destinationFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), 1);
+      if ((destinationPacketTime % intermediateFormat.GetFrameTime()) != 0) {
+        PTRACE(1, "Patch\tCould produce without buffered media format converting (which not implemented yet) for " << *sinkStream);
+        return false;
+      }
+      intermediateFormat.AddOption(new OpalMediaOptionUnsigned(OpalAudioFormat::TxFramesPerPacketOption(),
+                                                               true,
+                                                               OpalMediaOption::NoMerge,
+                                                               destinationPacketTime/intermediateFormat.GetFrameTime()),
+                                   true);
+    }
+
     sink->primaryCodec = OpalTranscoder::Create(sourceFormat, intermediateFormat, (const BYTE *)id, id.GetLength());
     sink->secondaryCodec = OpalTranscoder::Create(intermediateFormat, destinationFormat, (const BYTE *)id, id.GetLength());
 
     PTRACE(4, "Patch\tCreated two stage codec " << sourceFormat << "/" << intermediateFormat << "/" << destinationFormat << " with ID " << id);
 
-    if (!sinkStream->SetDataSize(sink->secondaryCodec->GetOptimalDataFrameSize(false), destinationFormat.GetFrameSize())) {
+    sink->primaryCodec->SetMaxOutputSize(sink->secondaryCodec->GetOptimalDataFrameSize(true));
+
+    if (!sinkStream->SetDataSize(sink->secondaryCodec->GetOptimalDataFrameSize(false), sourceFormat.GetFrameTime())) {
       PTRACE(1, "Patch\tSink stream " << *sinkStream << " cannot support data size "
               << sink->secondaryCodec->GetOptimalDataFrameSize(PFalse));
       return false;
@@ -214,7 +252,7 @@ PBoolean OpalMediaPatch::AddSink(const OpalMediaStreamPtr & sinkStream)
            << " and " << *sink->secondaryCodec << ", data size=" << sinkStream->GetDataSize());
   }
 
-  source.SetDataSize(sink->primaryCodec->GetOptimalDataFrameSize(true), sourceFormat.GetFrameSize());
+  source.SetDataSize(sink->primaryCodec->GetOptimalDataFrameSize(true), destinationFormat.GetFrameTime());
   return true;
 }
 
@@ -236,6 +274,9 @@ void OpalMediaPatch::RemoveSink(const OpalMediaStreamPtr & stream)
     }
   }
 
+  if (m_bypassFromPatch != NULL && sinks.IsEmpty())
+    m_bypassFromPatch->SetBypassPatch(NULL);
+
   inUse.EndWrite();
 }
 
@@ -349,11 +390,16 @@ OpalMediaPatch::Sink::~Sink()
 void OpalMediaPatch::AddFilter(const PNotifier & filter, const OpalMediaFormat & stage)
 {
   PWriteWaitAndSignal mutex(inUse);
-  
+
+  if (source.GetMediaFormat().GetMediaType() != stage.GetMediaType())
+    return;
+
   // ensures that a filter is added only once
   for (PList<Filter>::iterator f = filters.begin(); f != filters.end(); ++f) {
-    if (f->notifier == filter && f->stage == stage)
+    if (f->notifier == filter && f->stage == stage) {
+      PTRACE(3, "OpalCon\tFilter already added for stage " << stage);
       return;
+    }
   }
   filters.Append(new Filter(filter, stage));
 }
@@ -366,11 +412,12 @@ PBoolean OpalMediaPatch::RemoveFilter(const PNotifier & filter, const OpalMediaF
   for (PList<Filter>::iterator f = filters.begin(); f != filters.end(); ++f) {
     if (f->notifier == filter && f->stage == stage) {
       filters.erase(f);
-      return PTrue;
+      return true;
     }
   }
 
-  return PFalse;
+  PTRACE(2, "OpalCon\tNo filter to remove for stage " << stage);
+  return false;
 }
 
 
@@ -430,21 +477,18 @@ void OpalMediaPatch::SetCommandNotifier(const PNotifier & notifier, PBoolean fro
 }
 
 
-bool OpalMediaPatch::OnPatchStart()
+bool OpalMediaPatch::OnStartMediaPatch()
 {
-  source.OnPatchStart();
+  source.OnStartMediaPatch();
 
   if (source.IsSynchronous())
     return false;
 
-  bool isAudio = source.GetMediaFormat().GetMediaType() == OpalMediaType::Audio();
-
   PReadWaitAndSignal mutex(inUse);
 
   for (PList<Sink>::iterator s = sinks.begin(); s != sinks.end(); ++s) {
     if (s->stream->IsSynchronous()) {
-      if (isAudio) 
-        source.EnableJitterBuffer();
+      source.EnableJitterBuffer();
       return false;
     }
   }
@@ -457,7 +501,8 @@ void OpalMediaPatch::Main()
 {
   PTRACE(4, "Patch\tThread started for " << *this);
 	
-  bool asynchronous = OnPatchStart();
+  bool asynchronous = OnStartMediaPatch();
+  PTimeInterval lastTick;
 
   /* Note the RTP frame is outside loop so that a) it is more efficient
      for memory usage, the buffer is only ever increased and not allocated
@@ -486,33 +531,104 @@ void OpalMediaPatch::Main()
     inUse.EndRead();
 
     if (!written) {
-      PTRACE(4, "Patch\tThread ended because all sink writes failed failed");
+      PTRACE(4, "Patch\tThread ended because all sink writes failed");
       break;
     }
  
-    // Don't starve the CPU if we have idle frames and the no source or destination is synchronous.
-    // Note that performing a Yield is not good enough, as the media patch threads are
-    // high priority and will consume all available CPU if allowed.
-    if (asynchronous)
-      PThread::Sleep(5);
+    /* Don't starve the CPU if we have idle frames and the no source or
+       destination is synchronous. Note that performing a Yield is not good
+       enough, as the media patch threads are high priority and will consume
+       all available CPU if allowed. Also just doing a sleep each time around
+       the loop slows down video where you get clusters of packets thrown at
+       us, want to clear them as quickly as possible out of the UDP OS buffers
+       or we overflow and lose some. Best compromise is to every 100ms, sleep
+       for 10ms so can not use more than about 90% of CPU. */
+    if (asynchronous && PTimer::Tick() - lastTick > 100) {
+      PThread::Sleep(10);
+      lastTick = PTimer::Tick();
+    }
   }
 
-  source.OnPatchStop();
+  source.OnStopMediaPatch(*this);
 
   PTRACE(4, "Patch\tThread ended for " << *this);
 }
 
 
+bool OpalMediaPatch::SetBypassPatch(OpalMediaPatch * patch)
+{
+  PTRACE(4, "Patch\tSetting media patch bypass to " << patch << " on " << *this);
+
+  PWriteWaitAndSignal mutex(inUse);
+
+  if (!PAssert(m_bypassFromPatch == NULL, PLogicError))
+    return false; // Can't be both!
+
+  if (m_bypassToPatch == patch)
+    return true; // Already set
+
+  if (m_bypassToPatch != NULL) {
+    if (!PAssert(m_bypassToPatch->m_bypassFromPatch == this, PLogicError))
+      return false;
+
+    m_bypassToPatch->m_bypassFromPatch = NULL;
+    m_bypassToPatch->m_bypassEnded.Signal();
+  }
+
+  if (patch != NULL) {
+    if (!PAssert(patch->m_bypassFromPatch == NULL, PLogicError))
+      return false;
+
+    patch->m_bypassFromPatch = this;
+  }
+
+  m_bypassToPatch = patch;
+
+#if OPAL_VIDEO
+  m_bypassActive = m_bypassToPatch != NULL && !m_videoDecoder;
+#else
+  m_bypassActive = m_bypassToPatch != NULL;
+#endif
+
+  return true;
+}
+
+
+PBoolean OpalMediaPatch::PushFrame(RTP_DataFrame & frame)
+{
+  PReadWaitAndSignal mutex(inUse);
+  return DispatchFrame(frame);
+}
+
+
 bool OpalMediaPatch::DispatchFrame(RTP_DataFrame & frame)
 {
-  FilterFrame(frame, source.GetMediaFormat());    
+  if (m_bypassFromPatch != NULL) {
+    PTRACE(3, "Patch\tMedia patch bypass started by " << m_bypassFromPatch << " on " << *this);
+    inUse.EndRead();
+    m_bypassEnded.Wait();
+    PTRACE(4, "Patch\tMedia patch bypass ended on " << *this);
+    inUse.StartRead();
+    return true;
+  }
+
+  FilterFrame(frame, source.GetMediaFormat());
 
   bool written = false;
-  for (PList<Sink>::iterator s = sinks.begin(); s != sinks.end(); ++s) {
-    if (s->WriteFrame(frame))
-      written = true;
-    else {
-      PTRACE(2, "Patch\tWriteFrame failed");
+
+  // This is not an if/else as WriteFrame can change the state of m_bypassActive
+  if (!m_bypassActive) {
+    for (PList<Sink>::iterator s = sinks.begin(); s != sinks.end(); ++s) {
+      if (s->WriteFrame(frame))
+        written = true;
+    }
+  }
+
+  if (m_bypassActive) {
+    written = false;
+    for (PList<Sink>::iterator s = m_bypassToPatch->sinks.begin(); s != m_bypassToPatch->sinks.end(); ++s) {
+      if (s->stream->WritePacket(frame))
+        written = true;
     }
   }
 
@@ -639,12 +755,15 @@ bool OpalMediaPatch::Sink::WriteFrame(RTP_DataFrame & sourceFrame)
   if (!writeSuccessful)
     return false;
   
+  if (stream->IsPaused())
+    return true;
+
 #if OPAL_VIDEO
   if (rateController != NULL) {
     bool forceIFrame = false;
     bool s = RateControlExceeded(forceIFrame);
     if (forceIFrame)
-      stream->ExecuteCommand(OpalVideoUpdatePicture());
+      stream->ExecuteCommand(OpalVideoUpdatePicture2(sourceFrame.GetSequenceNumber(), sourceFrame.GetTimestamp()));
     if (s) {
       if (secondaryCodec == NULL) {
         bool wasIFrame = false;
@@ -677,6 +796,12 @@ bool OpalMediaPatch::Sink::WriteFrame(RTP_DataFrame & sourceFrame)
   }
 
 #if OPAL_VIDEO
+  if (!patch.m_bypassActive && patch.m_videoDecoder && patch.m_bypassToPatch != NULL &&
+                dynamic_cast<OpalVideoTranscoder *>(primaryCodec)->WasLastFrameIFrame()) {
+    PTRACE(3, "Patch\tActivating video patch bypass to " << patch.m_bypassToPatch << " on " << patch);
+    patch.m_bypassActive = true;
+  }
+
   if (secondaryCodec == NULL && rateController != NULL) {
     PTRACE(4, "Patch\tPushing " << intermediateFrames.GetSize() << " packet into RC");
     rateController->Push(intermediateFrames, ((OpalVideoTranscoder *)primaryCodec)->WasLastFrameIFrame());
@@ -747,21 +872,13 @@ OpalMediaPatch::Thread::Thread(OpalMediaPatch & p)
 
 
 OpalPassiveMediaPatch::OpalPassiveMediaPatch(OpalMediaStream & source)
-: OpalMediaPatch(source)
+  : OpalMediaPatch(source)
 {
 }
 
 
 void OpalPassiveMediaPatch::Start()
 {
-  source.OnPatchStart();
+  OnStartMediaPatch();
 }
 
-
-PBoolean OpalPassiveMediaPatch::PushFrame(RTP_DataFrame & frame)
-{
-  PReadWaitAndSignal mutex(inUse);
-  return DispatchFrame(frame);
-}
-
-
