@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22699 $
+ * $Revision: 23741 $
  * $Author: rjongbloed $
- * $Date: 2009-05-22 00:39:55 -0500 (Fri, 22 May 2009) $
+ * $Date: 2009-11-01 20:27:28 -0600 (Sun, 01 Nov 2009) $
  */
 
 #include <ptlib.h>
@@ -38,6 +38,8 @@
 
 #include <opal/pcss.h>
 
+#if OPAL_HAS_PCSS
+
 #if OPAL_PTLIB_AUDIO
 
 #if OPAL_VIDEO
@@ -46,10 +48,7 @@
 #endif
 
 #include <ptlib/sound.h>
-#include <codec/silencedetect.h>
-#include <codec/echocancel.h>
 #include <opal/call.h>
-#include <opal/patch.h>
 #include <opal/manager.h>
 
 #if OPAL_HAS_IM
@@ -76,27 +75,6 @@ class PCSSIMStream : public OpalIMMediaStream
       PAssertAlways("Cannot ReadData from OpalSIPIMMediaStream");
       return false;
     }
-
-    /**Write raw media data to the sink media stream.
-       The default behaviour writes to the PChannel object.
-      */
-    virtual PBoolean WritePacket(RTP_DataFrame & packet)
-    {
-      if (!IsOpen())
-        return false;
-
-      OpalConnection::IMInfo info;
-      info.sessionId   = sessionID;
-      info.mediaFormat = mediaFormat;
-      info.body        = T140String(packet.GetPayloadPtr(), packet.GetPayloadSize());
-
-      connection.OnReceiveIM(info);
-
-      return true;
-    }
-
-    virtual PBoolean RequiresPatchThread() const   { return !isSource; }
-
 };
 
 #endif  // OPAL_HAS_IM
@@ -106,18 +84,19 @@ class PCSSIMStream : public OpalIMMediaStream
 /////////////////////////////////////////////////////////////////////////////
 
 OpalPCSSEndPoint::OpalPCSSEndPoint(OpalManager & mgr, const char * prefix)
-  : OpalEndPoint(mgr, prefix, CanTerminateCall),
+  : OpalLocalEndPoint(mgr, prefix),
     soundChannelPlayDevice(PSoundChannel::GetDefaultDevice(PSoundChannel::Player)),
     soundChannelRecordDevice(PSoundChannel::GetDefaultDevice(PSoundChannel::Recorder))
 {
 #ifdef _WIN32
   // Windows MultMedia stuff seems to need greater depth due to enormous
   // latencies in its operation
-  soundChannelBuffers = 6;
+  m_soundChannelBufferTime = 120;
 #else
   // Should only need double buffering for Unix platforms
-  soundChannelBuffers = 2;
+  m_soundChannelBufferTime = 40;
 #endif
+  soundChannelBuffers = 2;
 
   PTRACE(3, "PCSS\tCreated PC sound system endpoint.\n" << setfill('\n')
          << "Players:\n"   << PSoundChannel::GetDeviceNames(PSoundChannel::Player)
@@ -130,12 +109,6 @@ OpalPCSSEndPoint::~OpalPCSSEndPoint()
   PTRACE(4, "PCSS\tDeleted PC sound system endpoint.");
 }
 
-OpalMediaFormatList OpalPCSSEndPoint::GetMediaFormats() const
-{
-  OpalMediaFormatList list = OpalEndPoint::GetMediaFormats();
-  list += OpalL16_STEREO_48KHZ;
-  return list;
-}
 
 static bool SetDeviceName(const PString & name,
                           PSoundChannel::Directions dir,
@@ -197,7 +170,9 @@ static bool SetDeviceNames(const PString & remoteParty, PString & playResult, PS
   PINDEX prefixLength = remoteParty.Find(':')+1;
 
   PString playDevice, recordDevice;
-  PINDEX separator = remoteParty.FindOneOf("\n\\", prefixLength);
+  PINDEX separator = remoteParty.Find('|', prefixLength);
+  if (separator == P_MAX_INDEX)
+    separator = remoteParty.Find('\\', prefixLength);
   if (separator == P_MAX_INDEX)
     playDevice = recordDevice = remoteParty.Mid(prefixLength);
   else {
@@ -223,17 +198,33 @@ static bool SetDeviceNames(const PString & remoteParty, PString & playResult, PS
 }
 
 
-PBoolean OpalPCSSEndPoint::MakeConnection(OpalCall & call,
-                                      const PString & remoteParty,
-                                      void * userData,
-                               unsigned int options,
-                               OpalConnection::StringOptions * stringOptions)
+PSafePtr<OpalConnection> OpalPCSSEndPoint::MakeConnection(OpalCall & call,
+                                                     const PString & remoteParty,
+                                                              void * userData,
+                                                        unsigned int options,
+                                     OpalConnection::StringOptions * stringOptions)
 {
+  PString deviceNames = remoteParty;
+  OpalConnection::StringOptions localStringOptions;
+
+  PINDEX semicolon = remoteParty.Find(';');
+  if (semicolon != P_MAX_INDEX) {
+    if (stringOptions == NULL)
+      stringOptions = &localStringOptions;
+
+    PStringToString params;
+    PURL::SplitVars(remoteParty.Mid(semicolon), params, ';', '=');
+    for (PINDEX i = 0; i < params.GetSize(); ++i)
+      stringOptions->SetAt(params.GetKeyAt(i), params.GetDataAt(i));
+
+    deviceNames.Delete(semicolon, P_MAX_INDEX);
+  }
+
   PString playDevice = soundChannelPlayDevice;
   PString recordDevice = soundChannelRecordDevice;
-  if (!SetDeviceNames(remoteParty, playDevice, recordDevice, "call")) {
+  if (!SetDeviceNames(deviceNames, playDevice, recordDevice, "call")) {
     call.Clear(OpalConnection::EndedByLocalBusy);
-    return false;
+    return NULL;
   }
 
   return AddConnection(CreateConnection(call, playDevice, recordDevice, userData, options, stringOptions));
@@ -294,27 +285,33 @@ PSoundChannel * OpalPCSSEndPoint::CreateSoundChannel(const OpalPCSSConnection &
 }
 
 
+bool OpalPCSSEndPoint::OnOutgoingCall(const OpalLocalConnection & connection)
+{
+  return OnShowOutgoing(dynamic_cast<const OpalPCSSConnection &>(connection));
+}
+
+
+bool OpalPCSSEndPoint::OnIncomingCall(OpalLocalConnection & connection)
+{
+  return OnShowIncoming(dynamic_cast<const OpalPCSSConnection &>(connection));
+}
+
+
+bool OpalPCSSEndPoint::OnUserInput(const OpalLocalConnection & connection, const PString & indication)
+{
+  return OnShowUserInput(dynamic_cast<const OpalPCSSConnection &>(connection), indication);
+}
+
+
 PBoolean OpalPCSSEndPoint::AcceptIncomingConnection(const PString & token)
 {
-  PSafePtr<OpalPCSSConnection> connection = GetPCSSConnectionWithLock(token, PSafeReadOnly);
-  if (connection == NULL) {
-    PTRACE(2, "PCSS\tCould not find connection using token \"" << token << '"');
-    return false;
-  }
-
-  connection->AcceptIncoming();
-  return true;
+  return AcceptIncomingCall(token);
 }
 
 
 PBoolean OpalPCSSEndPoint::RejectIncomingConnection(const PString & token)
 {
-  PSafePtr<OpalPCSSConnection> connection = GetPCSSConnectionWithLock(token, PSafeReadOnly);
-  if (connection == NULL)
-    return false;
-
-  connection->Release(OpalConnection::EndedByAnswerDenied);
-  return true;
+  return RejectIncomingCall(token);
 }
 
 
@@ -324,11 +321,6 @@ PBoolean OpalPCSSEndPoint::OnShowUserInput(const OpalPCSSConnection &, const PSt
 }
 
 
-void OpalPCSSEndPoint::OnPatchMediaStream(const OpalPCSSConnection & /*connection*/, PBoolean /*isSource*/, OpalMediaPatch & /*patch*/)
-{
-}
-
-
 PBoolean OpalPCSSEndPoint::SetSoundChannelPlayDevice(const PString & name)
 {
   return SetDeviceName(name, PSoundChannel::Player, soundChannelPlayDevice);
@@ -348,6 +340,13 @@ void OpalPCSSEndPoint::SetSoundChannelBufferDepth(unsigned depth)
 }
 
 
+void OpalPCSSEndPoint::SetSoundChannelBufferTime(unsigned depth)
+{
+  PAssert(depth >= 20, PInvalidParameter);
+  m_soundChannelBufferTime = depth;
+}
+
+
 /////////////////////////////////////////////////////////////////////////////
 
 OpalPCSSConnection::OpalPCSSConnection(OpalCall & call,
@@ -356,14 +355,17 @@ OpalPCSSConnection::OpalPCSSConnection(OpalCall & call,
                                        const PString & recordDevice,
                                        unsigned options,
                           OpalConnection::StringOptions * stringOptions)
-  : OpalConnection(call, ep, ep.GetManager().GetNextCallToken(), options, stringOptions),
-    endpoint(ep),
-    soundChannelPlayDevice(playDevice),
-    soundChannelRecordDevice(recordDevice),
-    soundChannelBuffers(ep.GetSoundChannelBufferDepth())
+  : OpalLocalConnection(call, ep, NULL, options, stringOptions, 'P')
+  , endpoint(ep)
+  , soundChannelPlayDevice(playDevice)
+  , soundChannelRecordDevice(recordDevice)
+  , soundChannelBuffers(ep.GetSoundChannelBufferDepth())
+  , m_soundChannelBufferTime(ep.GetSoundChannelBufferTime())
 {
   silenceDetector = new OpalPCM16SilenceDetector(endpoint.GetManager().GetSilenceDetectParams());
+#if OPAL_AEC
   echoCanceler = new OpalEchoCanceler;
+#endif
 
   PTRACE(4, "PCSS\tCreated PC sound system connection: token=\"" << callToken << "\" "
             "player=\"" << playDevice << "\" recorder=\"" << recordDevice << '"');
@@ -376,44 +378,6 @@ OpalPCSSConnection::~OpalPCSSConnection()
 }
 
 
-PBoolean OpalPCSSConnection::SetUpConnection()
-{
-  originating = true;
-
-  // Check if we are A-Party in thsi call, so need to do things differently
-  if (ownerCall.GetConnection(0) == this) {
-    SetPhase(SetUpPhase);
-    if (!OnIncomingConnection(0, NULL)) {
-      Release(EndedByCallerAbort);
-      return PFalse;
-    }
-
-    PTRACE(3, "PCSS\tOutgoing call routed to " << ownerCall.GetPartyB() << " for " << *this);
-    if (!ownerCall.OnSetUp(*this)) {
-      Release(EndedByNoAccept);
-      return PFalse;
-    }
-
-    return PTrue;
-  }
-
-  PTRACE(3, "PCSS\tSetUpConnection(" << remotePartyName << ')');
-  SetPhase(AlertingPhase);
-  OnAlerting();
-
-  return endpoint.OnShowIncoming(*this);
-}
-
-
-PBoolean OpalPCSSConnection::SetAlerting(const PString & calleeName, PBoolean)
-{
-  PTRACE(3, "PCSS\tSetAlerting(" << calleeName << ')');
-  SetPhase(AlertingPhase);
-  remotePartyName = calleeName;
-  return endpoint.OnShowOutgoing(*this);
-}
-
-
 bool OpalPCSSConnection::TransferConnection(const PString & remoteParty)
 {
   PString playDevice = soundChannelPlayDevice;
@@ -451,7 +415,7 @@ OpalMediaStream * OpalPCSSConnection::CreateMediaStream(const OpalMediaFormat &
     if (soundChannel == NULL)
       return NULL;
 
-    return new OpalAudioMediaStream(*this, mediaFormat, sessionID, isSource, soundChannelBuffers, soundChannel);
+    return new OpalAudioMediaStream(*this, mediaFormat, sessionID, isSource, soundChannelBuffers, m_soundChannelBufferTime, soundChannel);
   }
 
 #if OPAL_HAS_IM
@@ -464,40 +428,6 @@ OpalMediaStream * OpalPCSSConnection::CreateMediaStream(const OpalMediaFormat &
 }
 
 
-void OpalPCSSConnection::OnPatchMediaStream(PBoolean isSource,
-					    OpalMediaPatch & patch)
-{
-  endpoint.OnPatchMediaStream(*this, isSource, patch);
-
-  int clockRate;
-  if (patch.GetSource().GetMediaFormat().GetMediaType() == OpalMediaType::Audio()) {
-    PTRACE(3, "PCSS\tAdding filters to patch");
-    if (isSource)
-      patch.AddFilter(silenceDetector->GetReceiveHandler(), OpalPCM16);
-    clockRate = patch.GetSource().GetMediaFormat().GetClockRate();
-    echoCanceler->SetParameters(endpoint.GetManager().GetEchoCancelParams());
-    echoCanceler->SetClockRate(clockRate);
-    patch.AddFilter(isSource?echoCanceler->GetReceiveHandler():echoCanceler->GetSendHandler(), OpalPCM16);
-  }
-}
-
-
-OpalMediaStreamPtr OpalPCSSConnection::OpenMediaStream(const OpalMediaFormat & mediaFormat, unsigned sessionID, bool isSource)
-{
-#if OPAL_VIDEO
-  if ( isSource &&
-       mediaFormat.GetMediaType() == OpalMediaType::Video() &&
-      !ownerCall.IsEstablished() &&
-      !endpoint.GetManager().CanAutoStartTransmitVideo()) {
-    PTRACE(3, "PCSS\tOpenMediaStream auto start disabled, refusing video open");
-    return NULL;
-  }
-#endif
-
-  return OpalConnection::OpenMediaStream(mediaFormat, sessionID, isSource);
-}
-
-
 PBoolean OpalPCSSConnection::SetAudioVolume(PBoolean source, unsigned percentage)
 {
   PSafePtr<OpalAudioMediaStream> stream = PSafePtrCast<OpalMediaStream, OpalAudioMediaStream>(GetMediaStream(OpalMediaType::Audio(), source));
@@ -528,22 +458,9 @@ PSoundChannel * OpalPCSSConnection::CreateSoundChannel(const OpalMediaFormat & m
 }
 
 
-PBoolean OpalPCSSConnection::SendUserInputString(const PString & value)
-{
-  PTRACE(3, "PCSS\tSendUserInputString(" << value << ')');
-  return endpoint.OnShowUserInput(*this, value);
-}
-
-
-void OpalPCSSConnection::AcceptIncoming()
-{
-  if (LockReadWrite()) {
-    OnConnectedInternal();
-    UnlockReadWrite();
-  }
-}
-
 #endif // OPAL_PTLIB_AUDIO
 
+#endif // OPAL_HAS_PCSS
+
 
 /////////////////////////////////////////////////////////////////////////////
