@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 21989 $
+ * $Revision: 24280 $
  * $Author: rjongbloed $
- * $Date: 2009-02-04 06:34:43 +0000 (Wed, 04 Feb 2009) $
+ * $Date: 2010-04-25 23:18:25 -0500 (Sun, 25 Apr 2010) $
  */
 
 #include <ptlib.h>
@@ -86,22 +86,7 @@ OpalTranscoder::OpalTranscoder(const OpalMediaFormat & inputMediaFormat,
 bool OpalTranscoder::UpdateMediaFormats(const OpalMediaFormat & input, const OpalMediaFormat & output)
 {
   PWaitAndSignal mutex(updateMutex);
-
-  if (input.IsValid()) {
-    if (inputMediaFormat == input)
-      inputMediaFormat = input;
-    else if (!inputMediaFormat.Merge(input))
-      return false;
-  }
-
-  if (output.IsValid()) {
-    if (outputMediaFormat == output)
-      outputMediaFormat = output;
-    else if (!outputMediaFormat.Merge(output))
-      return false;
-  }
-
-  return true;
+  return inputMediaFormat.Update(input) && outputMediaFormat.Update(output);
 }
 
 
@@ -140,7 +125,12 @@ PBoolean OpalTranscoder::ConvertFrames(const RTP_DataFrame & input, RTP_DataFram
   }
 
   // set the output timestamp and marker bit
-  output.front().SetTimestamp(input.GetTimestamp()*outputMediaFormat.GetClockRate()/inputMediaFormat.GetClockRate());
+  unsigned timestamp = input.GetTimestamp();
+  unsigned inClockRate = inputMediaFormat.GetClockRate();
+  unsigned outClockRate = outputMediaFormat.GetClockRate();
+  if (inClockRate != outClockRate)
+    timestamp = (unsigned)((PUInt64)timestamp*outClockRate/inClockRate);
+  output.front().SetTimestamp(timestamp);
   output.front().SetMarker(input.GetMarker());
 
   // set the output payload type directly from the output media format
@@ -167,8 +157,8 @@ OpalTranscoder * OpalTranscoder::Create(const OpalMediaFormat & srcFormat,
 {
   OpalTranscoder * transcoder = OpalTranscoderFactory::CreateInstance(OpalTranscoderKey(srcFormat, destFormat));
   if (transcoder != NULL) {
+    transcoder->SetInstanceID(instance, instanceLen); // Make sure this is done first
     transcoder->UpdateMediaFormats(srcFormat, destFormat);
-    transcoder->SetInstanceID(instance, instanceLen);
   }
 
   return transcoder;
@@ -246,7 +236,8 @@ static bool MergeFormats(const OpalMediaFormatList & masterFormats,
   }
   else {
     srcFormat = *masterFormat;
-    PTRACE(5, "Opal\tInitial source format from master:\n" << setw(-1) << srcFormat);
+    PTRACE(5, "Opal\tInitial source format from master:\n" << setw(-1) << srcFormat
+                        << "Merging with capability\n" << setw(-1) << srcCapability);
     if (!srcFormat.Merge(srcCapability))
       return false;
   }
@@ -258,8 +249,14 @@ static bool MergeFormats(const OpalMediaFormatList & masterFormats,
   }
   else {
     dstFormat = *masterFormat;
-    PTRACE(5, "Opal\tInitial destination format from master:\n" << setw(-1) << dstFormat);
-    if (!dstFormat.Merge(dstCapability))
+    PTRACE(5, "Opal\tInitial destination format from master:\n" << setw(-1) << dstFormat
+                             << "Merging with capability\n" << setw(-1) << dstCapability);
+
+    OpalMediaFormat tmpCapability = dstCapability;
+    if (!tmpCapability.Merge(dstFormat))
+      return false;
+
+    if (!dstFormat.Update(tmpCapability))
       return false;
   }
 
@@ -273,7 +270,8 @@ static bool MergeFormats(const OpalMediaFormatList & masterFormats,
 }
 
 
-bool OpalTranscoder::SelectFormats(const OpalMediaFormatList & srcFormats,
+bool OpalTranscoder::SelectFormats(const OpalMediaType & mediaType,
+                                   const OpalMediaFormatList & srcFormats,
                                    const OpalMediaFormatList & dstFormats,
                                    const OpalMediaFormatList & allFormats,
                                    OpalMediaFormat & srcFormat,
@@ -285,7 +283,7 @@ bool OpalTranscoder::SelectFormats(const OpalMediaFormatList & srcFormats,
   // directly from the given format to a possible one with no transcoders.
   for (d = dstFormats.begin(); d != dstFormats.end(); ++d) {
     for (s = srcFormats.begin(); s != srcFormats.end(); ++s) {
-      if (*s == *d && MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
+      if (*s == *d && s->GetMediaType() == mediaType && MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
         return true;
     }
   }
@@ -293,11 +291,13 @@ bool OpalTranscoder::SelectFormats(const OpalMediaFormatList & srcFormats,
   // Search for a single transcoder to get from a to b
   for (d = dstFormats.begin(); d != dstFormats.end(); ++d) {
     for (s = srcFormats.begin(); s != srcFormats.end(); ++s) {
-      OpalTranscoderKey search(*s, *d);
-      OpalTranscoderList availableTranscoders = OpalTranscoderFactory::GetKeyList();
-      for (OpalTranscoderIterator i = availableTranscoders.begin(); i != availableTranscoders.end(); ++i) {
-        if (search == *i && MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
-          return true;
+      if (s->GetMediaType() == mediaType || d->GetMediaType() == mediaType) {
+        OpalTranscoderKey search(*s, *d);
+        OpalTranscoderList availableTranscoders = OpalTranscoderFactory::GetKeyList();
+        for (OpalTranscoderIterator i = availableTranscoders.begin(); i != availableTranscoders.end(); ++i) {
+          if (search == *i && MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
+            return true;
+        }
       }
     }
   }
@@ -305,9 +305,12 @@ bool OpalTranscoder::SelectFormats(const OpalMediaFormatList & srcFormats,
   // Last gasp search for a double transcoder to get from a to b
   for (d = dstFormats.begin(); d != dstFormats.end(); ++d) {
     for (s = srcFormats.begin(); s != srcFormats.end(); ++s) {
-      OpalMediaFormat intermediateFormat;
-      if (FindIntermediateFormat(*s, *d, intermediateFormat) && MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
-        return true;
+      if (s->GetMediaType() == mediaType || d->GetMediaType() == mediaType) {
+        OpalMediaFormat intermediateFormat;
+        if (FindIntermediateFormat(*s, *d, intermediateFormat) &&
+            MergeFormats(allFormats, *s, *d, srcFormat, dstFormat))
+          return true;
+      }
     }
   }
 
@@ -400,22 +403,10 @@ OpalMediaFormatList OpalTranscoder::GetPossibleFormats(const OpalMediaFormatList
 /////////////////////////////////////////////////////////////////////////////
 
 OpalFramedTranscoder::OpalFramedTranscoder(const OpalMediaFormat & inputMediaFormat,
-                                           const OpalMediaFormat & outputMediaFormat,
-                                           PINDEX inputBytes, PINDEX outputBytes)
+                                           const OpalMediaFormat & outputMediaFormat)
   : OpalTranscoder(inputMediaFormat, outputMediaFormat)
 {
-  PINDEX framesPerPacket = outputMediaFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), 1);
-  inputBytesPerFrame = inputBytes*framesPerPacket;
-  outputBytesPerFrame = outputBytes*framesPerPacket;
-
-  PINDEX inMaxTimePerFrame  = inputMediaFormat.GetOptionInteger(OpalAudioFormat::MaxFramesPerPacketOption()) * 
-                              inputMediaFormat.GetOptionInteger(OpalAudioFormat::FrameTimeOption());
-  PINDEX outMaxTimePerFrame = outputMediaFormat.GetOptionInteger(OpalAudioFormat::MaxFramesPerPacketOption()) * 
-                              outputMediaFormat.GetOptionInteger(OpalAudioFormat::FrameTimeOption());
-
-  PINDEX maxPacketTime = PMAX(inMaxTimePerFrame, outMaxTimePerFrame);
-
-  maxOutputDataSize = outputBytesPerFrame * (maxPacketTime / outputMediaFormat.GetOptionInteger(OpalAudioFormat::FrameTimeOption()));
+  CalculateSizes();
 }
 
 
@@ -430,7 +421,15 @@ bool OpalFramedTranscoder::UpdateMediaFormats(const OpalMediaFormat & input, con
   if (!OpalTranscoder::UpdateMediaFormats(input, output))
     return false;
 
-  unsigned framesPerPacket = outputMediaFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), 1);
+  CalculateSizes();
+  return true;
+}
+
+
+void OpalFramedTranscoder::CalculateSizes()
+{
+  unsigned framesPerPacket = outputMediaFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(),
+                              inputMediaFormat.GetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), 1));
   unsigned inFrameSize = inputMediaFormat.GetFrameSize();
   unsigned outFrameSize = outputMediaFormat.GetFrameSize();
   unsigned inFrameTime = inputMediaFormat.GetFrameTime();
@@ -447,8 +446,6 @@ bool OpalFramedTranscoder::UpdateMediaFormats(const OpalMediaFormat & input, con
   PINDEX maxPacketTime = PMAX(inMaxTimePerFrame, outMaxTimePerFrame);
 
   maxOutputDataSize = outputBytesPerFrame * (maxPacketTime / outputMediaFormat.GetOptionInteger(OpalAudioFormat::FrameTimeOption()));
-
-  return true;
 }
 
 
@@ -526,7 +523,8 @@ PBoolean OpalFramedTranscoder::Convert(const RTP_DataFrame & input, RTP_DataFram
     if (!ConvertFrame(inputPtr, consumed, outputPtr, created))
       return PFalse;
 
-    if (consumed == 0 && created == 0)
+    // If did not consume any data, codec must think it is rubbish, skip rest of packet.
+    if (consumed == 0)
       break;
 
     outputPtr   += created;
@@ -589,16 +587,20 @@ PBoolean OpalStreamedTranscoder::Convert(const RTP_DataFrame & input,
 {
   PINDEX i, bit, mask;
 
+  PINDEX samples = input.GetPayloadSize()*8/inputBitsPerSample;
+  PINDEX outputSize = (samples * outputBitsPerSample + 7) / 8;
+  output.SetPayloadSize(outputSize);
+
+  // The conversion algorithm for 5,3 & 2 bits per sample needs an extra
+  // couple of bytes at the end of the buffer to avoid lots of conditionals
+  output.SetMinSize(output.GetHeaderSize()+outputSize+2);
+
   const BYTE * inputBytes = input.GetPayloadPtr();
   const short * inputWords = (const short *)inputBytes;
 
   BYTE * outputBytes = output.GetPayloadPtr();
   short * outputWords = (short *)outputBytes;
 
-  PINDEX samples = input.GetPayloadSize()*8/inputBitsPerSample;
-  PINDEX outputSize = samples*outputBitsPerSample/8;
-  output.SetPayloadSize(outputSize);
-
   switch (inputBitsPerSample) {
     case 16 :
       switch (outputBitsPerSample) {
