@@ -26,9 +26,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 24327 $
- * $Author: ededu $
- * $Date: 2010-05-07 10:14:35 -0500 (Fri, 07 May 2010) $
+ * $Revision: 24125 $
+ * $Author: rjongbloed $
+ * $Date: 2010-03-13 00:11:13 -0600 (Sat, 13 Mar 2010) $
  */
 
 #include <ptlib.h>
@@ -52,24 +52,14 @@
 
 const unsigned SecondsFrom1900to1970 = (70*365+17)*24*60*60U;
 
-#define UDP_BUFFER_SIZE 32768
-
-namespace PWLibStupidLinkerHacks {
-extern int opalLoader;
-
-static class InstantiateMe
-{
-  public:
-    InstantiateMe()
-    {
-      opalLoader = 1;
-    }
-} instance;
-
-}; // namespace PWLibStupidLinkerHacks
+#define RTP_VIDEO_RX_BUFFER_SIZE 0x100000 // 1Mb
+#define RTP_AUDIO_RX_BUFFER_SIZE 0x4000   // 16kb
+#define RTP_DATA_TX_BUFFER_SIZE  0x2000   // 8kb
+#define RTP_CTRL_BUFFER_SIZE     0x1000   // 4kb
 
 
-static PFactory<RTP_Encoding>::Worker<RTP_Encoding> rtpAVPHandler("rtp/avp");
+PFACTORY_CREATE(PFactory<RTP_Encoding>, RTP_Encoding, "rtp/avp", false);
+
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -77,7 +67,8 @@ RTP_DataFrame::RTP_DataFrame(PINDEX payloadSz, PINDEX bufferSz)
   : PBYTEArray(std::max(bufferSz, MinHeaderSize+payloadSz))
 {
   payloadSize = payloadSz;
-  theArray[0] = '\x80';
+  theArray[0] = '\x80'; // Default to version 2
+  theArray[1] = '\x7f'; // Default to MaxPayloadType
 }
 
 
@@ -142,7 +133,7 @@ PINDEX RTP_DataFrame::GetHeaderSize() const
   PINDEX sz = MinHeaderSize + 4*GetContribSrcCount();
 
   if (GetExtension())
-    sz += 4 + GetExtensionSize();
+    sz += 4 + GetExtensionSizeDWORDs()*4;
 
   return sz;
 }
@@ -163,22 +154,22 @@ void RTP_DataFrame::SetExtensionType(int type)
     SetExtension(false);
   else {
     if (!GetExtension())
-      SetExtensionSize(0);
+      SetExtensionSizeDWORDs(0);
     *(PUInt16b *)&theArray[MinHeaderSize + 4*GetContribSrcCount()] = (WORD)type;
   }
 }
 
 
-PINDEX RTP_DataFrame::GetExtensionSize() const
+PINDEX RTP_DataFrame::GetExtensionSizeDWORDs() const
 {
   if (GetExtension())
-    return *(PUInt16b *)&theArray[MinHeaderSize + 4*GetContribSrcCount() + 2] * 4;
+    return *(PUInt16b *)&theArray[MinHeaderSize + 4*GetContribSrcCount() + 2];
 
   return 0;
 }
 
 
-PBoolean RTP_DataFrame::SetExtensionSize(PINDEX sz)
+PBoolean RTP_DataFrame::SetExtensionSizeDWORDs(PINDEX sz)
 {
   if (!SetMinSize(MinHeaderSize + 4*GetContribSrcCount() + 4+4*sz + payloadSize))
     return false;
@@ -223,7 +214,7 @@ void RTP_DataFrame::PrintOn(ostream & strm) const
 
   if (GetExtension())
     strm << "  Header Extension Type: " << GetExtensionType() << '\n'
-         << hex << setfill('0') << PBYTEArray(GetExtensionPtr(), GetExtensionSize(), false) << setfill(' ') << dec << '\n';
+         << hex << setfill('0') << PBYTEArray(GetExtensionPtr(), GetExtensionSizeDWORDs()*4, false) << setfill(' ') << dec << '\n';
 
   strm << hex << setfill('0') << PBYTEArray(GetPayloadPtr(), GetPayloadSize(), false) << setfill(' ') << dec;
 }
@@ -264,7 +255,9 @@ static const char * const PayloadTypesNames[RTP_DataFrame::LastKnownPayloadType]
   "H261",
   "MPV",
   "MP2T",
-  "H263"
+  "H263",
+  NULL, NULL, NULL,
+  "T38"
 };
 
 ostream & operator<<(ostream & o, RTP_DataFrame::PayloadTypes t)
@@ -454,9 +447,22 @@ OpalMediaStatistics::OpalMediaStatistics()
 
 #if OPAL_FAX
 OpalMediaStatistics::Fax::Fax()
+  : m_result(-2)
+  , m_bitRate(9600)
+  , m_compression(1)
+  , m_errorCorrection(false)
+  , m_txPages(-1)
+  , m_rxPages(-1)
+  , m_totalPages(0)
+  , m_imageSize(0)
+  , m_resolutionX(0)
+  , m_resolutionY(0)
+  , m_pageWidth(0)
+  , m_pageHeight(0)
+  , m_badRows(0)
+  , m_mostBadRows(0)
+  , m_errorCorrectionRetries(0)
 {
-  memset(this, 0, sizeof(*this));
-  m_result = -2;
 }
 #endif
 
@@ -491,14 +497,15 @@ void RTP_UserData::OnTxIntraFrameRequest(const RTP_Session & /*session*/) const
 /////////////////////////////////////////////////////////////////////////////
 
 RTP_Session::RTP_Session(const Params & params)
-  : canonicalName(PProcess::Current().GetUserName())
+  : m_timeUnits(params.isAudio ? 8 : 90)
+  , canonicalName(PProcess::Current().GetUserName())
   , toolName(PProcess::Current().GetName())
   , reportTimeInterval(0, 12)  // Seconds
   , reportTimer(reportTimeInterval)
   , failed(false)
 {
-  PAssert(params.id > 0 && params.id < 256, PInvalidParameter);
-  sessionID = (BYTE)params.id;
+  PAssert(params.id > 0, PInvalidParameter);
+  sessionID = params.id;
   isAudio = params.isAudio;
 
   userData = params.userData;
@@ -541,22 +548,24 @@ RTP_Session::~RTP_Session()
 {
   PTRACE_IF(3, packetsSent != 0 || packetsReceived != 0,
       "RTP\tSession " << sessionID << ", final statistics:\n"
-      "    packetsSent       = " << packetsSent << "\n"
-      "    octetsSent        = " << octetsSent << "\n"
-      "    averageSendTime   = " << averageSendTime << "\n"
-      "    maximumSendTime   = " << maximumSendTime << "\n"
-      "    minimumSendTime   = " << minimumSendTime << "\n"
-      "    packetsReceived   = " << packetsReceived << "\n"
-      "    octetsReceived    = " << octetsReceived << "\n"
-      "    packetsLost       = " << packetsLost << "\n"
-      "    packetsTooLate    = " << GetPacketsTooLate() << "\n"
-      "    packetOverruns    = " << GetPacketOverruns() << "\n"
-      "    packetsOutOfOrder = " << packetsOutOfOrder << "\n"
-      "    averageReceiveTime= " << averageReceiveTime << "\n"
-      "    maximumReceiveTime= " << maximumReceiveTime << "\n"
-      "    minimumReceiveTime= " << minimumReceiveTime << "\n"
-      "    averageJitter     = " << (jitterLevel >> 7) << "\n"
-      "    maximumJitter     = " << (maximumJitterLevel >> 7)
+      "    packetsSent        = " << packetsSent << "\n"
+      "    octetsSent         = " << octetsSent << "\n"
+      "    averageSendTime    = " << averageSendTime << "\n"
+      "    maximumSendTime    = " << maximumSendTime << "\n"
+      "    minimumSendTime    = " << minimumSendTime << "\n"
+      "    packetsLostByRemote= " << packetsLostByRemote << "\n"
+      "    jitterLevelOnRemote= " << jitterLevelOnRemote << "\n"
+      "    packetsReceived    = " << packetsReceived << "\n"
+      "    octetsReceived     = " << octetsReceived << "\n"
+      "    packetsLost        = " << packetsLost << "\n"
+      "    packetsTooLate     = " << GetPacketsTooLate() << "\n"
+      "    packetOverruns     = " << GetPacketOverruns() << "\n"
+      "    packetsOutOfOrder  = " << packetsOutOfOrder << "\n"
+      "    averageReceiveTime = " << averageReceiveTime << "\n"
+      "    maximumReceiveTime = " << maximumReceiveTime << "\n"
+      "    minimumReceiveTime = " << minimumReceiveTime << "\n"
+      "    averageJitter      = " << GetAvgJitterTime() << "\n"
+      "    maximumJitter      = " << GetMaxJitterTime()
      );
   if (autoDeleteUserData)
     delete userData;
@@ -572,6 +581,7 @@ void RTP_Session::ClearStatistics()
   packetsReceived = 0;
   octetsReceived = 0;
   packetsLost = 0;
+  packetsLostByRemote = 0;
   packetsOutOfOrder = 0;
   averageSendTime = 0;
   maximumSendTime = 0;
@@ -581,6 +591,7 @@ void RTP_Session::ClearStatistics()
   minimumReceiveTime = 0;
   jitterLevel = 0;
   maximumJitterLevel = 0;
+  jitterLevelOnRemote = 0;
   markerRecvCount = 0;
   markerSendCount = 0;
 
@@ -691,20 +702,22 @@ void RTP_Session::SetUserData(RTP_UserData * data, PBoolean autoDelete)
 void RTP_Session::SetJitterBufferSize(unsigned minJitterDelay,
                                       unsigned maxJitterDelay,
                                       unsigned timeUnits,
-                                        PINDEX stackSize)
+                                        PINDEX packetSize)
 {
+  if (timeUnits > 0)
+    m_timeUnits = timeUnits;
+
   if (minJitterDelay == 0 && maxJitterDelay == 0) {
-    PTRACE_IF(4, m_jitterBuffer != NULL, "InfLID\tSwitching off jitter buffer.");
+    PTRACE_IF(4, m_jitterBuffer != NULL, "InfLID\tSwitching off jitter buffer " << *m_jitterBuffer);
     m_jitterBuffer.SetNULL();
   }
   else {
     PTRACE(4, "InfLID\tSetting jitter buffer time from " << minJitterDelay << " to " << maxJitterDelay);
     SetIgnoreOutOfOrderPackets(false);
     if (m_jitterBuffer != NULL)
-      m_jitterBuffer->SetDelay(minJitterDelay, maxJitterDelay);
+      m_jitterBuffer->SetDelay(minJitterDelay, maxJitterDelay, packetSize);
     else
-      m_jitterBuffer = new RTP_JitterBuffer(*this, minJitterDelay, maxJitterDelay, timeUnits, stackSize);
-    m_jitterBuffer->Resume();
+      m_jitterBuffer = new RTP_JitterBuffer(*this, minJitterDelay, maxJitterDelay, m_timeUnits, packetSize);
   }
 }
 
@@ -716,13 +729,6 @@ unsigned RTP_Session::GetJitterBufferSize() const
 }
 
 
-unsigned RTP_Session::GetJitterTimeUnits() const
-{
-  JitterBufferPtr jitter = m_jitterBuffer; // Increase reference count
-  return jitter != NULL ? jitter->GetTimeUnits() : 0;
-}
-
-
 PBoolean RTP_Session::ReadBufferedData(RTP_DataFrame & frame)
 {
   JitterBufferPtr jitter = m_jitterBuffer; // Increase reference count
@@ -753,7 +759,7 @@ void RTP_Session::SetRxStatisticsInterval(unsigned packets)
 void RTP_Session::AddReceiverReport(RTP_ControlFrame::ReceiverReport & receiver)
 {
   receiver.ssrc = syncSourceIn;
-  receiver.SetLostPackets(packetsLost);
+  receiver.SetLostPackets(GetPacketsLost()+GetPacketsTooLate());
 
   if (expectedSequenceNumber > lastRRSequenceNumber)
     receiver.fraction = (BYTE)((packetsLostSinceLastRR<<8)/(expectedSequenceNumber - lastRRSequenceNumber));
@@ -764,7 +770,7 @@ void RTP_Session::AddReceiverReport(RTP_ControlFrame::ReceiverReport & receiver)
   receiver.last_seq = lastRRSequenceNumber;
   lastRRSequenceNumber = expectedSequenceNumber;
 
-  receiver.jitter = jitterLevel >> 4; // Allow for rounding protection bits
+  receiver.jitter = jitterLevel >> JitterRoundingGuardBits; // Allow for rounding protection bits
 
   // The following have not been calculated yet.
   receiver.lsr = 0;
@@ -862,8 +868,6 @@ RTP_Session::SendReceiveStatus RTP_Session::Internal_OnSendData(RTP_DataFrame &
   if (!SendReport())
     return e_AbortTransport;
 
-  frame.SetSize(frame.GetHeaderSize()+frame.GetPayloadSize());
-
   if (txStatisticsCount < txStatisticsInterval)
     return e_ProcessPacket;
 
@@ -966,10 +970,12 @@ RTP_Session::SendReceiveStatus RTP_Session::Internal_OnReceiveData(RTP_DataFrame
       if (allowAnySyncSource) {
         PTRACE(2, "RTP\tSession " << sessionID << ", SSRC changed from " << hex << frame.GetSyncSource() << " to " << syncSourceIn << dec);
         syncSourceIn = frame.GetSyncSource();
+        allowSequenceChange = true;
       } 
       else if (allowOneSyncSourceChange) {
         PTRACE(2, "RTP\tSession " << sessionID << ", allowed one SSRC change from SSRC=" << hex << syncSourceIn << " to =" << dec << frame.GetSyncSource() << dec);
         syncSourceIn = frame.GetSyncSource();
+        allowSequenceChange = true;
         allowOneSyncSourceChange = false;
       }
       else {
@@ -993,14 +999,13 @@ RTP_Session::SendReceiveStatus RTP_Session::Internal_OnReceiveData(RTP_DataFrame
           minimumReceiveTimeAccum = diff;
         rxStatisticsCount++;
 
-        // The following has the implicit assumption that something that has jitter
-        // is an audio codec and thus is in 8kHz timestamp units.
-        diff *= 8;
+        // As per RFC3550 Appendix 8
+        diff *= GetJitterTimeUnits(); // Convert to timestamp units
         long variance = diff - lastTransitTime;
         lastTransitTime = diff;
         if (variance < 0)
           variance = -variance;
-        jitterLevel += variance - ((jitterLevel+8) >> 4);
+        jitterLevel += variance - ((jitterLevel+(1<<(JitterRoundingGuardBits-1))) >> JitterRoundingGuardBits);
         if (jitterLevel > maximumJitterLevel)
           maximumJitterLevel = jitterLevel;
       }
@@ -1074,8 +1079,8 @@ RTP_Session::SendReceiveStatus RTP_Session::Internal_OnReceiveData(RTP_DataFrame
               " avgTime=" << averageReceiveTime <<
               " maxTime=" << maximumReceiveTime <<
               " minTime=" << minimumReceiveTime <<
-              " jitter=" << (jitterLevel >> 7) <<
-              " maxJitter=" << (maximumJitterLevel >> 7));
+              " jitter=" << GetAvgJitterTime() <<
+              " maxJitter=" << GetMaxJitterTime());
 
     if (userData != NULL)
       userData->OnRxStatistics(*this);
@@ -1095,7 +1100,7 @@ PBoolean RTP_Session::InsertReportPacket(RTP_ControlFrame & report)
     // Send RR as we are not transmitting
     report.StartNewPacket();
     report.SetPayloadType(RTP_ControlFrame::e_ReceiverReport);
-    report.SetPayloadSize(4 + sizeof(RTP_ControlFrame::ReceiverReport));  // length is SSRC of packet sender plus RR
+    report.SetPayloadSize(sizeof(PUInt32b) + sizeof(RTP_ControlFrame::ReceiverReport));  // length is SSRC of packet sender plus RR
     report.SetCount(1);
     BYTE * payload = report.GetPayloadPtr();
 
@@ -1110,7 +1115,7 @@ PBoolean RTP_Session::InsertReportPacket(RTP_ControlFrame & report)
     // send SR and RR
     report.StartNewPacket();
     report.SetPayloadType(RTP_ControlFrame::e_SenderReport);
-    report.SetPayloadSize(4 + sizeof(RTP_ControlFrame::SenderReport));  // length is SSRC of packet sender plus SR
+    report.SetPayloadSize(sizeof(PUInt32b) + sizeof(RTP_ControlFrame::SenderReport));  // length is SSRC of packet sender plus SR
     report.SetCount(0);
     BYTE * payload = report.GetPayloadPtr();
 
@@ -1118,7 +1123,7 @@ PBoolean RTP_Session::InsertReportPacket(RTP_ControlFrame & report)
     *(PUInt32b *)payload = syncSourceOut;
 
     // add the SR after the SSRC
-    RTP_ControlFrame::SenderReport * sender = (RTP_ControlFrame::SenderReport *)(payload+4);
+    RTP_ControlFrame::SenderReport * sender = (RTP_ControlFrame::SenderReport *)(payload+sizeof(PUInt32b));
     PTime now;
     sender->ntp_sec  = (DWORD)(now.GetTimeInSeconds()+SecondsFrom1900to1970); // Convert from 1970 to 1900
     sender->ntp_frac = now.GetMicrosecond()*4294; // Scale microseconds to "fraction" from 0 to 2^32
@@ -1134,9 +1139,9 @@ PBoolean RTP_Session::InsertReportPacket(RTP_ControlFrame & report)
            << " osent=" << sender->osent);
 
     if (syncSourceIn != 0) {
-      report.SetPayloadSize(4 + sizeof(RTP_ControlFrame::SenderReport) + sizeof(RTP_ControlFrame::ReceiverReport));
+      report.SetPayloadSize(sizeof(PUInt32b) + sizeof(RTP_ControlFrame::SenderReport) + sizeof(RTP_ControlFrame::ReceiverReport));
       report.SetCount(1);
-      AddReceiverReport(*(RTP_ControlFrame::ReceiverReport *)(payload+4+sizeof(RTP_ControlFrame::SenderReport)));
+      AddReceiverReport(*(RTP_ControlFrame::ReceiverReport *)(payload+sizeof(PUInt32b)+sizeof(RTP_ControlFrame::SenderReport)));
     }
   }
 
@@ -1191,14 +1196,14 @@ void RTP_Session::GetStatistics(OpalMediaStatistics & statistics, bool receiver)
 {
   statistics.m_totalBytes        = receiver ? GetOctetsReceived()     : GetOctetsSent();
   statistics.m_totalPackets      = receiver ? GetPacketsReceived()    : GetPacketsSent();
-  statistics.m_packetsLost       = receiver ? GetPacketsLost()        : 0;
+  statistics.m_packetsLost       = receiver ? GetPacketsLost()        : GetPacketsLostByRemote();
   statistics.m_packetsOutOfOrder = receiver ? GetPacketsOutOfOrder()  : 0;
   statistics.m_packetsTooLate    = receiver ? GetPacketsTooLate()     : 0;
   statistics.m_packetOverruns    = receiver ? GetPacketOverruns()     : 0;
   statistics.m_minimumPacketTime = receiver ? GetMinimumReceiveTime() : GetMinimumSendTime();
   statistics.m_averagePacketTime = receiver ? GetAverageReceiveTime() : GetAverageSendTime();
   statistics.m_maximumPacketTime = receiver ? GetMaximumReceiveTime() : GetMaximumSendTime();
-  statistics.m_averageJitter     = receiver ? GetAvgJitterTime()      : 0;
+  statistics.m_averageJitter     = receiver ? GetAvgJitterTime()      : GetJitterTimeOnRemote();
   statistics.m_maximumJitter     = receiver ? GetMaxJitterTime()      : 0;
 }
 #endif
@@ -1241,15 +1246,15 @@ RTP_Session::SendReceiveStatus RTP_Session::OnReceiveControl(RTP_ControlFrame &
     }
     switch (frame.GetPayloadType()) {
     case RTP_ControlFrame::e_SenderReport :
-      if (size >= sizeof(RTP_ControlFrame::SenderReport)) {
+      if (size >= sizeof(PUInt32b)+sizeof(RTP_ControlFrame::SenderReport)+frame.GetCount()*sizeof(RTP_ControlFrame::ReceiverReport)) {
         SenderReport sender;
         sender.sourceIdentifier = *(const PUInt32b *)payload;
-        const RTP_ControlFrame::SenderReport & sr = *(const RTP_ControlFrame::SenderReport *)(payload+4);
+        const RTP_ControlFrame::SenderReport & sr = *(const RTP_ControlFrame::SenderReport *)(payload+sizeof(PUInt32b));
         sender.realTimestamp = PTime(sr.ntp_sec-SecondsFrom1900to1970, sr.ntp_frac/4294);
         sender.rtpTimestamp = sr.rtp_ts;
         sender.packetsSent = sr.psent;
         sender.octetsSent = sr.osent;
-        OnRxSenderReport(sender, BuildReceiverReportArray(frame, sizeof(RTP_ControlFrame::SenderReport)));
+        OnRxSenderReport(sender, BuildReceiverReportArray(frame, sizeof(PUInt32b)+sizeof(RTP_ControlFrame::SenderReport)));
       }
       else {
         PTRACE(2, "RTP\tSession " << sessionID << ", SenderReport packet truncated");
@@ -1257,9 +1262,8 @@ RTP_Session::SendReceiveStatus RTP_Session::OnReceiveControl(RTP_ControlFrame &
       break;
 
     case RTP_ControlFrame::e_ReceiverReport :
-      if (size >= 4)
-        OnRxReceiverReport(*(const PUInt32b *)payload,
-        BuildReceiverReportArray(frame, sizeof(PUInt32b)));
+      if (size >= sizeof(PUInt32b)+frame.GetCount()*sizeof(RTP_ControlFrame::ReceiverReport))
+        OnRxReceiverReport(*(const PUInt32b *)payload, BuildReceiverReportArray(frame, sizeof(PUInt32b)));
       else {
         PTRACE(2, "RTP\tSession " << sessionID << ", ReceiverReport packet truncated");
       }
@@ -1359,8 +1363,7 @@ RTP_Session::SendReceiveStatus RTP_Session::OnReceiveControl(RTP_ControlFrame &
 }
 
 
-void RTP_Session::OnRxSenderReport(const SenderReport & PTRACE_PARAM(sender),
-           const ReceiverReportArray & PTRACE_PARAM(reports))
+void RTP_Session::OnRxSenderReport(const SenderReport & PTRACE_PARAM(sender), const ReceiverReportArray & reports)
 {
 #if PTRACING
   if (PTrace::CanTrace(3)) {
@@ -1371,11 +1374,11 @@ void RTP_Session::OnRxSenderReport(const SenderReport & PTRACE_PARAM(sender),
     strm << PTrace::End;
   }
 #endif
+  OnReceiverReports(reports);
 }
 
 
-void RTP_Session::OnRxReceiverReport(DWORD PTRACE_PARAM(src),
-             const ReceiverReportArray & PTRACE_PARAM(reports))
+void RTP_Session::OnRxReceiverReport(DWORD PTRACE_PARAM(src), const ReceiverReportArray & reports)
 {
 #if PTRACING
   if (PTrace::CanTrace(3)) {
@@ -1386,6 +1389,19 @@ void RTP_Session::OnRxReceiverReport(DWORD PTRACE_PARAM(src),
     strm << PTrace::End;
   }
 #endif
+  OnReceiverReports(reports);
+}
+
+
+void RTP_Session::OnReceiverReports(const ReceiverReportArray & reports)
+{
+  for (PINDEX i = 0; i < reports.GetSize(); i++) {
+    if (reports[i].sourceIdentifier == syncSourceOut) {
+      packetsLostByRemote = reports[i].totalLost;
+      jitterLevelOnRemote = reports[i].jitter;
+      break;
+    }
+  }
 }
 
 
@@ -1493,22 +1509,36 @@ void RTP_Session::AddFilter(const PNotifier & filter)
 
 /////////////////////////////////////////////////////////////////////////////
 
-static void SetMinBufferSize(PUDPSocket & sock, int buftype)
+static void SetMinBufferSize(PUDPSocket & sock, int buftype, int bufsz)
 {
   int sz = 0;
-  if (sock.GetOption(buftype, sz)) {
-    if (sz >= UDP_BUFFER_SIZE)
-      return;
-  }
-  else {
+  if (!sock.GetOption(buftype, sz)) {
     PTRACE(1, "RTP_UDP\tGetOption(" << buftype << ") failed: " << sock.GetErrorText());
+    return;
   }
 
-  if (!sock.SetOption(buftype, UDP_BUFFER_SIZE)) {
-    PTRACE(1, "RTP_UDP\tSetOption(" << buftype << ") failed: " << sock.GetErrorText());
-  }
+  // Already big enough
+  if (sz >= bufsz)
+    return;
 
-  PTRACE_IF(1, !sock.GetOption(buftype, sz) && sz < UDP_BUFFER_SIZE, "RTP_UDP\tSetOption(" << buftype << ") failed, even though it said it succeeded!");
+  for (; bufsz >= 1024; bufsz /= 2) {
+    // Set to new size
+    if (!sock.SetOption(buftype, bufsz)) {
+      PTRACE(1, "RTP_UDP\tSetOption(" << buftype << ',' << bufsz << ") failed: " << sock.GetErrorText());
+      continue;
+    }
+
+    // As some stacks lie about setting the buffer size, we double check.
+    if (!sock.GetOption(buftype, sz)) {
+      PTRACE(1, "RTP_UDP\tGetOption(" << buftype << ") failed: " << sock.GetErrorText());
+      return;
+    }
+
+    if (sz >= bufsz)
+      return;
+
+    PTRACE(1, "RTP_UDP\tSetOption(" << buftype << ',' << bufsz << ") failed, even though it said it succeeded!");
+  }
 }
 
 
@@ -1573,7 +1603,7 @@ PBoolean RTP_UDP::ModifyQOS(RTP_QOS * rtpqos)
   return retval;
 }
 
-PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
+PBoolean RTP_UDP::Open(PIPSocket::Address transportLocalAddress,
                    WORD portBase, WORD portMax,
                    BYTE tos,
                    PNatMethod * natMethod,
@@ -1583,7 +1613,7 @@ PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
 
   first = true;
   // save local address 
-  localAddress = _localAddress;
+  localAddress = transportLocalAddress;
 
   localDataPort    = (WORD)(portBase&0xfffe);
   localControlPort = (WORD)(localDataPort + 1);
@@ -1607,24 +1637,15 @@ PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
     PIPSocket::Address bindingAddress = localAddress;
     if (natMethod != NULL && natMethod->IsAvailable(localAddress)) {
       switch (natMethod->GetRTPSupport()) {
-        case PNatMethod::RTPIfSendMedia :
-          /* This NAT variant will work if we send something out through the
-             NAT port to "open" it so packets can then flow inward. We set
-             this flag to make that happen as soon as we get the remotes IP
-             address and port to send to.
-            */
-          localHasNAT = true;
-          // Then do case for full cone support and create STUN sockets
-
         case PNatMethod::RTPSupported :
           if (natMethod->CreateSocketPair(dataSocket, controlSocket, localAddress)) {
-            PTRACE(4, "RTP\tSession " << sessionID << ", " << natMethod->GetName() << " created STUN RTP/RTCP socket pair.");
+            PTRACE(4, "RTP\tSession " << sessionID << ", " << natMethod->GetName() << " created RTP/RTCP socket pair.");
             dataSocket->GetLocalAddress(localAddress, localDataPort);
             controlSocket->GetLocalAddress(localAddress, localControlPort);
           }
           else {
             PTRACE(2, "RTP\tSession " << sessionID << ", " << natMethod->GetName()
-                   << " could not create STUN RTP/RTCP socket pair; trying to create individual sockets.");
+                   << " could not create RTP/RTCP socket pair; trying to create RTP socket anyway.");
             if (natMethod->CreateSocket(dataSocket, localAddress) && natMethod->CreateSocket(controlSocket, localAddress)) {
               dataSocket->GetLocalAddress(localAddress, localDataPort);
               controlSocket->GetLocalAddress(localAddress, localControlPort);
@@ -1635,18 +1656,22 @@ PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
               dataSocket = NULL;
               controlSocket = NULL;
               PTRACE(2, "RTP\tSession " << sessionID << ", " << natMethod->GetName()
-                     << " could not create STUN RTP/RTCP sockets individually either, using normal sockets.");
+                     << " could not create RTP sockets individually either, using normal sockets.");
             }
           }
           break;
 
-        default :
+        case PNatMethod::RTPIfSendMedia :
           /* We canot use NAT traversal method (e.g. STUN) to create sockets
-             in the remaining modes as the NAT router will then not let us
-             talk to the real RTP destination. All we can so is bind to the
-             local interface the NAT is on and hope the NAT router is doing
-             something sneaky like symmetric port forwarding. */
-          natMethod->GetInterfaceAddress(bindingAddress);
+             as the NAT router will then not let us talk to the real RTP
+             destination. All we can so is bind to the local interface, which
+             telling the remote our address is the external address of the
+             NATrouter, and hope the remote is tolerant enough of things like
+             non adjacent RTP/RTCP ports etc. */
+          localHasNAT = natMethod->GetInterfaceAddress(bindingAddress);
+          break;
+
+        default :
           break;
       }
     }
@@ -1673,10 +1698,10 @@ PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
     }
 
     // Increase internal buffer size on media UDP sockets
-    SetMinBufferSize(*dataSocket,    SO_RCVBUF);
-    SetMinBufferSize(*dataSocket,    SO_SNDBUF);
-    SetMinBufferSize(*controlSocket, SO_RCVBUF);
-    SetMinBufferSize(*controlSocket, SO_SNDBUF);
+    SetMinBufferSize(*dataSocket,    SO_RCVBUF, isAudio ? RTP_AUDIO_RX_BUFFER_SIZE : RTP_VIDEO_RX_BUFFER_SIZE);
+    SetMinBufferSize(*dataSocket,    SO_SNDBUF, RTP_DATA_TX_BUFFER_SIZE);
+    SetMinBufferSize(*controlSocket, SO_RCVBUF, RTP_CTRL_BUFFER_SIZE);
+    SetMinBufferSize(*controlSocket, SO_SNDBUF, RTP_CTRL_BUFFER_SIZE);
 #   endif
   }
 
@@ -1689,7 +1714,6 @@ PBoolean RTP_UDP::Open(PIPSocket::Address _localAddress,
   PTRACE(3, "RTP_UDP\tSession " << sessionID << " created: "
          << localAddress << ':' << localDataPort << '-' << localControlPort
          << " ssrc=" << syncSourceOut);
-  
 
   return true;
 }
@@ -1703,20 +1727,31 @@ void RTP_UDP::Reopen(PBoolean reading)
     shutdownRead = false;
   else
     shutdownWrite = false;
+
+  badTransmitCounter = 0;
+
+  PTRACE(3, "RTP_UDP\tSession " << sessionID << " reopened for " << (reading ? "reading" : "writing"));
 }
 
 
-void RTP_UDP::Close(PBoolean reading)
+bool RTP_UDP::Close(PBoolean reading)
 {
   //SendBYE();
 
-  PWaitAndSignal mutex(dataMutex);
-
   if (reading) {
-    if (!shutdownRead) {
+    {
+      PWaitAndSignal mutex(dataMutex);
+
+      if (shutdownRead) {
+        PTRACE(4, "RTP_UDP\tSession " << sessionID << ", read already shut down .");
+        return false;
+      }
+
       PTRACE(3, "RTP_UDP\tSession " << sessionID << ", Shutting down read.");
+
       syncSourceIn = 0;
       shutdownRead = true;
+
       if (dataSocket != NULL && controlSocket != NULL) {
         PIPSocket::Address addr;
         controlSocket->GetLocalAddress(addr);
@@ -1725,11 +1760,20 @@ void RTP_UDP::Close(PBoolean reading)
         dataSocket->WriteTo("", 1, addr, controlSocket->GetPort());
       }
     }
+
+    SetJitterBufferSize(0, 0); // Kill jitter buffer too, but outside mutex
   }
   else {
-    PTRACE(3, "RTP_UDP\tSession " << sessionID << ", Shutting down write.");
+    if (shutdownWrite) {
+      PTRACE(4, "RTP_UDP\tSession " << sessionID << ", write already shut down .");
+      return false;
+    }
+
+    PTRACE(3, "RTP_UDP\tSession " << sessionID << ", shutting down write.");
     shutdownWrite = true;
   }
+
+  return true;
 }
 
 
@@ -1948,9 +1992,9 @@ RTP_Session::SendReceiveStatus RTP_UDP::ReadDataOrControlPDU(BYTE * framePtr,
       if (++badTransmitCounter == 1) 
         badTransmitStart = PTime();
       else {
-        if ((PTime()- badTransmitStart).GetMilliSeconds() < BAD_TRANSMIT_TIME_MAX * 1000)
+        if (badTransmitCounter < 5 || (PTime()- badTransmitStart).GetSeconds() < BAD_TRANSMIT_TIME_MAX)
           return RTP_Session::e_IgnorePacket;
-        PTRACE(2, "RTP_UDP\tSession " << sessionID << ", " << channelName << " " << BAD_TRANSMIT_TIME_MAX << "s of transmit fails - informing connection");
+        PTRACE(2, "RTP_UDP\tSession " << sessionID << ", " << channelName << " " << BAD_TRANSMIT_TIME_MAX << " seconds of transmit fails - informing connection");
         userData->SessionFailing(*this);
       }
       return RTP_Session::e_IgnorePacket;
@@ -2033,6 +2077,7 @@ RTP_Session::SendReceiveStatus RTP_UDP::ReadControlPDU()
   return OnReceiveControl(frame);
 }
 
+
 PBoolean RTP_UDP::WriteOOBData(RTP_DataFrame & frame, bool rewriteTimeStamp)
 {
   PWaitAndSignal m(dataMutex);
@@ -2053,12 +2098,12 @@ PBoolean RTP_UDP::WriteOOBData(RTP_DataFrame & frame, bool rewriteTimeStamp)
     frame.SetTimestamp(oobTimeStampOutBase + ((PTimer::Tick() - oobTimeStampBase).GetInterval() * 8));
 
   // write the data
-  return WriteData(frame);
+  return EncodingLock(*this)->WriteData(frame, true);
 }
 
 PBoolean RTP_UDP::WriteData(RTP_DataFrame & frame)
 {
-  return EncodingLock(*this)->WriteData(frame);
+  return EncodingLock(*this)->WriteData(frame, false);
 }
 
 
@@ -2237,7 +2282,7 @@ RTP_Session::SendReceiveStatus RTP_Encoding::OnSendData(RTP_DataFrame & frame)
   return rtpUDP->Internal_OnSendData(frame);
 }
 
-PBoolean RTP_Encoding::WriteData(RTP_DataFrame & frame)
+PBoolean RTP_Encoding::WriteData(RTP_DataFrame & frame, bool)
 {
   return rtpUDP->Internal_WriteData(frame);
 }
@@ -2249,7 +2294,7 @@ RTP_Session::SendReceiveStatus RTP_Encoding::OnSendControl(RTP_ControlFrame & fr
 
 bool RTP_Encoding::WriteDataPDU(RTP_DataFrame & frame)
 {
-  return rtpUDP->WriteDataOrControlPDU(frame.GetPointer(), frame.GetSize(), true);
+  return rtpUDP->WriteDataOrControlPDU(frame.GetPointer(), frame.GetHeaderSize()+frame.GetPayloadSize(), true);
 }
 
 RTP_Session::SendReceiveStatus RTP_Encoding::ReadDataPDU(RTP_DataFrame & frame)
