@@ -23,9 +23,9 @@
  *
  * Contributor(s): ______________________________________.
  *
- * $Revision: 22538 $
+ * $Revision: 24208 $
  * $Author: rjongbloed $
- * $Date: 2009-05-07 03:43:27 +0000 (Thu, 07 May 2009) $
+ * $Date: 2010-04-07 02:41:14 -0500 (Wed, 07 Apr 2010) $
  */
 
 #include <ptlib.h>
@@ -55,6 +55,8 @@ static const char SDPBandwidthPrefix[] = "SDP-Bandwidth-";
 //
 //#define FMTP_BEFORE_RTPMAP 1
 
+#define SDP_MIN_PTIME 10
+
 
 /////////////////////////////////////////////////////////
 //
@@ -71,7 +73,7 @@ OpalMediaType OpalMediaType::GetMediaTypeFromSDP(const std::string & sdp, const
       return OpalMediaType(*r);
   }
 
-  std::string s = sdp +"|" + transport;
+  std::string s = sdp + "|" + transport;
 
   for (r = mediaTypes.begin(); r != mediaTypes.end(); ++r) {
     if (OpalMediaType::GetDefinition(*r)->GetSDPType() == s)
@@ -112,8 +114,11 @@ static OpalTransportAddress ParseConnectAddress(const PStringArray & tokens, PIN
         else if (tokens[offset+2] == "0.0.0.0") {
           PTRACE(3, "SDP\tConnection address of 0.0.0.0 specified for HOLD request.");
         }
-        else
-          return OpalTransportAddress(tokens[offset+2], port, "udp");
+        else {
+          OpalTransportAddress address(tokens[offset+2], port, "udp");
+          PTRACE(4, "SDP\tParsed connection address " << address);
+          return address;
+        }
       }
       else
       {
@@ -165,7 +170,7 @@ SDPMediaFormat::SDPMediaFormat(SDPMediaDescription & parent, RTP_DataFrame::Payl
 
 
 SDPMediaFormat::SDPMediaFormat(SDPMediaDescription & parent, const OpalMediaFormat & fmt)
-  : mediaFormat(fmt)
+  : m_mediaFormat(fmt)
   , m_parent(parent) 
   , payloadType(fmt.GetPayloadType())
   , clockRate(fmt.GetClockRate())
@@ -177,104 +182,45 @@ SDPMediaFormat::SDPMediaFormat(SDPMediaDescription & parent, const OpalMediaForm
 
 void SDPMediaFormat::SetFMTP(const PString & str)
 {
-  if (str.IsEmpty())
-    return;
-
-  fmtp = str;
-  if (GetMediaFormat().IsEmpty()) // Use GetMediaFormat() to force creation of member
-    return;
-
-  mediaFormat.AddOption(new OpalMediaOptionString("RawFMTP", false, str), PTrue); // Save the 'fmtp=' line so it is available at the application level.
-
-  // See if standard format OPT=VAL;OPT=VAL
-  if (str.FindOneOf(";=") == P_MAX_INDEX) {
-    // Nope, just save the whole string as is
-    mediaFormat.SetOptionString("FMTP", str);
-    return;
-  }
-
-  // guess at the seperator
-  char sep = ';';
-  if (str.Find(sep) == P_MAX_INDEX)
-    sep = ' ';
-
-  // Parse the string for option names and values OPT=VAL;OPT=VAL
-  PINDEX sep1prev = 0;
-  do {
-    // find the next separator (' ' or ';')
-    PINDEX sep1next = str.Find(sep, sep1prev);
-    if (sep1next == P_MAX_INDEX)
-      sep1next--; // Implicit assumption string is not a couple of gigabytes long ...
-
-    // find the next '='. If past the next separator, then ignore it
-    PINDEX sep2pos = str.Find('=', sep1prev);
-    if (sep2pos > sep1next)
-      sep2pos = sep1next;
-
-    PCaselessString key = str(sep1prev, sep2pos-1).Trim();
-    if (key.IsEmpty()) {
-      PTRACE(2, "SDP\tBadly formed FMTP parameter \"" << str << '"');
-      break;
-    }
-
-    OpalMediaOption * option = mediaFormat.FindOption(key);
-    if (option == NULL || key != option->GetFMTPName()) {
-      for (PINDEX i = 0; i < mediaFormat.GetOptionCount(); i++) {
-        if (key == mediaFormat.GetOption(i).GetFMTPName()) {
-          option = const_cast<OpalMediaOption *>(&mediaFormat.GetOption(i));
-          break;
-        }
-      }
-    }
-    if (option != NULL) {
-      PString value = str(sep2pos+1, sep1next-1);
-      if (dynamic_cast< OpalMediaOptionOctets * >(option) != NULL) {
-        if (str.GetLength() % 2 != 0)
-          value = value.Trim();
-      } else {
-        // for non-octet string parameters, check for mixed separators
-        PINDEX brokenSep = str.Find(' ', sep2pos);
-        if (brokenSep < sep1next) {
-          sep1next = brokenSep;
-          value = str(sep2pos+1, sep1next-1);
-        }
-        value = value.Trim();
-        if (value.IsEmpty())
-          value = "1"; // Assume it is a boolean
-      }
-      if (!option->FromString(value)) {
-        PTRACE(2, "SDP\tCould not set FMTP parameter \"" << key << "\" to value \"" << value << '"');
-      }
-    }
-
-    sep1prev = sep1next+1;
-  } while (sep1prev != P_MAX_INDEX);
+  m_fmtp = str;
+  InitialiseMediaFormat(m_mediaFormat); // Initialise all the OpalMediaOptions from m_fmtp
 }
 
 
 PString SDPMediaFormat::GetFMTP() const
 {
-  if (GetMediaFormat().IsEmpty()) // Use GetMediaFormat() to force creation of member
+  // Use GetMediaFormat() to force creation of member
+  OpalMediaFormat mediaFormat = GetMediaFormat();
+  if (mediaFormat.IsEmpty())
+    return m_fmtp;
+
+  PString fmtp = mediaFormat.GetOptionString("FMTP");
+  if (!fmtp.IsEmpty())
     return fmtp;
 
-  PString str = mediaFormat.GetOptionString("FMTP");
-  if (!str.IsEmpty())
-    return str;
-
+  PStringStream strm;
   for (PINDEX i = 0; i < mediaFormat.GetOptionCount(); i++) {
     const OpalMediaOption & option = mediaFormat.GetOption(i);
     const PString & name = option.GetFMTPName();
     if (!name.IsEmpty()) {
       PString value = option.AsString();
-      if (value != option.GetFMTPDefault()) {
-        if (!str.IsEmpty())
-          str += ';';
-        str += name + '=' + value;
+      if (value.IsEmpty() && value != option.GetFMTPDefault())
+        strm << name;
+      else {
+        PStringArray values = value.Tokenise(';', false);
+        for (PINDEX v = 0; v < values.GetSize(); ++v) {
+          value = values[v];
+          if (value != option.GetFMTPDefault()) {
+            if (!strm.IsEmpty())
+              strm << ';';
+            strm << name << '=' << value;
+          }
+        }
       }
     }
   }
 
-  return !str ? str : fmtp;
+  return strm.IsEmpty() ? m_fmtp : strm;
 }
 
 
@@ -312,30 +258,62 @@ void SDPMediaFormat::PrintOn(ostream & strm) const
 
 const OpalMediaFormat & SDPMediaFormat::GetMediaFormat() const
 {
-  if (mediaFormat.IsEmpty())
-    const_cast<SDPMediaFormat *>(this)->InitialiseMediaFormat();
-  return mediaFormat;
+  if (m_mediaFormat.IsEmpty())
+    InitialiseMediaFormat(const_cast<SDPMediaFormat *>(this)->m_mediaFormat);
+  return m_mediaFormat;
 }
 
 
-void SDPMediaFormat::InitialiseMediaFormat()
+OpalMediaFormat & SDPMediaFormat::GetWritableMediaFormat()
 {
-  mediaFormat = OpalMediaFormat(payloadType, clockRate, encodingName, "sip");
+  if (m_mediaFormat.IsEmpty())
+    InitialiseMediaFormat(m_mediaFormat);
+  return m_mediaFormat;
+}
+
+
+void SDPMediaFormat::InitialiseMediaFormat(OpalMediaFormat & mediaFormat) const
+{
+  if (mediaFormat.IsEmpty()) {
+    OpalMediaFormatList masterFormats = OpalMediaFormat::GetMatchingRegisteredMediaFormats(payloadType, clockRate, encodingName, "sip");
+    for (OpalMediaFormatList::iterator iterFormat = masterFormats.begin(); iterFormat != masterFormats.end(); ++iterFormat) {
+      OpalMediaFormat adjustedFormat = *iterFormat;
+      SetMediaFormatOptions(adjustedFormat);
+      // skip formats whose fmtp don't match options
+      if (iterFormat->ValidateMerge(adjustedFormat)) {
+        PTRACE(3, "SIP\tRTP payload type " << encodingName << " matched to codec " << *iterFormat);
+        mediaFormat = adjustedFormat;
+        break;
+      }
+
+      PTRACE(4, "SIP\tRTP payload type " << encodingName << " not matched to codec " << *iterFormat);
+    }
+  }
+
   if (mediaFormat.IsEmpty())
     mediaFormat = OpalMediaFormat(encodingName);
+
   if (mediaFormat.IsEmpty()) {
     PTRACE(2, "SDP\tCould not find media format for \""
            << encodingName << "\", pt=" << payloadType << ", clock=" << clockRate);
     return;
   }
 
+  SetMediaFormatOptions(mediaFormat);
+}
+
+
+void SDPMediaFormat::SetMediaFormatOptions(OpalMediaFormat & mediaFormat) const
+{
   mediaFormat.MakeUnique();
   mediaFormat.SetPayloadType(payloadType);
+  mediaFormat.SetOptionInteger(OpalAudioFormat::ChannelsOption(), parameters.IsEmpty() ? 1 : parameters.AsUnsigned());
 
-  if (!parameters.IsEmpty() && (mediaFormat.GetMediaType() == OpalMediaType::Audio())) 
-    mediaFormat.SetOptionInteger(OpalAudioFormat::ChannelsOption(), parameters.AsUnsigned());
-  else
-    mediaFormat.SetOptionInteger(OpalAudioFormat::ChannelsOption(), 1);
+  // Set bandwidth limits, if present
+  for (SDPBandwidth::const_iterator r = m_parent.GetBandwidth().begin(); r != m_parent.GetBandwidth().end(); ++r) {
+    if (r->second > 0)
+      mediaFormat.AddOption(new OpalMediaOptionString(SDPBandwidthPrefix + r->first, false, r->second), true);
+  }
 
   // Fill in the default values for (possibly) missing FMTP options
   for (PINDEX i = 0; i < mediaFormat.GetOptionCount(); i++) {
@@ -343,24 +321,111 @@ void SDPMediaFormat::InitialiseMediaFormat()
     if (!option.GetFMTPName().IsEmpty() && !option.GetFMTPDefault().IsEmpty())
       option.FromString(option.GetFMTPDefault());
   }
-
-  for (SDPBandwidth::const_iterator r = m_parent.GetBandwidth().begin(); r != m_parent.GetBandwidth().end(); ++r) {
-    if (r->second > 0)
-      mediaFormat.AddOption(new OpalMediaOptionString(SDPBandwidthPrefix + r->first, false, r->second), true);
+  // Set the frame size options from media's ptime & maxptime attributes
+  if (m_parent.GetPTime() >= SDP_MIN_PTIME && mediaFormat.HasOption(OpalAudioFormat::TxFramesPerPacketOption())) {
+    unsigned frames = (m_parent.GetPTime() * mediaFormat.GetTimeUnits()) / mediaFormat.GetFrameTime();
+    if (frames < 1)
+      frames = 1;
+    mediaFormat.SetOptionInteger(OpalAudioFormat::TxFramesPerPacketOption(), frames);
   }
+  if (m_parent.GetMaxPTime() >= SDP_MIN_PTIME && mediaFormat.HasOption(OpalAudioFormat::RxFramesPerPacketOption())) {
+    unsigned frames = (m_parent.GetMaxPTime() * mediaFormat.GetTimeUnits()) / mediaFormat.GetFrameTime();
+    if (frames < 1)
+      frames = 1;
+    mediaFormat.SetOptionInteger(OpalAudioFormat::RxFramesPerPacketOption(), frames);
+  }
+
+  // If format has an explicit FMTP option then we only use that, no high level parsing
+  if (mediaFormat.SetOptionString("FMTP", m_fmtp))
+    return;
+
+  // No FMTP to parse, may as well stop here
+  if (m_fmtp.IsEmpty())
+    return;
+
+  // Save the raw 'fmtp=' line so it is available for information purposes.
+  mediaFormat.AddOption(new OpalMediaOptionString("RawFMTP", false, m_fmtp), true);
+
+  // guess at the seperator
+  char sep = ';';
+  if (m_fmtp.Find(sep) == P_MAX_INDEX)
+    sep = ' ';
+
+  // Parse the string for option names and values OPT=VAL;OPT=VAL
+  PINDEX sep1prev = 0;
+  do {
+    // find the next separator (' ' or ';')
+    PINDEX sep1next = m_fmtp.Find(sep, sep1prev);
+    if (sep1next == P_MAX_INDEX)
+      sep1next--; // Implicit assumption string is not a couple of gigabytes long ...
+
+    // find the next '='. If past the next separator, then ignore it
+    PINDEX sep2pos = m_fmtp.Find('=', sep1prev);
+    if (sep2pos > sep1next)
+      sep2pos = sep1next;
+
+    PCaselessString key = m_fmtp(sep1prev, sep2pos-1).Trim();
+    if (key.IsEmpty()) {
+      PTRACE(2, "SDP\tBadly formed FMTP parameter \"" << m_fmtp << '"');
+      break;
+    }
+
+    OpalMediaOption * option = mediaFormat.FindOption(key);
+    if (option == NULL || key != option->GetFMTPName()) {
+      for (PINDEX i = 0; i < mediaFormat.GetOptionCount(); i++) {
+        if (key == mediaFormat.GetOption(i).GetFMTPName()) {
+          option = const_cast<OpalMediaOption *>(&mediaFormat.GetOption(i));
+          break;
+        }
+      }
+    }
+    if (option != NULL) {
+      PString value = m_fmtp(sep2pos+1, sep1next-1);
+
+      if (dynamic_cast< OpalMediaOptionOctets * >(option) != NULL) {
+        if (m_fmtp.GetLength() % 2 != 0)
+          value = value.Trim();
+      }
+      else {
+        // for non-octet string parameters, check for mixed separators
+        PINDEX brokenSep = m_fmtp.Find(' ', sep2pos);
+        if (brokenSep < sep1next) {
+          sep1next = brokenSep;
+          value = m_fmtp(sep2pos+1, sep1next-1);
+        }
+        value = value.Trim();
+        if (value.IsEmpty())
+          value = "1"; // Assume it is a boolean
+      }
+
+      if (dynamic_cast< OpalMediaOptionString * >(option) != NULL) {
+        PString previous = option->AsString();
+        if (!previous.IsEmpty())
+          value = previous + ';' + value;
+      }
+
+      if (!option->FromString(value)) {
+        PTRACE(2, "SDP\tCould not set FMTP parameter \"" << key << "\" to value \"" << value << '"');
+      }
+    }
+
+    sep1prev = sep1next+1;
+  } while (sep1prev != P_MAX_INDEX);
 }
 
 
 bool SDPMediaFormat::PreEncode()
 {
-  mediaFormat.SetOptionString(OpalMediaFormat::ProtocolOption(), "SIP");
-  return mediaFormat.ToCustomisedOptions();
+  m_mediaFormat.SetOptionString(OpalMediaFormat::ProtocolOption(), "SIP");
+  return m_mediaFormat.ToCustomisedOptions();
 }
 
 
 bool SDPMediaFormat::PostDecode(unsigned bandwidth)
 {
-  if (GetMediaFormat().IsEmpty()) // Use GetMediaFormat() to force creation of member
+  // Use GetWritableMediaFormat() to force creation of member
+  OpalMediaFormat & mediaFormat = GetWritableMediaFormat();
+  if (mediaFormat.IsEmpty())
     return false;
 
   // try to init encodingName from global list, to avoid PAssert when media has no rtpmap
@@ -370,6 +435,14 @@ bool SDPMediaFormat::PostDecode(unsigned bandwidth)
   if (bandwidth > 0) {
     PTRACE(4, "SDP\tAdjusting format \"" << mediaFormat << "\" bandwidth to " << bandwidth);
     mediaFormat.SetOptionInteger(OpalMediaFormat::MaxBitRateOption(), bandwidth);
+
+    /* Some codecs (e.g. MPEG4, H.264) have difficulties with controlling max
+       bit rate. This is due to the very poor support of TIAS by SIP clients.
+       So, if we HAVE got a TIAS we indicate it with a special media option
+       so the codec can then trust the "MaxBitRate" option. If this is not
+       present then the codec has to play games with downgrading profile
+       levels to assure that a max bit rate is not exceeded. */
+    mediaFormat.AddOption(new OpalMediaOptionBoolean("SDP-Bandwidth-TIAS", false, OpalMediaOption::AndMerge, true), true);
   }
 
   mediaFormat.SetOptionString(OpalMediaFormat::ProtocolOption(), "SIP");
@@ -381,18 +454,6 @@ bool SDPMediaFormat::PostDecode(unsigned bandwidth)
 }
 
 
-void SDPMediaFormat::SetPacketTime(const PString & optionName, unsigned ptime)
-{
-  if (mediaFormat.HasOption(optionName)) {
-    unsigned frameTime = mediaFormat.GetFrameTime();
-    unsigned newCount = (ptime*mediaFormat.GetTimeUnits()+frameTime-1)/frameTime;
-    mediaFormat.SetOptionInteger(optionName, newCount);
-    PTRACE(4, "SDP\tMedia format \"" << mediaFormat << "\" option \"" << optionName
-           << "\" set to " << newCount << " packets from " << ptime << " milliseconds");
-  }
-}
-
-
 //////////////////////////////////////////////////////////////////////////////
 
 unsigned & SDPBandwidth::operator[](const PString & type)
@@ -441,8 +502,11 @@ void SDPBandwidth::SetMin(const PString & type, unsigned value)
 
 //////////////////////////////////////////////////////////////////////////////
 
-SDPMediaDescription::SDPMediaDescription(const OpalTransportAddress & address)
+SDPMediaDescription::SDPMediaDescription(const OpalTransportAddress & address, const OpalMediaType & type)
   : transportAddress(address)
+  , mediaType(type)
+  , ptime(0)
+  , maxptime(0)
 {
   direction = Undefined;
   port      = 0;
@@ -670,19 +734,6 @@ SDPMediaFormat * SDPMediaDescription::FindFormat(PString & params) const
 }
 
 
-void SDPMediaDescription::SetPacketTime(const PString & optionName, const PString & value)
-{
-  unsigned newTime = value.AsUnsigned();
-  if (newTime < 10) {
-    PTRACE(2, "SDP\tMalformed (max)ptime attribute value " << value);
-    return;
-  }
-
-  for (SDPMediaFormatList::iterator format = formats.begin(); format != formats.end(); ++format)
-   format->SetPacketTime(optionName, newTime);
-}
-
-
 bool SDPMediaDescription::PreEncode()
 {
   for (SDPMediaFormatList::iterator format = formats.begin(); format != formats.end(); ++format) {
@@ -699,8 +750,6 @@ void SDPMediaDescription::Encode(const OpalTransportAddress & commonAddr, ostrea
   PIPSocket::Address commonIP, transportIP;
   if (transportAddress.GetIpAddress(transportIP) && commonAddr.GetIpAddress(commonIP) && commonIP != transportIP)
     connectString = GetConnectAddressString(transportAddress);
-  if (connectString.IsEmpty())
-    connectString = GetConnectAddressString(transportAddress);
   PrintOn(strm, connectString);
 }
 
@@ -762,17 +811,12 @@ OpalMediaFormatList SDPMediaDescription::GetMediaFormats() const
   OpalMediaFormatList list;
 
   for (SDPMediaFormatList::const_iterator format = formats.begin(); format != formats.end(); ++format) {
-    OpalMediaFormat opalFormat = format->GetMediaFormat();
-    if (opalFormat.IsEmpty())
+    OpalMediaFormat mediaFormat = format->GetMediaFormat();
+    if (mediaFormat.IsValid())
+      list += mediaFormat;
+    else {
       PTRACE(2, "SIP\tRTP payload type " << format->GetPayloadType() 
              << ", name=" << format->GetEncodingName() << ", not matched to supported codecs");
-    else {
-      if (opalFormat.GetMediaType() == mediaType && 
-          opalFormat.IsValidForProtocol("sip") &&
-          opalFormat.GetEncodingName() != NULL) {
-        PTRACE(3, "SIP\tRTP payload type " << format->GetPayloadType() << " matched to codec " << opalFormat);
-        list += opalFormat;
-      }
     }
   }
 
@@ -786,26 +830,6 @@ void SDPMediaDescription::AddSDPMediaFormat(SDPMediaFormat * sdpMediaFormat)
 }
 
 
-void SDPMediaDescription::RemoveSDPMediaFormat(const SDPMediaFormat & sdpMediaFormat)
-{
-  OpalMediaFormat mediaFormat = sdpMediaFormat.GetMediaFormat();
-  const char * encodingName = mediaFormat.GetEncodingName();
-  unsigned clockRate = mediaFormat.GetClockRate();
-
-  if (!mediaFormat.IsValidForProtocol("sip") || encodingName == NULL || *encodingName == '\0')
-    return;
-
-  for (SDPMediaFormatList::iterator format = formats.begin(); format != formats.end(); ) {
-    if ((format->GetEncodingName() *= encodingName) && (format->GetClockRate() == clockRate)) {
-      PTRACE(3, "SDP\tRemoving format=" << encodingName << ", " << format->GetPayloadType());
-      formats.erase(format++);
-    }
-    else
-      ++format;
-  }
-}
-
-
 void SDPMediaDescription::AddMediaFormat(const OpalMediaFormat & mediaFormat)
 {
   if (!mediaFormat.IsTransportable() || !mediaFormat.IsValidForProtocol("sip")) {
@@ -814,13 +838,26 @@ void SDPMediaDescription::AddMediaFormat(const OpalMediaFormat & mediaFormat)
   }
 
   RTP_DataFrame::PayloadTypes payloadType = mediaFormat.GetPayloadType();
+  const char * encodingName = mediaFormat.GetEncodingName();
   unsigned clockRate = mediaFormat.GetClockRate();
 
   for (SDPMediaFormatList::iterator format = formats.begin(); format != formats.end(); ++format) {
-    if (format->GetPayloadType() == payloadType ||
-        ((format->GetEncodingName() *= mediaFormat.GetEncodingName()) && format->GetClockRate() == clockRate)
-        ) {
-      PTRACE(4, "SDP\tSDP not including " << mediaFormat << " as it is already included");
+    const OpalMediaFormat & sdpMediaFormat = format->GetMediaFormat();
+    if (mediaFormat == sdpMediaFormat) {
+      PTRACE(2, "SDP\tSDP not including " << mediaFormat << " as already included");
+      return;
+    }
+
+    if (format->GetPayloadType() == payloadType) {
+      PTRACE(2, "SDP\tSDP not including " << mediaFormat << " as it is has duplicate payload type " << payloadType);
+      return;
+    }
+
+    if (format->GetEncodingName() == encodingName &&
+        format->GetClockRate() == clockRate &&
+        mediaFormat.ValidateMerge(sdpMediaFormat)) {
+      PTRACE(2, "SDP\tSDP not including " << mediaFormat
+             << " as an equivalent (" << sdpMediaFormat << ") is already included");
       return;
     }
   }
@@ -832,6 +869,7 @@ void SDPMediaDescription::AddMediaFormat(const OpalMediaFormat & mediaFormat)
   AddSDPMediaFormat(sdpFormat);
 }
 
+
 void SDPMediaDescription::ProcessMediaOptions(SDPMediaFormat & /*sdpFormat*/, const OpalMediaFormat & /*mediaFormat*/)
 {
 }
@@ -847,8 +885,8 @@ void SDPMediaDescription::AddMediaFormats(const OpalMediaFormatList & mediaForma
 
 //////////////////////////////////////////////////////////////////////////////
 
-SDPRTPAVPMediaDescription::SDPRTPAVPMediaDescription(const OpalTransportAddress & address)
-  : SDPMediaDescription(address)
+SDPRTPAVPMediaDescription::SDPRTPAVPMediaDescription(const OpalTransportAddress & address, const OpalMediaType & mediaType)
+  : SDPMediaDescription(address, mediaType)
 {
 }
 
@@ -920,7 +958,7 @@ void SDPRTPAVPMediaDescription::SetAttribute(const PString & attr, const PString
 //////////////////////////////////////////////////////////////////////////////
 
 SDPAudioMediaDescription::SDPAudioMediaDescription(const OpalTransportAddress & address)
-  : SDPRTPAVPMediaDescription(address)
+  : SDPRTPAVPMediaDescription(address, OpalMediaType::Audio())
 {
 }
 
@@ -937,69 +975,146 @@ bool SDPAudioMediaDescription::PrintOn(ostream & str, const PString & connectStr
   if (!SDPRTPAVPMediaDescription::PrintOn(str, connectString))
     return false;
 
-#ifdef HAVE_PTIME
-  // Fill in the ptime  as maximum tx packets of all media formats
-  // and maxptime as minimum rx packets of all media formats
-  unsigned ptime = 0;
+  /* The ptime parameter is a recommendation to the remote that we want them
+     to send that number of milliseconds on audio in each RTP packet. OPAL
+     does not have an equivalent parameter anywhere, so we do not provide it
+     on outgoing SDP. We do try to honour it on incoing SDP, however.
+
+     The maxptime parameter can be represented by the RxFramesPerPacketOption,
+     so we go through all the codecs offered and calculate a maxptime based on
+     the smallest maximum rx packets of the codecs. Allowance must be made for
+     maxptime to be at least big enough for 1 frame per packet for the largest
+     frame size of those codecs.
+     
+     In practice this generally means if we mix GSM and G.723.1 then the
+     maxptime cannot be smaller than 30ms even if GSM wants one frame per
+     packet. That should still work as teh remote cannot send 2fpp as it would
+     exceed the 30ms.
+
+     However, certain combinations cannot be represented, e.g. if you want 2fpp
+     of G.729 (20ms) and 1fpp of G.723.1 (30ms) then the G.729 codec COULD
+     receive 3fpp. This is really a failing in SIP/SDP and the techniques for
+     woking around the limitation are for too complicated to be worth doing for
+     what should be rare cases.
+    */
+
+  unsigned largestFrameTime = 0;
   unsigned maxptime = UINT_MAX;
 
   // output attributes for each payload type
-  for (format = formats.begin(); format != formats.end(); ++format) {
-    const OpalMediaFormat & mediaFormat = formats[i].GetMediaFormat();
-    if (mediaFormat.HasOption(OpalAudioFormat::TxFramesPerPacketOption())) {
-      unsigned ptime1 = txFrames*mediaFormat.GetFrameTime()/mediaFormat.GetTimeUnits();
-      if (ptime < ptime1)
-        ptime = ptime1;
-    }
+  for (SDPMediaFormatList::const_iterator format = formats.begin(); format != formats.end(); ++format) {
+    const OpalMediaFormat & mediaFormat = format->GetMediaFormat();
     if (mediaFormat.HasOption(OpalAudioFormat::RxFramesPerPacketOption())) {
-      unsigned maxptime1 = mediaFormat.GetOptionInteger(OpalAudioFormat::RxFramesPerPacketOption())*mediaFormat.GetFrameTime()/mediaFormat.GetTimeUnits();
+      unsigned frameTime = mediaFormat.GetFrameTime()/mediaFormat.GetTimeUnits();
+      if (largestFrameTime < frameTime)
+        largestFrameTime = frameTime;
+
+      unsigned maxptime1 = mediaFormat.GetOptionInteger(OpalAudioFormat::RxFramesPerPacketOption())*frameTime;
       if (maxptime > maxptime1)
         maxptime = maxptime1;
     }
   }
 
-  // don't output ptime parameters, as some Cisco endpoints barf on it
-  // and it's not very well-defined anyway
-  //if (ptime > 0)
-  //  str << "a=ptime:" << ptime << "\r\n";
-
-  if (maxptime < UINT_MAX)
+  if (maxptime < UINT_MAX) {
+    if (maxptime < largestFrameTime)
+      maxptime = largestFrameTime;
     str << "a=maxptime:" << maxptime << "\r\n";
-#endif // HAVE_PTIME
+  }
 
   return true;
 }
 
+
 void SDPAudioMediaDescription::SetAttribute(const PString & attr, const PString & value)
 {
-#ifdef HAVE_PTIME
   if (attr *= "ptime") {
-    SetPacketTime(OpalAudioFormat::TxFramesPerPacketOption(), value);
+    unsigned newTime = value.AsUnsigned();
+    if (newTime < SDP_MIN_PTIME) {
+      PTRACE(2, "SDP\tMalformed ptime attribute value " << value);
+      return;
+    }
+    ptime = newTime;
     return;
   }
 
   if (attr *= "maxptime") {
-    SetPacketTime(OpalAudioFormat::RxFramesPerPacketOption(), value);
+    unsigned newTime = value.AsUnsigned();
+    if (newTime < SDP_MIN_PTIME) {
+      PTRACE(2, "SDP\tMalformed maxptime attribute value " << value);
+      return;
+    }
+    maxptime = newTime;
     return;
   }
-#endif // HAVE_PTIME
 
   return SDPRTPAVPMediaDescription::SetAttribute(attr, value);
 }
 
+
 //////////////////////////////////////////////////////////////////////////////
 
 SDPVideoMediaDescription::SDPVideoMediaDescription(const OpalTransportAddress & address)
-  : SDPRTPAVPMediaDescription(address)
+  : SDPRTPAVPMediaDescription(address, OpalMediaType::Video())
 {
 }
 
+
 PString SDPVideoMediaDescription::GetSDPMediaType() const 
 { 
   return "video"; 
 }
 
 
+static const char * const ContentRoleNames[] = { NULL, "slides", "main", "speaker", "sl" };
+
+
+bool SDPVideoMediaDescription::PrintOn(ostream & str, const PString & connectString) const
+{
+  // call ancestor
+  if (!SDPRTPAVPMediaDescription::PrintOn(str, connectString))
+    return false;
+
+#if OPAL_VIDEO
+  for (SDPMediaFormatList::const_iterator format = formats.begin(); format != formats.end(); ++format) {
+    PINDEX role = format->GetMediaFormat().GetOptionEnum(OpalVideoFormat::ContentRoleOption());
+    if (role > 0) {
+      str << "a=content:" << ContentRoleNames[role] << "\r\n";
+      break;
+    }
+  }
+#endif
+
+  return true;
+}
+
+
+void SDPVideoMediaDescription::SetAttribute(const PString & attr, const PString & value)
+{
+  if (attr *= "content") {
+    PINDEX role = 0;
+    PStringArray tokens = value.Tokenise(',');
+    for (PINDEX i = 0; i < tokens.GetSize(); ++i) {
+      for (role = PARRAYSIZE(ContentRoleNames)-1; role > 0; --role) {
+        if (tokens[i] *= ContentRoleNames[role])
+          break;
+      }
+      if (role > 0)
+        break;
+    }
+
+#if OPAL_VIDEO
+    for (SDPMediaFormatList::iterator format = formats.begin(); format != formats.end(); ++format)
+      format->GetWritableMediaFormat().SetOptionEnum(OpalVideoFormat::ContentRoleOption(), role);
+#endif
+
+    return;
+  }
+
+  return SDPRTPAVPMediaDescription::SetAttribute(attr, value);
+}
+
+
+
 bool SDPVideoMediaDescription::PreEncode()
 {
   if (!SDPRTPAVPMediaDescription::PreEncode())
@@ -1035,10 +1150,11 @@ bool SDPVideoMediaDescription::PreEncode()
 //////////////////////////////////////////////////////////////////////////////
 
 SDPApplicationMediaDescription::SDPApplicationMediaDescription(const OpalTransportAddress & address)
-  : SDPMediaDescription(address)
+  : SDPMediaDescription(address, "")
 {
 }
 
+
 PCaselessString SDPApplicationMediaDescription::GetSDPTransportType() const
 { 
   return "rtp/avp"; 
@@ -1050,11 +1166,13 @@ PString SDPApplicationMediaDescription::GetSDPMediaType() const
   return "application"; 
 }
 
+
 SDPMediaFormat * SDPApplicationMediaDescription::CreateSDPMediaFormat(const PString & portString)
 {
   return new SDPMediaFormat(*this, RTP_DataFrame::DynamicBase, portString);
 }
 
+
 PString SDPApplicationMediaDescription::GetSDPPortList() const
 {
   PStringStream str;
@@ -1073,44 +1191,21 @@ const PString & SDPSessionDescription::ConferenceTotalBandwidthType()      { sta
 const PString & SDPSessionDescription::ApplicationSpecificBandwidthType()  { static PString s = "AS";   return s; }
 const PString & SDPSessionDescription::TransportIndependentBandwidthType() { static PString s = "TIAS"; return s; }
 
-SDPSessionDescription::SDPSessionDescription(const OpalTransportAddress & address)
-  : sessionName(SIP_DEFAULT_SESSION_NAME),
-    ownerUsername('-'),
-    ownerAddress(address),
-    defaultConnectAddress(address)
+SDPSessionDescription::SDPSessionDescription(time_t sessionId, unsigned version, const OpalTransportAddress & address)
+  : sessionName(SIP_DEFAULT_SESSION_NAME)
+  , ownerUsername('-')
+  , ownerSessionId(sessionId)
+  , ownerVersion(version)
+  , ownerAddress(address)
+  , defaultConnectAddress(address)
 {
   protocolVersion  = 0;
-  ownerSessionId  = ownerVersion = (unsigned)PTime().GetTimeInSeconds();
   direction = SDPMediaDescription::Undefined;
 }
 
 
 void SDPSessionDescription::PrintOn(ostream & str) const
 {
-  OpalTransportAddress connectionAddress(defaultConnectAddress);
-
-  // see common connect address is needed
-  {
-    OpalTransportAddress descrAddress;
-    PINDEX matched = 0;
-    PINDEX descrMatched = 0;
-    PINDEX i;
-    for (i = 0; i < mediaDescriptions.GetSize(); i++) {
-      if (i == 0)
-        descrAddress = mediaDescriptions[i].GetTransportAddress();
-      if (mediaDescriptions[i].GetTransportAddress() == connectionAddress)
-        ++matched;
-      if (mediaDescriptions[i].GetTransportAddress() == descrAddress)
-        ++descrMatched;
-    }
-    if (connectionAddress != descrAddress) {
-      if ((descrMatched > matched))
-        connectionAddress = descrAddress;
-      else
-        connectionAddress.MakeEmpty();
-    }
-  }
-
   /* encode mandatory session information, note the order is important according to RFC!
      Must be vosiuepcbzkatrm and within the m it is icbka */
   str << "v=" << protocolVersion << "\r\n"
@@ -1121,8 +1216,8 @@ void SDPSessionDescription::PrintOn(ostream & str) const
       << "\r\n"
          "s=" << sessionName << "\r\n";
 
-  if (!connectionAddress.IsEmpty())
-    str << "c=" << GetConnectAddressString(connectionAddress) << "\r\n";
+  if (!defaultConnectAddress.IsEmpty())
+    str << "c=" << GetConnectAddressString(defaultConnectAddress) << "\r\n";
   
   str << bandwidth
       << "t=" << "0 0" << "\r\n";
@@ -1147,7 +1242,7 @@ void SDPSessionDescription::PrintOn(ostream & str) const
   // encode media session information
   for (PINDEX i = 0; i < mediaDescriptions.GetSize(); i++) {
     if (mediaDescriptions[i].PreEncode())
-      mediaDescriptions[i].Encode(connectionAddress, str);
+      mediaDescriptions[i].Encode(defaultConnectAddress, str);
   }
 }
 
@@ -1387,6 +1482,41 @@ void SDPSessionDescription::SetDefaultConnectAddress(const OpalTransportAddress
      ownerAddress = address;
 }
 
+
+static OpalMediaFormat GetNxECapabilities(const SDPMediaDescription & incomingMedia, const OpalMediaFormat & mediaFormat)
+{
+  // Find the payload type and capabilities used for telephone-event, if present
+  const SDPMediaFormatList & sdpMediaList = incomingMedia.GetSDPMediaFormats();
+  for (SDPMediaFormatList::const_iterator format = sdpMediaList.begin(); format != sdpMediaList.end(); ++format) {
+    if (format->GetEncodingName() == mediaFormat.GetEncodingName())
+      return format->GetMediaFormat();
+  }
+
+  return OpalMediaFormat();
+}
+
+
+OpalMediaFormatList SDPSessionDescription::GetMediaFormats() const
+{
+  OpalMediaFormatList formatList;
+
+  // Extract the media from SDP into our remote capabilities list
+  for (PINDEX i = 0; i < mediaDescriptions.GetSize(); ++i) {
+    SDPMediaDescription & mediaDescription = mediaDescriptions[i];
+
+    formatList += mediaDescription.GetMediaFormats();
+
+    // Find the payload type and capabilities used for telephone-event, if present
+    formatList += GetNxECapabilities(mediaDescription, OpalRFC2833);
+#if OPAL_T38_CAPABILITY
+    formatList += GetNxECapabilities(mediaDescription, OpalCiscoNSE);
+#endif
+  }
+
+  return formatList;
+}
+
+
 #endif // OPAL_SIP
 
 // End of file ////////////////////////////////////////////////////////////////
